(***********************************************************************
 * File: Cache.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies a VIPT Cache
 * 
 *************************************************************************)
Require Import Dict_bound Addressing Page.

Module Type CACHE (PG:PAGE).
  Parameter cache : Set.
  
  Definition cache_key : Set := (vadd*madd)%type.
  Axiom eq_cache_key : forall (k k':cache_key), {k = k'} + {k <> k'}.

  Parameter get_cache : cache -> cache_key -> option PG.page.
  Parameter add_cache : cache -> cache_key -> PG.page -> cache*option (cache_key*PG.page).
  Parameter remove_cache : cache -> cache_key -> cache.
  Parameter remove_ma_cache : cache -> madd -> cache*option (cache_key*PG.page).
  Parameter empty_cache : cache.
  Parameter is_empty_cache : cache -> Prop.
  Parameter is_full_cache : cache -> cache_key -> Prop.

  Axiom is_empty_cache_empty : is_empty_cache empty_cache.
  Axiom not_empty_add_cache : forall (c:cache) (k:cache_key) (pg:PG.page),
    PG.is_cacheable pg ->
    ~ is_empty_cache (fst (add_cache c k pg)).

  Axiom get_empty_cache : forall (k:cache_key),
    get_cache empty_cache k = None.
  Axiom get_add_cache_cacheable : forall (c:cache) (k:cache_key) (pg:PG.page),
    PG.is_cacheable pg ->
    Some pg = get_cache (fst (add_cache c k pg)) k.
  Axiom get_add_cache_non_cacheable : forall (c:cache) (k k':cache_key) (pg:PG.page),
    ~ PG.is_cacheable pg ->
    get_cache c k' = get_cache (fst (add_cache c k pg)) k'.

  Axiom is_full_cache_dec : forall (c:cache) (k:cache_key), {is_full_cache c k}+{~(is_full_cache c k)}.

  Axiom get_add_cache_distinct : forall (c:cache) (k k':cache_key) (pg pg':PG.page),
    let (c',entry) := add_cache c k pg in
    k <> k' -> (entry = None \/ (exists k'', exists pg'', entry = Some (k'',pg'') /\ k' <> k'')) ->
    Some pg' = get_cache c k' ->
    Some pg' = get_cache c' k'.

  Axiom get_add_cache_distinct_not_new : forall (c:cache) (k k':cache_key) (pg pg':PG.page),
    let (c',entry) := add_cache c k pg in
    k <> k' -> Some pg' = get_cache c' k' ->
    Some pg' = get_cache c k'.

  Axiom add_cache_entry_behavior_not_full : forall (c:cache) (k:cache_key) (pg:PG.page),
    let (c',entry) := add_cache c k pg in
      ~ is_full_cache c k -> entry = None.

  Axiom add_cache_entry_behavior_update : forall (c:cache) (k:cache_key) (pg:PG.page),
  let (c',entry) := add_cache c k pg in
    get_cache c k <> None -> entry = None.

  Axiom add_cache_entry_behavior_non_cacheable : forall (c:cache) (k:cache_key) (pg:PG.page),
  let (c',entry) := add_cache c k pg in
    ~ PG.is_cacheable pg -> entry = None.

  Axiom add_cache_entry_behavior_full : forall (c:cache) (k:cache_key) (pg:PG.page),
    let (c',entry) := add_cache c k pg in
      is_full_cache c k -> get_cache c k = None -> PG.is_cacheable pg ->
      exists k', exists pg',
        entry = Some (k',pg') /\
        get_cache c k' = Some pg' /\
        get_cache c' k' = None.

  Axiom get_remove_ma_cache_ne : forall c va' ma ma',
    ma <> ma' ->
    get_cache (fst (remove_ma_cache c ma)) (va',ma') = get_cache c (va',ma').

  Axiom get_remove_ma_cache_eq : forall c va ma,
    get_cache (fst (remove_ma_cache c ma)) (va,ma) = None.
  
  Axiom get_remove_ma_cache_some_entry : forall c va ma ma' pg,
    snd (remove_ma_cache c ma) = Some ((va,ma'),pg) ->
    ma = ma' /\ get_cache c (va,ma) = Some pg.

  Axiom get_remove_ma_cache_no_entry : forall c ma,
    snd (remove_ma_cache c ma) = None ->
    forall va, get_cache c (va,ma) = None.

  Axiom get_remove_cache : forall (c:cache) (k:cache_key),
    get_cache (remove_cache c k) k = None.

  Axiom get_remove_cache_distinct : forall (c:cache) (k k':cache_key) (pg:PG.page),
    k <> k' -> get_cache c k = Some pg -> get_cache (remove_cache c k') k = Some pg.

  Axiom get_remove_cache_old : forall (c:cache) (k k':cache_key) (pg:PG.page),
    get_cache (remove_cache c k') k = Some pg -> get_cache c k = Some pg.
End CACHE.

Module CACHE_REFINED (PG:PAGE) : CACHE PG.
  Definition page := PG.page.

  Declare Module D : DICT_BOUND
    with Definition key := (vadd*madd)%type
    with Definition info := page.

  Definition cache := D.dict.

  Definition cache_key : Set := (vadd*madd)%type.
  Lemma eq_cache_key (k k':cache_key) : {k = k'} + {k <> k'}.
    Proof.
      case (vadd_eq (fst k) (fst k')); intros.
      case (madd_eq (snd k) (snd k')); intros;
        [left | right]; auto.
      destruct k; destruct k';
        simpl in *; subst; auto.
      intro; subst k.
      elim n; auto.

      right; intro; subst k.
      elim n; auto.
    Defined.              
  
  Definition get_cache := D.get.

  Definition add_cache (c:cache) (k:cache_key) (pg:PG.page) :
      cache*option (cache_key*PG.page) :=
      match PG.page_cacheable pg with
        | true => D.set c k pg
        | false => (c,None)
      end.
    

  Definition remove_cache := D.remove.
  Definition remove_ma_cache (c:cache) (ma:madd) :=
    D.filter c (fun k => if madd_eq (snd k) ma then true else false).

  Definition empty_cache := D.empty.
  Definition is_empty_cache := D.is_empty.
  Definition is_full_cache (c:cache) (k:cache_key) :=
    D.is_full c.

  Definition is_full_cache_dec (c:cache) (k:cache_key) :=
    D.is_full_dec c.

  Definition is_empty_cache_empty := D.is_empty_empty.
  Lemma not_empty_add_cache : forall (c:cache) (k:cache_key) (pg:PG.page),
    PG.is_cacheable pg ->
    ~ is_empty_cache (fst (add_cache c k pg)).
  Proof.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg).
    apply D.not_empty_set.
    discriminate H.
  Qed.

  Definition get_empty_cache := D.get_empty.
  
  Lemma get_add_cache_cacheable : forall (c:cache) (k:cache_key) (pg:PG.page),
    PG.is_cacheable pg ->
    Some pg = get_cache (fst (add_cache c k pg)) k.
  Proof.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg).
    apply D.get_set.
    discriminate H.
  Qed.

  Lemma get_add_cache_non_cacheable : forall (c:cache) (k k':cache_key) (pg:PG.page),
    ~ PG.is_cacheable pg ->
    get_cache c k' = get_cache (fst (add_cache c k pg)) k'.
  Proof.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg); try congruence; auto.
  Qed.    

  Lemma get_add_cache_distinct : forall (c:cache) (k k':cache_key) (pg pg':PG.page),
    let (c',entry) := add_cache c k pg in
    k <> k' -> (entry = None \/ (exists k'', exists pg'', entry = Some (k'',pg'') /\ k' <> k'')) ->
    Some pg' = get_cache c k' ->
    Some pg' = get_cache c' k'.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    destruct new_cache.
    intros.
    unfold add_cache in Heqnew_cache.
    destruct (PG.page_cacheable pg).
    generalize (D.get_set_distinct c k k' pg pg'); intros.
    rewrite <- Heqnew_cache in H2; auto.

    injection Heqnew_cache; intros; subst; auto.
  Qed.    

  Lemma get_add_cache_distinct_not_new : forall (c:cache) (k k':cache_key) (pg pg':PG.page),
    let (c',entry) := add_cache c k pg in
    k <> k' -> Some pg' = get_cache c' k' ->
    Some pg' = get_cache c k'.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    induction new_cache; auto.
    intros.
    unfold add_cache in Heqnew_cache.
    induction (PG.page_cacheable pg).
    generalize (D.get_set_distinct_not_new c k k' pg pg'); intros.
    rewrite <- Heqnew_cache in H1; auto.
    injection Heqnew_cache; intros; subst a; auto.
  Qed.

  Lemma add_cache_entry_behavior_not_full : forall (c:cache) (k:cache_key) (pg:PG.page),
    let (c',entry) := add_cache c k pg in
      ~ is_full_cache c k -> entry = None.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    induction new_cache.
    intros.
    unfold add_cache in Heqnew_cache.
    induction (PG.page_cacheable pg).
    generalize (D.set_entry_behavior_not_full c k pg); intros.
    rewrite <- Heqnew_cache in H0; auto.
    injection Heqnew_cache; intros; subst b; auto.
  Qed.

  Lemma add_cache_entry_behavior_update : forall (c:cache) (k:cache_key) (pg:PG.page),
  let (c',entry) := add_cache c k pg in
    get_cache c k <> None -> entry = None.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    induction new_cache.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg).
    generalize (D.set_entry_behavior_update c k pg); intros.
    rewrite <- Heqnew_cache in H0; auto.
    injection Heqnew_cache; intros; subst b; auto.
  Qed.    

  Lemma add_cache_entry_behavior_non_cacheable : forall (c:cache) (k:cache_key) (pg:PG.page),
  let (c',entry) := add_cache c k pg in
    ~ PG.is_cacheable pg -> entry = None.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    induction new_cache.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg); congruence.
  Qed.    

  Lemma add_cache_entry_behavior_full : forall (c:cache) (k:cache_key) (pg:PG.page),
    let (c',entry) := add_cache c k pg in
      is_full_cache c k -> get_cache c k = None -> PG.is_cacheable pg ->
      exists k', exists pg',
        entry = Some (k',pg') /\
        get_cache c k' = Some pg' /\
        get_cache c' k' = None.
  Proof.
    intros.
    remember (add_cache c k pg) as new_cache; intros.
    induction new_cache.
    intros.
    unfold add_cache, PG.is_cacheable in *.
    induction (PG.page_cacheable pg); try congruence.
    generalize (D.set_entry_behavior_full c k pg); intros.
    rewrite <- Heqnew_cache in H2; auto.
  Qed.

  Definition get_remove_cache := D.get_remove.
  Definition get_remove_cache_distinct := D.get_remove_distinct.
  Definition get_remove_cache_old := D.get_remove_old.

  Lemma get_remove_ma_cache_ne : forall c va' ma ma',
    ma <> ma' ->
    get_cache (fst (remove_ma_cache c ma)) (va',ma') = get_cache c (va',ma').
  Proof.
    intros.
    unfold get_cache, remove_ma_cache.
    rewrite D.get_filter_notp; auto.
    simpl.
    case (madd_eq ma' ma); auto.
    intro H'; elim H; auto.
  Qed.

  Lemma get_remove_ma_cache_eq : forall c va ma,
    get_cache (fst (remove_ma_cache c ma)) (va,ma) = None.
  Proof.
    intros.
    unfold get_cache, remove_ma_cache.
    rewrite D.get_filter_p; auto.
    simpl.
    case (madd_eq ma ma); auto.
    intro H; elim H; auto.
  Qed.

  Lemma get_remove_ma_cache_some_entry : forall c va ma ma' pg,
    snd (remove_ma_cache c ma) = Some ((va,ma'),pg) ->
    ma = ma' /\ get_cache c (va,ma) = Some pg.
  Proof.
    intros.
    assert (ma = ma').
    remember (fun k : D.key => if madd_eq (snd k) ma then true else false).
    assert (b (va,ma') = true); auto.
    apply D.get_filter_some_entry_p with c pg.
    subst; unfold remove_ma_cache in H; auto.
    subst; simpl in *.
    remember (madd_eq ma' ma).
    destruct s; try discriminate; auto.

    split; auto.
    subst.
    erewrite D.get_filter_some_entry; eauto.
  Qed.    

  Lemma get_remove_ma_cache_no_entry : forall c ma,
    snd (remove_ma_cache c ma) = None ->
    forall va, get_cache c (va,ma) = None.
  Proof.
    intros.
    remember (get_cache c (va,ma)).
    destruct o; auto.
    unfold remove_ma_cache in H.
    remember (fun k : D.key => if madd_eq (snd k) ma then true else false).
    assert (b (va,ma) = false).
    eapply D.get_filter_no_entry; eauto.
    subst; simpl in *; auto.
    destruct (madd_eq ma ma); congruence.
  Qed.    
End CACHE_REFINED.
