(***********************************************************************
 * File: Cache_tactics.v
 * $Revision $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy $
 ***********************************************************************
 * Cache related tactics and lemmas.
 **********************************************************************)

Require Import Opt_monad Addressing Page Cache Tactics.

Module CACHE_TACTICS (PG:PAGE) (C:CACHE PG).
  Import C.

  (* Case analisys on the result of adding a page to cache c. *)
  Ltac destruct_add_cache c :=
    let opt_add_cache := fresh "opt_add_cache" in
    match goal with
      | [H: context[add_cache c ?k ?pg] |-_ ] =>
          remember (add_cache c k pg) as opt_add_cache;
            destruct opt_add_cache as [? [[[? ?] ?] |]]; crush
      | [|- context[add_cache c ?k ?pg]] =>
          remember (add_cache c k pg) as opt_add_cache;
            destruct opt_add_cache as [? [[[? ?] ?] |]]; crush
    end.

  (* Handle absurd case when a value was removed from the cache,
     but no entry was returned by add_cache with a cacheable_page. *)
  Ltac no_entry_but_removed :=
    match goal with
      | [H1: (?c', None) = add_cache ?c (?va,?ma) ?pg,
         H2: get_cache ?c (?va', ?ma') = Some ?pg',
         H3: get_cache ?c' (?va', ?ma') = None,
         H4: PG.is_cacheable ?pg |-_] =>
      let H5 := fresh "H" in
      let H6 := fresh "H" in
        generalize (get_add_cache_distinct c (va,ma) (va',ma') pg pg');
          intro H5; crush; rewrite <- H5 in H3; crush;
            generalize (get_add_cache_cacheable c (va', ma') pg); intros H6;
              apply H6 in H4; rewrite <- H1 in H4;
                simpl in *; rewrite <- H4 in H3; crush
    end.

  (* Prove that either cache is full, the page was cacheable or the
     value was not updated, given that an entry was removed from the cache. *)
  Ltac some_entry_add_cache :=
    match goal with
        | [H1: (?c', Some (?k, ?pg')) = add_cache ?c (?va, ?ma) ?pg |-_] =>
          let new_cache := fresh "new_cache" in
          let opt_entry := fresh "opt_entry" in
          let H_behavior := fresh "H_behavior" in
            assert (snd (add_cache c (va,ma) pg) = Some (k,pg'));
              [rewrite <- H1; crush;fail|];
          match goal with
            | [|- is_full_cache c (va,ma) ] =>
              generalize (add_cache_entry_behavior_not_full c (va,ma) pg)
            | [|- PG.is_cacheable pg ] =>
              generalize (add_cache_entry_behavior_non_cacheable c (va,ma) pg)
            | [|- get_cache c (va,ma) = None ] =>
              generalize (add_cache_entry_behavior_update c (va,ma) pg)
          end; intros H_behavior;
            remember (add_cache c (va,ma) pg) as new_cache;
              destruct new_cache as [new_cache opt_entry]; crush;
          match goal with
            | [|- is_full_cache c (va,ma)] =>
              destruct (is_full_cache_dec c (va,ma))
            | [|- PG.is_cacheable pg ] =>
              unfold PG.is_cacheable in *; destruct (PG.page_cacheable pg)
            | [|- get_cache c (va,ma) = None ] =>
              destruct (get_cache c (va,ma))
          end; crush; discriminate H_behavior; crush
      end.

  (* Prove that if add_cache returns an entry; it was in the cache and
     now it is not. *)
  Ltac simpl_some_cache_entry :=
    match goal with
      | [H: (?c', Some (?va',?ma',?pg')) = add_cache ?c (?va,?ma) ?pg |-_]=>
        let H' := fresh "H" in
          generalize (add_cache_entry_behavior_full c (va,ma) pg); intro H';
            crush; destruct H'; crush; try some_entry_add_cache
    end.

  (* If an entry was replaced and a value is no longer in the cache
     they must be the same key and page. *)
  Ltac removed_val_is_entry := 
    match goal with
      | [H1: (?c', Some (?va'', ?ma'', ?pg'')) = add_cache ?c (?va, ?ma) ?pg,
        H2: get_cache ?c (?va', ?ma') = Some ?pg',
        H3: get_cache ?c' (?va', ?ma') = None,
        H4: PG.is_cacheable ?pg
        |- _] =>
      let H5 := fresh "H" in
        assert (Some (va'',ma'',pg'') = Some (va', ma', pg'));
          [simpl_some_cache_entry;
            case (eq_cache_key (va'',ma'') (va', ma')); crush;
            case (eq_cache_key (va,ma) (va', ma')); crush;
              [ generalize (get_add_cache_cacheable c (va', ma') pg H4);
                intros H5; rewrite <- H1 in H5; simpl in H5; crush
              | generalize (get_add_cache_distinct c (va,ma) (va',ma') pg pg');
                intro H5; crush; rewrite <- H5 in H3; crush]
            | crush]
    end.

  (* If the value was not in old cache, but it is in the new one, it
     must be the read entry. *)
  Ltac new_entry_is_added :=
    match goal with
      | [H1: (?c', _) = add_cache ?c (?va1, ?ma1) ?pg1,
        H2: None = get_cache ?c (?va2, ?ma2),
        H3: Some ?pg2 = get_cache ?c' (?va2, ?ma2) |-_] =>
      destruct (eq_cache_key (va1,ma1) (va2,ma2));
        [|generalize (get_add_cache_distinct_not_new
          c (va1,ma1) (va2,ma2) pg1 pg2)]; crush
    end.

  (* If a cache entry was replaced, it was present in the original cache. *)
  Lemma cache_deleted_entry_was_there : forall c va ma k pg pg',
    (snd (add_cache c (va,ma) pg)) = Some (k, pg') ->
    get_cache c k = Some pg'.
  Proof.
    intros; destruct_add_cache c; simpl_some_cache_entry.
  Qed.

  Lemma cache_deleted_entry: forall c va va' ma ma' pg pg',
    PG.is_cacheable pg ->
    get_cache c (va',ma') = Some pg' ->
    get_cache (fst (add_cache c (va,ma) pg)) (va',ma') = None ->
    (snd (add_cache c (va,ma) pg)) = Some (va',ma', pg').
  Proof.
    intros; destruct_add_cache c; try no_entry_but_removed;
      removed_val_is_entry.
  Qed.

End CACHE_TACTICS.
