(** Libraries. *)
Require Import Coqlib.
Require Import Errors.
Require Import AST.
Require Import Smallstep.
(** Languages (syntax and semantics). *)
Require Csyntax.
Require Csem.
Require Cstrategy.
Require Cexec.
Require Clight.
Require Csharpminor.
Require Cminor.
Require CminorSel.
Require RTL.
Require LTL.
Require Linear.
Require Mach.
Require Asm.
(** Translation passes. *)
Require Initializers.
Require SimplExpr.
Require SimplLocals.
Require Cshmgen.
Require Cminorgen.
Require Selection.
Require RTLgen.
Require Tailcall.
Require Inlining.
Require Renumber.
Require Constprop.
Require CSE.
Require Allocation.
Require Tunneling.
Require Linearize.
Require CleanupLabels.
Require Stacking.
Require Asmgen.
(** Proofs of semantic preservation. *)
Require SimplExprproof.
Require SimplLocalsproof.
Require Cshmgenproof.
Require Cminorgenproof.
Require Selectionproof.
Require RTLgenproof.
Require Tailcallproof.
Require Inliningproof.
Require Renumberproof.
Require Constpropproof.
Require CSEproof.
Require Allocproof.
Require Tunnelingproof.
Require Linearizeproof.
Require CleanupLabelsproof.
Require Stackingproof.
Require Asmgenproof.
Require Import Compiler.
Require MachIR.
Require Import MachIRStealthType.
Require Import Stealth_secure_program.
Require Import Stealth_interleaving.

Definition compiler (p: Csyntax.program) : res MachIR.program :=
   OK p
  @@@ SimplExpr.transl_program
   @@ print print_Clight
  @@@ SimplLocals.transf_program
  @@@ Cshmgen.transl_program
  @@@ Cminorgen.transl_program
   @@ print print_Cminor
  @@@ Selection.sel_program
  @@@ RTLgen.transl_program
   @@ print print_RTL
   @@ Tailcall.transf_program
   @@ print print_RTL_tailcall
  @@@ Inlining.transf_program
   @@ Renumber.transf_program
   @@ print print_RTL_inline
   @@ Constprop.transf_program
   @@ Renumber.transf_program
   @@ print print_RTL_constprop
  @@@ CSE.transf_program
   @@ print print_RTL_cse
  @@@ Allocation.transf_program
   @@ print print_LTL
   @@ Tunneling.tunnel_program
  @@@ Linearize.transf_program
   @@ CleanupLabels.transf_program
  @@@ Stacking.transf_program
  @@@ MachIR.transl_program.

Module END_TO_END (S:SCHEDULER) (INT:INTERLEAVING S).

Module Secure := SecureMachIR S INT.
Module INT_PROOF := INTERLEAVING_PROOF S INT.

Import Stealth_traces STEALTH_TRACES
       Stealth_action_streams TRACES_VIC
       Stealth_equivalence STEALTH_EQ
       Stealth_valid VALID_SM_STATE
       ACTION_STREAMS Stealth_actions
       Addressing Ensembles Values
       Secure INT INT_PROOF.


Record virtual_mapping : Type := Make_vm {
    lift_address : val -> vadd;
    lift_address_not_stealth : forall v, C.stealth_va <> lift_address v;
    lift_value : val -> value
  }.

Definition regroup (vm:virtual_mapping) (p:MachIR.program) (Xh Xs:stealth_var) 
: Secure.L :=
  Make_L p vm.(lift_address) vm.(lift_address_not_stealth) vm.(lift_value) Xh Xs.

Theorem compiler_non_leakage : 
  forall prog_c prog_machir Xh Xs vm,
    compiler prog_c = OK prog_machir ->
    typecheck prog_machir Xh Xs = true ->
    let P := regroup vm prog_machir Xh Xs in
    forall s0 s0' Ls0 Ls0' trace trace' str str',
      valid_sm_state s0 ->
      valid_sm_state s0' ->
      low_var_eq P Ls0 Ls0' ->
      osi_stealth_equivalent s0 s0' osi_a ->
      get_trace P Ls0 trace -> get_trace P Ls0' trace' ->
      interleave s0 trace str ->
      interleave s0' trace' str' ->
      leak_osi_equivalent_traces osi_a str str'.
  Proof.
    Local Hint Resolve interleave_run.
    intros.
    apply Non_leakage; crush.
    eapply interleave_same_non_stealth_osi_a_actions
      with (s0:=s0) (s0':=s0'); crush.
    unfold get_trace in *.
    eapply transform_secure with (P:=P); eauto.
    erewrite interleave_hd; crush.
    erewrite interleave_hd; crush.
  Qed.

End END_TO_END.