Require Import Coqlib.
Require Import DLib.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Values.
Require Import Memory.
Require Import Globalenvs.
Require Import Events.
Require Import Smallstep.
Require Import Op.
Require Import Locations.
Require Import Conventions.
Require Import Mach.
Require Import Errors.


Notation "a ## b" := (List.map a b) (at level 1).

Definition node := positive.

Inductive instruction: Type :=
  | MIop: operation -> list mreg -> mreg -> node -> instruction
  | MIload: memory_chunk -> addressing -> list mreg -> mreg -> node -> instruction
  | MIstore: memory_chunk -> addressing -> list mreg -> mreg -> node -> instruction
  | MImemcpy_8_4: mreg -> mreg -> node -> instruction
  | MIgoto: node -> instruction
  | MIcond: condition -> list mreg -> node -> node -> instruction
  | MIreturn: instruction.

Definition code := PTree.t instruction.

Record function: Type := mkfunction
  { fn_sig: signature;
    fn_code: code;
    fn_entrypoint : node;
    fn_stacksize: Z;
    fn_link_ofs: int;
    fn_retaddr_ofs: int }.

Definition fundef := AST.fundef function.

Definition program := (function * Mach.program)%type.

Definition funsig (fd: fundef) :=
  match fd with
  | Internal f => fn_sig f
  | External ef => ef_sig ef
  end.

Definition successors (i: instruction) : list node :=
  match i with
  | MIop _ _ _ s 
  | MIload _ _ _ _ s 
  | MIstore _ _ _ _ s 
  | MImemcpy_8_4 _ _ s 
  | MIgoto s => s :: nil
  | MIcond _ _ ifso ifnot => ifso :: ifnot :: nil
  | MIreturn => nil
  end.

Section RELSEM.

Variable ge: genv.
Variable f: function.
Variable sp: val.

(** Mach execution states. *)

Inductive state: Type :=
  | State:
      forall (pc: node)
             (rs: regset)              (**r register state *)
             (m: mem),                 (**r memory state *)
      state.

Inductive step: state -> trace -> state -> Prop :=
  | exec_MIop:
      forall op args res rs m v pc pc',
      (fn_code f)!pc = Some (MIop op args res pc') ->
      eval_operation ge sp op rs##args m = Some v ->
      step (State pc rs m)
        E0 (State pc' (rs#res <- v) m)
  | exec_MIload:
      forall chunk addr args dst rs m a v pc pc',
      (fn_code f)!pc = Some (MIload chunk addr args dst pc') ->
      eval_addressing ge sp addr rs##args = Some a ->
      Mem.loadv chunk m a = Some v ->
      step (State pc rs m)
        E0 (State pc' (rs#dst <- v) m)
  | exec_MIstore:
      forall chunk addr args src rs m m' a pc pc',
      (fn_code f)!pc = Some (MIstore chunk addr args src pc') ->
      eval_addressing ge sp addr rs##args = Some a ->
      Mem.storev chunk m a (rs src) = Some m' ->
      step (State pc rs m)
        E0 (State pc' rs m')
  | exec_MIgoto:
      forall rs m pc pc',
      (fn_code f)!pc = Some (MIgoto pc') ->
      step (State pc rs m)
        E0 (State pc' rs m)
  | exec_MIcond_true:
      forall cond args rs m pc pc1 pc2,
      (fn_code f)!pc = Some (MIcond cond args pc1 pc2) ->
      eval_condition cond rs##args m = Some true ->
      step (State pc rs m)
        E0 (State pc1 rs m)
  | exec_MIcond_false:
      forall cond args rs m pc pc1 pc2,
      (fn_code f)!pc = Some (MIcond cond args pc1 pc2) ->
      eval_condition cond rs##args m = Some false ->
      step (State pc rs m)
        E0 (State pc2 rs m)
  | exec_MIreturn:
      forall rs m pc,
      (fn_code f)!pc = Some MIreturn ->
      step (State pc rs m)
        E0 (State pc rs m).

End RELSEM.


Definition find_pc_label (lbl: label) (f:Mach.function) : option node :=
  match find_label lbl (Mach.fn_code f) with
      | None => None
      | Some c => Some (Ppred (Pos.of_succ_nat (length c)))
  end.

Definition transl_instr (ge:Mach.genv) (f: Mach.function) (pc:node) (i: Mach.instruction) :=
  let pc' := Ppred pc in
  match i with
  | Mgetstack ofs ty dst => Some (MIload (chunk_of_type ty) (Ainstack ofs) nil dst pc')
  | Msetstack src ofs ty => Some (MIstore (chunk_of_type ty) (Ainstack ofs) nil src pc')
  | Mgetparam ofs ty dst => None
  | Mop op args res => Some (MIop op args res pc')
  | Mload chunk addr args dst => Some (MIload chunk addr args dst pc')
  | Mstore chunk addr args src => Some (MIstore chunk addr args src pc')
  | Mcall sig _ => None
  | Mtailcall sig _ => None
  | Mlabel lbl => Some (MIgoto pc')
  | Mgoto lbl =>
    match find_pc_label lbl f with
      | None => None
      | Some pc_lbl => Some (MIgoto pc_lbl)
    end
  | Mcond cond args lbl =>
    match find_pc_label lbl f with
      | None => None
      | Some pc_lbl => Some (MIcond cond args pc_lbl pc')
    end
  | Mjumptable arg tbl => None
  | Mreturn => Some MIreturn
  | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) =>
    Some (MImemcpy_8_4 dst src pc')
  | Mbuiltin _ _ _ => None
  | Mannot ef args => None
  end.

Fixpoint transl_fun_aux ge (f: Mach.function) (c:Mach.code) 
: option (node * code) :=
  match c with
    | nil => Some (xH, PTree.empty _)
    | i::q =>
      match transl_fun_aux ge f q with
        | None => None
        | Some (pc,c') =>
          match transl_instr ge f pc i with
            | None => None
            | Some i' => Some (Psucc pc, PTree.set pc i' c')
          end
      end
  end.

Lemma transl_fun_aux_length : forall ge f c pc c',
  transl_fun_aux ge f c = Some (pc,c') ->
  pc = Pos.of_succ_nat (length c).
Proof.
  induction c; simpl; intros.
  - inv H; go.
  - flatten H.
    rewrite <- (IHc _ _ (refl_equal _)).
    go.
Qed.
  
Definition transl_fun ge (f: Mach.function) : option (node*code) :=
  transl_fun_aux ge f f.(Mach.fn_code) .

Definition checker (ge:Mach.genv) (f:Mach.function) : bool :=
  List.forallb
    (fun ins =>
       match ins with
         | Mach.Mgetparam _ _ _
         | Mach.Mcall _ (inl _)
         | Mach.Mtailcall _ _
         | Mach.Mjumptable _ _
         | Mach.Mannot _ _ => false
         | Mach.Mgoto lbl 
         | Mach.Mcond _ _ lbl =>
           match find_label lbl f.(Mach.fn_code) with
             | Some _ => true
             | None => false
           end
         | Mach.Mcall _ _ => false
         | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) => true
         | Mach.Mbuiltin _ _ _ => false
         | _ => true
       end)
    f.(Mach.fn_code) &&
  (match f.(Mach.fn_code) with
       | nil => false | _ => true end).


Definition transl_function ge (f: Mach.function) : res function :=
  if checker ge f then
    match transl_fun ge f with
      | None => Errors.Error (Errors.MSG "MachIR conversion failed" :: nil)
      | Some (pc,c) =>
        Errors.OK (mkfunction
                     f.(Mach.fn_sig)
                     c
                     (Ppred pc)
                     f.(Mach.fn_stacksize)
                     f.(Mach.fn_link_ofs)
                     f.(Mach.fn_retaddr_ofs))
    end
  else Errors.Error (Errors.MSG "MachIR conversion failed : checker" :: nil).

Definition transl_program (p: Mach.program) : res program :=
  let ge := Genv.globalenv p in
  match Genv.find_symbol ge p.(prog_main) with
    | Some fb =>
      match Genv.find_funct_ptr ge fb with
        | Some (Internal f) => 
          match transl_function ge f with
            | OK tf => OK (tf, p)
            | Error e => Error e
          end
        | _ => Errors.Error (Errors.MSG "MachIR gen failed : no main in ge" :: nil)
      end
    | _ => Errors.Error (Errors.MSG "MachIR gen failed : no main" :: nil)
  end.



(** Proofs *)

Lemma nth_error_some_length: forall A l n (a:A),
    List.nth_error l n = Some a -> (n < length l)%nat.
Proof.
  induction l; destruct n; simpl; intros; try (inv H; fail).
  - omega.
  - apply IHl in H; omega.
Qed.

Lemma nth_error_none_length: forall A (l:list A) n,
    List.nth_error l n = None -> (n >= length l)%nat.
Proof.
  induction l; destruct n; simpl; intros; try (inv H; fail); try omega.
  apply IHl in H; omega.
Qed.

Lemma nth_error_length: forall A l1 l2 (a:A),
    List.nth_error (l1++a::l2) (length l1) = Some a.
Proof.
  induction l1; simpl; auto.
Qed.

Lemma transl_instr_some: forall ge pc0 f pc i,
  transl_instr ge f pc0 i <> None ->
  transl_instr ge f pc i <> None.
Proof.
  destruct i; simpl; flatten; try congruence.
Qed.

Lemma transl_fun_aux_some : forall ge f c pc0 cod,
  transl_fun_aux ge f c = Some (pc0,cod) ->        
  forall i pc, In i c -> transl_instr ge f pc i <> None.
Proof.
  induction c; simpl; intros; go.
  flatten H.
  destruct H0; go.
  subst; apply transl_instr_some with n; congruence.
Qed.

Lemma transl_fun_some : forall ge f pc cod,
  transl_fun ge f = Some (pc,cod) ->        
  forall i pc, In i (Mach.fn_code f) -> transl_instr ge f pc i <> None.
Proof.
  unfold transl_fun; intros.
  eapply transl_fun_aux_some; eauto.
Qed.

Lemma transl_fun_aux_correct : forall ge f c pc0 cod,
  transl_fun_aux ge f c = Some (pc0,cod) ->        
  pc0 = Pos.of_succ_nat (length c) /\
  forall c0 pc,
    Mach.fn_code f = c0 ++ c ->
    if Z_lt_dec (Zpos pc) (Zpos (Pos.of_succ_nat (length c))) then
      PTree.get pc cod = 
      match List.nth_error (Mach.fn_code f)   ((length (Mach.fn_code f)) - 
                                               (Pos.to_nat pc)) with
        | Some i' => transl_instr ge f pc i'
        | None => None
      end
    else 
      PTree.get pc cod = None.
Proof.
  induction c; intros; simpl transl_fun_aux in *; split; auto.
  - inv H; simpl; auto.
  - inv H; intros; flatten.
    + simpl in l. zify. omega.
    + simpl in l. zify. omega.
    + rewrite PTree.gempty; auto.
  - flatten H.
    destruct (IHc _ _ (refl_equal _)) as [IH1 IH2]; clear IHc.
    subst. go.
  - flatten H.
    destruct (IHc _ _ (refl_equal _)) as [IH1 IH2]; clear IHc.
    rewrite PTree.gsspec; destruct peq.
    * subst.
      { destruct Z_lt_dec.
        - rewrite H2.
          rewrite SuccNat2Pos.id_succ.
          rewrite app_length; simpl.
          replace (length c1 + Datatypes.S (length c) - Datatypes.S (length c))%nat
          with (length c1)
            by omega.
          rewrite nth_error_length; auto.
        - elim n; simpl; zify; omega. }
    * { exploit (IH2 (c1++a::nil) pc); clear IH2.
        - rewrite app_ass; auto.
        - destruct Z_lt_dec; destruct Z_lt_dec; go.
          + simpl in n1. elim n1.
            zify; omega.
          + simpl in *.
            elim n1.
            assert (Z.pos pc < (Z.pos (Pos.of_succ_nat (length c)))
                    \/ Z.pos pc = (Z.pos (Pos.of_succ_nat (length c))))
              by (zify; omega).
            destruct H; auto.
            go. }
Qed.

Lemma transl_fun_correct : forall ge f pc cod,
  transl_fun ge f = Some (pc,cod) ->        
  forall pc,
    (Pos.to_nat pc <= length (Mach.fn_code f))%nat ->
    PTree.get pc cod = 
      match List.nth_error (Mach.fn_code f) (length (Mach.fn_code f) - Pos.to_nat pc) with
        | Some i' => transl_instr ge f pc i'
        | None => None
      end.
Proof.
  unfold transl_fun; intros.
  destruct (transl_fun_aux_correct _ _ _ _ _ H) as [_ IH].
  exploit (IH nil pc0); eauto.
  flatten.
  elim n; simpl; flatten.
  zify; omega.
Qed.

Section checker_correct.

Let checker_aux (ge:Mach.genv) (l:list Mach.instruction) : bool :=
  List.forallb
    (fun ins =>
       match ins with
         | Mach.Mgetparam _ _ _
         | Mach.Mcall _ (inl _)
         | Mach.Mtailcall _ _
         | Mach.Mjumptable _ _
         | Mach.Mannot _ _ => false
         | Mach.Mgoto lbl 
         | Mach.Mcond _ _ lbl =>
           match find_label lbl l with
             | Some _ => true
             | None => false
           end
         | Mach.Mcall _ _ => false
         | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) => true
         | Mach.Mbuiltin _ _ _ => false
         | _ => true
       end) l.

Lemma checker_Hcode: forall ge l,
  checker_aux ge l = true ->
  forall ins, 
    In ins l ->
    match ins with
        | Mach.Mgetparam _ _ _
        | Mach.Mcall _ (inl _)
        | Mach.Mtailcall _ _
        | Mach.Mjumptable _ _
        | Mach.Mannot _ _ => False
        | Mach.Mcall _ _ => False
        | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) => True
        | Mach.Mbuiltin _ _ _ => False
        | _ => True
    end.
Proof.
  unfold checker_aux.
  intros.
  rewrite List.forallb_forall in H.
  generalize (H _ H0).
  destruct ins; go.
  flatten; auto.
  flatten; auto.
Qed.

Lemma Hcode_lbl_wf: forall ge l,
  checker_aux ge l = true ->
  forall ins, 
    In ins l ->
    match ins with
        | Mach.Mgoto lbl 
        | Mach.Mcond _ _ lbl =>
          find_label lbl l <> None
        | _ => True
    end.
Proof.
  unfold checker_aux.
  intros.
  rewrite List.forallb_forall in H.
  generalize (H _ H0).
  destruct ins; go.
  flatten; auto.
  flatten; auto.
Qed.

End checker_correct.

Section CORRECTNESS.

Variable tf: function.

Variable ge : Mach.genv.

Variable f : Mach.function.
Variable sp : block.

Variable rao: Mach.function -> Mach.code -> int -> Prop.

Definition match_regset (rs1 rs2:regset) :=
  forall r, Val.lessdef (rs1 r) (rs2 r).

Lemma match_regset_set : forall rs rs' dst v v',
   match_regset rs rs' ->
   Val.lessdef v v' ->
   match_regset (rs # dst <- v) (rs' # dst <- v').
Proof.
  unfold match_regset; intros.
  unfold Regmap.set.
  flatten; go.
Qed.

Lemma match_regset_undef_regs : forall l rs rs',
  match_regset rs rs' ->
  match_regset (undef_regs l rs) rs'.
Proof.
  induction l; simpl; intros; auto.
  intros r; unfold Regmap.set; flatten; auto.
  apply IHl; auto.  
Qed.

Lemma match_regset_list : forall l rs rs',
  match_regset rs rs' ->                            
  Val.lessdef_list (rs ## l) (rs' ## l).
Proof.
  induction l; auto.
  intros; simpl; constructor; auto.
Qed.

Hint Resolve match_regset_set match_regset_undef_regs match_regset_list.

Inductive match_states : Mach.state -> state -> Prop :=
  | match_state:
      forall fb c rs rs' m m' c0 pc
        (GEN: transl_function ge f = OK tf)
        (SS: fn_stacksize tf = Mach.fn_stacksize f)
        (LO: fn_link_ofs tf = Mach.fn_link_ofs f)
        (GF: Genv.find_funct_ptr ge fb = Some (Internal f))
        (CODE: Mach.fn_code f = c0 ++ c)
        (PC: Pos.to_nat pc = length c \/ (c=nil /\ pc=xH))
        (REGS: match_regset rs rs')
        (MEM: Mem.extends m m'),
      match_states (Mach.State nil fb (Vptr sp Int.zero) c rs m)
                   (State pc rs' m')
  | match_state_return: forall rs m pc' rs' m',
      match_states (Mach.Returnstate nil rs m) (State pc' rs' m').


Lemma find_label_app:
  forall lbl c c', find_label lbl c = Some c' -> exists c'', c = c'' ++ c'.
Proof.
  induction c; simpl; intros; go.
  flatten H.
  - exists (a::nil); auto.
  - edestruct IHc; eauto; subst.
    exists (a::x); auto.
Qed.

Lemma find_pc_label_find_label : forall lbl f n c',
  find_pc_label lbl f = Some n ->
  find_label lbl (Mach.fn_code f) = Some c' ->
  exists c0, Mach.fn_code f = c0++c' /\ 
             (Pos.to_nat n = length c' \/ c' = nil /\ n = 1%positive).
Proof.
  intros.
  unfold find_pc_label in *.
  flatten H.
  apply find_label_app in Eq.
  destruct Eq as (c'' & C1).
  exists c''; split; go.
  destruct c'; go.
  left.
  rewrite Pos2Nat.inj_pred.
  - rewrite SuccNat2Pos.id_succ; auto.
  - simpl. zify. go.
Qed.

Hypothesis Hcode:
  forall ins, 
    In ins (Mach.fn_code f) ->
    match ins with
        | Mach.Mgetparam _ _ _
        | Mach.Mcall _ (inl _)
        | Mach.Mtailcall _ _
        | Mach.Mjumptable _ _
        | Mach.Mannot _ _ => False
        | Mach.Mcall _ _ => False
        | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) => True
        | Mach.Mbuiltin _ _ _ => False
        | _ => True
    end.

Hypothesis Hcode_lbl_wf:
  forall ins, 
    In ins (Mach.fn_code f) ->
    match ins with
        | Mach.Mgoto lbl 
        | Mach.Mcond _ _ lbl =>
          find_label lbl f.(Mach.fn_code) <> None
        | _ => True
    end.


Lemma In_app_cons {A} (a:A) l1 l2 : In a (l1++a::l2).
Proof.
  rewrite in_app; simpl; auto.
Qed.

Lemma ineq1: forall c0 (i:Mach.instruction) c1,
  (length (i::c1) <= length (c0 ++ i :: c1))%nat.
Proof.
  intros.
  rewrite app_length; simpl; omega.
Qed.

Lemma eq1: forall c0 (i:Mach.instruction) c1,
  (length (c0 ++ i :: c1) - (length (i::c1)))%nat = length c0.
Proof.
  intros.
  rewrite app_length; simpl; omega.
Qed.

Lemma pc_step : forall pc (i:Mach.instruction) c,
  Pos.to_nat pc = length (i :: c) ->
  Pos.to_nat (Pos.pred pc) = length c \/
   c = nil /\ Pos.pred pc = 1%positive.
Proof.
  intros.
  generalize (Pos2SuccNat.id_succ pc); rewrite H.
  destruct c; simpl; intros.
  - assert (Ppred (Psucc pc) = 1%positive).
      rewrite <- H0; go.
    rewrite Pos.pred_succ in H1.  
    subst; simpl; intuition.
  - rewrite <- (Pos.pred_succ pc).
    rewrite <- H0.
    rewrite Pos.pred_succ.
    rewrite Pos.pred_succ.
    rewrite SuccNat2Pos.id_succ; auto.
Qed.
Hint Resolve pc_step.

Lemma transl_function_OK : forall f tf,
  transl_function ge f = OK tf ->
  exists pc0, transl_fun ge f = Some (pc0,fn_code tf).
Proof.
  intros f0 tf0.
  unfold transl_function.
  intros H; flatten H.
  simpl.
  eauto.
Qed.

Theorem transl_step_correct:
  forall S1 t S2, Mach.step rao ge S1 t S2 ->
  forall R1, match_states S1 R1 ->
  exists R2,
    step ge tf (Vptr sp Int.zero) R1 t R2 /\ match_states S2 R2.
Proof.
  intros S1 t S2 H R1 Hm.
  inv Hm; inv H;
  destruct (transl_function_OK _ _ GEN) as [pc3 GEN1];
  try (
     rewrite CODE in Hcode;
     generalize (Hcode _ (In_app_cons _ _ _));
     intuition; fail);

  match goal with
    | id: Mach.fn_code _ = ?l1 ++ ?a :: ?l2 |- _ => 
      assert (IS:(transl_instr ge f pc a) <> None); [try (simpl; congruence)|idtac]
  end;
  generalize (transl_fun_correct _ _ _ _ GEN1 pc);
  (destruct PC as [PC|PC]; [idtac|destruct PC;congruence;fail]);
    rewrite CODE; rewrite PC;
  intros IH; generalize (IH (ineq1 _ _ _)); clear IH;
  rewrite eq1;
  rewrite nth_error_length;
  intros INSTR; simpl in INSTR;
    match goal with
      | id: Mach.fn_code _ = ?l1 ++ ?a :: ?l2 |- _ =>
        replace (l1++a::l2) with ((l1++a::nil)++l2) in id; 
          [idtac| rewrite <- app_assoc; reflexivity]
    end;
    try (elim IS; reflexivity);

  try (econstructor; split; [
         econstructor (eauto;
                       fail) |
         econstructor; eauto]).

  - exploit Mem.loadv_extends; eauto; intros (v' & V1 & V2).
    econstructor; split.
    + eapply exec_MIload; eauto; simpl; auto.
    + econstructor; eauto.

  - exploit Mem.storev_extends; eauto; intros (m1' & M1 & M2).
    econstructor; split.
    + eapply exec_MIstore; eauto; simpl; auto.
    + econstructor; eauto.
     
  - edestruct eval_operation_lessdef with (3:=H8) as (v' & V1 & V2); eauto.
    econstructor; split.
    + eapply exec_MIop; eauto; simpl; eauto.
    + econstructor; eauto.

  - edestruct eval_addressing_lessdef with (2:=H8) as (v' & V1 & V2); eauto.
    exploit Mem.loadv_extends; eauto; intros (v'' & V1' & V2').
    econstructor; split.
    + eapply exec_MIload; eauto; simpl; auto.
    + econstructor; eauto.

  - edestruct eval_addressing_lessdef with (2:=H8) as (v' & V1 & V2); eauto.
    exploit Mem.storev_extends; eauto; intros (v'' & V1' & V2').
    econstructor; split.
    + eapply exec_MIstore; eauto; simpl; auto.
    + econstructor; eauto.

  - rewrite CODE in Hcode.
    rewrite app_ass in Hcode; simpl in Hcode.
    generalize (Hcode _ (In_app_cons _ _ _)).
    flatten; intuition.

  - admit. (* memcpy *)
  - admit. (* memcpy *)

  - simpl; flatten.
    assert (f0=f) by congruence; subst.
    unfold find_pc_label in Eq; flatten Eq.

  - assert (f0=f) by congruence; subst.
    flatten INSTR.
    + econstructor; split; go.
      edestruct find_pc_label_find_label as (c2 & C1 & C2); eauto.
      econstructor; eauto.
    + unfold find_pc_label in Eq.
      flatten Eq.

  - simpl; flatten.
    assert (f0=f) by congruence; subst.
    unfold find_pc_label in Eq; flatten Eq.

  - assert (f0=f) by congruence; subst.
    flatten INSTR.
    + econstructor; split.
      * eapply exec_MIcond_true; eauto.
        eapply eval_condition_lessdef; eauto.
      * edestruct find_pc_label_find_label as (c2 & C1 & C2); eauto.
        econstructor; eauto.
    + unfold find_pc_label in Eq.
      flatten Eq.

  - rewrite CODE in Hcode_lbl_wf.
    rewrite app_ass in Hcode_lbl_wf; simpl in Hcode_lbl_wf.
    simpl; flatten.
    generalize (Hcode_lbl_wf _ (In_app_cons _ _ _)).
    unfold find_pc_label in Eq.
    rewrite CODE in Eq.
    rewrite app_ass in Eq; simpl in Eq.
    flatten Eq.

  - flatten INSTR.
    + econstructor; split.
      * eapply exec_MIcond_false; eauto.
        eapply eval_condition_lessdef; eauto.
      * econstructor; eauto.
    + generalize (transl_fun_correct _ _ _ _ GEN1 pc).
      rewrite INSTR.
      rewrite CODE.
      rewrite PC.
      rewrite app_ass; simpl app.      
      intros T.
      generalize (T (ineq1 _ _ _)); clear T.
      rewrite eq1.
      rewrite nth_error_length.
      congruence.
Qed.

Inductive extended_state :=
 | NState (f:function) (sp:val) (s:state)
 | IState (f:function) (m:mem).

Inductive mevent :=
 | MRead (addr:val)
 | MWrite (addr:val)
 | MSilent.

Inductive get_event f sp: state -> mevent -> Prop :=
  | get_event_MIop:
      forall op args res rs m pc pc',
      (fn_code f)!pc = Some (MIop op args res pc') ->
      get_event f sp (State pc rs m) MSilent
  | get_event_MIload:
      forall chunk addr args dst rs m pc pc' b ofs,
      (fn_code f)!pc = Some (MIload chunk addr args dst pc') ->
      eval_addressing ge sp addr rs##args = Some (Vptr b ofs) ->
      get_event f sp (State pc rs m) (MRead (Vptr b ofs))
  | get_event_MIstore:
      forall chunk addr args src rs m pc pc' b ofs,
      (fn_code f)!pc = Some (MIstore chunk addr args src pc') ->
      eval_addressing ge sp addr rs##args = Some (Vptr b ofs) ->
      get_event f sp (State pc rs m) (MWrite (Vptr b ofs))
  | get_event_MIgoto:
      forall rs m pc pc',
      (fn_code f)!pc = Some (MIgoto pc') ->
      get_event f sp (State pc rs m) MSilent
  | get_event_MIcond:
      forall cond args rs m pc pc1 pc2,
      (fn_code f)!pc = Some (MIcond cond args pc1 pc2) ->
      get_event f sp (State pc rs m) MSilent
  | get_event_MIreturn:
      forall rs m pc,
      (fn_code f)!pc = Some MIreturn ->
      get_event f sp (State pc rs m) MSilent.

Inductive estep :  extended_state -> mevent -> extended_state -> Prop :=
  | estep_start : forall tf m0 m1 m2 m3 sp stk
    (SP:Mem.alloc m0 0 tf.(fn_stacksize) = (m1, stk))
    (SP_EQ:sp = Vptr stk Int.zero)
    (STORE1:store_stack m1 sp Tint (fn_link_ofs tf) (parent_sp nil) = Some m2)
    (STORE2:store_stack m2 sp Tint (fn_retaddr_ofs tf) (parent_ra nil) = Some m3),
    estep (IState tf m0) MSilent 
          (NState tf sp (State tf.(fn_entrypoint) (Regmap.init Vundef) m3))
  | estep_normal : forall f sp s1 e ev s2
    (EVENT: get_event f sp s1 e)                      
    (STEP: step ge f sp s1 ev s2),
    estep (NState f sp s1) e (NState f sp s2).

Inductive get_event_mach (ge:genv): Mach.state -> mevent -> Prop :=
  | get_event_mach_Mlabel:
      forall s f sp lbl c rs m,
      get_event_mach ge (Mach.State s f sp (Mlabel lbl :: c) rs m) MSilent
  | get_event_mach_Mgetstack:
      forall s f sp ofs ty dst c rs m,
      get_event_mach ge (Mach.State s f sp (Mgetstack ofs ty dst :: c) rs m)
                     (MRead (Val.add sp (Vint ofs)))
  | get_event_mach_Msetstack:
      forall s f sp src ofs ty c rs m ,
      get_event_mach ge (Mach.State s f sp (Msetstack src ofs ty :: c) rs m)
                     (MWrite (Val.add sp (Vint ofs)))
  | get_event_mach_Mgetparam:
      forall s fb sp ofs ty dst c rs m,
      get_event_mach ge (Mach.State s fb sp (Mgetparam ofs ty dst :: c) rs m) MSilent
  | get_event_mach_Mop:
      forall s f sp op args res c rs m,
      get_event_mach ge (Mach.State s f sp (Mop op args res :: c) rs m) MSilent
  | get_event_mach_Mload:
      forall s f sp chunk addr args dst c rs m b ofs,
      eval_addressing ge sp addr rs##args = Some (Vptr b ofs) ->
      get_event_mach ge (Mach.State s f sp (Mload chunk addr args dst :: c) rs m)
                     (MRead (Vptr b ofs))
  | get_event_mach_Mstore:
      forall s f sp chunk addr args src c rs m b ofs,
      eval_addressing ge sp addr rs##args = Some (Vptr b ofs) ->
      get_event_mach ge (Mach.State s f sp (Mstore chunk addr args src :: c) rs m)
                     (MWrite (Vptr b ofs))
  | get_event_mach_Mcall:
      forall s fb sp sig ros c rs m,
      get_event_mach ge (Mach.State s fb sp (Mcall sig ros :: c) rs m) MSilent
  | get_event_mach_Mtailcall:
      forall s fb stk soff sig ros c rs m,
      get_event_mach ge (Mach.State s fb (Vptr stk soff) (Mtailcall sig ros :: c) rs m) MSilent
  | get_event_mach_Mbuiltin:
      forall s f sp rs m ef args res b,
      get_event_mach ge (Mach.State s f sp (Mbuiltin ef args res :: b) rs m) MSilent
  | get_event_mach_Mannot:
      forall s f sp rs m ef args b,
      get_event_mach ge (Mach.State s f sp (Mannot ef args :: b) rs m) MSilent
  | get_event_mach_Mgoto:
      forall s fb sp lbl c rs m,
      get_event_mach ge (Mach.State s fb sp (Mgoto lbl :: c) rs m) MSilent
  | get_event_mach_Mcond:
      forall s fb sp cond args lbl c rs m,
      get_event_mach ge (Mach.State s fb sp (Mcond cond args lbl :: c) rs m) MSilent
  | get_event_mach_Mjumptable:
      forall s fb sp arg tbl c rs m,
      get_event_mach ge (Mach.State s fb sp (Mjumptable arg tbl :: c) rs m) MSilent
  | get_event_mach_Mreturn:
      forall s fb stk soff c rs m,
      get_event_mach ge (Mach.State s fb (Vptr stk soff) (Mreturn :: c) rs m) MSilent
  | get_event_mach_function:
      forall s fb rs m,
      get_event_mach ge (Callstate s fb rs m) MSilent
  | get_event_mach_return:
      forall s rs m,
      get_event_mach ge (Returnstate s rs m) MSilent.

Inductive estep_mach :  Mach.state -> mevent -> Mach.state -> Prop :=
  | estep_mach_def : forall s1 e ev s2
    (EVENT: get_event_mach ge s1 e)                      
    (STEP: Mach.step rao ge s1 ev s2),
    estep_mach s1 e s2.

End CORRECTNESS.

Inductive initial_state (p: program): extended_state -> Prop :=
  | initial_state_intro: forall b m0,
      let ge := Genv.globalenv (snd p) in
      Genv.init_mem (snd p) = Some m0 ->
      Genv.find_symbol ge (snd p).(prog_main) = Some b ->
      initial_state p (IState (fst p) m0).

CoInductive mtraceinf :=
  | Mconsinf: mevent -> mtraceinf -> mtraceinf.

Definition make_genv (p:program) := Genv.globalenv (snd p).

CoInductive get_trace (p:program) : extended_state -> mtraceinf -> Prop :=
  | get_trace_def s1 e s2 tr
    (STEP: estep (make_genv p) s1 e s2)
    (TR: get_trace p s2 tr):
    get_trace p s1 (Mconsinf e tr).

CoInductive get_trace_mach (p:Mach.program) rao : Mach.state -> mtraceinf -> Prop :=
  | get_trace_mach_def s1 e s2 tr
    (STEP: estep_mach (Genv.globalenv p) rao s1 e s2)
    (TR: get_trace_mach p rao s2 tr):
    get_trace_mach p rao s1 (Mconsinf e tr).


Inductive match_estates (p:Mach.program) : Mach.state -> extended_state -> Prop :=
 | match_estates_nstate: forall tf f sp s es,
   match_states tf (Genv.globalenv p) f sp s es ->
   match_estates p s (NState tf (Vptr sp Int.zero) es)
 | match_estates_istate: forall b f tf m
    (GF: Genv.find_funct_ptr (Genv.globalenv p) b = Some (Internal f))
    (TRANS:transl_function (Genv.globalenv p) f = OK tf),
   match_estates p (Callstate nil b (Regmap.init Vundef) m) (IState tf m).

Lemma transl_fun_check1 : forall ge f tf,
  transl_function ge f = OK tf ->
  forall ins, 
    In ins (Mach.fn_code f) ->
    match ins with
        | Mach.Mgetparam _ _ _
        | Mach.Mcall _ (inl _)
        | Mach.Mtailcall _ _
        | Mach.Mjumptable _ _
        | Mach.Mannot _ _ => False
        | Mach.Mcall _ _ => False
        | Mbuiltin (EF_memcpy 8%Z 4%Z) (dst::src::nil) (_::nil) => True
        | Mach.Mbuiltin _ _ _ => False
        | _ => True
    end.
Proof.
  unfold transl_function; intros.
  flatten H.
  unfold checker in *.
  rewrite andb_true_iff in Eq; destruct Eq.
  eapply checker_Hcode; eauto.
Qed.

Lemma transl_fun_check2 : forall ge f tf,
  transl_function ge f = OK tf ->
  forall ins, 
    In ins (Mach.fn_code f) ->
    match ins with
        | Mach.Mgoto lbl 
        | Mach.Mcond _ _ lbl =>
          find_label lbl f.(Mach.fn_code) <> None
        | _ => True
    end.
Proof.
  unfold transl_function; intros.
  flatten H.
  unfold checker in *.
  rewrite andb_true_iff in Eq; destruct Eq.
  eapply Hcode_lbl_wf; eauto.
Qed.

Lemma match_states_get_event: forall p tf f sp s es e rao t s',
  match_states tf (Genv.globalenv p) f sp s es ->
  Mach.step rao (Genv.globalenv p) s t s' ->
  get_event_mach (Genv.globalenv p) s e ->
  get_event (Genv.globalenv p) tf (Vptr sp Int.zero) es e.
Proof.
  intros p tf f sp s es e rao t s' H Hstep H0.
  inv H; inv H0;
  try (destruct (transl_function_OK _ _ _ GEN) as [pc3 GEN1];
    generalize (transl_fun_correct _ _ _ _ GEN1 pc);
    (destruct PC as [PC|PC]; [idtac|destruct PC;congruence;fail]);
    rewrite CODE; rewrite PC;
    intros IH; generalize (IH (ineq1 _ _ _)); clear IH;
    rewrite eq1;
    rewrite nth_error_length;
    simpl in *;
    intros; go);
  try (
      generalize (transl_fun_check1 _ _ _ GEN); rewrite CODE;
      intros T; elim (T _ (In_app_cons _ _ _)); fail).

   edestruct eval_addressing_lessdef with (2:=H7) as (v' & V1 & V2); eauto.
   eapply match_regset_list; eauto.
   eapply get_event_MIload; eauto.
   inv V2; auto.

   edestruct eval_addressing_lessdef with (2:=H7) as (v' & V1 & V2); eauto.
   eapply match_regset_list; eauto.
   eapply get_event_MIstore; eauto.
   inv V2; auto.

   flatten H;
   try (generalize (transl_fun_check1 _ _ _ GEN); rewrite CODE;
      intros T; generalize (T _ (In_app_cons _ _ _)); flatten;
      intuition; fail).

   admit. (* memcpy *)

   generalize (transl_fun_check2 _ _ _ GEN); rewrite CODE;
      intros T; generalize (T _ (In_app_cons _ _ _)); clear T; intros T.
   unfold find_pc_label in H.   
   rewrite <- CODE in *.
   flatten H.
   go.

   generalize (transl_fun_check2 _ _ _ GEN); rewrite CODE;
      intros T; generalize (T _ (In_app_cons _ _ _)); clear T; intros T.
   unfold find_pc_label in H.   
   rewrite <- CODE in *.
   flatten H.
   go.

   inv Hstep.
Qed.

Lemma preserves_trace: forall p tf rao s tr,
  get_trace_mach p rao s tr -> forall es,
  match_estates p s es ->
  get_trace (tf,p) es tr.
Proof.
  intros p tf rao.
  cofix COH.
  intros s tr H es Hm.
  inv H.
  assert (exists ss2, estep (Genv.globalenv p) es e ss2 /\ match_estates p s2 ss2).
    inv Hm.
    - inv STEP.
      exploit transl_step_correct; eauto.
      * eapply transl_fun_check1; eauto.
        inv H; eauto; inv STEP0.
      * eapply transl_fun_check2; eauto.
        inv H; eauto; inv STEP0.
      * intros (ss2 & T1 & T2).
        econstructor; split; econstructor; eauto.
        eapply match_states_get_event; eauto.
    - inv STEP.
      inv EVENT.
      inv STEP0.
      * assert (f0=f) by congruence; subst.
        generalize TRANS; intros TRANS2.
        unfold transl_function in TRANS; flatten TRANS.
        econstructor; split.
        + econstructor; simpl; eauto.
        + econstructor.
          econstructor 1 with (c0:=nil); go.
          simpl.
          left.
          unfold transl_fun in *.
          exploit transl_fun_aux_length; eauto.
          intros; subst.
          rewrite Pos.pred_of_succ_nat.
          rewrite Nat2Pos.id; auto.
          unfold checker in *.
          rewrite andb_true_iff in Eq; destruct Eq.
          flatten H0; go.
          apply match_regset_undef_regs.    
          intro; auto.
          apply Mem.extends_refl.
      * congruence.
    -  destruct H as (ss1 & T1 & T2).
       constructor 1 with ss1; auto.
       apply COH with s2; auto.
Qed.                            

Lemma match_initial : forall p tp s,
  transl_program p = OK tp ->                        
  Mach.initial_state p s ->
  exists es, initial_state tp es /\
             match_estates p s es.
Proof.
  intros p tp s H H0.
  inv H0.
  unfold transl_program in H; flatten H.
  econstructor; split.
  - econstructor; simpl; eauto.
  - econstructor; eauto.
    unfold ge in *.
    assert (b=fb) by congruence; subst.
    auto.
Qed.

Definition mach_program_get_trace rao (p:Mach.program) (tr:mtraceinf) : Prop :=
  exists s,
    Mach.initial_state p s /\ get_trace_mach p rao s tr.

Definition program_get_trace (p:program) (tr:mtraceinf) : Prop :=
  exists s,
    initial_state p s /\ get_trace p s tr.


Lemma transl_program_correct : forall rao p tp tr,
  transl_program p = OK tp ->                        
  mach_program_get_trace rao p tr ->
  program_get_trace tp tr.
Proof.
  intros rao p tp tr H (s & T1 & T2).
  edestruct match_initial as (es & S1 & S2); eauto.
  exists es; split; auto.
  unfold transl_program in H; flatten H.
  eapply preserves_trace; eauto.  
Qed.

