Require Import Coqlib.
Require Import Maps.
Require Import Errors.
Require Import AST.
Require Import Integers.
Require Import Floats.
Require Import Op.
Require Import Registers.
Require Import Lattice.
Require Import Kildall.
Require Import MachIR.
Require Import VarSets.
Require Import Machregs.
Require Import Conventions1.
Require Import DLib.

Inductive pt :=
  | Unknown
  | Symb (s:pointsto_symbol)
  | NoPtr.

Module Approx <: SEMILATTICE_WITH_TOP.
  Definition t := pt.
  Definition eq (x y: t) := (x = y).
  Definition eq_refl: forall x, eq x x := (@refl_equal t).
  Definition eq_sym: forall x y, eq x y -> eq y x := (@sym_equal t).
  Definition eq_trans: forall x y z, eq x y -> eq y z -> eq x z := (@trans_equal t).
  Lemma eq_dec: forall (x y: t), {x=y} + {x<>y}.
  Proof.
    repeat decide equality.
  Defined.
  Definition beq (x y: t) := if eq_dec x y then true else false.
  Lemma beq_correct: forall x y, beq x y = true -> x = y.
  Proof.
    unfold beq; intros.  destruct (eq_dec x y). auto. congruence.
  Qed.

  Definition ge (x y: t) : Prop := x = Unknown \/ y = NoPtr \/ x = y.

  Lemma ge_refl: forall x y, eq x y -> ge x y.
  Proof.
    unfold eq, ge; tauto.
  Qed.
  Lemma ge_trans: forall x y z, ge x y -> ge y z -> ge x z.
  Proof.
    unfold ge; intuition congruence.
  Qed.
  Lemma ge_compat: forall x x' y y', eq x x' -> eq y y' -> ge x y -> ge x' y'.
  Proof.
    unfold eq, ge; intros; congruence.
  Qed.
  Definition bot := NoPtr.
  Definition top := Unknown.
  Lemma ge_bot: forall x, ge x bot.
  Proof.
    unfold ge, bot; tauto.
  Qed.
  Lemma ge_top: forall x, ge top x.
  Proof.
    unfold ge, bot; tauto.
  Qed.
  Definition lub (x y: t) : t :=
    if eq_dec x y then x else
    match x, y with
    | NoPtr, _ => y
    | _, NoPtr => x
    | _, _ => Unknown
    end.
  Lemma ge_lub_left: forall x y, ge (lub x y) x.
  Proof.
    unfold lub; intros.
    flatten; unfold ge; 
    intuition congruence.
  Qed.
  Lemma ge_lub_right: forall x y, ge (lub x y) y.
  Proof.
    unfold lub; intros.
    flatten; unfold ge; 
    intuition congruence.
  Qed.

  Lemma least : forall x y z,
                  ge z x -> ge z y -> ge z (lub x y).
  Proof.
    unfold ge.
    destruct x; destruct y; destruct z; unfold lub; try tauto.
    flatten; intuition congruence.
    flatten; intuition congruence.
  Qed.

End Approx.

Module D <: SEMILATTICE_WITH_TOP.
  
  Module P := LPMapOLD Approx.

  Lemma gsspec: forall p q v x,
    P.get p (P.set q v x) = if peq p q then v else P.get p x.
  Proof.
    intros.
    destruct peq.
    - subst; rewrite P.gss; go.
    - rewrite P.gso; go.
  Qed.

  Definition z2p (z:Z) : positive :=
    match z with
      | Z0 => xH
      | Zpos p => xI p
      | Zneg p => xO p
    end.

  Definition p2z (p:positive) : Z :=
    match p with
      | xH => Z0
      | xI p => Zpos p
      | xO p => Zneg p
    end.

  Definition mr2p : mreg -> positive := Var_Type_Leibniz.mreg_to_pos.
  Definition v2p (v:var_type) : positive := 
    match v with
      | VReg mr => mr2p mr
      | VMem (Global id) => (15 + 2 * id)%positive
      | VMem (Stack i) => (15 + 2 * (z2p i) + 1)%positive
    end.

  Definition t := P.t.

  Definition get (v:var_type) (m:t) : Approx.t :=
    P.get (v2p v) m.

  (* strong update *)
  Definition sset (v:var_type) (a:Approx.t) (m:t) : t :=
    P.set (v2p v) a m.

  Definition set (v:var_type) (a:Approx.t) (m:t) : t :=
    match v with
      | VMem (Global _) =>
        let a' := Approx.lub a (get v m) in
        P.set (v2p v) a' m
      | _ =>  P.set (v2p v) a m
    end.

  Definition eq (x y: t) : Prop :=
    forall p, Approx.eq (get p x) (get p y).

  Lemma eq_refl: forall x, eq x x.
  Proof. 
    unfold eq; intros. apply Approx.eq_refl.
  Qed.

  Lemma eq_sym: forall x y, eq x y -> eq y x.
  Proof.
    unfold eq; intros. apply Approx.eq_sym; auto.
  Qed.

  Lemma eq_trans: forall x y z, eq x y -> eq y z -> eq x z.
  Proof.
    unfold eq; intros. eapply Approx.eq_trans; eauto.
  Qed.

  Definition beq: t -> t -> bool := P.beq.

  Lemma beq_correct: forall x y, beq x y = true -> eq x y.
  Proof.
    unfold eq; intros.
    apply P.beq_correct in H.
    apply (H (v2p p)).
  Qed.

  Definition ge (x y: t) : Prop :=
    forall p, Approx.ge (get p x) (get p y).

  Lemma ge_refl: forall x y, eq x y -> ge x y.
  Proof.
    unfold ge, eq; intros. apply Approx.ge_refl. auto.
  Qed.

  Lemma ge_trans: forall x y z, ge x y -> ge y z -> ge x z.
  Proof.
    unfold ge; intros. apply Approx.ge_trans with (get p y); auto.
  Qed.

  Definition bot: t := P.bot.

  Lemma ge_bot: forall x, ge x bot.
  Proof.
    unfold ge; intros.
    apply (P.ge_bot x (v2p p)).
  Qed.
  
  Definition lub: t -> t -> t := P.lub.

  Lemma ge_lub_left: forall x y, ge (lub x y) x.
  Proof.
    unfold ge; intros.
    apply (P.ge_lub_left x y (v2p p)).    
  Qed.

  Lemma ge_lub_right: forall x y, ge (lub x y) y.
  Proof.
    unfold ge; intros.
    apply (P.ge_lub_right x y (v2p p)).    
  Qed.

  Definition top := P.top.

  Lemma ge_top: forall x, ge top x.
  Proof.
    unfold ge; intros.
    apply (P.ge_top x (v2p p)).    
  Qed.

End D.

Definition add_ofs (n:int) (p:pt) :=
  match p with
  | Unknown => Unknown
  | Symb (Global x) => Symb (Global x)
  | Symb (Stack ofs) => 
    if Int.eq_dec 
         (Int.add_carry (Int.repr ofs) n Int.zero)
         Int.zero then
      Symb (Stack (ofs + (Int.unsigned n)))
    else Unknown
  | NoPtr => NoPtr
  end.

Notation "ab # x" := (D.get (VReg x) ab).

Definition add_val (ab:D.t) (x y:mreg) :=
  match ab#x, ab#y with
  | NoPtr, NoPtr => NoPtr

  | Symb (Global s), NoPtr
  | NoPtr, Symb (Global s) => Symb (Global s)

  | _, _ => Unknown
  end.

Definition add_sub (ab:D.t) (x y:mreg) :=
  match ab#x, ab#y with
  | NoPtr, NoPtr => NoPtr

  | Symb (Global s), NoPtr => Symb (Global s)

  | _, _ => Unknown
  end.


Definition addressing (ab:D.t) (addr:addressing) (args:list mreg) :=
  match addr, args with
  | Aindexed n, x::nil => add_ofs n ab#x
  | Aindexed2 n, x::y::nil => add_ofs n (add_val ab x y)
  | Aindexed2scaled sc ofs, x::y::nil => 
    match ab#x with
      | Symb (Stack _) => Unknown
      | a => a
    end
  | Aglobal s ofs, nil => Symb (Global s)
  | Abased s ofs, x::nil => Symb (Global s)
  | Abasedscaled sc s ofs, x::nil => Symb (Global s)
  | Ainstack ofs, nil => Symb (Stack (Int.unsigned ofs))
  | _, _ => NoPtr
  end.

Definition pointsto (ab:D.t) := addressing ab .

Definition operation (ab:D.t) (op:Op.operation) (args:list mreg) := 
  match op, args with
  | Omove, x::nil => ab#x
  | Oindirectsymbol s, nil =>  Symb (Global s)
  | Olea addr, _ => addressing ab addr args
  | Osub, x::y::nil => add_sub ab x y (* to be checked... *)
  | _, _ => NoPtr
  end.


Definition memcpy (ab:D.t) (dst:Z) (src:Z) : D.t :=
  D.set (VMem (Stack dst)) (D.get (VMem (Stack src)) ab) 
        (D.set (VMem (Stack (dst +4))) (D.get (VMem (Stack (src +4))) ab) ab).

Definition transfer (f: function) (pc: node) (before: D.t) :=
  match f.(fn_code)!pc with
  | None => before
  | Some i =>
      match i with
        | MIop op args res s =>
          D.set (VReg res) (operation before op args) before
        | MIload chunk addr args dst s =>
          match addressing before addr args with
            | NoPtr => before (* dead code *)
            | Symb s => D.set (VReg dst) (D.get (VMem s) before) before
            | Unknown => D.set (VReg dst) Unknown before
          end
        | MIstore chunk addr args src s =>
          match addressing before addr args with
            | NoPtr => before (* dead code *)
            | Symb s => D.set (VMem s) before#src before
            | Unknown => D.top (* if it happens, the program is too hard for us *)
          end          
        | MImemcpy_8_4 dst src s => 
          match before#dst, before#src with
            | NoPtr, _
            | _, NoPtr => before
            | Symb (Stack off_dst), Symb (Stack off_src) => 
              memcpy before off_dst off_src
            | _, _ => D.top
          end
        | MIreturn
        | MIgoto _ 
        | MIcond _ _ _ _ => before
      end
  end.

Module DS := Dataflow_Solver(D)(NodeSetForward).

Definition mk_init_data (id:init_data) :=
  match id with
    | Init_addrof symb ofs => Symb (Global symb)
    |  _ => NoPtr
  end.

Fixpoint mk_init_data_list (l:list init_data) :=
  match l with
    | nil => NoPtr
    | id::q => Approx.lub (mk_init_data id) (mk_init_data_list q)
  end.

Definition alloc_global (ab:D.t) (idg: ident * globdef Mach.fundef unit) :=
  match idg with
  | (id, Gfun f) => ab
  | (id, Gvar v) => D.sset (VMem (Global id)) (mk_init_data_list v.(gvar_init)) ab
  end.

Definition alloc_globals (ab:D.t) (gl: list (ident * globdef Mach.fundef unit)) :=
  fold_left alloc_global gl ab.

Lemma mr2p_inj : forall x y,
  D.mr2p x = D.mr2p y -> x=y.
Proof.
  unfold D.mr2p, Var_Type_Leibniz.mreg_to_pos; intros x y H.
  flatten H.
Qed.

Lemma mr2p_bound : forall x,
  1 <= Zpos (D.mr2p x) <= 16.
Proof.
  unfold D.mr2p, Var_Type_Leibniz.mreg_to_pos; intros x.
  flatten; omega.
Qed.


Lemma Zpos_z2p_unsigned : forall i,
  i >= 0 ->                           
  Z.pos (D.z2p i) = 2* i + 1.
Proof.
  destruct i; simpl;
  unfold D.z2p;
  flatten; zify; try omega.  
Qed.

Lemma z2p_inj : forall x y,
  D.z2p x = D.z2p y -> x=y.
Proof.
  destruct x; destruct y; simpl; zify; try omega.
Qed.

Lemma v2p_inj : forall x y,
  D.v2p x = D.v2p y -> x=y.
Proof.
  unfold D.v2p; intros x y H.
  destruct x; destruct y.
  - apply mr2p_inj in H; congruence.
  - generalize (mr2p_bound m); intro.
    destruct p; zify; omega.
  - generalize (mr2p_bound m); intro.
    destruct p; zify; omega.
  - destruct p; destruct p0; try (zify; omega).
    + zify.
      assert (Zpos i= Zpos i0) by omega.
      congruence.
    + assert (D.z2p z=D.z2p z0).
        zify; omega.
      apply z2p_inj in H0; congruence.
Qed.


Definition init (p:Mach.program) (size:Z) : D.t :=
  alloc_globals D.bot p.(prog_defs).

Definition find_fun (id0:ident) (o:option function) (idg: ident * globdef fundef unit) :=
  match idg with
  | (id, Gfun (Internal f)) => if peq id0 id then Some f else o
  | _ => o
  end.

Definition find_main (p:program) : function := (fst p).

Definition analyze (p:Mach.program) (f:function) : PMap.t D.t :=
  if list_norepet_dec peq (prog_defs_names p) then
    match DS.fixpoint (fn_code f) successors (transfer f) 
                      ((fn_entrypoint f, init p (fn_stacksize f)) :: nil) with
      | None => PMap.init D.top
      | Some res => res
    end
  else PMap.init D.top.

Definition query (m:PMap.t D.t) (n:node) (addr:Op.addressing) (regs:list mreg) 
  : option pointsto_symbol :=
  match addressing (PMap.get n m) addr regs with
    | Symb x => Some x
    | _ => None
  end.

Definition test (p:Mach.program) (f:function) : res unit :=
  let alias := analyze p f in
  PTree.fold
    (fun res n ins =>
       match ins with
         | MIload chunk addr args _ s 
         | MIstore chunk addr args _ s =>
           match query alias n addr args with
             | Some _ => res
             | None => Error (msg "PT test: query error")
           end
         | _ => res
       end)
    (fn_code f)
    (OK tt).

Require Import Values.
Require Import Globalenvs.
Require Import Memory.


Section CORRECTNESS.

Variable ge : Mach.genv.
Variable sp : ident.
Variable sp_not_symbol: forall i,
  Genv.find_symbol ge i <> Some sp.

Inductive gamma_val : pt -> val -> Prop :=
| gamma_Unknown v : gamma_val Unknown v
| gamma_glob b ofs id  
    (G1: Genv.find_symbol ge id = Some b):
    gamma_val (Symb (Global id)) (Vptr b ofs)
| gamma_stack b ofs1 ofs2 
    (G1: b=sp)
    (G2: ofs1=Int.unsigned ofs2):
    gamma_val (Symb (Stack ofs1)) (Vptr b ofs2)
| gamma_other v ab
    (G1: forall b ofs, v<>Vptr b ofs):
    gamma_val ab v.

Lemma gamma_NoPtr_Vundef: gamma_val NoPtr Vundef.
Proof.
  go.
Qed.
Hint Resolve gamma_NoPtr_Vundef.

Lemma gamma_val_monotone : forall ab1 ab2 v,
  Approx.ge ab2 ab1 ->                             
  gamma_val ab1 v ->
  gamma_val ab2 v.
Proof.
  intros.
  inv H; inv H0; intuition go.
Qed.

Definition gamma_regs (ab:D.t) (rs:Mach.regset) : Prop :=
  forall r, gamma_val (ab#r) (rs r).

Inductive gamma_mem (ab:D.t) (m:mem) : Prop :=
  | gamma_mem_def
  (GINV1: forall chunk b ofs b' ofs',
            Mem.load chunk m b ofs = Some (Vptr b' ofs') -> chunk = Mint32)
  (GDEF: forall a v s,
    Mem.loadv Mint32 m a = Some v ->
    gamma_val (Symb s) a ->
    gamma_val (D.get (VMem s) ab) v):
  gamma_mem ab m.

Lemma gamma_mem_monotone : forall ab1 ab2 m,
  D.ge ab2 ab1 ->                             
  gamma_mem ab1 m ->
  gamma_mem ab2 m.
Proof.
  intros.
  inv H0; constructor; intros; eauto.
  eapply gamma_val_monotone.
  - apply H.
  - eauto.
Qed.

Inductive gamma (ab:D.t) : state -> Prop :=
  gamma_def pc rs m
    (G1: gamma_mem ab m)
    (G2: gamma_regs ab rs)
    (G3: forall i : ident, Genv.find_symbol ge i <> Some sp):
    gamma ab (State pc rs m).

Lemma gamma_mem_set_reg : forall ab m r a,
  gamma_mem ab m ->
  gamma_mem (D.set (VReg r) a ab) m.
Proof.
  intros ab m r a H.
  inv H.
  constructor; auto.

  intros a0 v s Hl Hg.  
  unfold D.set, D.get.
  rewrite D.gsspec; auto.
  flatten.
  - apply v2p_inj in e; congruence.
  - apply GDEF with (1:=Hl); auto.
Qed.

Lemma gamma_regs_set_mem : forall ab rs s a,
  gamma_regs ab rs ->
  gamma_regs (D.set (VMem s) a ab) rs.
Proof.
  intros ab rs s a H r.
  unfold D.set, D.get.
  destruct s; rewrite D.gsspec; auto.
  flatten.
  apply v2p_inj in e; congruence.
  apply H.
  flatten.
  apply v2p_inj in e; congruence.
  apply H.
Qed.

Lemma gamma_regs_set_reg: forall ab abv r v rs,
  gamma_regs ab rs ->
  gamma_val abv v ->
  gamma_regs (D.set (VReg r) abv ab)  (Mach.Regmap.set r v rs).
Proof.
  intros ab abv r v rs H H1.
  unfold gamma_regs in *.
  intros y.
  unfold D.set, D.get.
  rewrite D.gsspec.
  - flatten.
    + apply v2p_inj in e; inv e.
      rewrite Mach.Regmap.gss; auto.
    + rewrite Mach.Regmap.gso; try congruence.
      apply H.
Qed.

Lemma add_val_correct : forall ab rs t t0,
  gamma_regs ab rs ->
  gamma_val (add_val ab t t0) (Val.add (rs t) (rs t0)).
Proof.
  intros.
  assert (Hg:=H t).
  assert (Hg0:=H t0).
  unfold add_val, Val.add; inv Hg; inv Hg0; flatten; go.
Qed.

Lemma addressing_correct: forall ab rs a args v,
  gamma_regs ab rs ->
  eval_addressing ge (Vptr sp Int.zero) a rs ## args = Some v ->
  gamma_val (addressing ab a args) v.
Proof.
  intros.
  unfold addressing, eval_addressing in *; flatten H0; flatten;
  try match goal with
      | id: rs## _ = _ :: _ |- _ => inv id
    end; go;
  try (destruct v1; destruct v0; simpl; go; fail);
  try (destruct v0; simpl; go; fail).

  generalize (H t).
  unfold add_ofs; flatten; go.
  intros T; inv T; simpl; go.
  unfold Val.add; flatten; go.
  intros T; inv T; simpl; go.
  constructor; auto.
  rewrite Int.unsigned_add_carry.
  rewrite Int.repr_unsigned in e.
  rewrite e.
  rewrite Int.unsigned_zero.
  omega.
  unfold Val.add; flatten; go.

  intros T; inv T; simpl; go.
  unfold Val.add; flatten; go.

  rewrite Val.add_commut.

  assert (gamma_val (add_val ab t t0) (Val.add (rs t) (rs t0)))
    by (eapply add_val_correct; eauto).
  unfold add_ofs; flatten; inv H0; simpl; flatten; go.
  econstructor; eauto.
  rewrite Int.unsigned_add_carry.
  rewrite Int.repr_unsigned in e.
  rewrite e.
  rewrite Int.unsigned_zero.
  omega.



  assert (Hg:=H t); rewrite Eq3 in *;
  unfold Val.mul; flatten; inv Hg; destruct (rs t); simpl; go.

  assert (Hg:=H t); rewrite Eq3 in *;
  unfold Val.mul; flatten; inv Hg; destruct (rs t); simpl; go.

  unfold symbol_address; flatten; go.

  unfold symbol_address, Val.add; flatten; go.
  flatten Eq; go.
  flatten Eq; go.

  unfold symbol_address, Val.add; flatten; go.
  flatten Eq; go.
  flatten Eq; go.

  constructor; go.
  rewrite Int.add_zero_l; go.
Qed.

Lemma get_top_Unkown: forall x, D.get x D.top = Unknown.
Proof.
  intros; unfold D.top, D.P.top, D.get.
  simpl; rewrite PTree.gempty; compute; go.  
Qed.

Lemma transfer_correct : forall f pc ab rs m e st2,
  gamma ab (State pc rs m) ->
  step ge f (Vptr sp Int.zero) (State pc rs m) e st2 ->
  gamma (transfer f pc ab) st2.
Proof.
  intros f pc ab rs m e st2 Hg Hs.
  inv Hg; inv Hs; 
  unfold transfer; flatten;
  constructor; auto.
  - apply gamma_mem_set_reg; auto.
  - apply gamma_regs_set_reg; auto.
    unfold eval_operation, operation in *.
    flatten H5; flatten;
    try match goal with
      | id: rs## _ = _ :: _ |- _ => inv id
    end; go;
    try (destruct v1; destruct v0; simpl; go; fail);
    try (destruct v0; simpl; go; fail).

    unfold symbol_address; flatten; go.

    unfold add_sub; flatten; go.
    
    generalize (G2 m0); rewrite Eq0.
    generalize (G2 m1); rewrite Eq2.
    intros T1 T2; inv T1; inv T2; simpl; go.
    flatten; go.
    unfold Val.sub; flatten; go.

    generalize (G2 m0); rewrite Eq0.
    generalize (G2 m1); rewrite Eq1.
    intros T1 T2; inv T1; inv T2; simpl; go.
    unfold Val.sub; flatten; go.

    destruct v1; destruct v0; inv H5; simpl; go.
    flatten H1; go.
    destruct v1; destruct v0; inv H5; simpl; go.
    flatten H1; go.
    destruct v1; destruct v0; inv H5; simpl; go.
    flatten H1; go.
    destruct v1; destruct v0; inv H5; simpl; go.
    flatten H1; go.
    destruct v1; destruct v0; simpl; go.
    flatten; go.
    destruct v0; simpl; flatten; go.
    destruct v0; simpl; flatten; go.
    destruct v0; simpl; flatten; go.
    destruct v0; inv H5; simpl; go.
    flatten H1; go.
    destruct v1; destruct v0; simpl; flatten; go.
    destruct v0; simpl; flatten; go.
    destruct v0; simpl; flatten; go.
    destruct v1; destruct v0; simpl; flatten; go.

    eapply addressing_correct; eauto.
    
    destruct v0; inv H5.
    unfold option_map in H1; flatten H1; go.

    unfold Val.of_optbool; flatten; compute; go.
  - apply gamma_mem_set_reg; auto.
  - apply gamma_regs_set_reg; go.
  - apply gamma_mem_set_reg; auto.
  - apply gamma_regs_set_reg; auto.
    destruct v; go.
    assert (chunk=Mint32).
      destruct a; inv H6.
      inv G1; eauto.
    inv G1.
    exploit (GDEF _ _ s H6); eauto.
    rewrite <- Eq1.
    eapply addressing_correct; eauto.    
  - exploit addressing_correct; eauto.
    rewrite Eq1.
    intros G; destruct a; inv G; inv H6.
    congruence.
  - exploit addressing_correct; eauto.
    constructor.
    + intros.
      destruct a; inv H6.
      exploit Mem.load_pointer_store; eauto.
      destruct 1.
      * intuition.
      * exploit Mem.load_store_other; eauto.
        intros E; rewrite E in *.
        inv G1; eauto.
    + intros; rewrite get_top_Unkown; go.
  - intros r.
    rewrite get_top_Unkown.
    go.
  - constructor.
    + intros.
      destruct a; inv H6.
      exploit Mem.load_pointer_store; eauto.
      destruct 1.
      * intuition.
      * exploit Mem.load_store_other; eauto.
        intros E; rewrite E in *.
        inv G1; eauto.
    + intros.
      destruct v; go.
      destruct a0; inv H.
      destruct a; inv H6.
      destruct Mem.load_pointer_store with (1:=H1) (2:=H2); eauto.
      * destruct H as (T1 & T2 & T3 & T4 & T5); subst.
        unfold D.set, D.get.
        inv G1.
        destruct s;
          rewrite D.gsspec; flatten; auto;
          try apply lub_not_bot; auto.
        eapply gamma_val_monotone.
        eapply Approx.ge_lub_left.
        rewrite <- T2; apply G2.
        exploit addressing_correct; eauto.
        rewrite Eq1.
        inv H0; intros T; inv T; try congruence.
        generalize (Genv.find_invert_symbol _ _ G0).
        generalize (Genv.find_invert_symbol _ _ G1).
        congruence.
        
        eelim sp_not_symbol; eauto.

        rewrite <- T2; apply G2.

        exploit addressing_correct; eauto.
        rewrite Eq1.
        inv H0; intros T; inv T; try congruence.
        eelim sp_not_symbol; eauto.

      * rewrite (Mem.load_store_other _ _ _ _ _ _ H1) in H2; auto.
        apply gamma_val_monotone with (D.get (VMem s0) ab).
        unfold D.set, D.get.
        destruct s.
        rewrite D.gsspec; auto.
        flatten; auto.
        rewrite e.
        apply Approx.ge_lub_right.
        apply Approx.ge_refl; auto.
        apply Approx.eq_refl; auto.
        rewrite D.gsspec; auto.
        flatten; auto.
        exploit addressing_correct; eauto.
        rewrite Eq1; intros.
        apply v2p_inj in e; inv e.
        inv H3; inv H0; try congruence.
        destruct chunk; simpl in *; intuition omega.
        apply Approx.ge_refl; auto.
        apply Approx.eq_refl; auto.
        inv G1; auto.
        apply GDEF with (Vptr b0 i0); simpl; eauto.
    - apply gamma_regs_set_mem; auto.
    - destruct a; inv H6.
      exploit addressing_correct; eauto.
      rewrite Eq1; intros T; inv T; congruence.
Qed.

End CORRECTNESS.

Lemma gamma_monotone: forall ge sp ab1 ab2,
  D.ge ab2 ab1 ->
  forall s, gamma ge sp ab1 s -> gamma ge sp ab2 s.
Proof.
  intros.
  inv H0.
  split.
  - inv G1; constructor.
    * eauto.
    * intros.
      eapply gamma_val_monotone; eauto.
  - intros r.
    eapply gamma_val_monotone; eauto.
  - auto.
Qed.

Definition get_pc (s:state) : node :=
 match s with
   | State pc _ _ => pc
 end.

Lemma analyze_step_postfix : 
  forall p f sp st1 e st2,  
  gamma (Genv.globalenv p) sp (PMap.get (get_pc st1) (analyze p f)) st1 ->
  step (Genv.globalenv p) f (Vptr sp Int.zero) st1 e st2 ->
  gamma (Genv.globalenv p) sp (PMap.get (get_pc st2) (analyze p f)) st2.
Proof.
  intros.
  destruct st1; simpl in *.
  exploit transfer_correct; eauto.
  - inv H; auto.
  - intros.
    assert ( State pc rs m = st2
            \/ exists ins,
              (fn_code f) ! pc = Some ins /\ In (get_pc st2) (successors ins)).
      inv H0; try (right; econstructor; split; [eassumption|simpl; auto;fail]). 
      auto.
    destruct H2.
    inv H2; simpl; auto.
    destruct H2 as (ins & T1 & T2).
    apply gamma_monotone with (2:=H1).
    unfold analyze.
    flatten.
    + eapply DS.fixpoint_solution; eauto.
    + rewrite PMap.gi.
      auto.
      apply D.ge_top.
    + rewrite PMap.gi.
      auto.
      apply D.ge_top.
Qed.

Inductive gamma_estate ge tf (ab:PMap.t D.t) : extended_state -> Prop :=
  | gamma_estate_IState : forall m
    (G1:forall sp m1 m2 m3, 
        Mem.alloc m 0 (fn_stacksize tf) = (m1, sp) ->
        Mach.store_stack m1 (Vptr sp Int.zero) Tint (fn_link_ofs tf) (Mach.parent_sp nil) = 
           Some m2 ->
        Mach.store_stack m2 (Vptr sp Int.zero) Tint (fn_retaddr_ofs tf) (Mach.parent_ra nil) = Some m3 ->
        gamma_mem ge sp (PMap.get (fn_entrypoint tf) ab) m3)
    (G2:forall i b, Genv.find_symbol ge i = Some b ->
        Mem.valid_block m b),
    gamma_estate ge tf ab (IState tf m)
  | gamma_estate_NState : forall sp st
    (G2: gamma ge sp (PMap.get (get_pc st) ab) st),
    gamma_estate ge tf ab (NState tf (Vptr sp Int.zero) st).


Lemma analyze_estep_postfix : 
  forall p f st1 e st2,  
  gamma_estate (Genv.globalenv p) f (analyze p f) st1 ->
  estep (Genv.globalenv p) st1 e st2 ->
  gamma_estate (Genv.globalenv p) f (analyze p f) st2.
Proof.
  intros p f st1 e st2 H H0.
  inv H.
  - inv H0; constructor.
    constructor.
    * simpl; eapply G1; eauto.
    * simpl.
      intros r.
      apply gamma_val_monotone with ((init p (fn_stacksize f))#r).      
      + unfold analyze.
        flatten.
        exploit DS.fixpoint_entry; eauto; go.
        rewrite PMap.gi.
        apply D.ge_top.
        rewrite PMap.gi.
        apply D.ge_top.
      + rewrite Mach.Regmap.gi.
        go.
    *  red; intros.
       exploit G2; eauto.
       exploit Mem.fresh_block_alloc; eauto.
  - inv H0; constructor.
    eapply analyze_step_postfix; eauto.
Qed.

Lemma load_valid_block: forall chunk m b ofs v,
  Mem.load chunk m b ofs = Some v ->
  Mem.valid_block m b.
Proof.
  intros.
  apply Mem.load_valid_access in H.
  destruct H.
  apply Mem.perm_valid_block with ofs Cur Readable.
  apply H.
  destruct chunk; go.
Qed.

Lemma store_init_data_list_load_chunk: forall (ge:Mach.genv) l s m2 b m3,
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m2 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32) ->
  Genv.store_init_data_list ge m2 b s l = Some m3 ->
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m3 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32).
Proof.
  induction l; simpl; intros.
  - inv H0; eauto.
  - flatten H0.
    eapply IHl with (2:=H0); eauto.
    clear H1 H0 IHl; intros.
    unfold Genv.store_init_data in Eq; flatten Eq; subst;
    try (exploit Mem.load_pointer_store; eauto; destruct 1;
         [intuition congruence; fail|
          erewrite Mem.load_store_other in H0; eauto]).
    eauto.
Qed.

Lemma store_zeros_data_list_load_chunk: forall m2 b a c m3,
  store_zeros m2 b a c = Some m3 ->
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m2 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32) ->
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m3 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32).
Proof.
  intros m2 b a c m3.
  functional induction (store_zeros m2 b a c); intros.
  - inv H; eauto.
  - eapply IHo; eauto.
    clear H1 H IHo.
    intros.
    exploit Mem.load_pointer_store; eauto; destruct 1.
    + intuition congruence.
    + erewrite Mem.load_store_other in H; eauto.
  - congruence.
Qed.

Lemma drop_perm_load_chunk: forall m b0 lo hi p m',
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32) ->
  (forall ofs k p, Mem.perm m b0 ofs k p -> lo <= ofs < hi) ->
  Mem.drop_perm m b0 lo hi p = Some m' ->
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m' b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32).
Proof.
  intros.
  assert (Mem.perm m' b ofs Cur Readable) by (eauto with mem).
  destruct (peq b0 b); subst.
  - assert (O:(lo <= ofs < hi)\/~(lo <= ofs < hi)) by omega.
    destruct O as [O|O].
    + exploit Mem.perm_drop_2; eauto; intros HP.
      erewrite Mem.load_drop in H2; eauto.
    + exploit Mem.perm_drop_4; eauto; intros HP.
      eelim O; eauto.
  - erewrite Mem.load_drop in H2; eauto.
Qed.

Lemma alloc_load_chunk: forall m lo hi m' b,
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32) ->
  Mem.alloc m lo hi = (m', b) ->
  (forall chunk b ofs b' ofs', 
      Mem.load chunk m' b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32).
Proof.
  intros.
  destruct (peq b0 b); subst.
  - exploit Mem.load_alloc_same; eauto.
    congruence.
  - exploit load_valid_block; eauto; intros.
    edestruct Mem.valid_block_alloc_inv with (1:=H0); eauto.
    * congruence.
    * erewrite Mem.load_alloc_unchanged in H1; eauto.
Qed.

Lemma store_init_data_same_perm : forall (ge:Mach.genv) l m b s m' ofs k p,
  Genv.store_init_data_list ge m b s l = Some m' ->
  Mem.perm m' b ofs k p ->
  Mem.perm m b ofs k p.
Proof.
  induction l; simpl; intros.
  - congruence.
  - flatten H.
    destruct a; simpl in Eq;
    try (eapply Mem.perm_store_2; eauto; fail).
    inv Eq; eauto.
    flatten Eq.
    eapply Mem.perm_store_2; eauto.
Qed.

Lemma store_init_data_list_load_other : forall (ge:Mach.genv) l m b s m' chunk b' ofs,
  Genv.store_init_data_list ge m b s l = Some m' ->
  b <> b' ->
  Mem.load chunk m' b' ofs = Mem.load chunk m b' ofs.
Proof.
  induction l; simpl; intros.
  - congruence.
  - flatten H.
    erewrite (IHl m0); eauto.
    destruct a; simpl in Eq;
    try (eapply Mem.load_store_other; eauto; fail).
    congruence.
    flatten eq.
    eapply Mem.load_store_other; eauto.
Qed.

Lemma store_init_data_list_nextblock : forall (ge:Mach.genv) l m b s m',
  Genv.store_init_data_list ge m b s l = Some m' ->
  Mem.nextblock m' = Mem.nextblock m.
Proof.
  induction l; simpl; intros.
  - congruence.
  - flatten H.
    erewrite (IHl m0); eauto.
    destruct a; simpl in Eq;
    try (eapply Mem.nextblock_store; eauto; fail).
    congruence.
    flatten eq.
    eapply Mem.nextblock_store; eauto.
Qed.


Lemma store_zeros_same_perm: forall m b a c m',
  store_zeros m b a c = Some m' -> forall ofs k p,
  Mem.perm m' b ofs k p ->
  Mem.perm m b ofs k p.
Proof.
  intros m b a c m'.
  functional induction (store_zeros m b a c).
  - intros T; inv T; auto.
  - intros.
    eapply Mem.perm_store_2; eauto.
  - congruence.
Qed.

Lemma store_zeros_load_other: forall m b a c m',
  store_zeros m b a c = Some m' -> forall chunk b' ofs,
  b <> b' ->
  Mem.load chunk m' b' ofs = Mem.load chunk m b' ofs.
Proof.
  intros m b a c m'.
  functional induction (store_zeros m b a c).
  - intros T; inv T; auto.
  - intros.
    rewrite IHo; auto.
    eapply Mem.load_store_other; eauto.
  - congruence.
Qed.

Lemma store_zeros_load_ptr: forall m b a c m',
  store_zeros m b a c = Some m' -> forall chunk b' ofs b0 ofs0,
  Mem.load chunk m' b' ofs = Some (Vptr b0 ofs0) ->
  Mem.load chunk m b' ofs = Some (Vptr b0 ofs0).
Proof.
  intros m b a c m'.
  functional induction (store_zeros m b a c).
  - intros T; inv T; auto.
  - intros.
    eapply IHo in H0; eauto.
    exploit Mem.load_pointer_store; eauto.
    destruct 1.
    + intuition congruence.
    + erewrite <- Mem.load_store_other; eauto.
  - congruence.
Qed.

Lemma store_zeros_nextblock: forall m b a c m',
  store_zeros m b a c = Some m' -> 
  Mem.nextblock m' = Mem.nextblock m.
Proof.
  intros m b a c m'.
  functional induction (store_zeros m b a c).
  - intros T; inv T; auto.
  - intros.
    rewrite IHo; auto.
    eapply Mem.nextblock_store; eauto.
  - congruence.
Qed.

Lemma alloc_globals_inv1: forall (p:Mach.program) l m1 m2,
   Genv.alloc_globals (Genv.globalenv p) m1 l = Some m2 ->
   (forall chunk b ofs b' ofs', 
      Mem.load chunk m1 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32) ->
   (forall chunk b ofs b' ofs', 
      Mem.load chunk m2 b ofs = Some (Vptr b' ofs') ->
      chunk = Mint32).
Proof.
  induction l; simpl; intros.
  - inv H; eauto.
  - flatten H.
    eapply IHl; eauto.    
    clear - H0 Eq.
    unfold Genv.alloc_global in Eq; flatten Eq; subst.
    + eapply drop_perm_load_chunk with (3:=Eq); eauto.
      * eapply alloc_load_chunk; eauto.
      * eapply Mem.perm_alloc_3; eauto.
    + eapply drop_perm_load_chunk with (3:=Eq); eauto. 
      * eapply store_init_data_list_load_chunk with (2:=Eq4); eauto.
        eapply store_zeros_data_list_load_chunk; eauto.
        eapply alloc_load_chunk; eauto.
      * intros ofs k p0 H.
        eapply store_init_data_same_perm in H; eauto.
        eapply store_zeros_same_perm in H; eauto.
        eapply Mem.perm_alloc_3; eauto.
Qed.

Lemma store_init_data_other : forall (ge:Mach.genv) l m b s m',
  Genv.store_init_data_list ge m b s l = Some m' ->
  forall b0 ofs b' ofs',
  ofs < s ->
  Mem.load Mint32 m' b0 ofs = Some (Vptr b' ofs') ->
  Mem.load Mint32 m b0 ofs = Some (Vptr b' ofs').
Proof.
  induction l; simpl; intros.
  - inv H; auto.
  - flatten H.
    eapply (IHl m0) in H1; eauto.
    + destruct a; simpl in Eq;
      flatten Eq;
      try (
          exploit Mem.load_pointer_store; eauto;
          destruct 1; [intuition congruence|idtac];
          erewrite Mem.load_store_other in H1; eauto).
      exploit Mem.load_pointer_store; eauto.
      destruct 1.
      * intuition subst; omega.
      * erewrite Mem.load_store_other in H1; eauto.
    + assert (GL:=Genv.init_data_size_pos a).
      omega.
Qed.

Lemma store_init_data_ptr : forall (ge:Mach.genv) l m b s m',
  Genv.store_init_data_list ge m b s l = Some m' ->
  forall ofs b' ofs',
  s <= ofs < s+ Genv.init_data_list_size l -> 
  Mem.load Mint32 m' b ofs = Some (Vptr b' ofs') ->
  Mem.load Mint32 m b ofs = Some (Vptr b' ofs') \/
  exists symb,
  In (Init_addrof symb ofs') l /\ Genv.find_symbol ge symb = Some b'.
Proof.
  induction l; simpl; intros.
  - inv H; left; congruence.
  - flatten H.
    destruct (Z_le_dec (s+ (Genv.init_data_size a)) ofs).
    + assert (GL:=Genv.init_data_size_pos a).
      edestruct (IHl m0); eauto; try omega.
      * left.
        destruct a; simpl in Eq;
        try (erewrite <- Mem.load_store_other; eauto; fail).
        congruence.
        flatten Eq.
        erewrite <- Mem.load_store_other; eauto.
      * destruct H2 as (symb & Hi & Hg); go. 
    + assert (ofs < s + Genv.init_data_size a) by omega.
      exploit store_init_data_other; eauto; intros T.
      clear H1 n H IHl.
      destruct a; simpl in Eq; flatten Eq;
      try (exploit Mem.load_pointer_store; eauto; destruct 1; 
           [intuition congruence|
            erewrite Mem.load_store_other in T; eauto]).
      left; auto.
      exploit Mem.load_pointer_store; eauto.
      destruct 1; 
           [idtac|
            erewrite Mem.load_store_other in T; eauto].
      destruct H as (T1 & T2 & T3 & T4 & T5); subst.
      inv T2.
      go.
Qed.

Lemma nextblock_genv_next_aux : forall (ge:Mach.genv) l m1 m2 ge1 ge2, 
  Genv.alloc_global ge m1 l = Some m2 ->
  Genv.add_global ge1 l = ge2 -> 
  Mem.nextblock m1 = Genv.genv_next ge1 ->
  Mem.nextblock m2 = Genv.genv_next ge2.
Proof.
  intros.
  unfold Genv.alloc_global in H; flatten H; subst; simpl.
  + erewrite Mem.nextblock_drop; eauto.
    erewrite Mem.nextblock_alloc; eauto.
    congruence.
  + erewrite Mem.nextblock_drop; eauto.
    erewrite store_init_data_list_nextblock; eauto.
    erewrite store_zeros_nextblock; eauto.
    erewrite Mem.nextblock_alloc; eauto.
    congruence.
Qed.

Lemma nextblock_genv_next_aux' : forall (ge:Mach.genv) l m1 m2 ge1 ge2, 
  Genv.alloc_globals ge m1 l = Some m2 ->
  Genv.add_globals ge1 l = ge2 -> 
  Mem.nextblock m1 = Genv.genv_next ge1 ->
  Mem.nextblock m2 = Genv.genv_next ge2.
Proof.
  induction l; simpl; intros.
  - congruence.
  - flatten H.
    eapply IHl; eauto.
    eapply nextblock_genv_next_aux; eauto.
Qed.

Lemma find_symbol_other : forall i l (ge:Mach.genv),
  ~ In i fst ## l ->
  Genv.find_symbol (Genv.add_globals ge l) i = Genv.find_symbol ge i.
Proof.
  induction l; simpl; go.
  intros ge H.
  rewrite IHl; go.
  assert (fst a <>i) by go.
  unfold Genv.add_global, Genv.find_symbol; simpl.
  rewrite PTree.gso; auto.
Qed.

Lemma load_alloc_globals_add_globals : forall (ge:Mach.genv) l m1 m2 ge1 ge2
  (N:list_norepet (List.map fst l))
  (T:Mem.nextblock m1 = Genv.genv_next ge1),
  Genv.alloc_globals ge m1 l = Some m2 ->
  Genv.add_globals ge1 l = ge2 -> forall chunk b ofs v,
  Mem.load chunk m2 b ofs = Some v ->
  Mem.load chunk m1 b ofs = Some v \/
  exists symb, Genv.find_symbol ge2 symb = Some b.
Proof.
  induction l; simpl; intros.
  - left; congruence.
  - flatten H.
    edestruct (IHl m); eauto.
    inv N; auto.
    eapply nextblock_genv_next_aux; eauto.
    clear H H1 IHl.
    unfold Genv.alloc_global in Eq; flatten Eq; subst; simpl in *.
    + destruct (peq b0 b); subst.
      * { assert (O:(0 <= ofs < 1)\/~(0 <= ofs < 1)) by omega.
          assert (Mem.perm m b ofs Cur Readable) by (eauto with mem).         
          destruct O as [O|O].
          - exploit Mem.perm_drop_2; eauto; intros HP.
            inv HP.
          - exploit Mem.perm_drop_4; eauto; intros HP.
            eelim O.
            eapply Mem.perm_alloc_3; eauto.
        }
      * erewrite (Mem.load_drop m0) in H2; eauto.
        exploit load_valid_block; eauto; intros VB.
        exploit Mem.valid_block_alloc_inv; eauto.
        destruct 1; try congruence.
        erewrite <- (Mem.load_alloc_unchanged m1); eauto.
    + destruct (peq b0 b); subst.
      * { assert (O:(0 <= ofs < (Genv.init_data_list_size (gvar_init v0)))\/~(0 <= ofs < (Genv.init_data_list_size (gvar_init v0)))) by omega.
          assert (Mem.perm m b ofs Cur Readable) by (eauto with mem).         
          destruct O as [O|O].
          - exploit Mem.perm_drop_2; eauto; intros HP.
            erewrite (Mem.load_drop m4) in H2; eauto.
            clear HP H O Eq.
            exploit Mem.alloc_result; eauto.
            intros; subst.
            right; exists i.
            inv N.
            rewrite find_symbol_other; auto.
            unfold Genv.find_symbol, Genv.add_global; simpl.
            rewrite PTree.gss.
            congruence.
          - exploit Mem.perm_drop_4; eauto; intros HP.
            eelim O.
            eapply store_init_data_same_perm in HP; eauto.
            eapply store_zeros_same_perm in HP; eauto.
            eapply Mem.perm_alloc_3; eauto.
        }
      * erewrite (Mem.load_drop m4) in H2; eauto.
        exploit load_valid_block; eauto; intros VB.
        erewrite (store_init_data_list_load_other ge (gvar_init v0) m3) in H2; eauto.
        erewrite (store_zeros_load_other m0) in H2; eauto.            
        exploit load_valid_block; eauto; intros.
        exploit Mem.valid_block_alloc_inv; eauto.
        destruct 1; try congruence.
        erewrite (Mem.load_alloc_unchanged m1) in H2; eauto.
Qed.

Lemma load_init_mem_globalenv : forall (p:Mach.program) m,
  list_norepet (prog_defs_names p) ->
  Genv.init_mem p = Some m ->
  forall chunk b ofs v,
  Mem.load chunk m b ofs = Some v ->
  exists symb, Genv.find_symbol (Genv.globalenv p) symb = Some b.
Proof.
  unfold Genv.init_mem, Genv.globalenv; intros.
  edestruct load_alloc_globals_add_globals; eauto.
  - rewrite Mem.nextblock_empty; go.
  - exploit Mem.load_valid_access; eauto; intros.
    eelim Mem.valid_access_empty; eauto.
Qed.
  
Lemma load_ptr_alloc_globals : forall (ge:Mach.genv) l (ge1 ge2:Mach.genv) m1 m2
  (N:list_norepet (List.map fst l))
  (T:Mem.nextblock m1 = Genv.genv_next ge1)
  (AG:Genv.add_globals ge1 l = ge2),
  Genv.alloc_globals ge m1 l = Some m2 ->
  forall b ofs b' ofs',
    Mem.load Mint32 m2 b ofs = Some (Vptr b' ofs') ->
    Mem.load Mint32 m1 b ofs = Some (Vptr b' ofs') \/
    exists id v symb ofs'', 
      In (id, Gvar v) l /\
      Genv.find_symbol ge2 id = Some b /\
      In (Init_addrof symb ofs'') v.(gvar_init) /\ Genv.find_symbol ge symb = Some b'.
Proof.
  induction l; simpl.
  - left.
    inv H; auto.
  - intros ge1 ge2 m1 m2 N T AG H1.
    flatten H1.
    intros.
    edestruct IHl with (4:=H1) (ge1:=Genv.add_global ge1 a); eauto.
    + inv N; auto.
    + eapply nextblock_genv_next_aux; eauto.
    + unfold Genv.alloc_global in Eq.
      flatten Eq; subst.
      * { assert (Mem.perm m b ofs Cur Readable) by (eauto with mem).
          destruct (peq b0 b); left; subst.
          - assert (O:(0 <= ofs < 1)\/~(0 <= ofs < 1)) by omega.
            destruct O as [O|O].
            + exploit Mem.perm_drop_2; eauto; intros HP.
              inv HP.
            + exploit Mem.perm_drop_4; eauto; intros HP.
              eelim O.
              eapply Mem.perm_alloc_3; eauto.
          - erewrite (Mem.load_drop m0) in H0; eauto.
            exploit load_valid_block; eauto; intros.
            exploit Mem.valid_block_alloc_inv; eauto.
            destruct 1; try congruence.
            erewrite (Mem.load_alloc_unchanged m1) in H0; eauto.
        }
      * { assert (Mem.perm m b ofs Cur Readable) by (eauto with mem).
          destruct (peq b0 b); subst.
          - assert (O:(0 <= ofs < (Genv.init_data_list_size (gvar_init v)))\/~(0 <= ofs < (Genv.init_data_list_size (gvar_init v)))) by omega.
            destruct O as [O|O].
            + exploit Mem.perm_drop_2; eauto; intros HP.
              erewrite (Mem.load_drop m4) in H0; eauto.
              clear Eq H HP.
              edestruct store_init_data_ptr; eauto.              
              * left.
                eapply (store_zeros_load_ptr m0) in H; eauto.
                exploit Mem.load_alloc_same; eauto.
                congruence.
              * destruct H as (symb & T1 & T2).
                right.
                exploit Mem.alloc_result; eauto.
                intros; subst.
                exists i; exists v; exists symb; exists ofs'; repeat split; auto.
                rewrite find_symbol_other.
                unfold Genv.find_symbol, Genv.add_global; simpl.
                rewrite PTree.gss.
                congruence.
                inv N; auto.
            + left.
              exploit Mem.perm_drop_4; eauto; intros HP.
              eelim O.
              eapply store_init_data_same_perm in HP; eauto.
              eapply store_zeros_same_perm in HP; eauto.
              elim O.
              eapply Mem.perm_alloc_3; eauto.
          - left.
            erewrite (Mem.load_drop m4) in H0; eauto.
            exploit load_valid_block; eauto; intros.
            erewrite (store_init_data_list_load_other ge (gvar_init v) m3) in H0; eauto.
            erewrite (store_zeros_load_other m0) in H0; eauto.            
            exploit load_valid_block; eauto; intros.
            exploit Mem.valid_block_alloc_inv; eauto.
            destruct 1; try congruence.
            erewrite (Mem.load_alloc_unchanged m1) in H0; eauto.
        }
    + destruct H0 as (id & v & symb & ofs'' & A1 & A2). go.
Qed.


Lemma load_ptr_init_mem : forall (p:Mach.program) m',
  list_norepet (prog_defs_names p) ->
  Genv.init_mem p = Some m' ->
  forall b ofs b' ofs',
    Mem.load Mint32 m' b ofs = Some (Vptr b' ofs') ->
    exists id v symb ofs'', 
      In (id, Gvar v) p.(prog_defs) /\
      In (Init_addrof symb ofs'') v.(gvar_init) /\
      Genv.find_symbol (Genv.globalenv p) id = Some b /\
      Genv.find_var_info (Genv.globalenv p) b = Some v /\
      Genv.find_symbol (Genv.globalenv p) symb = Some b'.
Proof.
  unfold Genv.init_mem; intros.
  edestruct load_ptr_alloc_globals with (ge1:=(Genv.empty_genv _ _:Mach.genv)); eauto.
  - rewrite Mem.nextblock_empty; go.
  - exploit Mem.load_valid_access; eauto; intros.
    eelim Mem.valid_access_empty; eauto.
  - destruct H2 as (id & v & symb & ofs'' & T1 & T2 & T3 & T4).
    repeat econstructor; eauto.
    assert (T5:Genv.find_symbol (Genv.globalenv p) id = Some b) by auto.
    clear T4 T2.
    exploit Genv.find_symbol_inversion; eauto.
    unfold prog_defs_names; intros.
    exploit list_in_map_inv; eauto.
    intros ((s,g) & B1 & B2).
    inv B1.
    clear H2 H3; simpl in *.
    exploit Genv.find_var_exists; eauto.
    intros (b'' & C1 & C2).
    congruence.
Qed.

Lemma get_alloc_globals_other: forall id gl ab,
  ~ In id fst ## gl ->
  D.get (VMem (Global id)) (alloc_globals ab gl) = D.get (VMem (Global id)) ab.
Proof.
  induction gl; go.
  intros.
  simpl.
  rewrite IHgl.
  - unfold alloc_global; flatten.
    unfold D.get, D.sset; auto.
    rewrite D.P.gso; auto.
    intro T.
    exploit v2p_inj; eauto.
    intros V; inv V.
    elim H; simpl; auto.
  - simpl in H; intuition.
Qed.

Lemma alloc_globals_In: forall gl ab id v,
  In (id, Gvar v) gl ->
  list_norepet (List.map fst gl) ->
  D.get (VMem (Global id)) (alloc_globals ab gl) = (mk_init_data_list v.(gvar_init)).
Proof.
  induction gl; simpl; intros.
  - intuition.
  - destruct H as [H|H].
    + inv H.
      simpl.
      inv H0.
      rewrite get_alloc_globals_other; auto.
      unfold D.get, D.sset.
      rewrite D.P.gss; auto.
    + erewrite IHgl; eauto.
      inv H0; auto.
Qed.

Lemma In_mk_init_data_list : forall symb ofs l,
  In (Init_addrof symb ofs) l ->
  Approx.ge (mk_init_data_list l) (Symb (Global symb)).
Proof.
  induction l; simpl.
  - intuition.
  - destruct 1; subst.
    + simpl.
      apply Approx.ge_lub_left.
    + eapply Approx.ge_trans.
      * eapply Approx.ge_lub_right.
      * auto.
Qed.

Lemma analyze_gamma_init:
  forall p f st,  
    initial_state (f,p) st ->
    gamma_estate (Genv.globalenv p) f (analyze p f) st.
Proof.
  intros p f st H.
  inv H. 
  simpl in *.
  constructor.
  - intros sp m1 m2 m3 H H2 H3.
    assert (HG1:forall (chunk : memory_chunk) (b0 : block) (ofs : Z) 
       (b' : block) (ofs' : int),
              Mem.load chunk m3 b0 ofs = Some (Vptr b' ofs') -> chunk = Mint32).
       intros.
       unfold Mach.store_stack in *.
       simpl in *.      
       exploit Mem.load_pointer_store; eauto.
        destruct 1; try (intuition; fail).
        erewrite Mem.load_store_other in H4; eauto.
        clear H5 H3.
        exploit Mem.load_pointer_store; eauto.
        destruct 1; try (intuition; fail).
        erewrite Mem.load_store_other in H4; eauto.
        eapply alloc_load_chunk; eauto.
        unfold Genv.init_mem in *.
        eapply alloc_globals_inv1; eauto.            
        intros.
        apply Mem.load_valid_access in H5.
        eelim Mem.valid_access_empty; eauto.
    unfold analyze; flatten.      
    eapply gamma_mem_monotone.
    + eapply DS.fixpoint_entry; eauto; go.
    + unfold init.
      unfold Mach.store_stack in *.
      simpl in *.
      constructor; auto; intros.
      destruct v; go.
        destruct a; inv H4.
        exploit Mem.load_pointer_store; eauto.
        unfold Vzero.
        destruct 1; try (intuition congruence; fail).
        erewrite Mem.load_store_other in H7; eauto.
        clear H3 H4.
        exploit Mem.load_pointer_store; eauto.
        unfold Vzero.
        destruct 1; try (intuition congruence; fail).
        erewrite Mem.load_store_other in H7; eauto.
        clear H2 H3.
        { destruct (peq sp b1); subst.
          - exploit Mem.load_alloc_same; eauto.
            congruence.
          - exploit load_valid_block; eauto; intros.
            exploit Mem.valid_block_alloc_inv; eauto.
            destruct 1; try congruence.
            clear H2; erewrite Mem.load_alloc_unchanged in H7; eauto.
            clear H H3.
            inv H5; try congruence.
            apply Genv.find_invert_symbol in G1.
            edestruct load_ptr_init_mem as (id0 & v & symb & ofs'' & T1 & T2 & T3 & T4 & T5);
              eauto.
            apply Genv.find_invert_symbol in T3.
            assert (id0=id) by congruence; subst.
            apply Genv.invert_find_symbol in T3.
            exploit alloc_globals_In; eauto; intros Eq2.
            rewrite Eq2.
            eapply gamma_val_monotone.
            eapply In_mk_init_data_list; eauto.
            go.
        }
    + rewrite PMap.gi.
      constructor; auto.
      intros. 
      rewrite get_top_Unkown.
      constructor.
    + rewrite PMap.gi.
      constructor; auto.
      intros. 
      rewrite get_top_Unkown.
      constructor.
  - unfold ge in *; intros.
    eapply Genv.find_symbol_not_fresh; eauto.
Qed.

Lemma query_correct: forall ge pt f bsp pc rs m addr args x a,
    gamma_estate ge f pt (NState f (Vptr bsp Int.zero) (State pc rs m)) ->
    query pt pc addr args = Some x ->
    eval_addressing ge (Vptr bsp Int.zero) addr rs##args = Some a ->
    gamma_val ge bsp (Symb x) a.
Proof.
  intros ge pt0 f bsp pc rs m addr args x a H H0 H1.
  inv H.
  unfold query in H0; flatten H0.
  inv G2.
  exploit addressing_correct; eauto.
  simpl.
  congruence.
Qed.