Require Import Axioms.
Require Import Coqlib.
Require Import DLib.
Require Import Registers.
Require Import AST.
Require Import Machregs.
Require Import Globalenvs.
Require Import Op.
Require Import Integers.
Require Import Values.
Require Import Memory.
Require Import Mach.
Require Import MachIR.
Require Import MachIRAlias.
Require Import Maps.
Require Import VarSets.
Require Import Stealth_interleaving.
Require Import Stealth_secure_program.

(* Misc definitions *)
Inductive sec_level := Low | High.
Definition var_level := VS.t.
Definition stealth_var := Regset.t.

Definition sec_le (s1 s2:sec_level) : Prop :=
  s1=Low \/ s2=High.

Definition update_var_level (vl:var_level) (x:var_type) (l:sec_level) : var_level :=
  match l with
    | Low => VS.remove x vl
    | High => VS.add x vl
  end.

Definition get_var_level vl x :=
  if VS.mem x vl
  then High
  else Low.

Definition get_global_level Xh id :=
  if Regset.mem id Xh
  then High
  else Low.

Fixpoint get_args_level (vl:var_level) (args:list mreg) : sec_level :=
  match args with
    | nil => Low
    | r::rs =>
      if VS.mem (VReg r) vl then
        High
      else
        get_args_level vl rs
  end.

Definition max_sec_level (l1 l2:sec_level) : sec_level :=
  match l1 with
    | High => l1
    | _ => l2
  end.

Notation "l1 ∪ l2" := (max_sec_level l1 l2) (at level 10).
Notation "vl '##t' x " := (get_args_level vl x) (at level 1).
Notation "vl # x " := (get_var_level vl x) (at level 1).
Notation "vl # x <- l " := (update_var_level vl x l).

Fixpoint max_levels (l:list sec_level) : sec_level :=
  match l with
    | nil => Low
    | lv::q => max_sec_level lv (max_levels q)
  end.

Definition stack_offsets (ofs:Z) (chunk: memory_chunk) : list Z :=
  match chunk with
  | Mint8signed 
  | Mint8unsigned => ofs::nil
  | Mint16signed 
  | Mint16unsigned => ofs::(ofs+1)::nil
  | Mint32 
  | Mfloat32 => ofs::(ofs+1)::(ofs+2)::(ofs+3)::nil
  | Mint64 
  | Mfloat64 
  | Mfloat64al32 => ofs::(ofs+1)::(ofs+2)::(ofs+3)::(ofs+4)::(ofs+5)::(ofs+6)::(ofs+7)::nil
  end.

Fixpoint tloadN (n: nat) (p: Z) (vl:var_level) {struct n}: list sec_level :=
  match n with
  | O => nil
  | S n' => vl#(VMem (Stack p)) :: tloadN n' (p + 1) vl
  end.

Definition tload (vl:var_level) (chunk:memory_chunk) (ofs:Z) : sec_level :=
  max_levels (tloadN (size_chunk_nat chunk) ofs vl).

Fixpoint tstoreN (t:sec_level) (n: nat) (p: Z) (vl:var_level) {struct n}: var_level :=
  match n with
  | O => vl
  | S n' => tstoreN t n' (p+1) (vl#(VMem (Stack p)) <- t)
  end.

Definition tstore (vl:var_level) (chunk:memory_chunk) (ofs:Z) (t:sec_level) : var_level :=
  tstoreN t (size_chunk_nat chunk) ofs vl.



Module Inference.

Require Import Lattice.
Require Import Kildall.


Module D <: SEMILATTICE.
  Definition t := var_level.
  Definition eq (x y: t) := VS.eq x y.

  Lemma eq_refl: forall x, eq x x.
  Proof.
    destruct VS.eq_equiv; auto.
  Qed.
  Lemma eq_sym: forall x y, eq x y -> eq y x.
  Proof.
    destruct VS.eq_equiv; auto.
  Qed.
  Definition eq_trans: forall x y z, eq x y -> eq y z -> eq x z.
  Proof.
    destruct VS.eq_equiv; auto.
  Qed.
  Definition eq_dec: forall (x y: t), {eq x y} + {~ eq x y} := VS.eq_dec.

  Definition beq (x y: t) := if eq_dec x y then true else false.
  Lemma beq_correct: forall x y, beq x y = true -> eq x y.
  Proof.
    unfold beq; intros.  destruct (eq_dec x y). auto. congruence.
  Qed.

  Definition ge (x y: t) : Prop := VS.Subset y x.

  Lemma ge_refl: forall x y, eq x y -> ge x y.
  Proof.
    intros x y H v.
    rewrite (H v); auto.
  Qed.
  Lemma ge_trans: forall x y z, ge x y -> ge y z -> ge x z.
  Proof.
    unfold ge; intros x y z H1 H2 r.
    generalize (H1 r); generalize (H2 r).
    intuition.
  Qed.
  Lemma ge_compat: forall x x' y y', eq x x' -> eq y y' -> ge x y -> ge x' y'.
  Proof.
    unfold eq, ge; intros.
    intros r. 
    rewrite <- (H r).
    rewrite <- (H0 r).
    auto.
  Qed.
  Definition bot := VS.empty.

  Lemma ge_bot: forall x, ge x bot.
  Proof.
    unfold ge, bot.
    intros x r H.
    elim (VS.empty_spec H).
  Qed.

  Definition lub (x y: t) : t := VS.union x y.

  Lemma ge_lub_left: forall x y, ge (lub x y) x.
  Proof.
    unfold lub; intros x y r.
    rewrite VS.union_spec; auto.
  Qed.
  Lemma ge_lub_right: forall x y, ge (lub x y) y.
  Proof.
    unfold lub; intros x y r.
    rewrite VS.union_spec; auto.
  Qed.

End D.

Module DS := Dataflow_Solver(D)(NodeSetForward).

Definition transfer (f: function) 
  (Xh:stealth_var)
  (PT: node -> Op.addressing -> list mreg -> option pointsto_symbol)
  (pc: node) (vl: D.t) : D.t := 
  match (fn_code f)!pc with
    | Some (MIop op args dest s) => (vl#(VReg dest) <- (vl##t args))
    | Some (MIload chunk addr args dest s) =>
      match PT pc addr args with
        | Some (Global id) =>
          vl#(VReg dest) <- (vl#(VMem (Global id)) ∪ vl##t args)
        | Some (Stack ofs) =>
          vl#(VReg dest) <- (tload vl chunk ofs)
        | None => vl
      end
    | Some (MIstore chunk addr args src s) => 
      match PT pc addr args with
        | Some (Global id) => 
          vl#(VMem (Global id)) <- (vl#(VMem (Global id)) ∪ (vl#(VReg src)) ∪ vl##t args)
        | Some (Stack ofs) => 
          tstore vl chunk ofs (vl#(VReg src))
        | None => vl
      end
    | Some (MImemcpy_8_4 dst src s) => 
      let pt_reg reg := PT pc (Aindexed Int.zero) (reg::nil)  in
      match pt_reg dst, pt_reg src with
        | Some (Stack ofs_dst), Some (Stack ofs_src) => 
          tstore
            (tstore vl Mint32 (ofs_dst +4) (tload vl Mint32 (ofs_src + 4)))
            Mint32 ofs_dst (tload vl Mint32 ofs_src)                
        | _ , _  => vl
      end
    | _ => vl
  end.

Definition build_global_aux (f: function) 
  (PT: node -> Op.addressing -> list mreg -> option pointsto_symbol)
  (vl: PMap.t D.t)
  (Xh:stealth_var)
  (pc: node) (ins:instruction) : stealth_var := 
  match ins with
    | MIstore chunk addr args src s => 
      match PT pc addr args with
        | Some (Global id) => 
          let vl := PMap.get pc vl in
          match vl#(VReg src) ∪ vl##t args with
            | High => Regset.add id Xh
            | Low => Xh
          end
        | _ => Xh
      end
    | _ => Xh
  end.

Definition build_global (f: function) 
  (PT: node -> Op.addressing -> list mreg -> option pointsto_symbol)
  (vl: PMap.t D.t) (Xh:stealth_var) : stealth_var := 
  PTree.fold (build_global_aux f PT vl) (fn_code f) Xh.

Definition init (Xh:stealth_var) : D.t :=
  Regset.fold (fun id vl => VS.add (VMem (Global id)) vl) Xh VS.empty.

Definition run (p:Mach.program) (f:function) 
           (Xh:stealth_var)
           (PT: node -> Op.addressing -> list mreg -> option pointsto_symbol)
: option (stealth_var * PMap.t D.t) :=
  match DS.fixpoint (fn_code f) successors (transfer f Xh PT) 
                    ((fn_entrypoint f, init Xh) :: nil) with
    | Some vl => 
      Some (build_global f PT vl Xh,vl)  
    | None => None
  end.

End Inference.

Section type_system.
Variable Xs:stealth_var.
Variable Xh:Regset.t.
Variable f:function.
Variable p:Mach.program.
Let ge: Mach.genv := (Genv.globalenv p).
Let pt: PMap.t D.t := analyze p f.
Let PT: node -> Op.addressing -> list mreg -> option pointsto_symbol
  := query pt.                                                     

Variable spb: block.

Inductive tstep : node -> var_level -> var_level -> Prop :=
| TR_op : forall vl i op args dest s
  (CODE:(fn_code f)!i = Some (MIop op args dest s)),
  tstep i vl (vl#(VReg dest) <- (vl##t args))

| TR_load_global : forall vl i chunk addr args dest s id
  (CODE:(fn_code f)!i = Some (MIload chunk addr args dest s))
  (HPT:PT i addr args = Some (Global id))
  (HXs:((vl##t args = High) -> Regset.In id Xs)),
  tstep i vl vl#(VReg dest) <- ((get_global_level Xh id) ∪ vl##t args)

| TR_load_stack : forall vl i chunk addr args dest s ofs
  (CODE:(fn_code f)!i = Some (MIload chunk addr args dest s))
  (HPT:PT i addr args = Some (Stack ofs)),
  tstep i vl vl#(VReg dest) <- (tload vl chunk ofs)

| TR_store_global : forall vl i chunk addr args src s id
  (CODE:(fn_code f)!i = Some (MIstore chunk addr args src s))
  (HPT:PT i addr args = Some (Global id))
  (HXs:((vl##t args = High) -> Regset.In id Xs))
  (Hle:(sec_le (vl#(VReg src) ∪ vl##t args) (get_global_level Xh id))),
  tstep i vl vl

| TR_store_stack : forall vl i chunk addr args src s ofs
  (CODE:(fn_code f)!i = Some (MIstore chunk addr args src s))
  (HPT:PT i addr args = Some (Stack ofs)),
  tstep i vl (tstore vl chunk ofs (vl#(VReg src)))

| TR_memcpy: forall vl i dst src pc' ofs_dst ofs_src
  (CODE:(fn_code f)!i = Some (MImemcpy_8_4 dst src pc'))
  (PT1:PT i (Aindexed Int.zero) (dst::nil) = Some (Stack ofs_dst))
  (PT2:PT i (Aindexed Int.zero) (src::nil) = Some (Stack ofs_src)),
  tstep i vl (tstore
                 (tstore vl Mint32 (ofs_dst +4) (tload vl Mint32 (ofs_src + 4)))
                 Mint32 ofs_dst (tload vl Mint32 ofs_src))

| TR_goto: forall vl i a1
  (CODE:(fn_code f)!i = Some (MIgoto a1)),
  tstep i vl vl
| TR_cond: forall vl i cond args pc1 pc2
  (CODE:(fn_code f)!i = Some (MIcond cond args pc1 pc2))
  (LOW_BRANCH:vl##t args = Low),
  tstep i vl vl
| TR_return: forall vl i
  (CODE:(fn_code f)!i = Some (MIreturn)),
  tstep i vl vl.

Definition sec_le_dec (s1 s2:sec_level) : {sec_le s1 s2}+{~ sec_le s1 s2}.
Proof.
  destruct s1; destruct s2; try (left; go; fail); right.
  destruct 1; congruence.
Qed.

Definition exec_tstep (ge:Mach.genv) (pt:PMap.t D.t)
           (i:node) (vl:var_level) : option var_level :=
  match (fn_code f)!i with
    | Some (MIop op args dest s) => Some (vl#(VReg dest) <- (vl##t args))
    | Some (MIload chunk addr args dest s) =>
      match query pt i addr args with
        | Some (Global id) =>
          let t_args := vl##t args in
          let vl' := vl#(VReg dest) <- ((get_global_level Xh id) ∪ t_args) in
          match t_args with
            | High => if  Regset.mem id Xs then Some vl' else None
            | _ => Some vl'
          end
        | Some (Stack ofs) =>
          Some vl#(VReg dest) <- (tload vl chunk ofs)
        | None => None
      end
    | Some (MIstore chunk addr args src s) => 
      match query pt i addr args with
        | Some (Global id) =>
          let t_args := vl##t args in
          if sec_le_dec (vl#(VReg src) ∪ vl##t args) (get_global_level Xh id) then
            match t_args with
              | High => if  Regset.mem id Xs then Some vl else None
              | _ => Some vl
            end
          else None
        | Some (Stack ofs) => Some (tstore vl chunk ofs (vl#(VReg src)))
        | None => None
      end
    | Some (MImemcpy_8_4 dst src _) => 
      let pt_reg reg := query pt i (Aindexed Int.zero) (reg::nil)  in
      match pt_reg dst, pt_reg src with
        | Some (Stack ofs_dst), Some (Stack ofs_src) => 
          Some (tstore
                  (tstore vl Mint32 (ofs_dst +4) (tload vl Mint32 (ofs_src + 4)))
                  Mint32 ofs_dst (tload vl Mint32 ofs_src))
        | _ , _  => None
      end

    | Some (MIgoto a1) => Some vl
    | Some (MIcond cond args pc1 pc2) =>
      match vl##t args with
        | Low => Some vl
        | High => None
      end
    | Some (MIreturn) => Some vl
    | _ => None 
  end.

Definition typecheck0 (ge:Mach.genv) (pt:PMap.t D.t) (vl:PMap.t var_level) : bool :=
  ptree_forall (fn_code f)
               (fun i ins => 
                    let vl_i := PMap.get i vl in
                    match exec_tstep ge pt i vl_i with
                      | None => false
                      | Some vl' => 
                        List.forallb
                          (fun j => VS.subset vl' (PMap.get j vl))
                          (successors ins)
                    end).

Lemma exec_tstep_correct: forall i vl1 vl2,
  exec_tstep ge pt i vl1 = Some vl2 ->
  tstep i vl1 vl2.
Proof.
  unfold exec_tstep; intros.
  flatten H; go.
  - rewrite <- Eq3; go.
  - rewrite <- Eq3; econstructor; eauto.
    intros.
    apply Regset.mem_2; auto.
  - econstructor; eauto; intros.
    * congruence.
    * congruence.
  - econstructor; eauto.
    * intros; apply Regset.mem_2; auto.
    * congruence.
Qed.

Definition indist_regset (vl:var_level) (rs1 rs2:regset) : Prop :=
  forall r, vl#(VReg r) = Low -> (Regmap.get r rs1) = (Regmap.get r rs2).

Definition equivalent_block (m1 m2:mem) (b:block) : Prop :=
  PMap.get b (Mem.mem_contents m1) = PMap.get b (Mem.mem_contents m2).

Definition indist_mem_global (m1 m2:mem) : Prop :=
  forall (id:ident) (b:block),
    get_global_level Xh id = Low ->
    Genv.find_symbol ge id = Some b ->
    equivalent_block m1 m2 b.

Definition indist_mem_stack (vl:var_level) (m1 m2:mem) : Prop :=
 forall ofs,
   vl#(VMem (Stack ofs)) = Low ->
   ZMap.get ofs (PMap.get spb (Mem.mem_contents m1)) 
   = ZMap.get ofs (PMap.get spb (Mem.mem_contents m2)).


(*
Definition indist_mem (vl:var_level) (m1 m2:mem) : Prop :=
  forall x a chunk v1 v2,
    gamma_val ge spb (Symb x) a ->
    vl#(VMem x) = Low ->
    Mem.loadv chunk m1 a = Some v1 ->
    Mem.loadv chunk m2 a = Some v2 -> 
    v1=v2.
*)

Let sp := Vptr spb Int.zero.

Inductive indist (vl:var_level) : state -> state -> Prop :=
| indist_def: forall pc rs1 rs2 m1 m2
  (sp_not_symbol: forall i,
    Genv.find_symbol ge i <> Some spb)
  (INDR: indist_regset vl rs1 rs2) 
  (INDMG: indist_mem_global m1 m2)
  (INDMS: indist_mem_stack vl m1 m2)
  (INVM: forall b, PMap.get b (Mem.mem_access m1) = PMap.get b (Mem.mem_access m2))
  (HPT1:gamma_estate ge f pt (NState f sp (State pc rs1 m1)))
  (HPT2:gamma_estate ge f pt (NState f sp (State pc rs2 m2))),
  indist vl (State pc rs1 m1) (State pc rs2 m2).

Lemma indist_monotone : forall vl1 vl2 s1 s2,
  indist vl1 s1 s2 ->
  VS.subset vl1 vl2 = true ->
  indist vl2 s1 s2.
Proof.
  intros vl1 vl2 s1 s2 Hi Hs.
  rewrite VS.subset_spec in Hs.
  inv Hi; constructor; eauto.
  - intros r Hr; apply INDR.
    unfold get_var_level in *.
    flatten Hr; flatten.
    rewrite VS.mem_spec in *.
    apply Hs in Eq0.
    rewrite <- VS.mem_spec in Eq0.
    congruence. 
  - intros ofs Ho.
    apply INDMS.
    unfold get_var_level in *.
    flatten Ho; flatten.
    rewrite VS.mem_spec in *.
    apply Hs in Eq0.
    rewrite <- VS.mem_spec in Eq0.
    congruence. 
Qed.

Lemma indist_mem_upd_reg: forall vl m1 m2 r t,
  indist_mem_stack vl m1 m2 ->
  indist_mem_stack vl # (VReg r) <- t m1 m2.
Proof.
  intros vl m1 m2 r t H.
  unfold indist_mem_stack in *.
  intros ofs Ho.
  apply H.
  unfold get_var_level in *.
  flatten Ho.
  assert (V:~ VS.mem (VMem (Stack ofs)) vl # (VReg r) <- t = true) by congruence.
  rewrite VS.mem_spec in V.
  unfold update_var_level in V.
  flatten.
  flatten V.
  - rewrite VS.remove_spec in *.
    rewrite VS.mem_spec in *.
    intuition congruence.
  - rewrite VS.add_spec in *.
    rewrite VS.mem_spec in *.
    intuition congruence.
Qed.
Hint Resolve indist_mem_upd_reg.

Remark false_not_true: forall b,
  b = false -> ~ b = true.
Proof.
  congruence.
Qed.

Lemma update_set_var_level: forall vl x y t,
  (vl # x <- t) # y = if VS.E.eq_dec x y then t else vl#y.
Proof.
  intros vl x y t.
  unfold get_var_level, update_var_level.
  flatten; flatten Eq; subst;
    try rewrite VS.mem_spec in *.
  - rewrite VS.remove_spec in *.
    intuition congruence.
  - apply false_not_true in Eq.
    rewrite VS.mem_spec in *.
    rewrite VS.add_spec in Eq.
    intuition congruence.
  - apply false_not_true in Eq0.
    rewrite VS.mem_spec in *.
    rewrite VS.remove_spec in Eq.
    intuition congruence.
  - rewrite VS.add_spec in *.
    apply false_not_true in Eq0.
    rewrite VS.mem_spec in *.
    intuition congruence.
  - apply false_not_true in Eq.
    rewrite VS.mem_spec in *.
    rewrite VS.remove_spec in Eq.
    intuition congruence.
  - apply false_not_true in Eq.
    rewrite VS.mem_spec in *.
    rewrite VS.add_spec in *.
        intuition congruence.
Qed.

Lemma indist_regset_set : forall vl r t rs1 rs2 v1 v2,
  indist_regset vl rs1 rs2 ->
  (t=Low -> v1=v2) ->
  indist_regset (vl# (VReg r) <- t) (Regmap.set r v1 rs1) (Regmap.set r v2 rs2).
Proof.
  intros vl r0 t rs1 rs2 v1 v2 H Hv r Hl.
  rewrite update_set_var_level in Hl. 
  flatten Hl.
  - repeat rewrite Regmap.gsspec; flatten; go.
  - repeat rewrite Regmap.gsspec; flatten; auto.
Qed.


Lemma regs_all_Low: forall vl l x,
  vl ##t l = Low ->
  In x l ->
  vl # (VReg x) = Low.
Proof.
  induction l; simpl; intuition.
  - subst; auto.
    destruct (VS.mem (VReg x) vl) eqn:E; try congruence.
    unfold get_var_level; rewrite E; auto.
  - apply IHl; auto.
    destruct (VS.mem (VReg a) vl) eqn:E; try congruence.
Qed.
Hint Resolve regs_all_Low.

(* Hint Resolve indist_mem_upd_reg. *)

Lemma indist_regset_maps: forall vl rs1 rs2 args,
  indist_regset vl rs1 rs2 ->
  vl##t args = Low ->
  rs1##args = rs2##args.
Proof.
  induction args; simpl; auto; intros.
  flatten H0.
  rewrite IHargs; auto.
  generalize (H a).
  unfold Regmap.get.
  intros T; rewrite T; auto.
  unfold get_var_level; rewrite Eq; auto.
Qed.

Lemma lub_low: forall x y,
  x ∪ y = Low -> x = Low /\ y = Low.
Proof.
  destruct x; destruct y; simpl; split; congruence.
Qed.

Lemma load_equivalent_block: forall m1 m2 chunk b ofs v1 v2,
  Mem.load chunk m2 b ofs = Some v2 ->
  Mem.load chunk m1 b ofs = Some v1 ->
  equivalent_block m1 m2 b ->
  v1 = v2.
Proof.
  intros m1 m2 chunk b ofs v1 v2 Hl1 Hl2 Heq.
  unfold equivalent_block in *.
  Transparent Mem.load.
  unfold Mem.load in *.
  flatten Hl1; flatten Hl2.
  Opaque Mem.load.
Qed.

Lemma sec_le_low: forall l1 l2,
  sec_le l1 l2 -> l2 = Low -> l1 = Low.
Proof.
  destruct 1; intuition congruence.
Qed.

Lemma store_equivalent_block: forall m1 m2 b0 m1' m2' chunk b ofs1 ofs2 v1 v2,
  equivalent_block m1 m2 b0 ->
  (b=b0 -> ofs1=ofs2 /\ v1=v2) ->
  Mem.store chunk m2 b ofs2 v2 = Some m2' ->
  Mem.store chunk m1 b ofs1 v1 = Some m1' ->
  equivalent_block m1' m2' b0.
Proof.
  intros m1 m2 b0 m1' m2' chunk b ofs1 ofs2 v1 v2 Hyp Hyp3 Hyp0 Hyp1.
  Transparent Mem.store.
  unfold equivalent_block in *.
  unfold Mem.store in *.
  destruct (peq b0 b); subst.
  - flatten Hyp0; flatten Hyp1.
    simpl.
    destruct Hyp3; subst; auto.
    repeat rewrite PMap.gss.
    congruence.
  - flatten Hyp0; flatten Hyp1.
    simpl.
    repeat rewrite PMap.gso; auto.
  Opaque Mem.store.  
Qed.

Lemma storev_same_mem_access : forall chunk m a v m',
  Mem.storev chunk m a v = Some m' ->
  Mem.mem_access m = Mem.mem_access m'.
Proof.
  intros.
  Transparent Mem.store.
  destruct a; inv H.
  unfold Mem.store in H1; flatten H1.
  simpl.
  auto.
Qed.

Inductive mevent' := ME (me:mevent) | StealthWrite | StealthRead.

Definition eff (e:mevent) : mevent' :=
  match e with
    | MRead (Vptr b ofs) =>
      match Genv.invert_symbol ge b with
          | Some id => 
            if Regset.mem id Xs
            then StealthRead
            else ME e
          | None => ME MSilent
      end
    | MWrite (Vptr b ofs) =>
      match Genv.invert_symbol ge b with
          | Some id => 
            if Regset.mem id Xs
            then StealthWrite
            else ME e
          | None => ME MSilent
      end
    | _ => ME MSilent
  end.

Lemma tstore_low_case : forall chunk vl ofs0 t ofs1,
 (tstore vl chunk ofs0 t) # (VMem (Stack ofs1)) = Low ->
 (vl # (VMem (Stack ofs1)) = Low /\ (ofs1 < ofs0 \/ ofs0 + Z_of_nat (size_chunk_nat chunk) <= ofs1))
 \/
 (t = Low /\ ofs0 <= ofs1 < ofs0 + Z_of_nat (size_chunk_nat chunk)).
Proof.
  unfold tstore.
  intros chunk; generalize (size_chunk_nat chunk); clear chunk.
  induction n; simpl; intros.
  - left; split; auto; omega.
  - exploit IHn; eauto.
    destruct 1 as [[T1 T2]|[T1 T2]].
    + rewrite update_set_var_level in T1; flatten T1.
      * right; split; auto; split; try omega.
        zify; omega.
      * left; split; auto.
        assert (ofs0<>ofs1) by congruence.
        destruct T2; try omega.
        right; zify; omega.
    + right; split; auto.
      zify; omega.
Qed.

Lemma get_setN : forall l ofs0 ofs1 m,
  (ZMap.get ofs1 (Mem.setN l ofs0 m) = ZMap.get ofs1 m /\ 
   (ofs1 < ofs0 \/ ofs0 + Z_of_nat (length l) <= ofs1))
  \/
  (exists n,
     nth_error l n = Some (ZMap.get ofs1 (Mem.setN l ofs0 m)) /\
     Z_of_nat n = (ofs1 - ofs0) /\
     (ofs0 <= ofs1 < ofs0 + Z_of_nat (length l))).
Proof.
  induction l; simpl; intros.
  - left; split; auto.
    omega.
  - destruct (IHl (ofs0+1) ofs1 (ZMap.set ofs0 a m)) as [[T1 T2]|(n & T1 & T2 & T3)].
    + rewrite T1.      
      destruct (Z_eq_dec ofs0 ofs1).
      * subst.
        right.
        exists O.
        repeat split; auto; try (zify; omega).
        rewrite ZMap.gss; auto.
      * rewrite ZMap.gso; auto.
        left; split; auto.
        zify; omega.
    + right.
      exists (S n); split; auto.
      zify; omega.
Qed.

Lemma load_low_tload: forall vl m1 m2 chunk ofs v v0,
  indist_mem_stack vl m1 m2 ->
  tload vl chunk ofs = Low ->
  Mem.load chunk m2 spb ofs = Some v0 ->
  Mem.load chunk m1 spb ofs = Some v ->
  v = v0.
Proof.
  intros vl m1 m2 chunk ofs v v0 H.
  Transparent Mem.load.
  unfold tload, Mem.load.
  flatten.
  assert (forall n ofs,
            max_levels (tloadN n ofs vl) = Low ->
            (Mem.getN n ofs (Mem.mem_contents m2) !! spb) =
            (Mem.getN n ofs (Mem.mem_contents m1) !! spb)).
    induction n; simpl; try congruence.
    intros ofs1 Hl.
    exploit lub_low; eauto.
    destruct 1.
    rewrite IHn; auto.
    f_equal.
    rewrite H; auto.
  intro.
  rewrite H0; auto.
  congruence.
Qed.

Lemma indist_regset_tstore : forall rs1 rs2 chunk ofs t vl,
  indist_regset vl rs1 rs2 ->
  indist_regset (tstore vl chunk ofs t) rs1 rs2.
Proof.
  unfold tstore.
  intros rs1 rs2 chunk.
  generalize (size_chunk_nat chunk).
  induction n; simpl; auto.
  intros.
  apply IHn.
  intro r.
  rewrite update_set_var_level; flatten.
  eauto.
Qed.

Lemma indist_step : forall st1 st2 a1 a2 st1' st2' vl vl',
  estep ge (NState f sp st1) a1 (NState f sp st1') ->
  estep ge (NState f sp st2) a2 (NState f sp st2') ->
  tstep (get_pc st1) vl vl' ->
  indist vl st1 st2 ->
  indist vl' st1' st2'.
Proof.
  intros st1 st2 a1 a2 st1' st2' vl vl' Hs1 Hs2 Ht Hi.
  assert (Hes1:=Hs1); inv Hs1; clear EVENT;
  assert (Hes2:=Hs2); inv Hs2; clear EVENT;
  inv Hi; inv STEP; inv STEP0; try congruence; simpl get_pc;
  simpl in Ht; inv Ht; try congruence;
  repeat match goal with
   | h1: (fn_code f)!?pc = Some ?i1,
     h2: (fn_code f)!?pc = Some ?i2 |- _ =>
     rewrite h1 in h2; inv h2
  end.
  - constructor; [auto|idtac|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
    apply indist_regset_set; auto.
    intros Hargs_low.
    unfold eval_operation, eval_condition in *.
    assert (Mem.valid_pointer m1 = Mem.valid_pointer m2).
    + apply extensionality.
      intros b; apply extensionality.
      intros ofs.
      unfold Mem.valid_pointer.
      repeat destruct Mem.perm_dec; auto.
      * unfold Mem.perm in *.
        elim n; rewrite <- INVM; auto.
      * unfold Mem.perm in *.
        elim n; rewrite INVM; auto. 
    + flatten H7;
      try (exploit indist_regset_maps; eauto; intro);
      unfold RegEq.t in *;
      try f_equal;
      try congruence.
      assert (i0=i1) by congruence; flatten.
      assert (i0=i1) by congruence; flatten.
  - constructor; [auto|idtac|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
    apply indist_regset_set; auto.
    intros Hargs_low.
    edestruct lub_low; eauto.
    try (exploit indist_regset_maps; eauto; intro).
    unfold RegEq.t in *;
    assert (a0=a) by congruence; subst.
    exploit query_correct; eauto; intros Hga.
    destruct a; inv H9; inv H6.
    inv Hga; try congruence.
    exploit INDMG; eauto; intros HEquiv.
    eapply load_equivalent_block; eauto.
  - constructor; [auto|idtac|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
    apply indist_regset_set; auto.
    try (exploit indist_regset_maps; eauto; intro).
    exploit query_correct; eauto; intros Hga.
    clear HPT2.
    exploit query_correct; eauto; intros Hga0.
    destruct a0; inv H9.
    destruct a; inv H6.
    inv Hga; inv Hga0; try congruence.
    intros.
    assert (i=i0).
        rewrite <- (Int.repr_unsigned i).
        rewrite <- (Int.repr_unsigned i0).
        congruence.
    subst.    
    eapply load_low_tload; eauto.
  - constructor; [auto|idtac|idtac|idtac|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
    + intros r Hr; apply INDR. 
      auto.
    + intros id0 b Hl Hb.
      exploit query_correct; eauto; intros Hga0.
      clear HPT2.
      exploit query_correct; eauto; intros Hga.
      destruct a; inv H6.
      destruct a0; inv H9.
        assert (b1=b0) by
          (inv Hga; inv Hga0; congruence); subst.      
      eapply store_equivalent_block with (3:=H1) (4:=H0); eauto.
      intros; subst.
      assert (id0=id).
        inv Hga; try congruence.
        exploit Genv.find_invert_symbol; eauto.
        exploit (Genv.find_invert_symbol ge id0); eauto.
        congruence.
      subst.  
      exploit sec_le_low; eauto; intros Hl1.
      edestruct lub_low as [L1 L2]; eauto; clear Hl1.
      exploit indist_regset_maps; eauto; intro.
      unfold RegEq.t in *;
        split; try congruence.
      eapply INDR; eauto.
    + intros ofs Ho.
      exploit query_correct; eauto; intros Hga0.
      clear HPT2.
      exploit query_correct; eauto; intros Hga.
      destruct a; inv H6.
      destruct a0; inv H9.
      inv Hga; inv Hga0; try congruence.
      Transparent Mem.store.
      unfold Mem.store in H0, H1.
      flatten H1; flatten H0.
      simpl in *.
      rewrite PMap.gso; try congruence.
      rewrite PMap.gso; try congruence.
      apply INDMS; auto.
    + intros b.
      exploit (storev_same_mem_access chunk1 m1); eauto.
      exploit (storev_same_mem_access chunk1 m2); eauto.
      generalize (INVM b); congruence.
  - constructor; [auto|idtac|idtac|idtac|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
    + apply indist_regset_tstore; auto.
    + intros id0 b Hl Hb.
      exploit query_correct; eauto; intros Hga0.
      clear HPT2.
      exploit query_correct; eauto; intros Hga.
      destruct a; inv H6.
      destruct a0; inv H9.
        assert (b1=b0) by
          (inv Hga; inv Hga0; congruence); subst.            
      eapply store_equivalent_block with (3:=H1) (4:=H0); eauto.
      intros.
      inv Hga; congruence.
    + intros ofs1 Ho.
      exploit query_correct; eauto; intros Hga0.
      clear HPT2.
      exploit query_correct; eauto; intros Hga.
      destruct a; inv H6.
      destruct a0; inv H9.
      inv Hga; inv Hga0; try congruence.
      Transparent Mem.store.
      unfold Mem.store in H0, H1.
      flatten H1; flatten H0.
      simpl in *.
      rewrite PMap.gss.
      rewrite PMap.gss.
      assert (i=i0).
        rewrite <- (Int.repr_unsigned i).
        rewrite <- (Int.repr_unsigned i0).
        congruence.
      subst.
      exploit tstore_low_case; eauto.
      destruct 1 as [(T1 & T2)|(T1 & T2)].
      * rewrite <- (encode_val_length chunk1 (rs1 src1)) in T2.
        destruct (get_setN (encode_val chunk1 (rs1 src1)) (Int.unsigned i0) ofs1 
                           ((Mem.mem_contents m1) !! spb)) as [(V1 & V2)|(n & V1 & V2 & V3)];
          [idtac|apply False_ind; omega].
        rewrite V1.
        destruct (get_setN (encode_val chunk1 (rs2 src1)) (Int.unsigned i0) ofs1 
                           ((Mem.mem_contents m2) !! spb)) as [(W1 & W2)|(n & W1 & W2 & W3)];
          [idtac|apply False_ind; try omega].
        rewrite W1.
        auto.
        rewrite encode_val_length in *.
        omega.
      * exploit (INDR src1); eauto.
        unfold Regmap.get; intros.
        destruct (get_setN (encode_val chunk1 (rs1 src1)) (Int.unsigned i0) ofs1 
                           ((Mem.mem_contents m1) !! spb)) as [(V1 & V2)|(n & V1 & V2 & V3)];
          [rewrite encode_val_length in *; apply False_ind; try omega|idtac].
        destruct (get_setN (encode_val chunk1 (rs2 src1)) (Int.unsigned i0) ofs1 
                           ((Mem.mem_contents m2) !! spb)) as [(W1 & W2)|(n' & W1 & W2 & W3)];
          [rewrite encode_val_length in *; apply False_ind; try omega|idtac].
        assert (n=n') by (zify; omega).
        congruence.
    + intros b.
      exploit (storev_same_mem_access chunk1 m1); eauto.
      exploit (storev_same_mem_access chunk1 m2); eauto.
      generalize (INVM b); congruence.
  - constructor; [auto|auto|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
  - constructor; [auto|auto|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
  - unfold eval_condition in H7, H5.
    assert (Mem.valid_pointer m1 = Mem.valid_pointer m2).
    + apply extensionality.
      intros b; apply extensionality.
      intros ofs.
      unfold Mem.valid_pointer.
      repeat destruct Mem.perm_dec; auto.
      * unfold Mem.perm in *.
        elim n; rewrite <- INVM; auto.
      * unfold Mem.perm in *.
        elim n; rewrite INVM; auto. 
    + flatten H7;
      try (exploit indist_regset_maps; eauto; intro);
      unfold RegEq.t in *;
      try f_equal;
      try congruence.
  - unfold eval_condition in H7, H5.
    assert (Mem.valid_pointer m1 = Mem.valid_pointer m2).
    + apply extensionality.
      intros b; apply extensionality.
      intros ofs.
      unfold Mem.valid_pointer.
      repeat destruct Mem.perm_dec; auto.
      * unfold Mem.perm in *.
        elim n; rewrite <- INVM; auto.
      * unfold Mem.perm in *.
        elim n; rewrite INVM; auto. 
    + flatten H7;
      try (exploit indist_regset_maps; eauto; intro);
      unfold RegEq.t in *;
      try f_equal;
      try congruence.
  - constructor; [auto|auto|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
  - constructor; [auto|auto|auto|auto|auto|
                  apply analyze_estep_postfix with (2:=Hes1); eauto|
                  apply analyze_estep_postfix with (2:=Hes2); eauto].
Qed.

Lemma indist_step_same_actions : forall st1 st2 a1 a2 st1' st2' vl vl',
  estep ge (NState f sp st1) a1 (NState f sp st1') ->
  estep ge (NState f sp st2) a2 (NState f sp st2') ->
  tstep (get_pc st1) vl vl' ->
  indist vl st1 st2 ->
  eff a1 = eff a2.
Proof.
  intros st1 st2 a1 a2 st1' st2' vl vl' Hs1 Hs2 Ht Hi.
  inv Hs1; 
  inv Hs2; 
  inv Hi; inv STEP; inv STEP0; try congruence;
  inv EVENT; try congruence;
  inv EVENT0; try congruence;
  simpl in Ht; inv Ht; try congruence;
  repeat match goal with
   | h1: (fn_code f)!?pc = Some ?i1,
     h2: (fn_code f)!?pc = Some ?i2 |- _ =>
     rewrite h1 in h2; inv h2
  end.
  - exploit query_correct; eauto; intros Hga0.
    clear HPT2.
    exploit query_correct; eauto; intros Hga.
    destruct a; try congruence.
    destruct a0; try congruence.
    assert (b2=b0) by congruence; subst.
    assert (b1=b) by congruence; subst.
    inv Hga; inv Hga0; try congruence.
    simpl.
    apply Genv.find_invert_symbol in G1.
    apply Genv.find_invert_symbol in G0.
    flatten.
    destruct (vl ##t args3) eqn:E.
    +  try (exploit indist_regset_maps; eauto; intro).
       unfold RegEq.t in *;
       congruence.
    + exploit HXs; auto; intros.
      apply Regset.mem_1 in H.
      congruence.
  - exploit query_correct; eauto; intros Hga0.
    clear HPT2.
    exploit query_correct; eauto; intros Hga.
    destruct a; try congruence.
    destruct a0; try congruence.
    assert (b2=b0) by congruence; subst.
    assert (b1=b) by congruence; subst.
    inv Hga; inv Hga0; try congruence.
    simpl.
    flatten.
    assert (i=i0).
      rewrite <- (Int.repr_unsigned i).
      rewrite <- (Int.repr_unsigned i0).
      congruence.
    congruence.
  - exploit query_correct; eauto; intros Hga0.
    clear HPT2.
    exploit query_correct; eauto; intros Hga.
    destruct a; try congruence.
    destruct a0; try congruence.
    assert (b2=b0) by congruence; subst.
    assert (b1=b) by congruence; subst.
    inv Hga; inv Hga0; try congruence.
    simpl.
    apply Genv.find_invert_symbol in G1.
    apply Genv.find_invert_symbol in G0.
    flatten.
    destruct (vl' ##t args3) eqn:E.
    +  try (exploit indist_regset_maps; eauto; intro).
       unfold RegEq.t in *;
       congruence.
    + exploit HXs; auto; intros.
      apply Regset.mem_1 in H.
      congruence.
  - exploit query_correct; eauto; intros Hga0.
    clear HPT2.
    exploit query_correct; eauto; intros Hga.
    destruct a; try congruence.
    destruct a0; try congruence.
    assert (b2=b0) by congruence; subst.
    assert (b1=b) by congruence; subst.
    inv Hga; inv Hga0; try congruence.
    simpl.
    flatten.
    assert (i=i0).
      rewrite <- (Int.repr_unsigned i).
      rewrite <- (Int.repr_unsigned i0).
      congruence.
    congruence.
Qed.

Variable vl : PMap.t var_level.
Variable typecheck_true: typecheck0 ge pt vl = true.

Lemma indist_step_typecheck : forall st1 st2 a1 a2 st1' st2',
  estep ge (NState f sp st1) a1 (NState f sp st1') ->
  estep ge (NState f sp st2) a2 (NState f sp st2') ->
  indist (PMap.get (get_pc st1) vl) st1 st2 ->
  indist (PMap.get (get_pc st1') vl) st1' st2' /\ eff a1 = eff a2.
Proof.
  intros.
  assert (Hi:exists ins, (fn_code f)!(get_pc st1) = Some ins).
    inv H; inv STEP; go.
  destruct Hi as [ins Hi].  
  exploit ptree_forall_correct1; eauto.
  simpl.
  flatten.
  apply exec_tstep_correct in Eq.
  rewrite forallb_forall.
  intros.
  split.
  - apply indist_monotone with (vl1:=v).
    * eapply indist_step; eauto.
    * inv H; inv STEP; simpl in Hi;
      try (
          apply H2;
          rewrite Hi in H; inv H; simpl; go; fail).
      simpl in Eq; inv Eq; try congruence.
      simpl.
      apply <- VS.subset_spec.
      intro; auto.
  - eapply indist_step_same_actions; eauto.
Qed.  

End type_system.

Definition typecheck (tp:MachIR.program) (Xh:stealth_var) (Xs:Regset.t) : bool :=
  let (f,p) := tp in
  let pt := analyze p f in
  let PT := query pt in
  let ge := Genv.globalenv p in
  match Inference.run p f Xh PT with
    | None => false
    | Some (Xh',vl) =>
      ptree_forall (fn_code f)
                   (fun i ins =>
                      let vl_i := vl !! i in
                      match exec_tstep Xs Xh' f ge pt i vl_i with
                        | Some vl' =>
                          forallb (fun j => VS.subset vl' vl !! j) (successors ins)
                        | None => false
                      end)
  end.

Lemma indist_mem_global_monotone_Xh : forall p Xh1 Xh2 m1 m2,
  Regset.Subset Xh1 Xh2 ->
  indist_mem_global Xh1 p m1 m2 ->
  indist_mem_global Xh2 p m1 m2.
Proof.
  intros p Xh1 Xh2 m1 m2.
  intros Hs INDMG id b T1 T2.
  apply INDMG with id; auto.
  unfold get_global_level in *.
  flatten T1; flatten.
  apply Regset.mem_2 in Eq0.
  generalize (Hs id Eq0); intros.
  apply Regset.mem_1 in H.
  congruence.
Qed.

Lemma indist_monotone_Xh : forall f p bsp Xh1 Xh2 vl s1 s2,
  Regset.Subset Xh1 Xh2 ->
  indist Xh1 f p bsp vl s1 s2 ->
  indist Xh2 f p bsp vl s1 s2.
Proof.
  intros f p bsp Xh1 Xh2 vl s1 s2 H H0.
  inv H0; constructor; auto.
  eapply indist_mem_global_monotone_Xh; eauto.
Qed.

Lemma indist_step_typecheck_main : forall f p Xh Xs,
  typecheck (f,p) Xh Xs = true -> 
  exists vl, exists Xh',
  Regset.Subset Xh Xh' /\
  forall bsp,
  let ge := Genv.globalenv p in
  let sp := Vptr bsp Int.zero in
  forall st1 st2 a1 a2 st1' st2',
  estep ge (NState f sp st1) a1 (NState f sp st1') ->
  estep ge (NState f sp st2) a2 (NState f sp st2') ->
  indist Xh' f p bsp (PMap.get (get_pc st1) vl) st1 st2 ->
  indist Xh' f p bsp (PMap.get (get_pc st1') vl) st1' st2' /\ 
  eff Xs p a1 = eff Xs p a2.
Proof.
  intros f p Xh Xs H.
  unfold typecheck in H; flatten H.
  exists t.
  exists s.
  assert (Regset.Subset Xh s).
    clear - Eq.
    unfold Inference.run in Eq.
    flatten Eq.
    unfold Inference.build_global.
    rewrite PTree.fold_spec.
    clear Eq0.
    assert (forall pt l Xh1 Xh2,
      Regset.Subset Xh1 Xh2 ->
      Regset.Subset Xh1
                    (fold_left
                       (fun a p0 =>
                            Inference.build_global_aux f pt t a 
                            (fst p0) (snd p0)) l Xh2)).
      induction l; simpl; auto.
      intros.
      apply IHl.
      unfold Inference.build_global_aux.
      flatten; auto.
      intros x X.
      apply Regset.add_2; auto.
    apply H.
    intro; auto.
  split.
  - auto.
  - intros.
    eapply indist_step_typecheck; eauto.
Qed.

Inductive eindist (p:Mach.program) f (Xh:stealth_var) (vl:PMap.t var_level) :
  extended_state -> extended_state -> Prop :=
  | eindist_N : forall bsp st1 st2,
    indist Xh f p bsp (PMap.get (get_pc st1) vl) st1 st2 ->
    eindist p f Xh vl 
            (NState f (Vptr bsp Int.zero) st1) (NState f (Vptr bsp Int.zero) st2)
  | eindist_I : forall m1 m2
    (G1:gamma_estate (Genv.globalenv p) f (analyze p f) (IState f m1))
    (G2:gamma_estate (Genv.globalenv p) f (analyze p f) (IState f m2)),
    indist_mem_global Xh p m1 m2 ->
    Mem.nextblock m1 = Mem.nextblock m2 ->
    (forall b, PMap.get b (Mem.mem_access m1) = PMap.get b (Mem.mem_access m2)) ->
    (forall id b, Genv.find_symbol (Genv.globalenv p) id = Some b ->
                  Mem.valid_block m1 b) ->
    eindist p f Xh vl (IState f m1) (IState f m2).

Lemma indist_mem_global_storev : forall p Xh chunk m1 a1 v1 m1' m2 a2 v2 m2',
  Mem.storev chunk m1 a1 v1 = Some m1' ->
  Mem.storev chunk m2 a2 v2 = Some m2' ->
  a1=a2 ->
  v1=v2 ->
  indist_mem_global Xh p m1 m2 ->
  indist_mem_global Xh p m1' m2'.
Proof.
  intros.
  subst.
  destruct a2; inv H0; inv H.
  unfold indist_mem_global in *; intros.
  eapply store_equivalent_block; eauto.
Qed.

Lemma indist_mem_global_alloc: forall p Xh m1 m2 a b m1' m2' b1 b2,
  indist_mem_global Xh p m1 m2 ->
  Mem.nextblock m1 = Mem.nextblock m2 ->
  Mem.alloc m1 a b = (m1', b1) ->
  Mem.alloc m2 a b = (m2', b2) ->
  indist_mem_global Xh p m1' m2'.
Proof.
  Transparent Mem.alloc.
  unfold indist_mem_global, equivalent_block; intros.
  unfold Mem.alloc in *.
  inv H1; inv H2; simpl.
  repeat rewrite PMap.gsspec; flatten.
  eauto.
  Opaque Mem.alloc.
Qed.

Lemma indist_mem_stack_alloc : forall vl b a c m1 m1' m2 m2',
  Mem.alloc m1 a c = (m1', b) ->
  Mem.alloc m2 a c = (m2', b) ->
  Mem.nextblock m1 = Mem.nextblock m2 ->
  indist_mem_stack b vl m1' m2'.
Proof.
  Transparent Mem.alloc.
  unfold indist_mem_stack, Mem.alloc; simpl; intros.
  inv H; inv H0; simpl.
  rewrite PMap.gss.
  rewrite PMap.gsspec; flatten.
  Opaque Mem.alloc.
Qed.

Lemma indist_mem_stack_store_stack : forall vl m1 m2 a b c d m1' m2',
  store_stack m1 (Vptr b Int.zero) a c d = Some m1' ->
  store_stack m2 (Vptr b Int.zero) a c d = Some m2' ->
  indist_mem_stack b vl m1 m2 ->
  indist_mem_stack b vl m1' m2'.
Proof.
  unfold store_stack, indist_mem_stack; intros.
  Transparent Mem.store.
  simpl in H, H0.
  unfold Mem.store in H, H0; flatten H; flatten H0; simpl in *.
  rewrite PMap.gsspec; flatten.
  rewrite PMap.gsspec; flatten.
  edestruct get_setN as [(HT1 & HT2)|(n1 & HT1 & HT2 & HT3 & HT4)]; try rewrite HT1.
  - edestruct get_setN as [(HT1' & HT2')|(n1' & HT1' & HT2' & HT3' & HT4')]; try rewrite HT1'.
    * auto.
    * omega.
  - apply sym_eq.
    edestruct get_setN as [(HT1' & HT2')|(n1' & HT1' & HT2' & HT3' & HT4')]; try rewrite HT1'.
    * omega.
    * assert (n1'=n1) by (zify; omega).
      congruence.
Qed.


Lemma eindist_step_typecheck : forall f p Xh Xs,
  let ge := Genv.globalenv p in
  typecheck (f,p) Xh Xs = true -> exists vl, exists Xh',
  Regset.Subset Xh Xh' /\
  forall st1 st2 a1 a2 st1' st2',
  estep ge st1 a1 st1' ->
  estep ge st2 a2 st2' ->
  eindist p f Xh' vl st1 st2 ->
  eindist p f Xh' vl st1' st2' /\ eff Xs p a1 = eff Xs p a2.
Proof.
  intros.
  destruct (indist_step_typecheck_main _ _ _ _ H) as (vl & Xh' & T1 & T2).
  exists vl; exists Xh'; split; auto.
  intros st1 st2 a1 a2 st1' st2' H0 H1 H2.
  generalize H0; intros HH0.
  generalize H1; intros HH1.
  inv H2.
  - destruct st1'; destruct st2'; try (inv H1; fail); try (inv H0; fail).
    unfold ge in *.
    assert (f1=f) by (inv H1; congruence); subst.
    assert (f0=f) by (inv H0; congruence); subst.
    assert (sp0=Vptr bsp Int.zero) by (inv H1; congruence); subst.
    assert (sp=Vptr bsp Int.zero) by (inv H0; congruence); subst.
    edestruct (T2 bsp _ _ _ _ _ _ H0 H1); eauto.
    split; go.
  - inv H1; inv H0; constructor; auto.
    exploit Mem.alloc_result; eauto.
    assert (~Mem.valid_block m1 stk0) by (eapply Mem.fresh_block_alloc; eauto).
    revert SP0.
    exploit Mem.alloc_result; eauto.
    intros; subst.
    rewrite H4.
    econstructor.
    constructor; auto.
    * intros i Hi.
      exploit H6; eauto.
      congruence.
    * simpl. intro. auto.
    * unfold store_stack in *.
      eapply indist_mem_global_storev; eauto; try congruence.
      congruence.
      eapply indist_mem_global_storev; eauto; try congruence.
      congruence.
      eapply indist_mem_global_alloc; eauto.
    * simpl.
      rewrite <- H4 in *.
      apply indist_mem_stack_store_stack with (1:=STORE3) (2:=STORE2); eauto.
      apply indist_mem_stack_store_stack with (1:=STORE0) (2:=STORE1); eauto.
      eapply indist_mem_stack_alloc; eauto.
    * unfold store_stack in *.
      intros b.
      erewrite <- storev_same_mem_access.
      2: eapply STORE3.
      erewrite <- storev_same_mem_access.
      2: eapply STORE0.
      symmetry.
      erewrite <- storev_same_mem_access.
      2: eapply STORE2.
      erewrite <- storev_same_mem_access.
      2: eapply STORE1.
      Transparent Mem.alloc.
      clear STORE0 STORE1 STORE2 STORE3.
      unfold Mem.alloc in *.
      inv SP; inv SP0.
      simpl.
      repeat rewrite PMap.gsspec; flatten.
      Opaque Mem.alloc.
      auto.
    * rewrite <- H4.
      eapply analyze_estep_postfix.
      2: eapply HH0.
      eauto.
    * eapply analyze_estep_postfix; eauto.
Qed.

Lemma eindist_monotone : forall p f Xh1 Xh2 vl st1 st2,
  eindist p f Xh1 vl st1 st2 ->
  Regset.Subset Xh1 Xh2 ->
  eindist p f Xh2 vl st1 st2.
Proof.
  intros.
  inv H; constructor; 
  eauto using
        indist_mem_global_monotone_Xh, indist_monotone_Xh.
Qed.

Module SecureMachIR (S:SCHEDULER) (INT:INTERLEAVING S) <: SECURE_P S INT.

  Import Stealth_action_streams ACTION_STREAMS Stealth_actions STEALTH_ACTIONS 
         Actions Streams TRACES_VIC.

  (* L is the type of the language *)
  Record _L := Make_L {
    prog : MachIR.program;                
    lift_address : val -> vadd;
    lift_address_not_stealth : forall v, C.stealth_va <> lift_address v;
    lift_value : val -> value;
    Xh : stealth_var;
    Xs:stealth_var
  }.
  Definition L := _L.

  (* Lstate is the type of states of the language *)
  Definition Lstate : Type := extended_state.

  Definition eff (p:L) (me:mevent) : mevent' :=
    eff p.(Xs) (snd p.(prog)) me.

  Definition lift_event (p:L) (me:mevent) : effect :=
    match eff p me with
              | ME (MRead a) => EFFRead (p.(lift_address) a)
              | StealthRead => EFFRead C.stealth_va
              | ME (MWrite a) => EFFWrite (p.(lift_address) a) 
              | StealthWrite => EFFWrite C.stealth_va
              | ME MSilent => EFFSilent
            end.

  (* Given a program, and an initial state, return the action stream generated
     by the execution of that program. *)
  CoInductive _get_trace (p:L) : Lstate -> effects -> Prop :=
    | get_trace_step : forall s1 me s2 tr,                         
        _get_trace p s2 tr ->
        estep (Genv.globalenv (snd p.(prog))) s1 me s2 ->
        _get_trace p s1 (Cons (lift_event p me) tr).
  Definition get_trace := _get_trace.

  (* Actions generated by the program execution are osi_actions *)
  (* Lemma get_trace_os_actions : forall P s0 str, *)
  (*   _get_trace P s0 str -> *)
  (*   osi_actions str osi_v. *)
  (* Proof. *)
  (*   intros P; cofix CHyp; intros s0 str Hg. *)
  (*   inv Hg. *)
  (*   constructor; eauto. *)
  (*   unfold lift_event; flatten; simpl; auto. *)
  (* Qed. *)

  (* In program P, states s and s' differ only in values of high
     variables. *)
  Definition low_var_eq (P:L) (s1 s2:Lstate) : Prop :=
    let (f,p) := P.(prog) in
    forall vl,
      eindist p f P.(Xh) vl s1 s2.
  (* FIXME: [vl] not necessary on initial states. *)

  (* A program is secure if starting with the same low variables, the
     sequence of non stealth actions are the same. *)
  Definition secure (P:L) :=
    forall s0 s0' str str',
      low_var_eq P s0 s0' ->
      get_trace P s0 str -> get_trace P s0' str' ->
      same_non_stealth_effects_str str str'.

  Hint Resolve lift_address_not_stealth.

  Lemma transform_secure_aux : forall P Xh' vl,
      (forall (st1 st2 : extended_state) (a1 a2 : mevent)
         (st1' st2' : extended_state),
       estep (Genv.globalenv (snd P.(prog))) st1 a1 st1' ->
       estep (Genv.globalenv (snd P.(prog))) st2 a2 st2' ->
       eindist (snd P.(prog)) (fst P.(prog)) Xh' vl st1 st2 ->
       eindist (snd P.(prog)) (fst P.(prog)) Xh' vl st1' st2' /\
       eff P a1 = eff P a2) ->
      Regset.Subset P.(Xh) Xh' ->
      forall s1 str1,
        _get_trace P s1 str1 -> 
        forall s2 str2,
        eindist (snd P.(prog)) (fst P.(prog)) Xh' vl s1 s2 ->
        _get_trace P s2 str2 ->
        same_non_stealth_effects_str str1 str2.
  Proof.
    intros P Xh' vl Ht Hs.
    cofix CHyp.
    intros s1 str1 Hg1 s2 str2 He Hg2.
    inv Hg1; inv Hg2.
    destruct (Ht _ _ _ _ _ _ H0 H2); auto.
    replace (lift_event P me0) with (lift_event P me).
    - assert (same_non_stealth_effects_str tr tr0).
      eapply CHyp; eauto. 
      assert (same_non_stealth_effects_str (Cons EFFSilent tr)
                                           (Cons EFFSilent tr0)).
      + constructor 2.
        * intro T; inv T.
        * intro T; inv T.
        * auto.
      + destruct me; unfold lift_event; flatten; go.
        * constructor 2; try (intros T; inv T); auto.
        * constructor 2; try (intros T; inv T); auto.
        * constructor 2; try (intros T; inv T); auto.
        * constructor 2; try (intros T; inv T); auto.
    - unfold lift_event; flatten; go.
  Qed.

  Theorem transform_secure : forall P,
      typecheck P.(prog) P.(Xh) P.(Xs) = true ->
      secure P.
  Proof.
    intros P Ht.
    destruct P; simpl in *.
    destruct prog0.
    destruct (eindist_step_typecheck _ _ _ _ Ht) as (vl & Hx' & T1 & T2).
    intros s1 s2 str1 str2 Hi Hg1 Hg2.
    eapply transform_secure_aux with (Xh':=Hx'); eauto.
    - unfold eff; simpl; auto.
      intros.
      exploit (T2 st1 st2); eauto.
    - simpl; auto.
    - simpl.
      unfold low_var_eq in *.
      simpl in *.
      eapply eindist_monotone; eauto.
  Qed.

End SecureMachIR.
