(***********************************************************************
 * File: Machine_mem.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the machine memory. 
 * 
 *************************************************************************)
Require Import Opt_monad Dict_unbound Addressing Page.

Module Type MACHINE_MEM (PG:PAGE).
  Parameter mem : Set.
  Definition key : Set := madd.
  Axiom eq_key : forall (k k':key), {k = k'} + {k <> k'}.

  Parameter get_mem : mem -> key -> option PG.page.
  Parameter add_mem : mem -> key -> PG.page -> mem.
  Parameter remove_mem : mem -> key -> mem.
  Parameter empty_mem : mem.
  Parameter is_empty_mem : mem -> Prop.

  Axiom is_empty_mem_empty : is_empty_mem empty_mem.
  Axiom not_empty_add_mem : forall (m:mem) (k:key) (pg:PG.page),
    ~ is_empty_mem (add_mem m k pg).

  Axiom get_empty_mem : forall (k:key),
    get_mem empty_mem k = None.
  Axiom get_add_mem : forall (m:mem) (k:key) (pg:PG.page),
    Some pg = get_mem (add_mem m k pg) k.

  Axiom get_add_mem_distinct : forall (m:mem) (k k':key) (pg:PG.page),
    k <> k' ->
    get_mem (add_mem m k pg) k' = get_mem m k'.

  
  Axiom get_remove_mem : forall (m:mem) (k:key),
    get_mem (remove_mem m k) k = None.

  Axiom get_remove_mem_distinct : forall (m:mem) (k k':key) (pg:PG.page),
    k <> k' -> get_mem m k = Some pg -> get_mem (remove_mem m k') k = Some pg.

  Axiom get_remove_mem_old : forall (m:mem) (k k':key) (pg:PG.page),
    get_mem (remove_mem m k') k = Some pg -> get_mem m k = Some pg.

End MACHINE_MEM.

Module MACHINE_MEM_REFINED (PG:PAGE) : MACHINE_MEM PG.

  Declare Module D : DICT_UNBOUND
    with Definition key := madd
    with Definition info := PG.page.

  Definition mem := D.dict.

  Definition key : Set := madd.
  Definition eq_key := madd_eq.
  
  Definition get_mem := D.get.
  Definition add_mem := D.set.
  Definition remove_mem := D.remove.
  Definition empty_mem := D.empty.
  Definition is_empty_mem := D.is_empty.

  Definition is_empty_mem_empty := D.is_empty_empty.
  Definition not_empty_add_mem := D.not_empty_set.
  Definition get_empty_mem := D.get_empty.
  Definition get_add_mem := D.get_set.
  Definition get_add_mem_distinct := D.get_set_distinct.
  Definition get_remove_mem := D.get_remove.
  Definition get_remove_mem_distinct := D.get_remove_distinct.
  Definition get_remove_mem_old := D.get_remove_old.

End MACHINE_MEM_REFINED.
  
Module MACHINE_MEM_EXT (PG:PAGE) (MM:MACHINE_MEM PG).
  Import PG.
  Import MM.

  Definition madd_type (m:mem) (ma: madd) : option ptype :=
    pg <- get_mem m ma;
    match page_content pg with
      | RW_c _ => Some RW
      | PT_c _ => Some PT
      | _ => None
    end.

  Definition madd_owner (m:mem) (ma:madd) : page_owner :=
    match get_mem m ma with
      | None => No_Owner
      | Some pg => page_owned_by pg
    end.

  Definition madd_page_table (m:mem) (ma:madd) : option page_table :=
    pg <- get_mem m ma;
    match page_content pg with
      | PT_c pt => Some pt
      | _ => None
    end.

  Definition madd_of_vadd (m:mem) (pt_addr:madd) (va:vadd) : option madd :=
    pt <- madd_page_table m pt_addr;
    PT.get pt va.

  Definition add_page_table_entry (m:mem) (pt_addr:madd)
      (va:vadd) (ma:madd) : option mem :=
    pt_page <- get_mem m pt_addr;
    match page_content pt_page with
      | PT_c pt => 
        Some (add_mem m pt_addr (Page (PT_c (PT.set pt va ma))
                                   (page_owned_by pt_page)
                                   (page_cacheable pt_page)))
      | _ => None
    end.

  Definition remove_page_table_entry (m:mem) (pt_addr:madd) (va:vadd) : option mem :=
    pt_page <- get_mem m pt_addr;
    match page_content pt_page with
      | PT_c pt => 
        Some (add_mem m pt_addr (Page (PT_c (PT.remove pt va))
                                  (page_owned_by pt_page)
                                  (page_cacheable pt_page)))
      | _ => None
    end.

  Definition mark_page_as_non_cacheable (m:mem) ma : option mem :=
    pg <- get_mem m ma;
    Some (add_mem m ma (Page (page_content pg) (page_owned_by pg) false)).

  Definition update_cache_entry (m:mem) (entry:option ((vadd*madd)*PG.page)) : mem :=
    match entry with
      | None => m
      | Some ((va,ma),pg) => add_mem m ma pg
    end.

End MACHINE_MEM_EXT.