(** Pretty-printer for Mach code *)

open Printf
open Camlcoq
open Datatypes
open Maps
open AST
open Integers
open Locations
open Machregsaux
open Machregs
open Mach
open PrintAST
open PrintOp
open PrintMach
open VarSets
open MachIRAlias
open MachIR
open Clflags

let print_address = function
  | Unknown -> 
    Printf.eprintf "Warning: unknown memory access\n";
    "?"
  | Symb (Global id) -> (extern_atom id)
  | Symb (Stack i) -> 
    Printf.sprintf "ST[%ld]" (camlint_of_coqint i)
  | NoPtr -> "NOPTR!"

let chunk2string = function
| Mint8signed -> "Mint8signed"
| Mint8unsigned -> "Mint8unsigned"
| Mint16signed -> "Mint16signed"
| Mint16unsigned -> "Mint16unsigned"
| Mint32 -> "Mint32"
| Mint64 -> "Mint64"
| Mfloat32 -> "Mfloat32"
| Mfloat64 -> "Mfloat64"
| Mfloat64al32 -> "Mfloat64al32"

let print_chunk addr chunk =
  match addr with
  | Symb (Global _) -> ()
  | _ ->
    Printf.printf "%s : %s\n" 
      (print_address addr)
      (chunk2string chunk)

let print_mreg = function
  | AX -> "AX"
  | BX -> "BX"
  | CX -> "CX"
  | DX -> "DX"
  | SI -> "SI"
  | DI -> "DI"
  | BP -> "BP"
  | X0 -> "X0"
  | X1 -> "X1"
  | X2 -> "X2"
  | X3 -> "X3"
  | X4 -> "X4"
  | X5 -> "X5"
  | X6 -> "X6"
  | X7 -> "X7"
  | FP0 -> "FP0"

let print_var_type (v:var_type) : string =
  match v with
  | VReg r -> print_mreg r
  | VMem (Global id) -> extern_atom id
  | VMem (Stack i) -> Printf.sprintf "ST[%ld]" (camlint_of_coqint i)

(*
let print_vars s = 
  Printf.sprintf "{ %s}"
    (List.fold_left (fun s v -> Printf.sprintf "%s%s " s (print_var_type v)) ""
       (VS.elements s))
*)

let print_stealth s = 
  Printf.sprintf "{ %s}"
    (List.fold_left (fun s v -> Printf.sprintf "%s%s " s (extern_atom v)) ""
       (Registers.Regset.elements s))

(*
let print_pt = function
  | None -> "?"
  | Some (Global id) -> extern_atom id
  | Some (Stack ofs) -> Printf.sprintf "ST[%ld]" (camlint_of_coqint ofs)
*)

let print_pt = function
  | Symb (Global id) -> extern_atom id
  | Symb (Stack ofs) -> Printf.sprintf "ST[%ld]" (camlint_of_coqint ofs)
  | NoPtr -> "num"
  | Unknown -> "?"

let string_of_external_call = function
  | EF_external _  -> "external"
  | EF_builtin _   -> "builtin"
  | EF_vload _       -> "volatile load"
  | EF_vstore _      ->  "volatile store"
  | EF_vload_global _ -> "volatile load global"
  | EF_vstore_global _ -> "volatile store global"
  | EF_malloc            -> "malloc"
  | EF_free              -> "free"
  | EF_memcpy _      -> "memcpy"
  | EF_annot _   -> "annnot text"
  | EF_annot_val _-> "annot val"
  | EF_inline_asm _    -> "inlin asm"


let print_instruction (pp:out_channel) pt load ptv vl vl_args debug (pc,i) () =
  fprintf pp "%5ld: " (P.to_int32 pc);
  match i with
  | MIop(op, args, res, s) ->
      fprintf pp "%a{%s} = %a{%s} goto %ld\n"
         PrintMach.reg res
	(vl pc (VReg res))
	(PrintOp.print_operation reg) (op, args)
	(vl_args pc args)
	(P.to_int32 s)
  | MIload(chunk, addr, args, dst, s) ->
      fprintf pp "%a{%s} = %s[%a{%s}]{%s} (LOAD = %s{%s}) goto %ld\n"
        reg dst 
	(vl s (VReg dst))
	(name_of_chunk chunk)
        (PrintOp.print_addressing reg) (addr, args)
	(vl_args pc args)
	(print_pt (pt pc addr args))
	(print_pt (load pc addr args))
	(match pt pc addr args with
	| Symb p -> vl pc (VMem p)
	| _ -> "???")
	(P.to_int32 s)
  | MIstore(chunk, addr, args, src, s) ->
      fprintf pp "%s[%a{%s}] = %a{%s}{%s} (STORE = %s) goto %ld\n"
         (name_of_chunk chunk)
         (PrintOp.print_addressing reg) (addr, args)
	(vl_args pc args)
         reg src
	(vl pc (VReg src))
	(print_pt (pt pc addr args))
	(print_pt (ptv pc src))
	(P.to_int32 s)
  | MImemcpy_8_4(dst, src, s) ->
      fprintf pp "memcpy %a{%s} %a{%s} goto %ld\n"
         PrintMach.reg dst
	(vl pc (VReg dst))
         PrintMach.reg src
	(vl pc (VReg src))
	(P.to_int32 s)
  | MIgoto lbl ->
      fprintf pp "goto %ld\n" (P.to_int32 lbl)
  | MIcond(cond, args, lbl1, lbl2) ->
      fprintf pp "if (%a){%s} goto %ld else goto %ld\n"
        (PrintOp.print_condition reg) (cond, args)
	(vl_args pc args)
        (P.to_int32 lbl1)
        (P.to_int32 lbl2)
  | MIreturn ->
      fprintf pp "return\n"

let show ab =
  Printf.sprintf "[AX=%s BX=%s CX=%s DX=%s SI=%s DI=%s BP=%s]"
    (print_pt (D.get (VReg AX) ab))
    (print_pt (D.get (VReg BX) ab))
    (print_pt (D.get (VReg CX) ab))
    (print_pt (D.get (VReg DX) ab))
    (print_pt (D.get (VReg SI) ab))
    (print_pt (D.get (VReg DI) ab))
    (print_pt (D.get (VReg BP) ab))

(*
let manage_history instrs =
  let h = Array.of_list (!MachIRAlias.DS.history) in 
  let n = Array.length h in
  Array.iteri
    (fun i (_,_,j) -> if P.to_int j = 61 then Printf.printf "%d " (n-i)) h;
  print_newline ();
  let string_is_int s =
    try let _ = int_of_string s in true
    with Failure "int_of_string" -> false in
  Printf.printf "%d iterations" n;
  let read = ref "rien" in
  let last = ref 0 in
  while (!read<>"q") do
    while (!read <> "q" && !read <> "" && (not (string_is_int !read) || n < int_of_string !read  || 1 > int_of_string !read)) do
      Printf.printf "\n?";
      read := read_line () 
    done;
    if !read<>"q" then 
      begin
	if !read="" then incr last else last := int_of_string !read;
	let (i,ab,j) = h.(n- !last) in
	Printf.printf "\n\nITERATION[%d] NODE(%5ld) ==> NODE(%5ld)\n" !last
	  (P.to_int32 i)
	  (P.to_int32 j);
	let range i = 55 <= i && i < 65 in
        iter_list
	  (fun (pc,_) -> if range (P.to_int pc) then
	    Printf.printf "  %5ld:\n    [snow_alpha_mul]=%s\n    %s\n"
	      (P.to_int32 pc)
	      (print_pt (D.get (VMem (Global (intern_string "snow_alpha_mul"))) (PMap.get pc ab)))
	      (show (PMap.get pc ab)))
	  instrs;
	read := "rien"
      end
    done

let manage_history instrs =
  let h = Array.of_list (!MachIRAlias.DS.history) in 
  let n = Array.length h in
  for k=n -1 downto 0 do
	let (i,ab,j) = h.(k) in
	Printf.printf "\n\nITERATION[%d] NODE(%5ld) ==> NODE(%5ld)\n" (n-k)
	  (P.to_int32 i)
	  (P.to_int32 j);
	List.fold_right (print_instruction2 stdout ab) instrs ()
 done
  
*)
open MachIRStealthType

let print_function (pp:out_channel) f ptmap typechecked vl =
(*  let size = PTree.fold (fun n _ _ -> n+1) (fn_code f) 0 in *)
(*  Printf.printf "the function has %d nodes\n" size; *)
  let (Some (_,vl)) = vl in  
(*  let vl = (PMap.init VS.empty) in (* REMOVE ME FOR DEBUG *) *)
  Printf.printf "\x1B[0;34;48mtypechecking: %b\x1B[0;0;m\n\n" typechecked;
  let pt n addr regs =  addressing (PMap.get n ptmap) addr regs in
  fprintf pp "main() {\n";
  let load n addr regs =
    match addressing (PMap.get n ptmap) addr regs with
    | Symb s -> D.get (VMem s) (PMap.get n ptmap) 
    | _ -> Unknown in
  let ptv n reg =  D.get (VReg reg) (PMap.get n ptmap) in
  let vl_args n args =
    (match get_args_level (PMap.get n vl) args with
    | High -> "H"
    | Low -> "L" ) in
  let vl n x =
    (match get_var_level (PMap.get n vl) x with
    | High -> "H"
    | Low -> "L" ) in
  let debug n = D.get (VMem (Stack (Z.of_sint 48))) (PMap.get n ptmap) in
  let instrs = PTree.elements f.fn_code in
  let instrs = List.sort
    (fun (pc1,_) (pc2,_) -> Int32.compare (P.to_int32 pc1) (P.to_int32 pc2))
	instrs in
(*  manage_history instrs; *)
  List.fold_right (print_instruction pp pt load ptv vl vl_args debug) instrs ();
  fprintf pp "}"    

let destination : string option ref = ref None
let print_if optdest f pt typechecked vl =
  match !optdest with
  | None -> ()
  | Some name ->
      let oc = open_out name in
      print_function oc f pt typechecked vl;
      close_out oc

let print f pt typechecked vl = 
  print_if destination f pt typechecked vl;
  Errors.OK ()

let get_secrets () = List.map intern_string !secret_options
let get_stealths () = List.map intern_string !stealth_options




