(***********************************************************************
 * File: State.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the system state. 
 * 
 *************************************************************************)
Require Export Env System_mem Actions Page VCMaps.

Require Import Opt_monad.

Declare Module PG:PAGE.
Declare Module T:TLB.
Declare Module MM:MACHINE_MEM PG.
(* Declare Module SMEM:SYSTEM_MEM PG C T MM. *)
Declare Module OSS:OSS_MAP. 
Declare Module P2M:PHYSICAL_MAP. 
Declare Module HYP:HYPERVISOR P2M.

Module Type STATE (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM).
  Import PG.
  Import MM.
  Import SMEM.

  Record state :=
    State {
      (** Active OS. *)
      st_active_os : os_ident; 
      (** OSs *)
      st_oss : OSS.dict;
      (** Execution mode: ([usr]/[svc]). *)
      st_exec_mode : exec_mode;
      (** Activity of Active OS. When the Hypervisor is executing, 
          the corresponding Active OS is waiting *)
      st_aos_activity : os_activity; 
      (** hypervisor *)
      st_hypervisor : HYP.dict;
      (* The memory *)	
      st_memory : smem
    }.


  (* Auxiliary definitions *)
  Definition osi_current_page_table (s:state) (osi:os_ident) : option madd :=
    os <- OSS.get (st_oss s) osi;
    p2m <- HYP.get (st_hypervisor s) osi;
    P2M.get p2m (os_current_page os).

  Definition aos_current_page_table (s:state) : option madd :=
    osi_current_page_table s (st_active_os s).

  Definition osi_page_of_vadd_smem (s:state) (osi:os_ident) (va:vadd) :
       option page :=
    pt_addr <- osi_current_page_table s osi;
    if (os_ident_eq (st_active_os s) osi)
      then page_of_vadd_smem (st_memory s) pt_addr va
      else ma <- madd_of_vadd (smem_mm (st_memory s)) pt_addr va;
           get_mem (smem_mm (st_memory s)) ma.
        
  Definition aos_page_of_vadd_smem (s:state) (va:vadd) : option page :=
    pt_addr <- aos_current_page_table s;
    page_of_vadd_smem (st_memory s) pt_addr va.

  Definition aos_page_owner (s:state) (va:vadd) : option page_owner :=
    pg <- aos_page_of_vadd_smem s va;
    Some (page_owned_by pg).

  Definition os_accessible va :=
    ctxt_vadd_accessible ctxt va = true.

  Definition os_running s :=
    st_aos_activity s = OSA_running.

  Definition hyper_running s :=
    st_aos_activity s = OSA_waiting.

  Definition osi_pa_to_ma (s:state) (osi:os_ident) (pa:padd) : option madd :=
    p2m <- HYP.get (st_hypervisor s) osi;
    P2M.get p2m pa.

  Definition osi_pa_to_pg (s:state) (osi:os_ident) (pa:padd) : option page :=
    p2m <- HYP.get (st_hypervisor s) osi;
    ma <- P2M.get p2m pa;
    MM.get_mem (smem_mm (st_memory s)) ma.

  Definition osi_hypercall (s:state) (osi:os_ident) : option Hyperv_call :=
    os <- OSS.get (st_oss s) osi;
    os_hcall os.

  
  Definition shared_page (s:state) (va:vadd) (ma:madd) : Prop :=
    exists pt_padd, exists osi, exists pt_page,
    exists mp, exists va',
      osi_pa_to_pg s osi pt_padd = Some pt_page /\
      PG.page_content pt_page = PG.PT_c mp /\
      va <> va' /\
      PG.PT.get mp va' = Some ma.

  Axiom shared_page_dec : forall s va ma,
    {shared_page s va ma}+{~ shared_page s va ma}.


End STATE.

(* Declare Module ST : STATE. *)

Module Type ACTIONS (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM)
  (ST:STATE C SMEM).

  Import PG.
  Import T.
  Import C.
  Import MM.
  Import SMEM.
  Import ST.

  Parameter silent : state -> state.
  Parameter read : state -> vadd -> state.
  Parameter read_hyper : state -> vadd -> state.
  Parameter write  : state ->  vadd -> value -> state.       
  Parameter write_hyper  : state ->  vadd -> value -> state. 
  Parameter new_untrusted : state ->  os_ident -> vadd -> padd  -> state.
  Parameter new_hyper   : state ->  vadd -> madd -> state.   
  Parameter del_untrusted  : state ->  os_ident -> vadd -> state.  
  Parameter del_hyper : state -> vadd -> state.              
  Parameter switch : state -> os_ident -> state.                   
  Parameter lswitch_untrusted : state -> padd -> os_ident -> state. 
  Parameter hcall : state -> Hyperv_call -> state.            
  Parameter ret_ctrl : state -> state.           
  Parameter chmod : state -> state.              
  Parameter page_pin_untrusted : state -> os_ident -> padd -> ptype -> state.
  Parameter page_unpin_untrusted : state -> os_ident -> padd -> state.                                       

  (** **** Silent *)
  (** represents an action (some OS sub-specified action, i.e., silent) 
      in which the properties relevant to our state abstraction don't change  *)
  Definition pre_silent (s:state) := True.
  Definition post_silent (s s':state) := s = s'.

  (** **** Read *)
  (** Read in the memory : 
      the virtual address can be mapped to any kind of page *)

  (** For all different read actions, we can make some general pre/post *)
  Definition pre_all_read (va:vadd) (s:state) :=
    exists pg,
      aos_page_of_vadd_smem s va = Some pg /\ is_rw_page pg.

  Definition post_all_read (va:vadd) (s s':state) :=
    forall pt_addr, aos_current_page_table s = Some pt_addr ->
      let new_mem := page_smem_smem (get (st_memory s) va pt_addr) in

        s' = State (st_active_os s) (st_oss s) (st_exec_mode s)
                   (st_aos_activity s) (st_hypervisor s) new_mem. 
  
  (** [read] 
     corresponds to an OS reading virtual address va from its address
     space (ie it is not reserved for the Hypervisor) *)
  Definition pre_read va s := 
    pre_all_read va s 
    /\ os_accessible va
    /\ os_running s.

  Definition post_read := post_all_read.

  (** [read_hyper]
     corresponds to the Hypervisor reading virtual address va.*)
  Definition pre_read_hyper va s := 
    pre_all_read va s /\ hyper_running s.

  Definition post_read_hyper := post_all_read.


  (** **** Write *)
  
  (** [write] represents an OS writing value val in virtual address va 
     from its address space.*)

  Definition pre_all_write va (val:value) s :=
    exists pg,
      aos_page_of_vadd_smem s va = Some pg /\ is_rw_page pg.

  Definition post_all_write va (val:value) s s' :=
    forall pt_addr pg,
      aos_current_page_table s = Some pt_addr ->
      aos_page_of_vadd_smem s va = Some pg ->
      let new_mem := set (st_memory s) va pt_addr val in

        s' = State (st_active_os s) (st_oss s) (st_exec_mode s)
                   (st_aos_activity s) (st_hypervisor s) new_mem. 
    

  Definition pre_write va (val : value) s :=
   (** guest OS is executing *)
    os_running s
   (** va is not reserved for the Hypervisor *)
    /\ os_accessible va
    /\ pre_all_write va val s.

  Definition post_write := post_all_write.


  (** [write_hyper]
     represents the Hypervisor writing value val in virtual address va. *)

  Definition pre_write_hyper va (val:value) s :=
  (** The Hypervisor is executing *)
    hyper_running s
    /\ pre_all_write va val s 
  (* machine address ma belongs to the Hypervisor *)
    /\ aos_page_owner s va = Some Hyp_PO.
  
  (** Here we restricted the Hypervisor to write memory values in pages 
     of his own. *)
  Definition post_write_hyper := post_all_write.


  (** [new] corresponds to a new action done by the Hypervisor
     for an untrusted OS that had asked for it. *)
  Definition pre_new osi va pa s := 
       ctxt_oss ctxt osi = false (* untrusted osi *) 
    (** pa should be mapped to a madd owned and pinned by osi. *)
    /\ exists ma, (osi_pa_to_ma s osi pa = Some ma /\
         madd_type (smem_mm (st_memory s)) ma <> None)
       (** va is not reserved for the Hypervisor ; As before this can be discarded *)
    /\ os_accessible va
       (** Untrusted OS has a 'new' hypercall *)
    /\ osi_hypercall s osi = Some (Hyperv_call_new va pa)
    /\ hyper_running s.

  Definition new (s:state) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd) :=
    let smem' :=
      if (shared_page_dec s va ma)
      then new_shared (st_memory s) pt_addr va ma
      else new_not_shared (st_memory s) pt_addr va ma in
    let tlb' :=
      if (os_ident_eq (st_active_os s) osi)
      then remove_tlb (smem_t smem') va
      else smem_t smem' in
    SMem (smem_c smem') tlb' (smem_mm smem').
   

  Definition post_all_new osi (va:vadd) (pa:padd) (s s':state) :=
  (* The mapping va->ma is added to the PT in pt_addr.
     hcall changes to None with an untrusted OS, otherwise it remain unchanged *)
  exists ma, exists pt_addr, exists os,
    osi_pa_to_ma s osi pa = Some ma /\
    osi_current_page_table s osi = Some pt_addr /\
    OSS.get (st_oss s) osi = Some os /\
    let newmem :=  new s osi pt_addr va ma in
    let newos := OS None (os_current_page os) in
    let newoss := (OSS.set (st_oss s) osi newos) in
      s' = State (st_active_os s) newoss (st_exec_mode s) (st_aos_activity s)
      (st_hypervisor s) newmem.


  Definition post_new osi va pa s s' :=
    (** The current memory mapping in [s'] maps [va] to [ma], 
       where pa is mapped to ma by hypervisor *)
    post_all_new osi va pa s s'.

  (** [new_hyper] corresponds  to the action of the Hypervisor adding a 
      new mapping to the Active Memory Mapping for its own needs. *)

  Definition pre_new_hyper va ma s :=
  (** Hypervisor is running *)
    hyper_running s
    /\ madd_type (smem_mm (st_memory s)) ma <> None
    /\ madd_owner (smem_mm (st_memory s)) ma = Hyp_PO
  (** va is reserved for the Hypervisor *)
    /\ ~(os_accessible va).

  Definition post_new_hyper va ma s s' :=
  (** The Active memory mapping in s' maps va to pa*)
    exists pt_addr,
      aos_current_page_table s = Some pt_addr /\
       let newmem :=  new s (st_active_os s) pt_addr va ma in
         s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
         (st_hypervisor s) newmem.

  (** **** Switch *)

  (** [switch] corresponds to a changing of Active OS by the Hypervisor. 
      At the moment this happens only when some OS returns control to the 
      Hypervisor, because we haven't specified an interruption mechanism yet.  *)

  Definition pre_switch osi s :=
  (** Hypervisor is running and osi has no pending request *)
    hyper_running s /\ osi_hypercall s osi = None.

  Definition post_switch osi s s' :=
  (** Hypervisor still running after switching. *)
    let newmem := ctxt_switch (st_memory s) in
      s' = State osi (st_oss s) (st_exec_mode s) (st_aos_activity s)
        (st_hypervisor s) newmem.

  (** **** Local switch *)

  (** [lswitch] corresponds to a process context switch inside some 
      Active OS executed by the Hypervisor on behalf of  the untrusted 
      OSes and by the trusted OSes.   *)

  Definition pre_lswitch pa osi s := 
  (** The new page table address belongs to the OS *)
    exists ma, (osi_pa_to_ma s osi pa = Some ma /\
      madd_type (smem_mm (st_memory s)) ma = Some PT) /\
    ctxt_oss ctxt osi = false (* untrusted osi *) /\
    hyper_running s /\
    osi_hypercall s osi = Some (Hyperv_call_lswitch pa).

  Definition post_lswitch pa osi s s' :=
    let os' := OS None pa in
    let new_oss := OSS.set (st_oss s) osi os' in
    let newmem := ctxt_switch (st_memory s) in
      s' = State (st_active_os s) new_oss (st_exec_mode s) (st_aos_activity s)
        (st_hypervisor s) newmem.

  (** [delete] corresponds to a delete action done by the Hypervisor
      for an untrusted OS that had asked for it. *)
  Definition pre_all_del (osi:os_ident) (va:vadd) (s:state) := 
    (** va is defined in the Active memory mapping *)
    exists pt_addr, exists ma,
      osi_current_page_table s osi = Some pt_addr /\
      madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.

  Definition pre_del (osi:os_ident) (va:vadd) (s:state) :=
    ctxt_oss ctxt osi = false (* untrusted osi *) 
    /\ osi_hypercall s osi = Some (Hyperv_call_del va)
    /\ hyper_running s
    (** va is not reserved for the Hypervisor *)
    /\ os_accessible va
    /\ pre_all_del osi va s.

  Definition delete_smem (osi:os_ident) (pt_addr:madd) (va:vadd) (s:state) :=
    let smem' := delete (st_memory s) pt_addr va in
    let tlb' :=
      if (os_ident_eq (st_active_os s) osi)
      then remove_tlb (smem_t smem') va
      else smem_t smem' in
    SMem (smem_c smem') tlb' (smem_mm smem').


  Definition post_del (osi:os_ident) (va:vadd) (s s':state) :=
  (* The mapping va->ma is deleted from the PT in pt_addr.
     hcall changes to None *)
  exists pt_addr, exists os,
    osi_current_page_table s osi = Some pt_addr /\
    OSS.get (st_oss s) osi = Some os /\
    let newmem := delete_smem osi pt_addr va s in
    let newos := OS None (os_current_page os) in
    let newoss := (OSS.set (st_oss s) osi newos) in
      s' = State (st_active_os s) newoss (st_exec_mode s) (st_aos_activity s)
      (st_hypervisor s) newmem.

  (** [del_hyper] corresponds  to the action of the Hypervisor removing a
      mapping from the Active Memory Mapping for its own needs. *)
  Definition pre_del_hyper va s :=
    pre_all_del (st_active_os s) va s 
    /\ hyper_running s
    (** va is reserved for the Hypervisor. *)
    /\ ~(os_accessible va).

  Definition post_del_hyper va s s' :=
    exists pt_addr,
      osi_current_page_table s (st_active_os s) = Some pt_addr /\
      let newmem := delete_smem (st_active_os s) pt_addr va s in
        s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
          (st_hypervisor s) newmem.

 
  (** **** Hypercall *)
  
  (** [hcall]
      corresponds to an untrusted OS making an hypercall of type hc. 
      This is the case when the untrusted OSes need the Hypervisor to run 
      some service that needs privilege on their behalf. 
      So this transition changes the execution mode from usr to svc 
      and sets the required service *)            

  Definition pre_hcall s :=
    os_running s /\ ctxt_oss ctxt (st_active_os s) = false.

  Definition post_hcall hc s s' :=
    exists aos, OSS.get (st_oss s) (st_active_os s) = Some aos /\
      let newos := OS (Some hc) (os_current_page aos) in
      let newoss := OSS.set (st_oss s) (st_active_os s) newos in
        s' = State (st_active_os s) newoss EXMD_svc OSA_waiting
        (st_hypervisor s) (st_memory s).

  (** **** Changing control *)

  (** [ret_ctrl] corresponds to Hypervisor getting back the control.
      This can be seen both like a timer interruption 
      or a call from an OS when it is going idle.
      So this transition changes the execution mode from user to supervisor mode *)

  Definition pre_ret_ctrl s := os_running s.

  Definition post_ret_ctrl s s' :=
    s' = State (st_active_os s) (st_oss s) EXMD_svc OSA_waiting
      (st_hypervisor s) (st_memory s).

  (** [chmod] corresponds  to the Hypervisor changing the execution mode, 
      particularly to let an OS execute. 
      It could be the case after a context switch or any Hypervisor execution
      was carried out by the Hypervisor on behalf of an untrusted OS. *)            

  Definition pre_chmod s :=
    hyper_running s
   (** There is no request to the Hypervisor pending in s *)
    /\ osi_hypercall s (st_active_os s) = None.

  Definition post_chmod s s' :=
    s' = State (st_active_os s) (st_oss s) EXMD_usr OSA_running
      (st_hypervisor s) (st_memory s).


  (** **** Page classification/registration *)

  (** [page_pin] corresponds to the registration of a page 
      (and classification with type [t]) by a trusted OS 
      or by the Hypervisor for the untrusted OS. 
      The Hypervisor doesn't use it for its own purposes. *)

  (* No osi PT maps any va to ma *)
  Definition no_va_mapped_to_ma osi ma s : Prop :=
    forall pt_padd pt_page va mp,
      osi_pa_to_pg s osi pt_padd = Some pt_page ->
      page_content pt_page = PT_c mp ->
      PG.PT.get mp va <> Some ma.

  Definition va_mapped_to_ma osi ma s : Prop :=
    exists pt_padd pt_page va mp,
      osi_pa_to_pg s osi pt_padd = Some pt_page /\
      page_content pt_page = PT_c mp /\
      PG.PT.get mp va = Some ma.

  (* The implementation should be able to search for ma in osi
     mappings to decide this predicate *)
  Axiom no_va_mapped_to_ma_dec :
    forall osi ma s,
      {no_va_mapped_to_ma osi ma s}+
      {va_mapped_to_ma osi ma s}.

  Definition pre_pin (osi:os_ident) (pa:padd) (type:ptype) (s:state) :=
    (** osi is not trusted *)
    (ctxt_oss ctxt) osi = false
    (** The pin hcall is pending *)
    /\ osi_hypercall s osi = Some (Hyperv_call_pin pa type)
    /\ hyper_running s
    (** There is at least an emtpy madd to assign to pa. *)
    /\ exists ma, (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg /\
         page_owned_by pg = No_Owner /\ page_content pg = Other_c /\ page_cacheable pg = true)
    /\ osi_pa_to_ma s osi pa = None.

  Definition post_pin (osi:os_ident) (pa:padd) (type:ptype) (s s':state) :=
  (** Get a free madd (it exists because of pre). *)
    exists ma, (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg /\
      page_owned_by pg = No_Owner /\ page_content pg = Other_c /\ page_cacheable pg = true) /\
    exists p2m, HYP.get (st_hypervisor s) osi = Some p2m /\
    exists system, OSS.get (st_oss s) osi = Some system /\
    (** The new page is registered with type type and owner osi in s' *)
    let newpage := new_page (Some type) (Os_PO osi) in
    (** Update the new page in system memory *)
    let newmem := SMem (smem_c (st_memory s)) (smem_t (st_memory s))
                    (add_mem (smem_mm (st_memory s)) ma newpage) in
    (** pa is added to the hypervisor mapping for osi *)
    let newp2m := P2M.set p2m pa ma in
    let newhyper := HYP.set (st_hypervisor s) osi newp2m in
    let newos := OS None (os_current_page system) in
    let newoss := OSS.set (st_oss s) osi newos in
      s' = State (st_active_os s) newoss (st_exec_mode s)
        (st_aos_activity s) newhyper newmem.

  (** [page_unpin] corresponds to the de-registration 
      (and declassification) of previously a registered/classified page. *) 
  Definition pre_unpin osi pa s :=
    (ctxt_oss ctxt) osi = false (* untrusted osi *)
    (** If the OS is untrusted then the corresponding request is pending  *)
    /\ osi_hypercall s osi = Some (Hyperv_call_unpin pa)
    /\ hyper_running s
    /\ exists system, exists ma,
      (** pa is not the osi current page physical address *)
      OSS.get (st_oss s) osi = Some system /\
      pa <> (os_current_page system) /\
      (** The page is registered in s *)
      osi_pa_to_ma s osi pa = Some ma /\
      madd_type (smem_mm (st_memory s)) ma <> None /\
      (** If it is a PT, it is empty *)
      (forall pg pt, get_mem (smem_mm (st_memory s)) ma = Some pg ->
                  page_content pg = PT_c pt ->
                  pt = PT.empty) /\ 
      (** And ma does not appear in any osi PT *)
      no_va_mapped_to_ma osi ma s.

  Definition post_unpin osi pa s s' :=
    (** The page is not registered with in s' *)
    exists p2m, exists ma, exists system,
      HYP.get (st_hypervisor s) osi = Some p2m /\
      P2M.get p2m pa = Some ma /\
      OSS.get (st_oss s) osi = Some system /\
      (** Delete pa from s' hypervisor mapping *)
      let newp2m := P2M.remove p2m pa in 
      let newhyper := HYP.set (st_hypervisor s) osi newp2m in
      (** Free the page in system memory *)
      let newpage := new_page None No_Owner in
      let newmem := SMem (smem_c (st_memory s)) (smem_t (st_memory s))
                    (add_mem (smem_mm (st_memory s)) ma newpage) in
      let newos := OS None (os_current_page system) in
      let newoss := OSS.set (st_oss s) osi newos in
        s' = State (st_active_os s) newoss (st_exec_mode s)
          (st_aos_activity s) newhyper newmem.


  (** Action semantics *)
  Axiom silent_sem : forall s,
    pre_silent s -> post_silent s (silent s).

  Axiom read_sem : forall s va,
    pre_read va s -> post_read va s (read s va).

  Axiom read_hyper_sem : forall s va,
    pre_read_hyper va s -> post_read_hyper va s (read_hyper s va).

  Axiom write_sem : forall s va val,
    pre_write va val s -> post_write va val s (write s va val).

  Axiom write_hyper_sem : forall s va val,
    pre_write_hyper va val s -> post_write_hyper va val s (write_hyper s va val).

  Axiom new_sem : forall s osi va pa,
    pre_new osi va pa s -> post_new osi va pa s (new_untrusted s osi va pa).

  Axiom switch_sem : forall s osi,
    pre_switch osi s -> post_switch osi s (switch s osi).

  Axiom lswitch_sem : forall s osi pa,
    pre_lswitch pa osi s -> post_lswitch pa osi s (lswitch_untrusted s pa osi).

  Axiom delete_sem : forall s osi va,
    pre_del va osi s -> post_del va osi s (del_untrusted s va osi).

  Axiom delete_hyper_sem : forall s va,
    pre_del_hyper va s -> post_del_hyper va s (del_hyper s va).

  Axiom hcall_sem : forall (s:state) (hc:Hyperv_call),
    pre_hcall s -> post_hcall hc s (hcall s hc).

  Axiom ret_ctrl_sem : forall (s:state),
    pre_ret_ctrl s -> post_ret_ctrl s (ret_ctrl s).

  Axiom chmod_sem : forall (s:state),
    pre_chmod s -> post_chmod s (chmod s).

  Axiom page_pin_sem : forall (s:state) (osi:os_ident) (pa:padd) (type:ptype),
    pre_pin osi pa type s -> post_pin osi pa type s (page_pin_untrusted s osi pa type).

  Axiom page_unpin_sem : forall (s:state) (osi:os_ident) (pa:padd),
    pre_unpin osi pa s -> post_unpin osi pa s (page_unpin_untrusted s osi pa).

  (** Get the precondition of an action *)
  Definition pre_action (act:action) : state -> Prop := 
    match act with
      | Silent => pre_silent
      | Read va => pre_read va
      | Read_hyper va => pre_read_hyper va
      | Write va val => pre_write va val
      | Write_hyper va val => pre_write_hyper va val
      | New_untrusted osi va pa => pre_new osi va pa 
      | New_hyper va ma => pre_new_hyper va ma 
      | Del_untrusted osi va => pre_del osi va
      | Del_hyper va => pre_del_hyper va
      | Switch osi => pre_switch osi
      | Lswitch_untrusted osi pa => pre_lswitch pa osi
      | Hcall c => pre_hcall
      | Ret_ctrl => pre_ret_ctrl
      | Chmod => pre_chmod
      | Page_unpin_untrusted osi pa => pre_unpin osi pa
      | Page_pin_untrusted osi pa pt => pre_pin osi pa pt
    end.

  (** The family of post-conditions of actions *)
  Definition post_action (act:action): state -> state -> Prop := 
    match act with
      | Silent => post_silent
      | Read va => post_read va
      | Read_hyper va => post_read_hyper va
      | Write va val => post_write va val
      | Write_hyper va val => post_write_hyper va val
      | New_untrusted osi va pa => post_new osi va pa
      | New_hyper va ma => post_new_hyper va ma
      | Del_untrusted osi va => post_del osi va
      | Del_hyper va => post_del_hyper va
      | Switch osi => post_switch osi
      | Lswitch_untrusted osi pa => post_lswitch pa osi
      | Hcall c => post_hcall c
      | Ret_ctrl => post_ret_ctrl
      | Chmod => post_chmod
      | Page_unpin_untrusted osi pa => post_unpin osi pa
      | Page_pin_untrusted osi pa pt => post_pin osi pa pt
    end.

  (** [valid_exec] specifies a transition from one state to another 
       executing a given action. *)
  Definition valid_exec (s:state) (act:action) (s':state) : Prop :=
      pre_action act s /\ post_action act s s'.

End ACTIONS.

(* Declare Module ACT : ACTIONS. *)
