(***********************************************************************
 * File: State_equivalence.v
 * $Revision $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy $
 ***********************************************************************
 * Definition of equivalence between states (for the AOS and a given
   osi). Proofs of equivalence properties and hypervisor behaviour
   related to equivalence.
 **********************************************************************) 
Require Import State State_valid.


Module STEQ (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM)
  (ST:STATE C SMEM) (ACT:ACTIONS C SMEM ST)
  (ST_VALID:ST_VALID C SMEM ST).

Import PG.
Import C.
Import SMEM.
Import ST.
Import ST_VALID.

(* Module STAC := STAC C SMEM ST ACT. *)

(* Two states s1 and s2 have equivalent PT pages for osi, if for
   all PT of osi, if it maps a va to an arbitrary ma with a RW
   page, the content is the same in the other state. Also, there is a
   pa that maps to that ma. *)
Definition osi_pt_pages_eq (s1 s2:state) (osi:os_ident) : Prop :=
  forall pt_padd pt_page1 mp1,
    (* pt_padd maps to pt_page1, which is a PT page with map mp1. *)
    osi_pa_to_pg s1 osi pt_padd = Some pt_page1 ->
    PT_c mp1 = (page_content pt_page1) ->
      (* Then, for s1 and s2 to be equivalent, pt_padd must map
         to a pt_page2, of type PT *)
      (exists pt_page2, osi_pa_to_pg s2 osi pt_padd = Some pt_page2 /\
       exists mp2, PT_c mp2 = (page_content pt_page2) /\
        (forall va pa ma1,
          (* If in mp1, va maps to ma1 *)
          PT.get mp1 va = Some ma1 ->
          (* and pa is _the_ physical address that maps to ma1 *)
          osi_pa_to_ma s1 osi pa = Some ma1 ->
            (* Then, in mp2, va must map to some ma2 *)
            exists ma2, PT.get mp2 va = Some ma2 /\
              (* pa must map to ma2 *)
              osi_pa_to_ma s2 osi pa = Some ma2 /\
              (* If va maps to a RW page, then the page in ma2 must be
                 the same as that mapped to ma1 in s1 *)
              (forall pg,
                 page_of_vadd_madd_smem (st_memory s1) va ma1 = Some pg ->
                 is_rw_page pg ->
                 page_of_vadd_madd_smem (st_memory s2) va ma2 = Some pg))).

Definition osi_equivalent_smem (s1 s2:state) (osi:os_ident): Prop :=
  (osi_pt_pages_eq s1 s2 osi) /\ (osi_pt_pages_eq s2 s1 osi).

Definition osi_hyper_eq (s1 s2:state) (osi:os_ident) : Prop :=
  forall pa ma1 pg,
    is_rw_page pg ->
    osi_pa_to_ma s1 osi pa = Some ma1 ->
    ACT.no_va_mapped_to_ma osi ma1 s1 ->
    MM.get_mem (smem_mm (st_memory s1)) ma1 = Some pg ->
    exists ma2,
      osi_pa_to_ma s2 osi pa = Some ma2 /\
      ACT.no_va_mapped_to_ma osi ma2 s2 /\
      MM.get_mem (smem_mm (st_memory s2)) ma2 = Some pg.

Definition osi_equivalent_hyper (s1 s2:state) (osi:os_ident): Prop :=
  (osi_hyper_eq s1 s2 osi) /\ (osi_hyper_eq s2 s1 osi).

(** Two states are osi equivalent, 
    if they are equal regarding only 
    the part of the states that corresponds 
    to osi.
*)
Definition osi_equivalent (s1 s2:state) (osi:os_ident):= 
  ((st_active_os s1 = osi /\ st_active_os s2 = osi /\ st_aos_activity s1 = st_aos_activity s2) \/
  (st_active_os s1 <> osi /\ st_active_os s2 <> osi)) /\

  OSS.get (st_oss s1) osi = OSS.get (st_oss s2) osi /\
  (osi_equivalent_smem s1 s2 osi) /\
  (osi_equivalent_hyper s1 s2 osi).

Lemma osi_equivalent_refl : forall (s:state) (osi:os_ident),
  osi_equivalent s s osi.
Proof.
  intros; unfold osi_equivalent,
    osi_equivalent_smem, osi_pt_pages_eq;
    split; [case (os_ident_eq (st_active_os s) osi); auto|];
    repeat ( try split; intros; auto;
             try eexists; eauto).
Qed.

Lemma osi_equivalent_smem_symm :
  forall (s1 s2:state) (osi:os_ident),
    STEQ.osi_equivalent_smem s1 s2  osi ->
    STEQ.osi_equivalent_smem s2 s1 osi.
Proof.
  unfold STEQ.osi_equivalent_smem; intros;
    destruct H; repeat split; auto.
Qed.

Lemma osi_equivalent_hyper_symm :
  forall (s1 s2:state) (osi:os_ident),
    STEQ.osi_equivalent_hyper s1 s2  osi ->
    STEQ.osi_equivalent_hyper s2 s1 osi.
Proof.
  unfold STEQ.osi_equivalent_hyper; intros;
    destruct H; repeat split; auto.
Qed.


Lemma osi_equivalent_symm : forall (s1 s2:state) (osi:os_ident),
  osi_equivalent s1 s2 osi -> osi_equivalent s2 s1 osi.
Proof.
  intros; unfold osi_equivalent,
    osi_equivalent_smem, osi_equivalent_hyper in *; 
  repeat split; auto; 
    repeat (match goal with
      | [H: _ /\ _ |-_] => destruct H
      | [H: _ \/ _ |-_] => destruct H
    end; auto).
Qed.


Lemma osi_eq_pt_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_pt_pages_eq s1 s2 osi -> osi_pt_pages_eq s2 s3 osi ->
  osi_pt_pages_eq s1 s3 osi.
Proof.
  Ltac destruct_and :=
  repeat match goal with
    | [H:_/\_|-_]=>destruct H
    | [ H : ex _ |- _] => destruct H
  end.

  intros; unfold osi_pt_pages_eq in *.
  repeat (try split; intros; eauto).
  destruct H with pt_padd pt_page1 mp1; auto;
    destruct_and.
  destruct H0 with pt_padd x x0; auto;
    destruct_and.
  exists x1; repeat split; intros; auto.
  exists x2; repeat split; intros; auto.
  destruct H5 with va pa ma1; auto;
    destruct_and.
  destruct H8 with va pa x3; auto;
    destruct_and.
  exists x4; repeat split; intros; auto.
Qed.  

Lemma osi_equivalent_smem_trans : forall (s1 s2 s3:state) (osi:os_ident),
  STEQ.osi_equivalent_smem s1 s2 osi -> STEQ.osi_equivalent_smem s2 s3 osi ->
  STEQ.osi_equivalent_smem s1 s3 osi.
Proof.
  unfold STEQ.osi_equivalent_smem in *; intros;
    split; destruct H; destruct H0.
  apply osi_eq_pt_trans with s2; auto.
  apply osi_eq_pt_trans with s2; auto.
Qed.

Lemma osi_eq_hyper_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_hyper_eq s1 s2 osi -> osi_hyper_eq s2 s3 osi ->
  osi_hyper_eq s1 s3 osi.
Proof.
  intros; unfold osi_hyper_eq in *.
  repeat (try split; intros; eauto).
  destruct H with pa ma1 pg; auto;
    destruct_and.
  destruct H0 with pa x pg; auto;
    destruct_and.
  exists x0; repeat split; intros; auto.
Qed.  

Lemma osi_equivalent_hyper_trans : forall (s1 s2 s3:state) (osi:os_ident),
  STEQ.osi_equivalent_hyper s1 s2 osi -> STEQ.osi_equivalent_hyper s2 s3 osi ->
  STEQ.osi_equivalent_hyper s1 s3 osi.
Proof.
  unfold STEQ.osi_equivalent_hyper in *; intros;
    split; destruct H; destruct H0.
  apply osi_eq_hyper_trans with s2; auto.
  apply osi_eq_hyper_trans with s2; auto.
Qed.

Lemma osi_equivalent_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_equivalent s1 s2 osi -> osi_equivalent s2 s3 osi -> osi_equivalent s1 s3 osi.
Proof.
  Local Hint Resolve osi_equivalent_smem_trans
    osi_equivalent_hyper_trans.
  intros; unfold osi_equivalent in *; split;
    destruct_and.
  destruct H; destruct_and;
  destruct H0; destruct_and;
  try (left; repeat split; auto; fail);
  try (right; repeat split; auto; fail).
  left; repeat split; auto.
  transitivity (st_aos_activity s2); auto.

  split; eauto.
  transitivity (OSS.get (st_oss s2) osi); auto.
Qed.  

End STEQ.
