(***********************************************************************
 * File: State_valid.v
 * $Revision: 12918 $
 * $Date: 2016-09-21 22:27:32 -0300 (Wed, 21 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Valid state definition.
 * 
 *************************************************************************)
Require Import State.

Module Type ST_VALID (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM).
  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  
  Definition running_os_not_hypercall
    (aosi:os_ident) (activity:os_activity) (ossmap:OSS.dict) : Prop :=
    forall aos:os,
      Some aos = OSS.get ossmap aosi -> activity = OSA_running ->
      os_hcall aos = None.

  Definition valid_hyper_exec_mode
    (activity:os_activity) (mode:exec_mode) : Prop :=
    activity = OSA_waiting -> mode = EXMD_svc.

  Definition valid_os_exec_mode (aosi:os_ident) (activity:os_activity)
    (mode:exec_mode) (ossmap:OSS.dict) : Prop :=
    activity = OSA_running -> mode = EXMD_usr.

  (* All OS identifiers exist in the state *)
  Definition valid_os_idents (oss:OSS.dict) (hyper:HYP.dict) : Prop :=
    forall osi,
      exists os, OSS.get oss osi = Some os /\
      exists p2m, HYP.get hyper osi = Some p2m.

  (* For every possible physical address of an OS, the machine address
     is owned by the OS *)
  Definition valid_hypervisor (hyper:HYP.dict) (mm:mem) : Prop :=
    forall (osi:os_ident) (p2m:P2M.dict) (pa:padd) (ma:madd),
      Some p2m = HYP.get hyper osi ->
      Some ma = P2M.get p2m pa ->
      madd_owner mm ma = Os_PO osi.

  Definition valid_pt_non_cacheable (mm:mem) : Prop :=
    forall (pt_addr:madd) (pt_page:page),
      get_mem mm pt_addr = Some pt_page ->
      is_pt_page pt_page ->
      ~ is_cacheable pt_page.

  Definition non_cacheable_synonyms (hyper:HYP.dict) (mm:mem) : Prop :=
    forall pt_addr pt_page mp va ma pg
      pt_addr' pt_page' mp' va',

      va <> va' -> 

      (exists p2m, exists osi, exists pt_padd, 
         HYP.get hyper osi = Some p2m /\
         P2M.get p2m pt_padd = Some pt_addr) ->

      (exists p2m, exists osi, exists pt_padd,
         HYP.get hyper osi = Some p2m /\
         P2M.get p2m pt_padd = Some pt_addr') ->
      
      get_mem mm pt_addr = Some pt_page ->
      page_content pt_page = PT_c mp ->
      PT.get mp va = Some ma ->

      get_mem mm pt_addr' = Some pt_page' ->
      page_content pt_page' = PT_c mp' ->
      PT.get mp' va' = Some ma ->

      get_mem mm ma = Some pg ->

      ~ is_cacheable pg.



  (* If a mapped virtual address is accessible by an OS, it is owned
     by that OS. If a mapped virtual address is not accessible by an
     OS, it is owned by the Hypervisor. *)
  Definition valid_virtual_mapping (mm:mem) (hyper:HYP.dict) : Prop :=
    forall (osi:os_ident) (pt_addr:madd) (pt_page:page)
      (pt:PT.dict) (va:vadd) (ma:madd),
      
      (exists p2m, exists pt_padd, 
        Some p2m = HYP.get hyper osi /\
        P2M.get p2m pt_padd = Some pt_addr) ->

      get_mem mm pt_addr = Some pt_page ->
      page_owned_by pt_page = Os_PO osi ->
      PT_c pt = page_content pt_page ->
      PT.get pt va = Some ma ->
      exists pg,
        Some pg = get_mem mm ma /\
        (os_accessible va -> page_owned_by pg = Os_PO osi) /\
        (~os_accessible va -> page_owned_by pg = Hyp_PO).


  (* Current page table of any OS is of type PT, and owned by that OS. *)
  Definition valid_current_page
    (hyper:HYP.dict) (mm:mem) (ossmap:OSS.dict) : Prop :=
    forall (osi:os_ident) (system:os) (p2m:P2M.dict), 
      Some system = OSS.get ossmap osi ->
      Some p2m = HYP.get hyper osi ->
      exists pt_addr, exists pt_page,
        Some pt_addr = P2M.get p2m (os_current_page system) /\
        Some pt_page = get_mem mm pt_addr /\
        is_pt_page pt_page /\
        page_owned_by pt_page = Os_PO osi.

  (* Hypervisors mappings must be injective *)
  Definition injective_hyper_mappings (hyper:HYP.dict) : Prop :=
    forall (osi:os_ident) (p2m:P2M.dict) (pa1 pa2:padd) (ma:madd),
      Some p2m = HYP.get hyper osi ->
      P2M.get p2m pa1 = Some ma -> P2M.get p2m pa2 = Some ma ->
      pa1 = pa2.

  (* madds in page tables, must be assigned to a pinned padd by hypervisor. *)
  Definition va_has_valid_pa (mm:mem) (hyper:HYP.dict) : Prop :=
    forall (osi:os_ident) (p2m:P2M.dict) (pt_addr:madd) 
           (pt_page:page) (pt:PT.dict) (va:vadd) (ma:madd),
      
      Some p2m = HYP.get hyper osi ->
      (exists pt_padd, P2M.get p2m pt_padd = Some pt_addr) ->
      get_mem mm pt_addr = Some pt_page ->

      PT_c pt = page_content pt_page ->
      PT.get pt va = Some ma ->
      os_accessible va ->
      exists pa, P2M.get p2m pa = Some ma.

  (* All cache keys are related in a page table mapping. *)
  Definition valid_cache_va_ma (s:state) :=
    let cache := smem_c (st_memory s) in
      forall va ma pg_cache,
        Some pg_cache = get_cache cache (va,ma) ->
        (exists osi, exists pt_padd, exists pt_page, exists mp,
          Some pt_page = osi_pa_to_pg s osi pt_padd /\
          page_content pt_page = PT_c mp /\
          PT.get mp va = Some ma).
  
  (* All cache pages have the same owner and type (RW) than those in
     machine memory. *)
  Definition valid_cache (s:state) :=
    let mmem := smem_mm (st_memory s) in
    let cache := smem_c (st_memory s) in
      forall va ma pg_cache,
        Some pg_cache = get_cache cache (va,ma) ->
        is_rw_page pg_cache /\ is_cacheable pg_cache /\
        (exists pg_mem, Some pg_mem = get_mem mmem ma /\
          is_rw_page pg_mem /\ is_cacheable pg_mem /\
          page_owned_by pg_cache = page_owned_by pg_mem).

  (* The tlb is included in the current PT of the AOS. *)
  Definition valid_tlb (s:state) :=
    let aos := st_active_os s in
    let mmem := smem_mm (st_memory s) in
    let tlb := smem_t (st_memory s) in
      forall va ma pt_addr, Some ma = get_tlb tlb va ->
        aos_current_page_table s = Some pt_addr ->
        madd_of_vadd mmem pt_addr va = Some ma.


  Definition valid_state (st:state) : Prop :=
    let vadd_accessible := ctxt_vadd_accessible ctxt in
    let trusted_os := ctxt_oss ctxt in
    let smem := st_memory st in
    let mm := smem_mm smem in
    let ossmap := st_oss st in
    let aos := st_active_os st in
    let activity := st_aos_activity st in
    let mode := st_exec_mode st in
    let hyper := st_hypervisor st in

    (* Trusted OSs do not make hypercalls *)
    (* trusted_os_not_hypercall oss ossmap /\ *)

    (* If active OS is running, it should not have a hypercall. *)
    running_os_not_hypercall aos activity ossmap /\
    (* Hypervisor always runs in supervisor mode *)
    valid_hyper_exec_mode activity mode /\
    (* Trusted OSs also run in supervisor mode *)
    (* valid_trusted_os_exec_mode aos mode oss /\  *)

    (* OSs must run in user mode (if they are not waiting) *)
    valid_os_exec_mode aos activity mode ossmap /\

    valid_os_idents ossmap hyper /\

    (* Hypervisor should map an OS padd to a madd owned by that OS (in
       memory). *)
    valid_hypervisor hyper mm /\

    valid_pt_non_cacheable mm /\

    non_cacheable_synonyms hyper mm /\

    (* If a mapped virtual address is accessible by an OS, it is owned
       by that OS. If a mapped virtual address is not accessible by an
       OS, it is owned by the Hypervisor. *)
    valid_virtual_mapping mm hyper /\

    (* Current page table of any OS is of type PT, and owned by that OS. *)
    valid_current_page hyper mm ossmap /\
    (*Hypervisor Mappings must be injective for all os*)
    injective_hyper_mappings hyper /\
    (*Virtual memory must be assigned pinned Phisical memory*)
    va_has_valid_pa mm hyper /\

    (* Cache and TLB validity *)
    valid_cache st /\
    valid_cache_va_ma st /\
    valid_tlb st.

End ST_VALID.  

(* Declare Module ST_VALID:VALID_STATE. *)