(***********************************************************************
 * File: Stealth_actions.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * New StealthMem actions.
 *************************************************************************)
Require Export Actions Addressing Stealth_cache
  Opt_monad State State_valid State_tactics
  Exec_invariant Write_isolation OS_Isolation.

Declare Module C : STEALTH_CACHE PG P2M HYP.
Declare Module SMEM : SYSTEM_MEM PG C T MM.
Declare Module ST : STATE C SMEM.
Declare Module ACT : ACTIONS C SMEM ST.
Declare Module ST_VALID : ST_VALID C SMEM ST.

Module STAC := STAC C SMEM ST ACT ST_VALID.
Module EXEC_INV := EXEC_INV C SMEM ST ACT ST_VALID.
Module WRITE_ISO := WRITE_ISOLATION C SMEM ST ACT ST_VALID.
Module OS_ISO := OS_ISOLATION C SMEM ST ACT ST_VALID.

Module STEALTH_ACTIONS.
  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ACT.

  (* New *)
  Definition pre_all_new osi va s := 
       ctxt_oss ctxt osi = false (* untrusted osi *) 
       (** va is not reserved for the Hypervisor *)
    /\ os_accessible va
    /\ hyper_running s.

  Definition non_stealth_ma (s:state) (osi:os_ident) (ma:madd) : Prop :=
    forall pt_padd pt_addr,
      osi_pa_to_ma s osi pt_padd = Some pt_addr ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va <> Some ma.

  Definition pre_new_normal osi va pa s :=
       ~ in_stealth_va_line va
       (** pa should be mapped to a madd pinned by osi. *)
    /\ exists ma, (osi_pa_to_ma s osi pa = Some ma /\
         madd_type (smem_mm (st_memory s)) ma <> None /\
         non_stealth_ma s osi ma)

       (** Untrusted OS has a 'new' hypercall *)
    /\ osi_hypercall s osi = Some (Hyperv_call_new va pa)
    /\ pre_all_new osi va s.

  Definition pre_new_sm osi pa s :=
       (** stealth_va is free in the current PT *)
       exists pt_addr, (osi_current_page_table s osi = Some pt_addr
    /\ madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = None) 
       (** There is no alias pointing to ma, and the page is a
           cacheable RW page *)
    /\ exists ma, exists pg, (osi_pa_to_ma s osi pa = Some ma /\
         ~ shared_page s stealth_va ma /\
         get_mem (smem_mm (st_memory s)) ma = Some pg /\
         is_rw_page pg /\ is_cacheable pg)
       (** Untrusted OS has a 'new_sm' hypercall *)
    /\ osi_hypercall s osi = Some (Hyperv_call_new stealth_va pa)
    /\ pre_all_new osi stealth_va s.

  Definition new_sm (s:state) (osi:os_ident) (pt_addr:madd) (ma:madd) :=
    let smem' := new_not_shared (st_memory s) pt_addr stealth_va ma in
    let cache' : cache :=
        if (os_ident_eq osi (st_active_os s)) then
          match page_of_vadd_madd_smem (st_memory s) stealth_va ma with
            | None => (smem_c smem')
            | Some pg => fst (add_cache (smem_c (st_memory s)) (stealth_va,ma) pg)
          end
        else (smem_c smem')
      in

    let tlb' :=
      if (os_ident_eq (st_active_os s) osi)
      then fst (add_tlb (remove_tlb (smem_t smem') stealth_va) stealth_va ma)
      else smem_t smem' in
    SMem cache' tlb' (smem_mm smem').


  Definition post_new_sm osi (pa:padd) (s s':state) :=
  (* The mapping stealth_va->ma is added to the PT in pt_addr. hcall
     changes to None with an untrusted OS, otherwise it remain
     unchanged *)
  exists ma, exists pt_addr, exists os,
    osi_pa_to_ma s osi pa = Some ma /\
    osi_current_page_table s osi = Some pt_addr /\
    OSS.get (st_oss s) osi = Some os /\
    let newmem := new_sm s osi pt_addr ma in
    let newos := OS None (os_current_page os) in
    let newoss := (OSS.set (st_oss s) osi newos) in
      s' = State (st_active_os s) newoss (st_exec_mode s) (st_aos_activity s)
      (st_hypervisor s) newmem.

  
  (* Switch *)
  Definition remove_osi_stealth_page mem pt_addr : smem :=
    match get_mem (smem_mm mem) pt_addr with
      | Some pt_page =>
        match page_content pt_page with
          | PT_c pt =>
            match PT.get pt stealth_va with
              | Some ma => flush_ma mem ma
              | None => mem
            end
          | _ => mem
        end
      | None => mem
    end.

  Definition ctxt_switch_new_smem s new_pt_addr : smem :=
    match aos_current_page_table s with
      | Some aos_pt_addr =>
        let mem' := ctxt_switch (st_memory s) in
        let mem'' := remove_osi_stealth_page mem' aos_pt_addr in
          page_smem_smem (get mem'' stealth_va new_pt_addr)
      | None => st_memory s
    end.            

  Definition post_sm_switch osi s s' :=
    forall pt_addr,
      osi_current_page_table s osi = Some pt_addr ->
      let newmem := ctxt_switch_new_smem s pt_addr in
        s' = State osi (st_oss s) (st_exec_mode s)
               (st_aos_activity s) (st_hypervisor s) newmem.

  Definition post_sm_lswitch pa osi s s' :=
    forall pt_addr,
      osi_pa_to_ma s osi pa = Some pt_addr ->
    let os' := OS None pa in
    let new_oss := OSS.set (st_oss s) osi os' in
    let newmem := if (os_ident_eq osi (st_active_os s)) then
                    ctxt_switch_new_smem s pt_addr
                  else (st_memory s) in
      s' = State (st_active_os s) new_oss (st_exec_mode s)
             (st_aos_activity s) (st_hypervisor s) newmem.
  

  Inductive sm_action : Set :=
    (** action not modifying observable state *)
  | Action : action -> sm_action
  | New_normal : os_ident -> vadd -> padd -> sm_action
  | New_sm : os_ident -> padd -> sm_action
  | Switch_sm : os_ident -> sm_action
  | LSwitch_untrusted_sm : os_ident -> padd -> sm_action.

    (** Get the precondition of an sm_action *)
  Definition pre_sm_action (act:sm_action) : state -> Prop := 
    match act with
      | Action a => pre_action a
      | New_normal osi va pa => pre_new_normal osi va pa
      | New_sm osi pa => pre_new_sm osi pa
      | Switch_sm osi => pre_switch osi
      | LSwitch_untrusted_sm osi pa => pre_lswitch pa osi
    end.

  (** The family of post-conditions of sm_actions *)
  Definition post_sm_action (act:sm_action): state -> state -> Prop := 
    match act with
      | Action a => post_action a
      | New_normal osi va pa => post_new osi va pa (* Same postcondition *)
      | New_sm osi pa => post_new_sm osi pa 
      | Switch_sm osi => post_sm_switch osi
      | LSwitch_untrusted_sm osi pa => post_sm_lswitch pa osi
    end.

  (* The set of actions that are used defined as sm_actions:
     all except new, del, switch and lswitch. *)
  Definition defined_action (act:sm_action) : Prop :=
    match act with
      | Action a =>
        match a with
          | New_untrusted _ _ _ => False
          | New_hyper _ _ => False
          | Del_hyper _ => False
          | Switch _ => False
          | Lswitch_untrusted _ _ => False
          | _ => True
        end
      | _ => True
    end.

  Definition valid_sm_exec (s:state) (act:sm_action) (s':state) : Prop :=
    defined_action act /\ pre_sm_action act s /\
    post_sm_action act s s'.


  Ltac sm_action_induction a :=
    induction a; [
      Case "Action"%string
      | Case "New_normal"%string
      | Case "New_sm"%string
      | Case "Switch"%string
      | Case "Lswitch_untrusted"%string].

  (* Unfold sm_action pre and post-conditions,
     and simplify hypothesis, but do not unfold normal
     actions. *)
  Ltac destruct_sm_action_sem :=
    match goal with
      | [H: valid_sm_exec ?s ?a ?s' |-_] =>
        let Hdef := fresh "Hdef" in
        let Hpre := fresh "Hpre" in
        let Hpost := fresh "Hpost" in
        unfold valid_sm_exec, pre_sm_action, post_sm_action,
          pre_new_normal, pre_new_sm, pre_all_new,
          post_new, post_new_sm, post_all_new,
          post_sm_switch, post_sm_lswitch
        in H;
        destruct H as [Hdef [Hpre Hpost]]
    end; crush.

  (* Simplify all actions semantics (sm and normal) *)
  Ltac destruct_sm_and_normal_action_sem :=
    destruct_sm_action_sem;
    try match goal with
          | [ Hpre : pre_action ?a ?s,
              Hpost : post_action ?a ?s ?s' |- _ ] =>
          let Hexec := fresh "Hexec" in
            assert (valid_exec s a s');
              [unfold valid_exec; crush |
                match goal with
                  | [Case := "Action"%string : string|-_] =>
                    clear Case
                end; clear Hpre Hpost;
                STAC.action_induction a;
                STAC.destruct_action_sem]
        end; simpl in *.
End STEALTH_ACTIONS.



