(***********************************************************************
 * File: Stealth_cache.v
 * $Revision: 12920 $
 * $Date: 2016-09-22 18:44:11 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Refinement of the Cache for Stealth Mem
 * 
 *************************************************************************)

Require Import Arith Opt_monad Tactics List Sets.Constructive_sets.
Require Export Env Cache Page VCMaps.

Module Type STEALTH_CACHE (PG:PAGE) (P2M:PHYSICAL_MAP) (HYP:HYPERVISOR P2M) <: CACHE PG.
  Import P2M.
  Import HYP.
  Import PG.
  Include (CACHE PG).

  (* The set of valid cache lines. *)
  Parameter line : Set.
  (* Given a virtual address, returns the index where it should be cached. *)
  Parameter get_line : vadd -> line.

  Axiom line_dec : forall (l1 l2:line), {l1 = l2}+{l1 <> l2}.

  (* The stealth vadd (one per core) *)
  Parameter stealth_va : vadd.
  Axiom stealth_va_accessible : ctxt_vadd_accessible ctxt stealth_va = true.

  Definition in_stealth_va_line (va:vadd) : Prop :=
    get_line va = get_line stealth_va.

  Axiom cache_inertia : forall c va va' ma ma' pg pg',
      snd (add_cache c (va,ma) pg) = Some (va', ma', pg') ->
      get_line va = get_line va'.

  Definition observation : Type := Ensemble vadd.
  Definition empty_observation (o:observation) :=
    o = Empty_set vadd.

  (* Returns the entries of c owned by osi that are not in c', plus
     the ones from o *)
  Parameter get_cache_observation :
    os_ident -> cache -> cache -> observation -> observation.

  Axiom get_cache_observation_spec :
    forall osi c c' o,
      let o' := get_cache_observation osi c c' o in
        (forall va, In vadd o' va ->
          In vadd o va \/
          (exists ma, exists pg, 
            get_cache c (va,ma) = Some pg /\
            PG.page_owned_by pg = PG.Os_PO osi /\
            get_cache c' (va,ma) = None)) /\
        (forall va ma pg,
          get_cache c (va,ma) = Some pg ->
          PG.page_owned_by pg = PG.Os_PO osi ->
          get_cache c' (va,ma) = None ->
          In vadd o' va) /\
        (forall va,
          In vadd o va ->
          In vadd o' va).

  (* Cache entries are exactly the same in the given line, in a way
     that add_cache will select the same entry in both caches for
     replacement. This is cache implementation (and replacement
     policy) specific. *)
  Parameter same_cache_line : cache -> HYP.dict ->
                              cache -> HYP.dict -> line -> Prop.

  Axiom same_cache_line_entries_refl : forall c hyp hyp' l,
    same_cache_line c hyp c hyp' l.
  Axiom same_cache_line_entries_symm : forall c1 hyp1 c2 hyp2 l,
    same_cache_line c1 hyp1 c2 hyp2 l ->
    same_cache_line c2 hyp2 c1 hyp1 l.
  Axiom same_cache_line_entries_trans : forall c1 hyp1 c2 hyp2 c3 hyp3 l,
    same_cache_line c1 hyp1 c2 hyp2 l ->
    same_cache_line c2 hyp2 c3 hyp3 l ->
    same_cache_line c1 hyp1 c3 hyp3 l.

  Definition osi_pa_to_ma_hyp (hyp:HYP.dict) (osi:os_ident) (pa:padd) :=
    p2m <- HYP.get hyp osi;
    P2M.get p2m pa.
  
  Axiom same_cache_line_entries_some_pg :
    forall c1 c2 hyp1 hyp2 osi va va' pa ma1 ma2 pg1,
      get_line va = get_line va' ->
      same_cache_line c1 hyp1 c2 hyp2 (get_line va') ->
      osi_pa_to_ma_hyp hyp1 osi pa = Some ma1 ->
      osi_pa_to_ma_hyp hyp2 osi pa = Some ma2 ->
      get_cache c1 (va, ma1) = Some pg1 ->
      page_owned_by pg1 = Os_PO osi ->
      exists pg2, (get_cache c2 (va, ma2) = Some pg2 /\
        same_page_but_value pg1 pg2).

  Axiom same_cache_line_entries_no_pg :
    forall c1 c2 hyp1 hyp2 va pa ma1 ma2,
      same_cache_line c1 hyp1 c2 hyp2 (get_line va) ->
      (exists osi, osi_pa_to_ma_hyp hyp1 osi pa = Some ma1 /\
                   osi_pa_to_ma_hyp hyp2 osi pa = Some ma2) ->
      get_cache c1 (va, ma1) = None ->
      get_cache c2 (va, ma2) = None.

  (* Replaced cache entries after add is either Some entry with the
     same key, or None in both caches. *)
  Definition same_replaced_cache_entry
    (e1 e2:cache*option (cache_key*PG.page)) : Prop :=
    match e1,e2 with
      | (_, Some (key1,_)), (_, Some (key2,_)) =>
        key1 = key2
      | (_, None), (_, None) => True
      | _, _ => False
    end.

  Axiom same_cache_line_add_cache : forall c1 hyp1 c2 hyp2 va ma1 ma2 pg1 pg2,
    same_cache_line c1 hyp1 c2 hyp2 (get_line va) ->
    same_page_but_value pg1 pg2 ->
    same_replaced_cache_entry (add_cache c1 (va,ma1) pg1)
                              (add_cache c2 (va,ma2) pg2).

  (* same_cache_line is invariant under cache modification functions *)
  (* add_cache *)
  Axiom same_cache_line_add_cache_inv :
    forall c1 hyp1 c2 hyp2 l va ma1 ma2 pg1 pg2,
      same_cache_line c1 hyp1 c2 hyp2 l ->
      same_page_but_value pg1 pg2 ->
      same_cache_line (fst (add_cache c1 (va,ma1) pg1)) hyp1
                      (fst (add_cache c2 (va,ma2) pg2)) hyp2 l.
  Axiom same_cache_line_add_cache_other_line : forall c hyp va va' ma' pg,
    get_line va <> get_line va' ->
    same_cache_line c hyp (fst (add_cache c (va',ma') pg)) hyp (get_line va).

  (* remove_cache *)
  Axiom same_cache_line_remove_cache_inv :
    forall c1 hyp1 c2 hyp2 osi va va' pa ma1 ma2,
    same_cache_line c1 hyp1 c2 hyp2 (get_line va) ->
    osi_pa_to_ma_hyp hyp1 osi pa = Some ma1 ->
    osi_pa_to_ma_hyp hyp2 osi pa = Some ma2 ->
    same_cache_line (remove_cache c1 (va',ma1)) hyp1
                    (remove_cache c2 (va',ma2)) hyp2 (get_line va).
  Axiom same_cache_line_remove_cache_other_line : forall c hyp va va' ma',
    get_line va <> get_line va' ->
    same_cache_line c hyp (remove_cache c (va',ma')) hyp (get_line va).

  (* remove_ma_cache *)
  Axiom same_cache_line_remove_ma_cache_inv :
    forall c1 hyp1 c2 hyp2 va va' ma1 ma2 pg1 pg2,
      get_line va = get_line va' ->
      same_cache_line c1 hyp1 c2 hyp2 (get_line va) ->
      snd (remove_ma_cache c1 ma1) = Some ((va',ma1),pg1) ->
      snd (remove_ma_cache c2 ma2) = Some ((va',ma2),pg2) ->
      same_cache_line (fst (remove_ma_cache c1 ma1)) hyp1
                      (fst (remove_ma_cache c2 ma2)) hyp2
                      (get_line va).
  Axiom same_cache_line_remove_ma_cache_no_entry_inv :
    forall c1 hyp1 c2 hyp2 ma1 ma2 va,
      same_cache_line c1 hyp1 c2 hyp2 (get_line va) ->
      snd (remove_ma_cache c1 ma1) = None ->
      snd (remove_ma_cache c2 ma2) = None ->
      same_cache_line (fst (remove_ma_cache c1 ma1)) hyp1
                      (fst (remove_ma_cache c2 ma2)) hyp2
                      (get_line va).
  Axiom same_cache_line_remove_ma_cache_other_line :
    forall c hyp1 hyp2 va va' ma pg,
    snd (remove_ma_cache c ma) = Some ((va',ma),pg) ->
    get_line va <> get_line va' ->
    same_cache_line c hyp1 (fst (remove_ma_cache c ma)) hyp2 (get_line va).


End STEALTH_CACHE.
