(***********************************************************************
 * File: Stealth_traces.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Definition of traces with observations and stealth action semantics.
 *************************************************************************)

Require Import Sets.Constructive_sets Stealth_actions
  Stealth_valid Stealth_isolation Stealth_equivalence Streams
  Stealth_non_stealth_equivalence Stealth_valid_inv.

Module STEALTH_TRACES.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE VALID_SM_STATE_INV
         STEALTH_ISOLATION STEALTH_EQ STEALTH_NON_STEALTH_EQ.

  Definition leak_action : Type := observation -> sm_action.

  Definition leak_function : Type :=
    state -> observation -> state -> observation.
  Record leak_state : Type :=
  Leak_State {
    leak_state_state:state;
    leak_state_obs:observation
  }.


  Definition leak_valid_state (s:leak_state) : Prop :=
    valid_sm_state (leak_state_state s).


  Inductive leak_valid_exec : leak_state -> leak_action -> leak_function -> leak_state -> Prop :=
    Leak_valid_exec : forall (ls ls':leak_state) (f:leak_function) (A:leak_action),
      let s := (leak_state_state ls) in
      let s' := (leak_state_state ls') in
      let o := (leak_state_obs ls) in
      let o' := (leak_state_obs ls') in
      let a := A o in
        valid_sm_exec s a s' ->
        f s o s' = o' ->
        leak_valid_exec ls A f ls'.


  Record leak_exec_state : Type :=
    Leak_Ex_State
    {l_ex_state : leak_state;
     l_ex_action  : leak_action;
     l_ex_function : leak_function
  }.

  (* A valid initial state for the trace has a valid state and an
     empty observation *)
  Definition leak_init_state (execst:leak_exec_state) : Prop :=
    let s := l_ex_state execst in
    let o := leak_state_obs s in
      leak_valid_state s /\ empty_observation o.

  Definition leak_system_transition
    (src tgt : leak_exec_state) : Prop :=
      let s := l_ex_state src in
      let A := l_ex_action src in
      let f := l_ex_function src in
      let s' := l_ex_state tgt in
        leak_valid_exec s A f s'.

  Definition leak_state_prop := leak_exec_state -> Prop.

  (* Invariants *)

  Definition leak_leads_to (P Q : leak_state_prop) : Prop := 
    forall s s': leak_exec_state,
      P s ->  
      leak_system_transition s s' ->  
      Q s'.

  Definition leak_invariant: leak_state_prop -> Prop :=
    fun P => leak_leads_to P P.

  (* Streams of states, formulas and temporal operators *)
  Definition leak_state_stream := (Stream leak_exec_state).
  Definition leak_stream_prop := leak_state_stream->Prop.
  Definition leak_binary_state_prop :=
    leak_state -> leak_state -> Prop.
  Definition leak_lift_unary (P:leak_state_prop) :
    leak_stream_prop := fun str => P (hd str).
  Definition leak_lift_binary (P:leak_binary_state_prop) :
    leak_stream_prop := 
      fun str => P (l_ex_state (hd str)) (l_ex_state (hd (tl str))).

  (****** Traces and runs with leakage *********)

  (* A trace is a stream of execution states (si,Ai,fi) for i=1..,
   s.t. si -Ai,fi-> s(i+1), where -> is leak_valid_exec *)

  (* A  co-inductive fdefinition of trace *) 
  CoInductive leak_trace : leak_state_stream -> Prop :=
    Leak_trace : forall (s:leak_exec_state) (str:leak_state_stream),
      leak_trace str ->
      leak_system_transition s (hd str) ->
      leak_trace (Cons s str).

  (* leak_run: a leak_trace starting from an initial state *)       
  Definition leak_run (str:leak_state_stream): Prop := 
    leak_init_state (hd str) /\ leak_trace str.

  (* Runs and valid state *)
  Definition leak_valid_state_stream (str:leak_state_stream) := 
    leak_valid_state (l_ex_state (hd str)).

  Lemma state_of_run_is_valid: 
    forall str, leak_run str -> leak_valid_state_stream str.
  Proof.
    intro str; case str; crush.
    inversion H; unfold leak_init_state in *; crush.
  Qed.

  (* Invariant: All the states of a run are are valid states *)
  Lemma leak_states_of_run_are_valid : 
    forall str, leak_run str -> ForAll leak_valid_state_stream str.
  Proof.
    Lemma leak_state_of_trace_are_valid :
      forall str, leak_trace str ->
        leak_valid_state (l_ex_state (hd str)) ->
        ForAll leak_valid_state_stream str.
    Proof.
      cofix; intros.
      constructor; crush.
      apply leak_state_of_trace_are_valid;
        inversion H; crush.

      unfold leak_system_transition in H2.
      inversion_clear H2; subst.
      unfold leak_valid_state in *.
      apply valid_sm_state_inv
        with (leak_state_state (l_ex_state s))
             (l_ex_action s o); crush.
    Qed.

    intros.
    inversion H.
    unfold leak_init_state in H0; crush.
    clear H H2.
    apply leak_state_of_trace_are_valid; crush.
  Qed.


  (* A Property is safe for a run of the system if it holds for all
     state of the execution trace *)

  Definition leak_safe
    (P:leak_state -> leak_state -> Prop) : Prop := 
      forall str:leak_state_stream,
        leak_run str -> ForAll (leak_lift_binary P) str.

  (* cache_test *)
  (* Returns the observations corresponding to cache
     entries in s. *)
  Definition cache_test (osi:os_ident) (s:state) (o:observation) (s':state) :
    observation :=
      get_cache_observation osi (smem_c (st_memory s)) (smem_c (st_memory s')) o.
  

  CoInductive leak_osi_equivalent_traces (osi:os_ident) : leak_state_stream -> leak_state_stream -> Prop :=
    eq_trace_osi_action: forall (str1 str2:leak_state_stream)
      (A1 A2:leak_action) (s1 s2:leak_state),
      leak_osi_equivalent_traces osi str1 str2 -> 
      osi_sm_action osi (leak_state_state s1) (A1 (leak_state_obs s1)) ->
      osi_sm_action osi (leak_state_state s2) (A2 (leak_state_obs s2)) ->
      osi_stealth_equivalent (leak_state_state s1) (leak_state_state s2) osi ->
      leak_osi_equivalent_traces osi (Cons (Leak_Ex_State s1 A1 (cache_test osi)) str1)
                                 (Cons (Leak_Ex_State s2 A2 (cache_test osi)) str2)
  | eq_trace_not_osi_action1: forall (str1 str2:leak_state_stream)
    (A:leak_action) (s1 s2:leak_state),
    leak_osi_equivalent_traces osi str1 str2 -> 
    ~(osi_sm_action osi (leak_state_state s1) (A (leak_state_obs s1))) ->
    leak_osi_equivalent_traces osi (Cons (Leak_Ex_State s1 A (cache_test osi)) str1) str2
  | eq_trace_not_osi_action2: forall (str1 str2:leak_state_stream)
    (A:leak_action) (s1 s2:leak_state),
    leak_osi_equivalent_traces osi str1 str2 -> 
    ~(osi_sm_action osi (leak_state_state s2) (A (leak_state_obs s2))) ->
    leak_osi_equivalent_traces osi str1 (Cons (Leak_Ex_State s2 A (cache_test osi)) str2). 

End STEALTH_TRACES.


