(***********************************************************************
 * File: Machine_mem.v
 * $Revision: 12640 $
 * $Date: 2014-05-20 19:11:24 -0300 (Tue, 20 May 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the machine memory. 
 * 
 *************************************************************************)
Require Export Addressing Page Cache Tlb Machine_mem.
Require Import Opt_monad.

Module Type SYSTEM_MEM (PG:PAGE) (C:CACHE PG) (T:TLB) (MM:MACHINE_MEM PG).
  Import PG.
  Import C.
  Import T.
  Import MM.
  Module MM_EXT := MACHINE_MEM_EXT PG MM.
  Include MM_EXT.

  Record smem :=
    SMem {
      smem_c:cache;
      smem_t:tlb;
      smem_mm:mem
    }.


  Record page_smem :=
    Page_smem {
      page_smem_page : option page;
      page_smem_smem : smem
    }.

  (* Parameter read_RW : smem -> vadd -> madd -> page_smem. *)
  (* Parameter read_PT : smem -> vadd -> madd -> page_smem. *)
  Parameter get : smem -> vadd -> madd -> page_smem.

  Parameter set : smem -> vadd -> madd -> value -> smem.

  Parameter new_shared : smem -> madd -> vadd -> madd -> smem.
  Parameter new_not_shared : smem -> madd -> vadd -> madd -> smem.
  Parameter delete : smem -> madd -> vadd -> smem.

  Parameter pin : smem -> madd -> content -> page_owner -> smem .
  Parameter unpin : smem -> madd -> smem.

  Parameter ctxt_switch : smem -> smem.

  (* Just removes the given ma from the cache *)
  Parameter flush_ma : smem -> madd -> smem.


  (* Searches va in the TLB first, and then in memory *)
  Definition madd_of_vadd_smem (mem:smem) (pt_addr:madd) (va:vadd) : option madd :=
    match get_tlb (smem_t mem) va with
      | Some ma => Some ma
      | _ => madd_of_vadd (smem_mm mem) pt_addr va
    end.

  (* Searches for the page in the cache first, and then in memory *)
  Definition page_of_vadd_madd_smem (mem:smem) (va:vadd) (ma:madd) : option page :=
    match get_cache (smem_c mem) (va,ma) with
      | Some pg => Some pg
      | _ => get_mem (smem_mm mem) ma
    end.
  

  (* Searches for the madd in TLB or PT, then the page in the cache
     first, and then in memory *)
  Definition page_of_vadd_smem (mem:smem) (pt_addr:madd) (va:vadd) : option page :=
    ma <- madd_of_vadd_smem mem pt_addr va;
    page_of_vadd_madd_smem mem va ma.

  Axiom get_page_result : forall mem pt_addr va pg,
    page_of_vadd_smem mem pt_addr va = Some pg ->
    page_smem_page (get mem va pt_addr) = Some pg.

  (* The replaced entry of the cache is in the new mem. *)
  Axiom get_cache_entry1 : forall mem pt_addr va va' ma ma' pg pg',
    snd (add_cache (smem_c mem) (va,ma) pg) = Some (va',ma',pg') ->
    let mem' := page_smem_smem (get mem va pt_addr) in
      get_mem (smem_mm mem') ma' = Some pg'.

  (* All other memory pages remain unchanged *)
  Axiom get_cache_entry2 : forall mem pt_addr va va' ma ma' ma'' pg pg',
    snd (add_cache (smem_c mem) (va,ma) pg) = Some (va',ma',pg') ->
    ma' <> ma'' ->
    let mem' := page_smem_smem (get mem va pt_addr) in
      get_mem (smem_mm mem') ma'' = get_mem (smem_mm mem) ma''.

  (* If no cache entry is replaced, machine memory does not change. *)
  Axiom get_mem_unchanged : forall mem pt_addr va ma pg,
    snd (add_cache (smem_c mem) (va,ma) pg) = None ->
    let mem' := page_smem_smem (get mem va pt_addr) in
      smem_mm mem' = smem_mm mem.

  Axiom get_invalid : forall mem pt_addr va,
    page_of_vadd_smem mem pt_addr va = None ->
    let mem' := page_smem_smem (get mem va pt_addr) in
      mem' = mem.

  Axiom get_new_cache : forall mem pt_addr va ma pg val,
    madd_of_vadd_smem mem pt_addr va = Some ma ->
    page_smem_page (get mem va pt_addr) = Some pg ->
    page_content pg = RW_c (Some val) ->
    let mem' := page_smem_smem (get mem va pt_addr) in
      smem_c mem' = fst (add_cache (smem_c mem) (va,ma) pg).

  Axiom get_new_tlb : forall mem pt_addr va ma,
    madd_of_vadd_smem mem pt_addr va = Some ma -> 
    let mem' := page_smem_smem (get mem va pt_addr) in
      smem_t mem' = fst (add_tlb (smem_t mem) va ma).

  
  (* Set *)
  Axiom set_new_mmem : forall mem pt_addr va ma val old_pg,
    let mem' := set mem va pt_addr val in
      page_of_vadd_smem mem pt_addr va = Some old_pg ->
      madd_of_vadd_smem mem pt_addr va = Some ma ->
      exists mem_pg, get_mem (smem_mm mem') ma = Some mem_pg /\
        same_page_type mem_pg old_pg = true /\
        page_owned_by mem_pg = page_owned_by old_pg /\
        page_cacheable mem_pg = page_cacheable old_pg.

  Axiom set_new_mmem_non_cacheable : forall mem pt_addr va ma val old_pg,
      page_of_vadd_smem mem pt_addr va = Some old_pg ->
      madd_of_vadd_smem mem pt_addr va = Some ma ->
      ~ is_cacheable old_pg ->
      let mem' := set mem va pt_addr val in
      let new_pg := Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg) in
      smem_mm mem' = add_mem (smem_mm mem) ma new_pg.

  Axiom set_mem_unchanged1 : forall mem pt_addr va va' ma ma' ma'' val pg pg',
    let mem' := set mem va pt_addr val in
      madd_of_vadd_smem mem pt_addr va = Some ma ->
      snd (add_cache (smem_c mem) (va,ma) pg) = Some (va',ma',pg') ->
      ma'' <> ma -> ma'' <> ma' ->
      get_mem (smem_mm mem') ma'' = get_mem (smem_mm mem) ma''.

  Axiom set_mem_unchanged2 : forall mem pt_addr va ma ma' val pg,
    let mem' := set mem va pt_addr val in
      madd_of_vadd_smem mem pt_addr va = Some ma ->
      snd (add_cache (smem_c mem) (va,ma) pg) = None ->
      ma' <> ma -> 
      get_mem (smem_mm mem') ma' = get_mem (smem_mm mem) ma'.

  (* The replaced entry of the cache is in the new mem. *)
  Axiom set_cache_entry1 : forall mem pt_addr val va va' ma ma' pg pg',
    snd (add_cache (smem_c mem) (va,ma) pg) = Some (va',ma',pg') ->
    let mem' := set mem va pt_addr val in
      get_mem (smem_mm mem') ma' = Some pg'.

  Axiom set_invalid : forall mem pt_addr va val,
    page_of_vadd_smem mem pt_addr va = None ->
    let mem' := set mem va pt_addr val in
      mem' = mem.

  Axiom set_new_cache : forall mem pt_addr va ma pg old_val new_val,
    madd_of_vadd_smem mem pt_addr va = Some ma ->
    page_of_vadd_smem mem pt_addr va = Some pg ->
    page_content pg = RW_c (Some old_val) ->
    let mem' := set mem va pt_addr new_val in
    let new_pg := Page (RW_c (Some new_val)) (page_owned_by pg) (page_cacheable pg) in
      smem_c mem' = fst (add_cache (smem_c mem) (va,ma) new_pg).

  Axiom set_new_tlb : forall mem pt_addr va ma val,
    madd_of_vadd_smem mem pt_addr va = Some ma -> 
    let mem' := set mem va pt_addr val in
      smem_t mem' = fst (add_tlb (smem_t mem) va ma).

  (* New *)
  Definition remove_new_va_cache (mem:smem) (pt_addr:madd) (va:vadd) :=
    match madd_of_vadd (smem_mm mem) pt_addr va with
      | None => smem_c mem
      | Some ma' =>
    match get_cache (smem_c mem) (va,ma') with
      | None => smem_c mem
      | Some pg => remove_cache (smem_c mem) (va,ma')
    end end.
  
  Definition remove_new_va_mem (mem old_mem:mem) (c:cache) (pt_addr:madd) (va:vadd) :=
    match madd_of_vadd old_mem pt_addr va with
      | None => mem
      | Some ma' =>
    match get_cache c (va,ma') with
      | None => mem
      | Some pg => add_mem mem ma' pg
    end end.
  
  Axiom new_mem_not_shared : forall mem mm' pt_addr va ma,
    add_page_table_entry (smem_mm mem) pt_addr va ma = Some mm' ->
    smem_mm (new_not_shared mem pt_addr va ma) =
      remove_new_va_mem mm' (smem_mm mem) (smem_c mem) pt_addr va.

  Axiom new_mem_shared : forall mem mm' mm'' mm''' pt_addr va ma,
    update_cache_entry (smem_mm mem) (snd (remove_ma_cache (smem_c mem) ma)) = mm' ->
    mark_page_as_non_cacheable mm' ma = Some mm'' ->
    add_page_table_entry mm'' pt_addr va ma = Some mm''' ->
    smem_mm (new_shared mem pt_addr va ma) =
      remove_new_va_mem mm''' (smem_mm mem) (smem_c mem) pt_addr va.

  Axiom new_mem_not_shared_invalid : forall mem pt_addr va ma,
    add_page_table_entry (smem_mm mem) pt_addr va ma = None ->
    smem_mm (new_not_shared mem pt_addr va ma) = smem_mm mem.

  Axiom new_mem_shared_invalid : forall mem pt_addr va ma,
    add_page_table_entry (smem_mm mem) pt_addr va ma = None ->
    smem_mm (new_shared mem pt_addr va ma) = smem_mm mem.

  Axiom new_cache_not_shared : forall mem pt_addr va ma,
      smem_c (new_not_shared mem pt_addr va ma) = remove_new_va_cache mem pt_addr va.

  Axiom new_cache_shared : forall mem pt_addr va ma,
    smem_c (new_shared mem pt_addr va ma) = (fst (remove_ma_cache (remove_new_va_cache mem pt_addr va) ma)).

  Axiom new_tlb_not_shared : forall mem pt_addr va ma,
      smem_t (new_not_shared mem pt_addr va ma) = smem_t mem.

  Axiom new_tlb_shared : forall mem pt_addr va ma,
      smem_t (new_shared mem pt_addr va ma) = smem_t mem.

  (* Delete *)
  Axiom delete_mem_some_entry : forall mem mm' pt_addr va ma pg,
    madd_of_vadd (smem_mm mem) pt_addr va = Some ma ->
    get_cache (smem_c mem) (va,ma) = Some pg ->
    remove_page_table_entry (smem_mm mem) pt_addr va = Some mm' ->
    smem_mm (delete mem pt_addr va) = add_mem mm' ma pg.

  Axiom delete_mem_no_entry : forall mem mm' pt_addr va ma,
    madd_of_vadd (smem_mm mem) pt_addr va = Some ma ->
    get_cache (smem_c mem) (va,ma) = None ->
    remove_page_table_entry (smem_mm mem) pt_addr va = Some mm' ->
    smem_mm (delete mem pt_addr va) = mm'.

  Axiom delete_mem_not_mapped : forall mem pt_addr va,
    madd_of_vadd (smem_mm mem) pt_addr va = None ->
    smem_mm (delete mem pt_addr va) = smem_mm mem.

  Axiom delete_mem_invalid : forall mem pt_addr va,
    remove_page_table_entry (smem_mm mem) pt_addr va = None ->
    smem_mm (delete mem pt_addr va) = smem_mm mem.

  Axiom delete_cache : forall mem pt_addr va ma,
    madd_of_vadd (smem_mm mem) pt_addr va = Some ma ->
    let mem' := delete mem pt_addr va in
      smem_c mem' = remove_cache (smem_c mem) (va,ma).

  Axiom delete_tlb : forall mem pt_addr va,
    let mem' := delete mem pt_addr va in
      smem_t mem' = (smem_t mem).
  

  (* Pin *)
  Axiom pin_mem : forall mem ma type owner,
    let mem' := pin mem ma type owner in 
    let pg := Page type owner true in
      smem_mm mem' = add_mem (smem_mm mem) ma pg.

  Axiom pin_cache : forall mem ma type owner,
    let mem' := pin mem ma type owner in 
      smem_c mem' = smem_c mem.

  Axiom pin_tlb : forall mem ma type owner,
    let mem' := pin mem ma type owner in 
      smem_t mem' = smem_t mem.

  (* Unpin *)
  Axiom unpin_mem : forall mem ma,
    let mem' := unpin mem ma in 
    let pg := Page Other_c No_Owner true in
      smem_mm mem' = add_mem (smem_mm mem) ma pg.

  Axiom unpin_cache : forall mem ma,
    let mem' := unpin mem ma in 
      smem_c mem' = smem_c mem.

  Axiom unpin_tlb : forall mem ma,
    let mem' := unpin mem ma in 
      smem_t mem' = smem_t mem.
  
  (* Contex switch *)
  Axiom ctxt_switch_mem : forall mem,
    let mem' := ctxt_switch mem in
      smem_mm mem' = smem_mm mem.

  Axiom ctxt_switch_cache : forall mem,
    let mem' := ctxt_switch mem in
      smem_c mem' = smem_c mem.

  Axiom ctxt_switch_tlb : forall mem,
    let mem' := ctxt_switch mem in
      smem_t mem' = empty_tlb.

  (* Flush ma *)
  Axiom flush_ma_mem : forall mem ma,
    smem_mm (flush_ma mem ma) =
    update_cache_entry (smem_mm mem)
      (snd (remove_ma_cache (smem_c mem) ma)).

  Axiom flush_ma_cache : forall mem ma,
    smem_c (flush_ma mem ma) = fst (remove_ma_cache (smem_c mem) ma).

  Axiom flush_ma_tlb : forall mem ma,
    smem_t (flush_ma mem ma) = smem_t mem.

End SYSTEM_MEM.
  
