(***********************************************************************
 * File: Tactics.v
 * $Revision $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *******************************************************************
 * This file specifies general automation tactics
 * 
 ******************************************************************)
Require Export Omega Arith List String.
Require Import Opt_monad Page.

(* Fails if hypothesis P is present in the current context. *)
Ltac notHyp P :=
  match goal with
    | [ _ : P |- _ ] => fail 1
    | _ =>
      match P with
        | ?P1 /\ ?P2 => first [ notHyp P1 | notHyp P2 | fail 2 ]
        | _ => idtac
      end
  end.

Ltac move_to_top x :=
  match reverse goal with
  | H : _ |- _ => try move x after H
  end.

(* Add string to context with description of current case for case
   analisys proofs. *)
Tactic Notation "assert_eq" ident(x) constr(v) :=
  let H := fresh in
  assert (x = v) as H by reflexivity;
  clear H.

Tactic Notation "Case_aux" ident(x) constr(name) :=
  first [
    set (x := name); move_to_top x
  | assert_eq x name; move_to_top x
  | fail 1 "because we are working on a different case" ].

Tactic Notation "Case" constr(name) := Case_aux Case name.
Tactic Notation "SCase" constr(name) := Case_aux SCase name.

Ltac clear_all_hyp :=
  repeat (match goal with
            | [ H: _ |- _ ] => clear H
          end).

(* Option monad crush tactic *)
Ltac opt_monad_crush :=
  repeat (match goal with
      | [H: ?e = Some ?v
         |- context[option_bind ?f ?e]] =>
          rewrite H; simpl
      | [H: Some ?v = ?e 
         |- context[option_bind ?f ?e]] =>
          rewrite <- H; simpl
      | [H1: Some ?v = ?e,
         H2: context[option_bind ?f ?e] |- _] =>
          rewrite <- H1 in H2; simpl in H2
      | [H1: ?e = Some ?v,
         H2: context[option_bind ?f ?e] |- _] =>
          rewrite H1 in H2; simpl in H2
  end); simpl_monad.

Ltac unify_opt_val :=
  try (match goal with
      | [H1: ?e = Some ?val1,
         H2: Some ?val2 = ?e |-_] =>
          rewrite H1 in H2; injection H2;
            intros; subst val2; clear H2
      | [H1: Some ?val1 = ?e,
         H2: Some ?val2 = ?e |-_] =>
          rewrite <- H1 in H2; injection H2;
            intros; subst val2; clear H2
      | [H1: ?e = Some ?val1,
         H2: ?e = Some ?val2 |-_] =>
          rewrite H1 in H2; injection H2;
            intros; subst val2; clear H2
      | [H1: ?e = Some ?val1,
         H2: None = ?e |-_] =>
          rewrite H1 in H2; discriminate H2
      | [H1: ?e = Some ?val1,
         H2: ?e = None |-_] =>
          rewrite H1 in H2; discriminate H2
      | [H1: Some ?val1 = ?e,
         H2: None = ?e |-_] =>
          rewrite <- H1 in H2; discriminate H2
  end).


Ltac crush :=
  let my_intuition := intros; try match goal with | [|- ~ ?E] => intro end;
      eauto; (* firstorder; *)
      try subst; simpl in *; try subst; (*intuition; *)
      try congruence; opt_monad_crush; unify_opt_val; auto in
    my_intuition; repeat (my_intuition;
      match goal with
        | [ H: ?P /\ ?Q |- _ ] => destruct H
        | [ |- context[?P /\ ?Q]] => split
        | [ H1: ?P -> ?Q, H2: ?P |-_] => apply H1 in H2
        | [ H: Some _ = Some _ |- _] => try (injection H; intros); clear H

        | [ H : ex _ |- _] => destruct H
        | [ H: ex2 _ _ |-_] => destruct H
        | [ |- ex (fun x => (?P ?i = ?P x)) ] => exists i
        | [ |- ex (fun x => (?P x = ?P ?i)) ] => exists i
        | [H: ?e = Some ?X |- ex (fun x => (?e = Some x /\ _)) ] =>
          exists X; split
        | [H: Some ?X = ?e |- ex (fun x => (?e = Some x /\ _)) ] =>
          exists X; split

        | [H: (?v1,?v2) = (?val1,?val2) |-_ ] =>
          injection H; intros; clear H
        | [ |- context[if ?E then _ else _]] =>
          let val := fresh "val" in
            remember E as val; destruct val
        | [ H : context[if ?E then _ else _] |- _] =>
          let val := fresh "val" in
            remember E as val; destruct val
        | [ H : context[match ?E with Some _ => _ | None => _ end] |- _] =>
          let val := fresh "val" in
            remember E as val; destruct val
        | [ |- context[match ?E with Some _ => _ | None => _ end]] =>
          let val := fresh "val" in
            remember E as val; destruct val
 
       | [ H: context[match ?E with pair _ _ => _ end] |-_] =>
         match goal with
           | [H': (?x,?y) = ?E |-_]=> rewrite <- H' in H
           | [H': ?E = (?x,?y) |-_]=> rewrite H' in H
           | |-_ =>
             let val := fresh "val" in
               remember E as val; destruct val
         end
        | [ |- context[match ?E with pair _ _ => _ end]] =>
         match goal with
           | [H': (?x,?y) = ?E |-_]=> rewrite <- H'
           | [H': ?E = (?x,?y) |-_]=> rewrite H'
           | |-_ =>
          let val := fresh "val" in
            remember E as val; destruct val
         end
        | [H1: (?x,?y) = ?E, H2: context[(fst ?E)]|- _] =>
          rewrite <- H1 in H2
        | [H1: (?x,?y) = ?E, H2: context[(snd ?E)]|- _] =>
          rewrite <- H1 in H2          
        | [H1: (?x,?y) = ?E |- context[(fst ?E)]] =>
          rewrite <- H1
        | [H1: (?x,?y) = ?E |- context[(snd ?E)]] =>
          rewrite <- H1

        | [H: False |-_] => elim H

      end; my_intuition); my_intuition; try omega.


