Require Import Coqlib.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Values.
Require Import Events.
Require Import Memory.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Op.
Require Import Registers.
Require Import Locations.
Require Import Conventions.
Require Import MachIR.
Require Import MSets.
Require Import MSetAVL.

Inductive pointsto_symbol :=
| Global : ident -> pointsto_symbol
| Stack  : Z -> pointsto_symbol.

Inductive var_type :=
| VReg : mreg -> var_type
| VMem : pointsto_symbol -> var_type.

Module Var_Type_Leibniz.
  Definition t := var_type.

  (* Equality *)
  Definition eq (v1 v2:t):= v1 = v2.

  Lemma eq_equiv : Equivalence eq.
  Proof.
    constructor; unfold Reflexive, Symmetric, Transitive, eq; auto.
    intros; transitivity y; auto.
  Qed.
  
  (* Less than *)
  Definition mreg_to_pos (r:mreg) : positive :=
    (match r with
       | AX => 1 | BX => 2 | SI => 3 | DI => 4 | BP => 5
       | X0 => 6 | X1 => 7 | X2 => 8 | X3 => 9 | X4 => 10 | X5 => 11 | X6 => 12 | X7 => 13
       | CX => 14 | DX => 15
       | FP0 => 16
     end)%positive. 

  Definition MRlt (r1 r2:mreg) : Prop :=
    Plt (mreg_to_pos r1) (mreg_to_pos r2).

  Definition lt (v1 v2:t) := 
    match v1, v2 with
      | VReg r1, VReg r2 =>
        MRlt r1 r2
      | VReg _, VMem _ => True
      | VMem s1, VMem s2 =>
        match s1, s2 with
          | Global i1, Global i2 =>
            Plt i1 i2
          | Stack z1, Stack z2 =>
            z1 < z2
          | Global _, Stack _ => True
          | _, _ => False
        end
      | _, _ => False
    end.

  Lemma lt_strorder : StrictOrder lt.
  Proof.
    constructor; unfold Irreflexive, Reflexive, Transitive, complement;
      intros; destruct x; try destruct y; try destruct z; simpl in *; eauto;
      try destruct p; try destruct p0; try destruct p1; simpl in *; eauto;
      try (unfold MRlt in *; eapply Plt_strict; eauto; fail);
      try (eapply Z.lt_strorder; eauto; fail);
      try match goal with | [H: False |-_] => elim H end;
      unfold MRlt in *; apply Plt_trans with (mreg_to_pos m0); auto.
  Qed.    

  Lemma lt_compat : Proper (eq ==> eq ==> iff) lt.
  Proof.
    now unfold eq; split; simpl in *; subst; auto.
  Qed.

  Definition compare v1 v2 :=
    match v1, v2 with
      | VReg r1, VReg r2 =>
        (mreg_to_pos r1 ?= mreg_to_pos r2)%positive
      | VReg _, VMem _ => Lt
      | VMem s1, VMem s2 =>
        match s1, s2 with
          | Global i1, Global i2 =>
            (i1 ?= i2)%positive
          | Stack z1, Stack z2 =>
            z1 ?= z2
          | Global _, Stack _ => Lt
          | Stack _, Global _ => Gt
        end
      | VMem _, VReg _ => Gt
    end.

  Lemma compare_spec : forall x y, CompSpec eq lt x y (compare x y).
  Proof.
   intros; case_eq (compare x y); constructor;
     unfold eq, lt; destruct x; try destruct y; simpl in *; auto;
     try discriminate.
   assert (mreg_to_pos m = mreg_to_pos m0); [apply Pos.compare_eq |]; subst; auto.
   destruct m; destruct m0; auto; try inversion H0.
   destruct p; try destruct p0; simpl in *; auto; try discriminate.
   assert (i = i0); [apply Pos.compare_eq |]; subst; auto.
   assert (z=z0); [apply Z.compare_eq |]; subst; auto.
   destruct p; try destruct p0; simpl in *; auto; try discriminate.
   apply Pos.compare_gt_iff; auto.
   destruct p; destruct p0; simpl in *; auto.
   apply Pos.compare_gt_iff; auto.
   discriminate.
   elim (Z.compare_gt_iff z z0); auto.
  Qed.

  Lemma eq_dec : forall x y, {eq x y}+{~eq x y}.
  Proof.
    intros; destruct x; destruct y; unfold eq;
    repeat decide equality; auto.
  Qed.

  Definition eq_leibniz a b (H:eq a b) := H.

End Var_Type_Leibniz.

Module VS := MakeWithLeibniz Var_Type_Leibniz.

