(***********************************************************************
 * File: State_tactics.v
 * $Revision: 12918 $
 * $Date: 2016-09-21 22:27:32 -0300 (Wed, 21 Sep 2016) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * State related tactics.
 **********************************************************************)
Require Export Opt_monad Tactics.
Require Import State State_valid Cache_tactics.

Module STAC (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM)
  (ST:STATE C SMEM) (ACT:ACTIONS C SMEM ST) (ST_VALID : ST_VALID C SMEM ST).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import ACT.

  Include CACHE_TACTICS PG C.
  

  Ltac destruct_valid_state_H H :=
    let Hvalid := fresh "H" in
    let H_running_nothypcall := fresh "H_running_nothypcall" in
    let H_valid_exmd := fresh "H_valid_exmd" in
    let H_valid_osexmd := fresh "H_valid_osexmd" in
    let H_valid_os_idents := fresh "H_valid_os_idents" in
    let H_valid_hyper := fresh "H_valid_hyper" in
    let H_valid_pt_non_cacheable := fresh "H_valid_pt_non_cacheable" in
    let H_syn := fresh "H_syn" in
    let H_valid_virtmap := fresh "H_valid_virtmap" in
    let H_valid_currpg := fresh "H_valid_currpg" in
    let H_injective_hyper := fresh "H_injective_hyper" in
    let H_va_has_pa := fresh "H_va_has_pa" in
    let H_valid_cache := fresh "H_valid_cache" in
    let H_valid_cache_va_ma := fresh "H_valid_cache_va_ma" in
    let H_valid_tlb := fresh "H_valid_tlb" in
      generalize H; intro Hvalid;
        destruct Hvalid as 
          (H_running_nothypcall,
          (H_valid_exmd,
          (H_valid_osexmd,
          (H_valid_os_idents,
          (H_valid_hyper,
          (H_valid_pt_non_cacheable,
          (H_syn,
          (H_valid_virtmap,
          (H_valid_currpg,
          (H_injective_hyper,
          (H_va_has_pa,
          (H_valid_cache,
          (H_valid_cache_va_ma,
           H_valid_tlb))))))))))))).
    

  (* Unfold valid_state hypothesis, and make one hypothesis for each
     condition. *)
  Ltac destruct_valid_state :=
    match goal with
      | [H: valid_state ?s |-_] =>
        destruct_valid_state_H H
    end.

  Ltac action_induction a :=
    induction a; [
      Case "Silent"%string
      | Case "Read"%string
      | Case "Read_hyper"%string
      | Case "Write"%string
      | Case "Write_hyper"%string
      | Case "New_untrusted"%string
      | Case "New_hyper"%string
      | Case "Del_untrusted"%string
      | Case "Del_hyper"%string
      | Case "Switch"%string
      | Case "Lswitch_untrusted"%string
      | Case "Hcall"%string
      | Case "Ret_ctrl"%string
      | Case "Chmod"%string
      | Case "Page_pin_untrusted"%string
      | Case "Page_unpin_untrusted"%string].

  (* Unify variables in read postcondition using the precondition. *)
  Ltac unify_read_post :=
    match goal with
      | [H1: aos_page_of_vadd_smem ?s ?v = Some ?pg,
         H2: forall pt_addr:madd, aos_current_page_table ?s = Some pt_addr
           -> ?s' = _ |- _] =>
      let pt_addr := fresh "pt_addr" in
      let H3 := fresh "H" in
        generalize H1; intro H3; unfold aos_page_of_vadd_smem in H1;
          unfold aos_page_of_vadd_smem in H1;
            remember (aos_current_page_table s) as opt_pt_addr;
              destruct opt_pt_addr as [pt_addr|]; crush;
                specialize (H2 pt_addr (eq_refl (Some pt_addr)))
    end.

  (* Unify variables in write postcondition using the precondition. *)
  Ltac unify_write_post :=
    match goal with
      | [H1: aos_page_of_vadd_smem ?s ?v = Some ?pg,
         H2: forall (pt_addr:madd) (pg:page),
             aos_current_page_table ?s = Some pt_addr ->
             aos_page_of_vadd_smem ?s ?v = Some pg ->
               ?s' = _ |- _] =>
      let pt_addr := fresh "pt_addr" in
      let H3 := fresh "H" in
        generalize H1; intro H3; unfold aos_page_of_vadd_smem in H1;
            remember (aos_current_page_table s) as opt_pt_addr;
              destruct opt_pt_addr as [pt_addr|]; crush;
                specialize (H2 pt_addr pg (eq_refl (Some pt_addr)) H3)
    end.

  (* Unfold action pre and post-conditions, and simplify hypothesis. *)
  Ltac destruct_action_sem :=
    match goal with
      | [H: valid_exec ?s ?a ?s' |-_] =>
        let Hpre := fresh "Hpre" in
        let Hpost := fresh "Hpost" in
        unfold valid_exec, pre_action,
          post_action, pre_silent, post_silent,
          pre_read, post_read, pre_read_hyper,
          post_read_hyper, pre_all_read, post_all_read,
          pre_write, post_write, pre_write_hyper, pre_all_write,
          post_write_hyper, post_all_write,
          pre_new, post_new,
          pre_new_hyper, post_new_hyper, post_all_new,
          pre_switch, post_switch, 
          pre_lswitch, post_lswitch,
          pre_del, post_del, 
          pre_del_hyper, post_del_hyper, pre_all_del, 
          pre_hcall, post_hcall, pre_ret_ctrl, post_ret_ctrl,
          pre_chmod, post_chmod, pre_pin, post_pin,
          pre_unpin, post_unpin
        in H;
        destruct H as [Hpre Hpost]
    end; crush; try unify_read_post; try unify_write_post; crush.


  (* Generates hypothesis with the AOS record and p2m mapping. *)
  Ltac get_aos_data :=
      match goal with
        | [ H1: valid_os_idents (st_oss ?s) (st_hypervisor ?s) |- _ ] =>
          let H2 := fresh "H" in
          let H_oss := fresh "H_oss" in
          let H_hyp := fresh "H_hyp" in
          let aos := fresh "aos" in
          let p2m := fresh "p2m" in
          destruct H1 with (st_active_os s) as [aos H2];
            destruct H2 as [H_oss [p2m H_hyp]]
      end.
  
  (* Generate hypothesis for the current PT of the AOS, given
     valid_current_page. *)
  Ltac get_aos_current_pt_page :=
      match goal with
        | [ H1: valid_current_page (st_hypervisor ?s)
                 (smem_mm (st_memory ?s)) (st_oss ?s),
            aos: OSS.info,
            H_oss: OSS.get (st_oss ?s) (st_active_os ?s) = Some ?aos,
            p2m: HYP.info,
            H_hyp: HYP.get (st_hypervisor ?s) (st_active_os ?s) = Some ?p2m
          |- _ ] =>
            let pt_addr := fresh "pt_addr" in
            let pt_page := fresh "pt_page" in
            let pt := fresh "pt" in
            let H2 := fresh "H" in
            let H3 := fresh "H" in
            let H4 := fresh "H" in
            let H5 := fresh "H" in
            let H6 := fresh "H" in
          destruct H1 with (st_active_os s) aos p2m as
            [pt_addr [pt_page [H2 [H3 [[pt H5] H6]]]]]; auto
        | [ H1: valid_current_page (st_hypervisor ?s)
                 (smem_mm (st_memory ?s)) (st_oss ?s),
            aos: OSS.info,
            H_oss: Some ?aos = OSS.get (st_oss ?s) (st_active_os ?s),
            p2m: HYP.info,
            H_hyp: Some ?p2m = HYP.get (st_hypervisor ?s) (st_active_os ?s)
          |- _ ] =>
            let pt_addr := fresh "pt_addr" in
            let pt_page := fresh "pt_page" in
            let pt := fresh "pt" in
            let H2 := fresh "H" in
            let H3 := fresh "H" in
            let H4 := fresh "H" in
            let H5 := fresh "H" in
            let H6 := fresh "H" in
          destruct H1 with (st_active_os s) aos p2m as
            [pt_addr [pt_page [H2 [H3 [[pt H5] H6]]]]]; auto
      end; unify_opt_val.


  (* Generate hypothesis for the current PT of some osi, given
     valid_current_page. *)
    Ltac get_osi_current_pt_page osi :=
      match goal with
        | [ H1: valid_current_page (st_hypervisor ?s)
                 (smem_mm (st_memory ?s)) (st_oss ?s),
            os: OSS.info,
            H_oss: OSS.get (st_oss ?s) osi = Some ?os,
            p2m: HYP.info,
            H_hyp: HYP.get (st_hypervisor ?s) osi = Some ?p2m
          |- _ ] =>
            let pt_addr := fresh "pt_addr" in
            let pt_page := fresh "pt_page" in
            let pt := fresh "pt" in
            let H2 := fresh "H" in
            let H3 := fresh "H" in
            let H4 := fresh "H" in
            let H5 := fresh "H" in
            let H6 := fresh "H" in
          destruct H1 with osi os p2m as
            [pt_addr [pt_page [H2 [H3 [[pt H5] H6]]]]]; auto
        | [ H1: valid_current_page (st_hypervisor ?s)
                 (smem_mm (st_memory ?s)) (st_oss ?s),
            os: OSS.info,
            H_oss: OSS.get (st_oss ?s) osi = Some ?os,
            p2m: HYP.info,
            H_hyp: Some ?p2m = HYP.get (st_hypervisor ?s) osi
          |- _ ] =>
            let pt_addr := fresh "pt_addr" in
            let pt_page := fresh "pt_page" in
            let pt := fresh "pt" in
            let H2 := fresh "H" in
            let H3 := fresh "H" in
            let H4 := fresh "H" in
            let H5 := fresh "H" in
            let H6 := fresh "H" in
          destruct H1 with osi os p2m as
            [pt_addr [pt_page [H2 [H3 [[pt H5] H6]]]]]; auto
      end; unify_opt_val.


  (* If pt_addr is the AOS current page, prove that
     Some pt_addr = aos_current_page_table s. *)
  Ltac get_aos_current_pt_page_f :=
      match goal with
        | [ H_oss: OSS.get (st_oss ?s) (st_active_os ?s) = Some ?aos,
            H_hyp: HYP.get (st_hypervisor ?s) (st_active_os ?s) = Some ?p2m,
            H_pt_addr_map: Some ?pt_addr = P2M.get ?p2m (os_current_page ?aos)
          |- _ ] =>
        assert (Some pt_addr = aos_current_page_table s);
          [unfold aos_current_page_table, osi_current_page_table;
            rewrite H_oss; simpl; rewrite H_hyp; simpl; auto; fail
              | idtac]
      end; unify_opt_val.


  (* Simplify aos_page_of_vadd_smem by unifying with the
     aos_current_page_table. *)
  Ltac simpl_aos_page_of_vadd_smem :=
      match goal with
        | [ H1: Some ?pt_addr = aos_current_page_table ?s,
            H2: aos_page_of_vadd_smem ?s ?va = Some ?pg |-_ ] =>
        unfold aos_page_of_vadd_smem in H2;
          rewrite <- H1 in H2; simpl in H2
      end.

  (* Make explicit the PT mapping that the madd_of_vadd predicate implies. *)
  Ltac destruct_madd_of_vadd H :=
    match type of H with
      | madd_of_vadd (smem_mm (st_memory ?s)) ?pt_addr ?va = Some ?ma =>
        let opt_get_mem := fresh "opt_get_mem" in
          let pt_page := fresh "pt_page" in
            let pt_content := fresh "pt_content" in
              unfold madd_of_vadd, madd_page_table in H; crush;
                remember (get_mem (smem_mm (st_memory s)) pt_addr) as opt_get_mem;
                  destruct opt_get_mem as [pt_page|]; crush;
                    try (remember (page_content pt_page) as pt_content;
                      destruct pt_content; crush)
    end.

  (* If va is mapped to ma in smem, then it is mapped to ma in the
     current PT of the AOS. *)
  Ltac from_smem_to_mem_virt_map :=
    match goal with
      | [H1: Some ?ma = madd_of_vadd_smem (st_memory ?s) ?pt_addr ?va,
         H2: aos_current_page_table ?s = Some ?pt_addr,
         H3: valid_tlb ?s |-_] =>
         let opt_ma := fresh "opt_ma" in 
           assert (Some ma = madd_of_vadd (smem_mm (st_memory s)) pt_addr va);
           [unfold madd_of_vadd_smem in H1;
            remember (get_tlb (smem_t (st_memory s)) va) as opt_ma;
            induction opt_ma; [|auto; fail];
            injection H4; intros; subst ma; erewrite H3; auto; fail|]
    end.

  (* Generate the pg that is mapped from va in the virtmap of an osi. *)
  Ltac get_curr_pt_va_pg :=
    match goal with
      | [H_virtmap: valid_virtual_mapping (smem_mm (st_memory ?s))
                                          (st_hypervisor ?s),
         H_oss: Some ?os = OSS.get (st_oss ?s) ?osi,
         H_hyp: Some ?p2m = HYP.get (st_hypervisor ?s) ?osi,
         H_pt_addr_map: P2M.get ?p2m (os_current_page ?os) = Some ?pt_addr,
         H_pt_page_map: Some ?pt_page = get_mem (smem_mm (st_memory ?s)) ?pt_addr,
         H_pt_page_content: page_content ?pt_page = PT_c ?pt,
         H_ma_map: Some ?ma = madd_of_vadd (smem_mm (st_memory ?s)) ?pt_addr ?va |-_] =>
          destruct H_virtmap with osi pt_addr pt_page
            pt va ma as [pg [H_pg_map [H_acc H_nacc]]]; [crush;fail|auto;fail|auto;fail|auto;fail|
          unfold madd_of_vadd, madd_page_table in H_ma_map;
          rewrite <- H_pt_page_map in H_ma_map; simpl in H_ma_map;
          rewrite H_pt_page_content in H_ma_map; auto; fail
          | ]
      end.

  (* If va maps to ma in smem for the AOS, then its owner is AOS. *)
  Lemma va_aos_curr_pg_owner : forall s ma va pt_addr,
      os_accessible va ->
      valid_virtual_mapping (smem_mm (st_memory s)) (st_hypervisor s) ->
      valid_current_page (st_hypervisor s) (smem_mm (st_memory s)) (st_oss s) ->
      valid_tlb s ->
      aos_current_page_table s = Some pt_addr ->
      Some ma = madd_of_vadd_smem (st_memory s) pt_addr va ->
      madd_owner (smem_mm (st_memory s)) ma = Os_PO (st_active_os s).
  Proof.
    intros; from_smem_to_mem_virt_map.
    unfold aos_current_page_table, osi_current_page_table in *.
    simpl_monad.
    get_aos_current_pt_page.
    get_curr_pt_va_pg.
    unfold madd_owner; crush.
  Qed.

  (* If a madd maps to two pages pg pg' with two vadd, the owners are
     the same. *)
  Lemma page_same_owner_cache_mem : forall s va va' ma pg pg',
    valid_state s ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg ->
    page_of_vadd_madd_smem (st_memory s) va' ma = Some pg' ->
    page_owned_by pg = page_owned_by pg'.
  Proof.
    intros.
    destruct_valid_state.
    unfold page_of_vadd_madd_smem in *; crush.
    unfold valid_cache in H_valid_cache.
    destruct H_valid_cache with va ma pg; crush.
    destruct H_valid_cache with va' ma pg'; crush.
    destruct H_valid_cache with va' ma pg'; crush.
    destruct H_valid_cache with va ma pg; crush.
  Qed.    


  (* Try to instantiate va_aos_curr_pg_owner lemma with proper
     variables to prove that the owner of a page is the AOS. *)
  Ltac apply_va_aos_curr_pg_owner :=
      match goal with
        | [ H1: os_accessible ?va,
            H2: valid_virtual_mapping (smem_mm (st_memory ?s)) (st_hypervisor ?s),
            H3: valid_current_page (st_hypervisor ?s)
                  (smem_mm (st_memory ?s)) (st_oss ?s),
            H4: valid_tlb ?s,
            H5: Some ?ma = madd_of_vadd_smem (st_memory ?s) ?pt_addr ?va,
            H6: Some ?pt_page = get_mem (smem_mm (st_memory ?s)) ?pt_addr,
            H7: page_owned_by ?pt_page = Os_PO ?osi,
            H8: page_of_vadd_madd_smem (st_memory ?s) ?va ?ma = Some ?pg
            |- page_owned_by ?pg = Os_PO ?osi] =>

        let H_madd_owner := fresh "H_madd_owner" in
        let opt_pg := fresh "opt_pg" in

        assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO (st_active_os s))
               as H_madd_owner;
            [apply va_aos_curr_pg_owner with va pt_addr; auto; fail | idtac];
        unfold madd_owner in H_madd_owner;
        rewrite <- H7 in H_madd_owner;
        remember (get_mem (smem_mm (st_memory s)) ma) as opt_pg;
        induction opt_pg; [|rewrite H7 in H_madd_owner; discriminate]
      end.


  (* Prove that the owner of (va,ma) in smem is the same as the owner
     of ma in mmem, given a valid cache. *)
  Ltac same_cache_mem_owner :=
      match goal with
        | [ H1: page_of_vadd_madd_smem (st_memory ?s) ?va ?ma =
                  Some ?pg1,
            H2: Some ?pg2 = get_mem (smem_mm (st_memory ?s)) ?ma,
            H3: valid_cache ?s |-_] =>
        let opt_pg := fresh "opt_pg" in
        let pg1' := fresh "pg1'" in
        let pg2' := fresh "pg2'" in
        let H_pg1'_type := fresh "H_pg1'_type" in
        let H_pg2'_map := fresh "H_pg2'_map" in
        let H_pg2'_type := fresh "H_pg2'_type" in
        let H_pg2'_owner := fresh "H_pg2'_owner" in

        assert (page_owned_by pg1 = page_owned_by pg2);
        [  
        unfold page_of_vadd_madd_smem in H1;
        remember (get_cache (smem_c (st_memory s)) (va,ma))
          as opt_pg;
        induction opt_pg as [pg1'|];
          [(* (va,ma) is mapped to pg in the cache *)
           injection H1; intros; subst pg1'; clear H1;
           unfold valid_cache in H3;
           destruct H3 with va ma pg1
             as [H_pg1'_type [H_pg1'_cacheable
               [pg2' [H_pg2'_map [H_pg2'_type [H_pg2'_cacheable H_pg2'_owner]]]]]];
             clear H3; auto;
           rewrite <- H_pg2'_map in H2;
             injection H2; intros; subst; auto
           
          |(* ma is mapped to pg in memory *)
            rewrite <- H2 in H1; injection H1;
              intros; subst; auto]
          | idtac]
      end.



  Lemma valid_state_aos_page_owner :
    forall s va pg, valid_state s ->
      aos_page_of_vadd_smem s va = Some pg ->
      os_accessible va ->
      page_owned_by pg = Os_PO (st_active_os s).
  Proof.
    intros.
    destruct_valid_state.
    get_aos_data.
    get_aos_current_pt_page.
    get_aos_current_pt_page_f.
    simpl_aos_page_of_vadd_smem.
    unfold page_of_vadd_smem in *; simpl_monad.
    apply_va_aos_curr_pg_owner.
    same_cache_mem_owner; crush.
  Qed.

  Ltac syn_non_cacheable :=
    match goal with
      | [H_pt_page1_map: Some ?pt_page1 = get_mem (smem_mm (st_memory ?s)) ?pt_addr1,
        H_mp1_map: PT_c ?mp1 = page_content ?pt_page1,
        H_ma_map1: PT.get ?mp1 ?va1 = Some ?ma,

        H_pt_page2_map: Some ?pt_page2 = get_mem (smem_mm (st_memory ?s)) ?pt_addr2,
        H_mp2_map: page_content ?pt_page2 = PT_c ?mp2,
        H_ma_map2: PT.get ?mp2 ?va2 = Some ?ma,

        H_pg_map: Some ?pg = get_mem (smem_mm (st_memory ?s)) ?ma,
        H_syn: non_cacheable_synonyms (st_hypervisor ?s) (smem_mm (st_memory ?s)) |-_]=>
        assert (~PG.is_cacheable pg); [
        generalize H_syn; intro;
          apply (H_syn pt_addr1 pt_page1 mp1 va1 ma pg
            pt_addr2 pt_page2 mp2 va2); crush|];
      try (unfold aos_current_page_table, osi_current_page_table, osi_pa_to_pg, osi_pa_to_ma in *; crush; match goal with 
               |[H1: P2M.get ?p2m ?pt_padd = Some ?pt_addr,
                 H2: Some ?p2m = HYP.get (st_hypervisor ?s) ?osi |-
                 ex (fun p2m' => 
                 ex (fun osi' => 
                 ex (fun pt_padd' =>
                       (HYP.get (st_hypervisor ?s) osi' = Some p2m' /\
                        P2M.get p2m' pt_padd' = Some ?pt_addr))))] =>
                exists p2m; exists osi; exists pt_padd; crush
               |[H1: P2M.get ?p2m ?pt_padd = Some ?pt_addr,
                 H2: HYP.get (st_hypervisor ?s) ?osi = Some ?p2m |-
                 ex (fun p2m' => 
                 ex (fun osi' => 
                 ex (fun pt_padd' =>
                       (HYP.get (st_hypervisor ?s) osi' = Some p2m' /\
                        P2M.get p2m' pt_padd' = Some ?pt_addr))))] =>
                exists p2m; exists osi; exists pt_padd; crush
               |[H1: Some ?pt_addr = P2M.get ?p2m ?pt_padd,
                 H2: Some ?p2m = HYP.get (st_hypervisor ?s) ?osi |-
                 ex (fun p2m' => 
                 ex (fun osi' => 
                 ex (fun pt_padd' =>
                       (HYP.get (st_hypervisor ?s) osi' = Some p2m' /\
                        P2M.get p2m' pt_padd' = Some ?pt_addr))))] =>
                exists p2m; exists osi; exists pt_padd; crush
           end; fail)
    end.


  Ltac aux_non_cacheable_pg_in_cache :=
    match goal with
      | [H_non_cacheable: ~ PG.is_cacheable ?pg,
        H_pg_map: Some ?pg = get_mem (smem_mm (st_memory ?s)) ?ma,
        H_ma_map: PT.get ?mp ?va = Some ?ma,
        H_pg_cache: get_cache (smem_c (st_memory ?s)) (?va, ?ma) = Some ?pg_cache,
        H_valid_cache : valid_cache ?s |-_] =>
      generalize H_valid_cache; intro;
        specialize (H_valid_cache va ma pg_cache (eq_sym H_pg_cache));
          crush
    end.
           
  Ltac synonym_cached :=
      match goal with 
        | [H_ma_map : madd_of_vadd (smem_mm (st_memory ?s)) ?pt_addr ?va1 = Some ?ma,
           H_ma_cache : get_cache (smem_c (st_memory ?s)) (?va2, ?ma) = Some ?pg,
           H_va1_neq_va2 : ?va1 <> ?va2,
           H_valid_cache : valid_cache ?s,
           H_valid_cache_va_ma : valid_cache_va_ma ?s |- _] =>

        let opt_pg := fresh "opt_pg" in
          unfold valid_cache_va_ma in H_valid_cache_va_ma;
            destruct H_valid_cache_va_ma with va2 ma pg;
              unfold osi_pa_to_pg in *; crush;
              destruct_madd_of_vadd H_ma_map;
              remember (get_mem (smem_mm (st_memory s)) ma) as opt_pg;
                destruct opt_pg;
                  [syn_non_cacheable; aux_non_cacheable_pg_in_cache|
                    destruct H_valid_cache with va2 ma pg; crush]
      end.


Lemma no_cache_alias : forall s va va' ma pg pg', valid_state s ->
  get_cache (smem_c (st_memory s)) (va,ma) = Some pg -> 
  get_cache (smem_c (st_memory s)) (va',ma) = Some pg' -> 
  va = va' /\ pg = pg'.
Proof. 
  intros. 
  assert (va=va'); [case (vadd_eq va va')|]; crush.
  STAC.destruct_valid_state. 
  generalize H_valid_cache_va_ma. 
  unfold valid_cache_va_ma in H_valid_cache_va_ma. 
  destruct H_valid_cache_va_ma with va ma pg; unfold osi_pa_to_pg in *; crush.
  cut (madd_of_vadd (smem_mm (st_memory s)) i0 va = Some ma); intros.
  STAC.synonym_cached.
  unfold madd_of_vadd, madd_page_table; crush;
    rewrite H3; crush. 
Qed.
  
Lemma madd_of_vadd_smem_then_mmem : forall s pt_addr va ma,
  valid_state s ->
  aos_current_page_table s = Some pt_addr ->
  madd_of_vadd_smem (st_memory s) pt_addr va = Some ma ->
  madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.
Proof.
  intros.
  unfold madd_of_vadd_smem in *; crush.
  destruct_valid_state.
  apply H_valid_tlb; crush.
Qed.


(* Read lemmas *)

  (* After a read, all pages have the same value (pages can move
     around, but not change). *)
  Lemma read_same_mem : forall s va ma va' ma' pt_addr pt_addr' pg val, 
    valid_state s ->

    (exists osi, exists pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
    (exists osi', exists pt_padd', osi_pa_to_ma s osi' pt_padd' = Some pt_addr') ->
   
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd_smem (st_memory s) pt_addr' va' = Some ma' ->
    page_of_vadd_madd_smem (st_memory s) va' ma' = Some pg ->
    page_content pg = RW_c (Some val) ->

    page_of_vadd_madd_smem (st_memory s) va ma =
    page_of_vadd_madd_smem
    (page_smem_smem (get (st_memory s) va' pt_addr')) va ma.
  Proof.
    intros.
    unfold page_of_vadd_madd_smem; 
      rewrite (get_new_cache (st_memory s) pt_addr' va' ma' pg val);
        auto; [| apply get_page_result; unfold page_of_vadd_smem]; 
          crush;
          try (remember (page_cacheable pg);
           destruct b;
            [symmetry in Heqb; fold (is_cacheable pg) in Heqb |
             rewrite get_add_cache_non_cacheable with (smem_c (st_memory s))
                 (va',ma') (va,ma) pg in Heqval0; crush; fail]).
                 
  (* Cache, Cache *)
    case (eq_cache_key (va',ma') (va,ma)); crush.
    rewrite <- get_add_cache_cacheable with (smem_c (st_memory s)) (va,ma) pg in Heqval1; crush.
    unfold page_of_vadd_madd_smem in H4; crush.
    destruct_add_cache (smem_c (st_memory s));
      generalize (get_add_cache_distinct (smem_c (st_memory s))
        (va',ma') (va,ma) pg p); intros.
    
    rewrite <- Heqopt_add_cache in H6.
    rewrite <- H6 in Heqval1; clear H6; crush.
    right; exists (v,m); exists p1; crush.
    (* entry is in the new cache... absurd *)
    generalize (add_cache_entry_behavior_full (smem_c (st_memory s)) (va',ma') pg);
 intros.
    rewrite <- Heqopt_add_cache in H6.
    destruct H6; try some_entry_add_cache; crush.

    (* No entry returned *)
    rewrite <- Heqopt_add_cache in H6.
    rewrite <- H6 in Heqval1; clear H6; crush.
    
  (* Cache, Memory *)
    symmetry in Heqval0.
    symmetry in Heqval1.
    destruct_add_cache (smem_c (st_memory s)).
    removed_val_is_entry.
    generalize (get_cache_entry1 (st_memory s) pt_addr'
      va' va ma' ma pg p); crush.
    rewrite H6; auto; clear H6.
    no_entry_but_removed.

  (* Memory, Cache *)
    unfold page_of_vadd_madd_smem in H4;
      crush; destruct_add_cache (smem_c (st_memory s)); new_entry_is_added;
      generalize (get_add_cache_cacheable (smem_c (st_memory s)) (va,ma) pg);
        crush;
        generalize (get_add_cache_non_cacheable (smem_c (st_memory s))
          (va,ma) (va,ma) pg);
      crush; unfold is_cacheable in *;
        destruct (page_cacheable pg);
          try (rewrite <- H6 in Heqval1; crush; fail);
            try (rewrite <- H7 in Heqval1; crush; fail).

    destruct_add_cache (smem_c (st_memory s)).
    symmetry.
    apply (get_cache_entry2 (st_memory s) pt_addr' va' v ma' m ma pg p);
      crush.

    2: generalize (get_mem_unchanged (st_memory s) pt_addr' va' ma' pg); crush;
      try rewrite H6; crush.
    case (vadd_eq va v); crush.
    rewrite cache_deleted_entry_was_there with (smem_c (st_memory s)) 
      va' ma' (v,ma) pg p in Heqval0; crush.
    (* The entry and our page were synonyms *)
    generalize (cache_deleted_entry_was_there
      (smem_c (st_memory s)) va' ma' (v,ma) pg p); intros;
    crush.
    specialize (H6 (eq_refl (Some (v,ma,p)))).
    destruct_valid_state.
    synonym_cached.
  Qed.      

  Lemma read_same_madd_of_vadd : forall s va va' pt_addr pt_addr',
     valid_state s ->
     aos_current_page_table s = Some pt_addr' ->
     madd_of_vadd (smem_mm (st_memory s)) pt_addr va =
     madd_of_vadd
       (smem_mm (page_smem_smem (get (st_memory s) va' pt_addr')))
       pt_addr va.
  Proof.
    intros.
    unfold madd_of_vadd, madd_page_table; crush.
    remember (page_of_vadd_smem (st_memory s) pt_addr' va').
    destruct o; crush.
    remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va').
    destruct o; crush.
    remember (add_cache (smem_c (st_memory s)) (va',m) p).
    destruct p0; crush.
    destruct o; crush.
    destruct p0. destruct c0.
    STAC.simpl_some_cache_entry.
    remember (madd_eq m0 pt_addr).
    destruct s0; crush.
    rewrite get_cache_entry1 with (va':=v) (ma:=m) (ma':=pt_addr) (pg:=p) (pg':=x0); crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with v pt_addr x0; crush.
    unfold is_rw_page in *; crush.
    rewrite H6; rewrite H1; crush.
    rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=p) (pg':=x0); crush.
    remember (get_mem (smem_mm (st_memory s)) pt_addr).
    destruct o; crush.
    rewrite get_mem_unchanged with (ma:=m) (pg:=p); crush.
    rewrite get_mem_unchanged with (ma:=m) (pg:=p); crush.
    rewrite <- Heqo1; crush.

    unfold page_of_vadd_smem in *; crush.
    unfold madd_of_vadd_smem in *; crush.
    
    STAC.destruct_valid_state.
    unfold valid_tlb in H_valid_tlb.
    specialize H_valid_tlb with va' m0 pt_addr'; crush.
    rewrite get_invalid; crush.
  Qed.

Lemma read_no_va_mapped_same_mem :
  forall s osi va' pt_addr' pa ma,
    valid_state s ->
    aos_current_page_table s = Some pt_addr' ->
    ACT.no_va_mapped_to_ma osi ma s ->
    osi_pa_to_ma s osi pa = Some ma ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (page_smem_smem (get (st_memory s) va' pt_addr'))) ma.
Proof.
  intros.
  remember (page_of_vadd_smem (st_memory s) pt_addr' va').
  destruct o; crush.
  remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va').
  destruct o; crush.
  remember (add_cache (smem_c (st_memory s)) (va',m) p).
  destruct p0; crush.
  destruct o; crush.
  destruct p0. destruct c0.
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache_va_ma with v ma x0; crush.
  destruct (os_ident_eq x osi); crush.
  destruct H1 with x1 x2 v x3; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap with x i1 x2 x3 v ma; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt v) eqn:?; unfold os_accessible in *; auto.
  rewrite H10 in *; crush.
  rewrite H11 in *; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=p) (pg':=x0); crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=p); crush.
  rewrite get_invalid; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in Heqo0; crush.
  rewrite <- Heqo0; crush.
  rewrite get_invalid; crush.
Qed.  

Lemma read_no_va_mapped :
  forall s s' va' osi ma,
    valid_state s ->
    ACT.pre_all_read va' s ->
    ACT.post_all_read va' s s' ->
    (ACT.no_va_mapped_to_ma osi ma s <->
    ACT.no_va_mapped_to_ma osi ma s').
Proof.
  unfold pre_all_read, post_all_read; crush;
  unify_read_post; crush.
  destruct_valid_state.
  get_aos_data; crush.
  get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.

  split; intros; unfold no_va_mapped_to_ma; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_of_vadd (smem_mm
    (page_smem_smem (get (st_memory s) va' pt_addr0))) i0 va = Some ma).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H9; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 va = Some ma).
  erewrite read_same_madd_of_vadd; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H5 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.

  unfold osi_pa_to_pg in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 va = Some ma).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H9; crush.
  assert (madd_of_vadd (smem_mm
    (page_smem_smem (get (st_memory s) va' pt_addr0))) i0 va = Some ma).
  erewrite <- read_same_madd_of_vadd; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H5 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.
Qed.

Lemma read_no_pt_change : 
  forall s pt_addr pt_addr' va' ma' pg',
    valid_state s ->
    madd_of_vadd_smem (st_memory s) pt_addr' va' = Some ma' ->
    page_of_vadd_madd_smem (st_memory s) va' ma' = Some pg' ->
    is_rw_page pg' ->

    madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
    get_mem (smem_mm (st_memory s)) pt_addr = 
    get_mem (smem_mm (page_smem_smem (get (st_memory s) va' pt_addr'))) pt_addr.
Proof.
  intros.
  STAC.destruct_valid_state.

  destruct (snd(add_cache (smem_c (st_memory s)) (va',ma') pg')) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  rewrite get_cache_entry2 with (va':=v) (ma:=ma') (ma':=m) (pg:=pg') (pg':=p0); crush.
  STAC.simpl_some_cache_entry.
  destruct H_valid_cache with v pt_addr x0; crush.
  unfold is_rw_page, madd_type in *; crush.
  rewrite H9 in *; crush.
  rewrite get_mem_unchanged with (ma:=ma') (pg:=pg'); crush.
Qed.

Lemma read_same_madd_type : 
  forall s pt_addr' va' ma ma' pg',
    valid_state s ->
    madd_of_vadd_smem (st_memory s) pt_addr' va' = Some ma' ->
    page_of_vadd_madd_smem (st_memory s) va' ma' = Some pg' ->
    is_rw_page pg' ->

    madd_type (smem_mm (st_memory s)) ma = 
    madd_type (smem_mm (page_smem_smem (get (st_memory s) va' pt_addr'))) ma.
Proof.
  unfold madd_type; crush.
  STAC.destruct_valid_state.

  destruct (snd(add_cache (smem_c (st_memory s)) (va',ma') pg')) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m ma); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=ma') (pg:=pg') (pg':=x0); crush.
  destruct H_valid_cache with v ma x0; crush.
  unfold is_rw_page in *; crush; rewrite H3, H8; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=ma') (ma':=m) (pg:=pg') (pg':=x0); crush.
  rewrite get_mem_unchanged with (ma:=ma') (pg:=pg'); crush.
Qed.


(* Write lemmas *)
Lemma write_same_madd_type :
  forall s ma pt_addr va val,
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    madd_type (smem_mm (set (st_memory s) va pt_addr val)) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.
  unfold madd_type; crush.
  remember (page_of_vadd_smem (st_memory s) pt_addr va).
  destruct o; crush.
  remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr va).
  destruct o; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m)
    (Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))).
  destruct p0; crush.
  destruct o; crush.
  destruct p0. destruct c0.
  assert (get_cache (smem_c (st_memory s)) (v,m0) = Some p0).
  apply STAC.cache_deleted_entry_was_there with va m
      (Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p)); crush.
  destruct (madd_eq m0 ma); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H5; crush.
  rewrite set_cache_entry1 with (va':=v) (ma:=m)
    (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))) (pg':=p0); crush.
  rewrite H2; crush.
  destruct (madd_eq m ma); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val p; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in Heqo0; crush.
  unfold page_of_vadd_smem in Heqo; crush.
  assert (m = ma); crush.
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m1) in Heqo0; crush.
  
  unfold page_of_vadd_madd_smem in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H9; crush.
  unfold same_page_type in *; crush.
  rewrite H6 in *; crush.
  destruct (page_content x); crush.
  unfold same_page_type in *; destruct (page_content x);
    destruct (page_content p); crush.

  erewrite set_mem_unchanged1 with (va':=v) (pg':=p0) (ma:=m)
    (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m1) in *; crush.

  destruct (madd_eq m ma); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val p; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in Heqo0; crush.
  unfold page_of_vadd_smem in Heqo; crush.
  assert (m = ma); crush.
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in *; crush.
  unfold page_of_vadd_madd_smem in Heqo; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma p0; crush.
  unfold is_rw_page, same_page_type in *; crush.
  rewrite H8 in *; crush.
  rewrite H5 in *; crush.
  destruct (page_content x); crush.
  unfold same_page_type in *; crush;
    destruct (page_content x); destruct (page_content p); crush.

  rewrite set_mem_unchanged2 with (ma:=m)
    (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in *; crush.

  rewrite set_invalid; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in *; crush.
  rewrite <- Heqo0; crush.

  rewrite set_invalid; crush.
Qed.

  Lemma write_same_madd_of_vadd_ne_pt_addr : forall s va va' pt_addr pt_addr' val,
     valid_state s ->
     aos_current_page_table s = Some pt_addr' ->
     madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' <> Some pt_addr ->
     madd_of_vadd (smem_mm (st_memory s)) pt_addr va =
     madd_of_vadd
       (smem_mm (set (st_memory s) va' pt_addr' val))
       pt_addr va.
  Proof.
    intros.
    unfold madd_of_vadd, madd_page_table; crush.
    remember (page_of_vadd_smem (st_memory s) pt_addr' va').
    destruct o; crush.
    remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va').
    destruct o; crush.
    remember (add_cache (smem_c (st_memory s)) (va',m)
      (Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))).
    destruct p0; crush.
    destruct o; crush.
    destruct p0. destruct c0.
    assert (get_cache (smem_c (st_memory s)) (v,m0) = Some p0).
    apply STAC.cache_deleted_entry_was_there with va' m
      (Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p)); crush.
    remember (madd_eq m0 pt_addr).
    destruct s0; crush.
    rewrite set_cache_entry1 with (va':=v) (ma:=m) (ma':=pt_addr)
      (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))) (pg':=p0); crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with v pt_addr p0; crush.
    unfold is_rw_page in *; crush.
    rewrite H6; rewrite H3; crush.
    remember (madd_eq pt_addr m).
    destruct s0; crush.
    rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0)
      (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))) (pg':=p0); crush.
    unfold madd_of_vadd_smem; crush.
    STAC.destruct_valid_state.
    rewrite (H_valid_tlb va' m1 pt_addr') in Heqo0; crush.

    rewrite set_mem_unchanged2 with (ma:=m)
      (pg:=(Page (RW_c (Some val)) (page_owned_by p) (page_cacheable p))); crush.
    unfold madd_of_vadd_smem; crush.
    STAC.destruct_valid_state.
    rewrite H_valid_tlb with (ma:=m0) in Heqo0; crush.

    unfold page_of_vadd_smem in Heqo; crush.
    unfold madd_of_vadd_smem in *; crush.
    STAC.destruct_valid_state.
    rewrite H_valid_tlb with (ma:=m0) in Heqo0; crush.

    rewrite set_invalid; crush.
  Qed.    

  Lemma write_same_madd_of_vadd : forall s va va' pt_addr pt_page pt_addr' old_pg val,
     valid_state s ->
     aos_current_page_table s = Some pt_addr' ->
     aos_page_of_vadd_smem s va' = Some old_pg ->
     is_rw_page old_pg ->
     get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
     is_pt_page pt_page ->
     madd_of_vadd (smem_mm (st_memory s)) pt_addr va =
     madd_of_vadd
       (smem_mm (set (st_memory s) va' pt_addr' val))
       pt_addr va.
    Proof.
      intros.
      apply write_same_madd_of_vadd_ne_pt_addr; crush.
      unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
      unfold madd_of_vadd_smem in *; crush.
      STAC.destruct_valid_state.
      rewrite H_valid_tlb with (ma:=m0) in H5; crush.
      unfold page_of_vadd_madd_smem in *; crush.
      destruct H_valid_cache with va' pt_addr old_pg; crush.
      unfold is_rw_page, is_pt_page in *; crush.
      unfold is_rw_page, is_pt_page in *; crush.
      unfold page_of_vadd_madd_smem in *; crush.
      STAC.destruct_valid_state.
      destruct H_valid_cache with va' pt_addr old_pg; crush.
      unfold is_rw_page, is_pt_page in *; crush.
      unfold is_rw_page, is_pt_page in *; crush.
    Qed.


Lemma absurd_cache_get_entry_pt :
  forall s va ma pg va_removed ma_removed pg_removed pg_mem c mp,
    valid_state s ->
    (c, Some (va_removed, ma_removed, pg_removed)) =
      add_cache (smem_c (st_memory s)) (va, ma) pg ->
    get_mem (smem_mm (st_memory s)) ma_removed = Some pg_mem ->
    ~ PT_c mp = page_content pg_mem.
Proof.
  intros.
  assert (get_cache (smem_c (st_memory s)) (va_removed, ma_removed) = Some pg_removed).
  apply STAC.cache_deleted_entry_was_there with va ma pg; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va_removed ma_removed pg_removed; crush.
  unfold is_rw_page in *; crush.
Qed.

Ltac apply_absurd_cache_get_entry_pt :=
  match goal with 
    | [H_valid: valid_state ?s,
       H_cache: (?c, Some (?va_removed, ?ma_removed, ?pg_removed)) =
                add_cache (smem_c (st_memory ?s)) (?va, ?ma) ?pg,
       H_get_mem: get_mem (smem_mm (st_memory ?s)) ?ma_removed = Some ?pg_mem,
       H_pt_page: PT_c ?mp = page_content ?pg_mem |- _ ] =>
      elim absurd_cache_get_entry_pt
        with s va ma pg va_removed ma_removed pg_removed pg_mem c mp;
        crush
  end.

Lemma write_smem_mem_neq : forall s va ma va' ma' pg pt_addr pt_addr' val,
   valid_state s ->

   (exists osi, exists pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->

   madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
   aos_current_page_table s = Some pt_addr' ->
   madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma' ->
   page_of_vadd_madd_smem (st_memory s) va' ma' = Some pg ->
   is_rw_page pg ->
   ma <> ma' ->
   page_of_vadd_madd_smem (set (st_memory s) va' pt_addr' val) va ma =
   page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  intros.
  unfold page_of_vadd_madd_smem; crush.
  (* Cache, Cache *)
  generalize (get_add_cache_distinct_not_new (smem_c (st_memory s)) (va',ma') (va,ma)
    (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)) p);
    crush.
  rewrite <- H7 in Heqval1; crush; clear H7.
  assert (c = fst (add_cache (smem_c (st_memory s)) (va', ma')
         (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))));
    crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=ma') (pg:=pg) (old_val:=x1) in Heqval0; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.

  (* Mem, Cache *)
  generalize (get_add_cache_distinct_not_new (smem_c (st_memory s)) (va',ma') (va,ma)
    (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)) p);
    crush.
  rewrite <- H7 in Heqval1; crush; clear H7.
  assert (c = fst (add_cache (smem_c (st_memory s)) (va', ma')
         (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))));
    crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=ma') (pg:=pg) (old_val:=x1) in Heqval0; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.

  (* Cache, Mem *)
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=ma') (pg:=pg) (old_val:=x1) in Heqval0; crush.
  apply set_cache_entry1 with (va':=va) (ma:=ma')
    (pg:=(Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))); crush.
  apply STAC.cache_deleted_entry; crush.
  destruct (page_cacheable pg) eqn:?; crush.
  unfold is_cacheable; crush.
  rewrite <- get_add_cache_non_cacheable in Heqval0; crush.
  unfold is_cacheable in *; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.

  (* Mem, Mem *)
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=ma') (pg:=pg) (old_val:=x1) in Heqval0; crush.
  destruct (add_cache (smem_c (st_memory s)) (va', ma')
    (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))) eqn:?;
    crush.
  destruct o. destruct p. destruct c0.
  apply set_mem_unchanged1 with (va':=v) (ma:=ma') (ma':=m) (pg':=p)
    (pg:=Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in H3; crush.
  rewrite Heqp; crush.
  assert (get_cache (smem_c (st_memory s)) (v, m) = Some p); crush.
  rewrite STAC.cache_deleted_entry_was_there with (va:=va') (ma:=ma')
    (pg:=Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)) (pg':=p); crush.
  rewrite Heqp; crush.  
  destruct (vadd_eq va v); crush.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  apply set_mem_unchanged2 with (ma:=ma')
    (pg:=Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
  rewrite Heqp; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H3; crush.
Qed.

Lemma write_smem_mem_eq : forall s va ma va' pt_addr pt_addr' val old_pg,
   valid_state s ->
   
   (exists osi, exists pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr') ->   

   madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma ->
   madd_of_vadd_smem (st_memory s) pt_addr va' = Some ma ->
   aos_current_page_table s = Some pt_addr ->
   page_of_vadd_madd_smem (st_memory s) va ma = Some old_pg ->
   is_rw_page old_pg ->
   page_of_vadd_madd_smem (set (st_memory s) va' pt_addr val) va ma =
     Some (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg)).
Proof.
  intros.
  unfold page_of_vadd_madd_smem in H4; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va' = Some ma).
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=ma); crush.
  (* Cache, * *)
  destruct (vadd_eq va' va); crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite set_new_cache with (ma:=ma) (pg:=old_pg) (old_val:=x1) in Heqval1; crush.
  destruct (page_cacheable old_pg) eqn:?; crush.
  rewrite <- get_add_cache_cacheable in Heqval1; crush.
  unfold is_cacheable; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma old_pg; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.

  destruct (add_cache (smem_c (st_memory s)) (va, ma)
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) eqn:?;
    crush.
  generalize (add_cache_entry_behavior_update (smem_c (st_memory s)) (va,ma)
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))); crush.
  rewrite H6 in Heqp; crush; clear H6.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=ma) (pg:=old_pg) (old_val:=x1) in Heqval1; crush.
  destruct (page_cacheable old_pg) eqn:?; crush.
  rewrite <- get_add_cache_cacheable in Heqval1; crush.
  unfold is_cacheable; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma old_pg; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  symmetry in Heqval0; STAC.destruct_valid_state;
    STAC.synonym_cached.

  (* Mem, * *)
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va' = Some ma).
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=ma); crush.

  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite set_new_cache with (ma:=ma) (pg:=old_pg) (old_val:=x1) in Heqval1; crush.
  destruct (vadd_eq va va'); crush.
  rewrite <- get_add_cache_cacheable in Heqval1; crush.
  unfold is_cacheable; simpl.
  destruct (page_cacheable old_pg) eqn:?; auto.
  rewrite <- get_add_cache_non_cacheable in Heqval1; unfold is_cacheable; simpl; crush.
  generalize (get_add_cache_distinct_not_new (smem_c (st_memory s)) (va',ma) (va,ma)
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg)) p);
    crush.
  rewrite <- H7 in Heqval0; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  destruct (vadd_eq va va'); crush.
  symmetry in Heqval2; STAC.destruct_valid_state;
    STAC.synonym_cached.

  rewrite set_new_cache with (ma:=ma) (pg:=old_pg) (old_val:=x1) in Heqval1; crush.
  destruct (page_cacheable old_pg) eqn:?; crush.  
  destruct (vadd_eq va va'); crush.
  rewrite <- get_add_cache_cacheable in Heqval1; crush.
  unfold is_cacheable; auto.
  STAC.destruct_valid_state.
  generalize H1, H6; intros.
  symmetry in H4.
  unfold madd_of_vadd, madd_page_table in H7, H8; crush.
  destruct (page_content p1) eqn:?; crush.
  destruct (page_content p2) eqn:?; crush.
  symmetry in Heqc.
  syn_non_cacheable.
  rewrite set_new_mmem_non_cacheable with (ma:=ma) (old_pg:=old_pg); crush.
  rewrite <- get_add_mem; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  destruct (vadd_eq va va'); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' ma p; crush.

  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  destruct (vadd_eq va va'); crush.
  symmetry in Heqval2; STAC.destruct_valid_state;
    STAC.synonym_cached.
Qed.

Lemma set_mem_old_pt_page : forall s va ma pt_addr pt_addr' pt_page' old_pg val,
  valid_state s ->
  aos_current_page_table s = Some pt_addr ->
  madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
  page_of_vadd_madd_smem (st_memory s) va ma = Some old_pg ->
  is_rw_page old_pg ->
  get_mem (smem_mm (set (st_memory s) va pt_addr val)) pt_addr' = Some pt_page' ->
  is_pt_page pt_page' ->
  get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page'.
Proof.
  intros.
  destruct (add_cache (smem_c (st_memory s)) (va,ma)
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) eqn:?; crush.
  destruct o; crush.
  destruct p. destruct c0.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=ma) (ma':=m)
    (pg:=(Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) (pg':=p) in H4;
    crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in H1; crush.
  rewrite Heqp; crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val old_pg; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in H1; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in H1; crush.
  unfold same_page_type in H6; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  rewrite H5, H3 in H6; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  symmetry; apply STAC.cache_deleted_entry_was_there with va ma
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg)); crush.
  rewrite Heqp; crush.
  rewrite set_cache_entry1 with (va':=v) (ma:=ma) (pg':=p) 
    (pg:=(Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) in H4;
    crush.
  unfold is_rw_page, is_pt_page in *; crush.
  rewrite Heqp; crush.
  rewrite set_mem_unchanged2 with (ma:=ma)
    (pg:=(Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) in H4;
    crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H1; crush.
  rewrite Heqp; crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val old_pg; crush.
  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H1; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H1; crush.
  unfold same_page_type in H6; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  rewrite H5, H3 in H6; crush.
Qed.

Lemma set_mem_unchanged_pt_page : forall s va ma pt_addr pt_addr' pt_page' old_pg val,
  valid_state s ->
  aos_current_page_table s = Some pt_addr ->
  madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
  page_of_vadd_madd_smem (st_memory s) va ma = Some old_pg ->
  is_rw_page old_pg ->
  get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
  is_pt_page pt_page' ->
  get_mem (smem_mm (set (st_memory s) va pt_addr val)) pt_addr' = Some pt_page'.
Proof.
  intros.
  destruct (add_cache (smem_c (st_memory s)) (va,ma)
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) eqn:?; crush.
  destruct o; crush.
  destruct p. destruct c0.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=ma) (ma':=m)
    (pg:=(Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg))) (pg':=p); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in H1; crush.
  rewrite Heqp; crush.
  unfold page_of_vadd_madd_smem in H2; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma old_pg; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  symmetry; apply STAC.cache_deleted_entry_was_there with va ma
    (Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg)); crush.
  rewrite Heqp; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  
  rewrite set_mem_unchanged2 with (ma:=ma)
    (pg:=Page (RW_c (Some val)) (page_owned_by old_pg) (page_cacheable old_pg)); crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in H1; crush.
  rewrite Heqp; crush.
  unfold page_of_vadd_madd_smem in H2; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma old_pg; crush.
  unfold is_rw_page, is_pt_page in *; crush.
  unfold is_rw_page, is_pt_page in *; crush.
Qed.

Lemma page_of_vadd_madd_smem_same_madd :
  forall s va1 va2 ma pt_addr1 pt_addr2,
    valid_state s ->
    
    (exists osi1, exists pt_padd1, osi_pa_to_ma s osi1 pt_padd1 = Some pt_addr1) ->   
    (exists osi2, exists pt_padd2, osi_pa_to_ma s osi2 pt_padd2 = Some pt_addr2) ->   

    madd_of_vadd (smem_mm (st_memory s)) pt_addr1 va1 = Some ma ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr2 va2 = Some ma ->
    page_of_vadd_madd_smem (st_memory s) va1 ma =
    page_of_vadd_madd_smem (st_memory s) va2 ma.
Proof.    
  intros.
  destruct (vadd_eq va1 va2); crush.
  assert (va2 <> va1); crush.
  unfold page_of_vadd_madd_smem; crush;
    symmetry in Heqval, Heqval0.
  destruct no_cache_alias with s va1 va2 ma p p0; crush.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
Qed.  

Lemma write_same_madd_of_vadd_s' : forall s va va' pt_addr pt_page pt_addr' old_pg val,
     valid_state s ->
     aos_current_page_table s = Some pt_addr' ->
     aos_page_of_vadd_smem s va' = Some old_pg ->
     is_rw_page old_pg ->
     get_mem (smem_mm (set (st_memory s) va' pt_addr' val)) pt_addr = Some pt_page ->
     is_pt_page pt_page ->
     madd_of_vadd (smem_mm (st_memory s)) pt_addr va =
     madd_of_vadd
       (smem_mm (set (st_memory s) va' pt_addr' val))
       pt_addr va.
Proof.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold is_pt_page in *; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  rewrite set_mem_old_pt_page with (va:=va') (ma:=m) (old_pg:=old_pg)
    (pt_addr:=pt_addr') (pt_page':=pt_page) (val:=val); crush.
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0); crush.
  unfold is_pt_page; crush.
Qed.

Lemma write_no_va_mapped_same_mem :
  forall s osi va' pt_addr' val pa ma,
    valid_state s ->
    aos_current_page_table s = Some pt_addr' ->
    ACT.no_va_mapped_to_ma osi ma s ->
    osi_pa_to_ma s osi pa = Some ma ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (set (st_memory s) va' pt_addr' val)) ma.
Proof.
  intros.
  remember (page_of_vadd_smem (st_memory s) pt_addr' va').
  destruct o; crush.
  remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va').
  destruct o; crush.
  remember (add_cache (smem_c (st_memory s)) (va',m) p).
  destruct p0; crush.
  destruct o; crush.
  destruct p0. destruct c0.
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache_va_ma with v ma x0; crush.
  destruct (os_ident_eq x osi); crush.
  destruct H1 with x1 x2 v x3; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap with x i1 x2 x3 v ma; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt v) eqn:?; unfold os_accessible in *; auto.
  rewrite H10 in *; crush.
  rewrite H11 in *; crush.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) (pg:=p) (pg':=x0); crush.
  unfold madd_of_vadd_smem; crush.
  destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m1) in Heqo0; crush.
  unfold madd_of_vadd, madd_page_table in Heqo0; crush.
  destruct (page_content p1) eqn:?; crush.
  destruct_valid_state.
  get_aos_data.
  get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite Heqc0 in H8; injection H8; crush; clear H8.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) m = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) m = Os_PO (st_active_os s)); crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr' p1 pt va' m; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H8 in H3; inversion H3; crush.
  rewrite H3 in H6; injection H6; crush; clear H6.
  destruct H1 with (os_current_page aos) p1 va' pt; crush.
  unfold osi_pa_to_pg; crush.

  rewrite set_mem_unchanged2 with (ma:=m) (pg:=p); crush.
  unfold madd_of_vadd_smem; crush.
  destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m0) in Heqo0; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct_valid_state.
  get_aos_data.
  get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H6 in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) m = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) m = Os_PO (st_active_os s)); crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr' pt_page pt va' m; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H9 in H3; inversion H3; crush.
  rewrite H3 in H5; injection H5; crush; clear H5.
  destruct H1 with (os_current_page aos) pt_page va' pt; crush.
  unfold osi_pa_to_pg; crush.

  rewrite set_invalid; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in Heqo0; crush.
  rewrite <- Heqo0; crush.
  rewrite set_invalid; crush.
Qed.

Lemma write_no_va_mapped_old :
  forall s s' va val osi ma,
    valid_state s ->
    ACT.pre_all_write va val s ->
    ACT.post_all_write va val s s' ->
    (ACT.no_va_mapped_to_ma osi ma s <->
    ACT.no_va_mapped_to_ma osi ma s').
Proof.    
  intros.
  unfold pre_all_write, post_all_write in *; crush.
  unify_write_post; crush.
  split; intros; unfold no_va_mapped_to_ma; crush.

  unfold osi_pa_to_pg in *; crush.
  assert (madd_type (smem_mm (set (st_memory s) va pt_addr0 val)) i0 = Some PT).
  unfold madd_type; crush.
  rewrite H5; crush.
  assert (madd_type (smem_mm (st_memory s)) i0 = Some PT).
  rewrite write_same_madd_type in H7; crush.
  clear H7.
  unfold madd_type in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold page_of_vadd_smem in *; crush.
  erewrite set_mem_unchanged_pt_page
    with (ma:=m) (pt_page':=p) (old_pg:=x) in H4; crush.
  destruct H1 with pt_padd pt_page va0 mp; crush.
  unfold osi_pa_to_pg; crush.
  apply madd_of_vadd_smem_then_mmem; crush.
  unfold is_pt_page; crush.

  unfold osi_pa_to_pg in *; crush.
  destruct H1 with pt_padd pt_page va0 mp; crush.
  unfold osi_pa_to_pg; crush.
  unfold page_of_vadd_smem in *; crush.
  rewrite set_mem_unchanged_pt_page
    with (ma:=m) (pt_page':=pt_page) (old_pg:=x); crush.
  apply madd_of_vadd_smem_then_mmem; crush.
  unfold is_pt_page; crush.
Qed.

Lemma write_no_pt_change : 
  forall s pt_addr pt_addr' va' val' pg',
    valid_state s ->

    aos_current_page_table s = Some pt_addr' ->
    aos_page_of_vadd_smem s va' = Some pg' ->
    is_rw_page pg' ->

    madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
    get_mem (smem_mm (st_memory s)) pt_addr = 
    get_mem (smem_mm (set (st_memory s) va' pt_addr' val')) pt_addr.
Proof.
  intros.
  STAC.destruct_valid_state.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (st_memory s)) (va',m) (Page (RW_c (Some val'))
                                                           (page_owned_by pg')
                                                           (page_cacheable pg'))) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) 
    (pg:=(Page (RW_c (Some val'))
               (page_owned_by pg')
               (page_cacheable pg'))) (pg':=x0); crush.
  unfold madd_type, page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va' m pg'; crush.
  unfold is_rw_page in *; crush.
  rewrite H7 in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H1 in *; crush.
  destruct H_valid_cache with v m0 x0; crush.
  unfold madd_type, is_rw_page in *; crush.
  rewrite H9 in *; crush.
  rewrite set_mem_unchanged2 with (ma:=m)
    (pg:=(Page (RW_c (Some val'))
               (page_owned_by pg')
               (page_cacheable pg'))); crush.
  unfold madd_type, page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va' m pg'; crush.
  unfold is_rw_page in *; crush.
  rewrite H5 in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H1 in *; crush.
Qed.  
  


(* New lemmas *)

Lemma same_page_type_refl : forall pg,
  same_page_type pg pg = true.
Proof.
  unfold same_page_type.
  intros.
  remember (page_content pg) as content.
  destruct content; auto.
  destruct o; crush.
Qed.

Lemma same_page_type_update : forall pg new_owner new_cacheable,
  same_page_type (Page (page_content pg) new_owner new_cacheable) pg = true.
Proof.
  unfold same_page_type.
  intros.
  remember (page_content pg) as content.
  destruct content; auto.
  destruct o; crush.
Qed.

Lemma same_page_type_rw : forall pg pg',
  is_rw_page pg -> is_rw_page pg' ->
  same_page_type pg pg' = true.
Proof.
  unfold same_page_type, is_rw_page; intros.
  destruct H; destruct H0; rewrite H; rewrite H0; reflexivity.
Qed.    

Lemma madd_of_vadd_new_not_shared : forall s pt_addr va ma,
  valid_state s ->
  madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
  madd_of_vadd (smem_mm (new_not_shared (st_memory s) pt_addr va ma))
               pt_addr va = Some ma.
Proof.
  intros.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o; crush.
  rewrite new_mem_not_shared with (mm':=m); crush.
  unfold remove_new_va_mem; crush.
  unfold add_page_table_entry in Heqo; crush.
  remember (page_content p0).
  destruct c; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  remember (madd_eq m0 pt_addr).
  destruct s0; crush.
  rewrite <- get_add_mem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr p; crush.
  unfold is_rw_page in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.
  unfold add_page_table_entry in Heqo; crush.
  remember (page_content p).
  destruct c; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.
  unfold add_page_table_entry in Heqo; crush.
  remember (page_content p).
  destruct c; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.

  unfold madd_type, add_page_table_entry in *; crush.
  remember (page_content p).
  destruct c; crush.
Qed.    

Lemma madd_of_vadd_new_shared : forall s pt_addr va ma,
  valid_state s ->
  madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
  (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
  madd_of_vadd (smem_mm (new_shared (st_memory s) pt_addr va ma))
               pt_addr va = Some ma.
Proof.
  intros.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm''|].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush.
  unfold madd_type in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold remove_new_va_mem; crush.
  assert (m <> pt_addr); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr p1; crush.
  unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p2) eqn:?; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- PT.get_set; crush.
  assert (madd_type mm' pt_addr = Some PT).
  unfold madd_type in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite Heqc; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p2; crush.
  unfold is_rw_page in *; crush.
  rewrite Heqc; crush.
  assert (madd_type mm'' pt_addr = Some PT); crush.
  unfold madd_type in H2; crush.
  destruct (page_content p) eqn:?; crush.
  unfold madd_type; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem; crush.
  rewrite Heqc; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite Heqc; crush.
  unfold add_page_table_entry in *; crush.
  unfold madd_type in H3; crush.
  destruct (page_content p) eqn:?; crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.  

Lemma madd_of_vadd_new_updated : forall s pt_addr osi va ma,
  valid_state s ->
  madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
  (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
  madd_of_vadd (smem_mm (ACT.new s osi pt_addr va ma))
               pt_addr va = Some ma.
Proof.
  crush.
  apply madd_of_vadd_new_shared; crush.
  apply madd_of_vadd_new_not_shared; crush.
Qed.

Lemma madd_of_vadd_new_not_shared_distinct_pt :
  forall s osi pt_addr pt_addr' va va' ma,
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    pt_addr <> pt_addr' ->
    madd_of_vadd (smem_mm (new_not_shared (st_memory s) pt_addr va ma)) pt_addr' va'
    = madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'.
Proof.  
  intros.
  destruct (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma) eqn:?;
    crush.
  rewrite new_mem_not_shared with (mm':=m); crush.

  assert (madd_of_vadd m pt_addr' va' =
          madd_of_vadd (smem_mm (st_memory s)) pt_addr' va').
  unfold add_page_table_entry in Heqo; crush.
  destruct (page_content p) eqn:?; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite get_add_mem_distinct; crush.

  unfold remove_new_va_mem; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq m0 pt_addr'); crush.
  rewrite <- get_add_mem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr' p; crush.
  unfold is_rw_page in *; crush.
  rewrite H3, H6; crush.
  rewrite get_add_mem_distinct; crush.

  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H7 in *; crush.
Qed.  

Lemma madd_of_vadd_new_shared_distinct_pt :
  forall s osi pt_addr pt_addr' va va' ma,
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    pt_addr <> pt_addr' ->
    madd_of_vadd (smem_mm (new_shared (st_memory s) pt_addr va ma)) pt_addr' va'
    = madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'.
Proof.  
  intros.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  assert (madd_of_vadd mm' pt_addr' va' =
          madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'); [crush|].
  unfold madd_of_vadd, madd_page_table; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct (madd_eq m pt_addr'); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr' p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H6, H3; crush.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm''|].
  assert (madd_of_vadd mm'' pt_addr' va' =
          madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'); [crush|].
  rewrite <- H3.
   unfold madd_of_vadd, madd_page_table; crush.
  unfold mark_page_as_non_cacheable in Heqo; crush.
  destruct (madd_eq ma pt_addr'); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  assert (madd_of_vadd mm''' pt_addr' va' =
          madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'); [crush|].
  rewrite <- H4.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- H5.
  unfold madd_of_vadd, madd_page_table; crush.
  erewrite new_mem_shared; crush.
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m pt_addr'); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr' p; crush.
  unfold is_rw_page in *; crush.
  rewrite H6; crush.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  destruct (get_mem mm''' pt_addr') eqn:?; crush.
  rewrite H9 in *; crush.
  unfold add_page_table_entry in  *; crush.
  destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold mark_page_as_non_cacheable in Heqo; crush.
  destruct (madd_eq ma pt_addr); crush;
    [rewrite <- get_add_mem in Heqo0 |
     rewrite get_add_mem_distinct in Heqo0]; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  destruct (madd_eq m pt_addr); crush;
    [rewrite <- get_add_mem in Heqopt_res |
     rewrite get_add_mem_distinct in Heqopt_res]; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct H_valid_cache with v pt_addr p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H10 in *; crush.
  rewrite H10 in *; crush.
  unfold update_cache_entry in Heqo0; crush.
  destruct (madd_eq m pt_addr); crush;
    [rewrite <- get_add_mem in Heqo0 |
     rewrite get_add_mem_distinct in Heqo0]; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite H10 in *; crush.
  rewrite H10 in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.  

Lemma madd_of_vadd_new_not_shared_distinct :
  forall s osi pt_addr pt_addr' va va' ma,
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    va <> va' ->
    madd_of_vadd (smem_mm (new_not_shared (st_memory s) pt_addr va ma)) pt_addr' va'
    = madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'.
Proof.  
  intros.
  destruct (madd_eq pt_addr' pt_addr);
     [|apply madd_of_vadd_new_not_shared_distinct_pt with osi]; crush.

  destruct (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma) eqn:?;
    crush.
  rewrite new_mem_not_shared with (mm':=m); crush.

  assert (madd_of_vadd m pt_addr va' =
          madd_of_vadd (smem_mm (st_memory s)) pt_addr va').
  unfold add_page_table_entry in Heqo; crush.
  destruct (page_content p) eqn:?; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- get_add_mem; crush.
  rewrite Heqc; crush.
  rewrite PT.get_set_distinct; crush.

  unfold remove_new_va_mem; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq m0 pt_addr); crush.
  rewrite <- get_add_mem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr p; crush.
  unfold is_rw_page in *; crush.
  rewrite H3, H6; crush.
  rewrite get_add_mem_distinct; crush.

  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H7 in *; crush.
Qed.

Lemma new_page_shared_entry_updated_page :
  forall s osi pt_addr v va va' ma ma' pg',
    valid_state s ->
    pt_addr <> ma ->
    madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
    shared_page s va ma ->
    snd (remove_ma_cache (smem_c (st_memory s)) ma) = Some (v,ma',pg') ->
    page_of_vadd_madd_smem (ACT.new s osi pt_addr va ma) va' ma =
    Some (Page (page_content pg') (page_owned_by pg') false).
Proof.
  unfold page_of_vadd_madd_smem; crush;
    unfold new in *; crush;
    clear H2 Heqval;
    edestruct get_remove_ma_cache_some_entry; crush;
    rename ma' into ma.

  (* Cache *)
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_eq in *; crush.

  (* Mem *)
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold mark_page_as_non_cacheable in *; crush.
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  destruct (vadd_eq va v); crush.
  destruct s0; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 x3 = Some ma); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H5; crush.
  assert (x3 <> v); crush.
  clear dependent x1.
  destruct_valid_state.
  synonym_cached.
  destruct no_cache_alias with s v va ma pg' p0; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in Heqo; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  rewrite <- get_add_mem in *; crush.
Qed.  
  
Lemma new_page_shared_no_entry_updated_page : forall s osi pt_addr va va' ma pg,
  valid_state s ->
  pt_addr <> ma ->
  madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
  shared_page s va ma ->
  snd (remove_ma_cache (smem_c (st_memory s)) ma) = None ->
  page_of_vadd_madd_smem (st_memory s) va' ma = Some pg ->
  page_of_vadd_madd_smem (ACT.new s osi pt_addr va ma) va' ma =
  Some (Page (page_content pg) (page_owned_by pg) false).
Proof.
  unfold page_of_vadd_madd_smem, ACT.new; crush.

  (* Cache, Cache *)
  rewrite new_cache_shared in Heqval1; crush.
  rewrite get_remove_ma_cache_eq in Heqval1; crush.

  (* Cache, Mem *)
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.

  (* Mem, Cache *)
  rewrite new_cache_shared in Heqval1; crush.
  rewrite get_remove_ma_cache_eq in Heqval1; crush.

  (* Mem, Mem *)
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq m ma); crush.
  rewrite get_remove_ma_cache_no_entry in Heqval3; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  unfold madd_type in *; crush.
  destruct (page_content p); crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
Qed.  

Lemma new_page_not_shared_updated_page :
  forall s osi pt_addr va ma,
    valid_state s ->
    pt_addr <> ma ->
    ~ shared_page s va ma ->
    page_of_vadd_madd_smem (ACT.new s osi pt_addr va ma) va ma =
    page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  unfold page_of_vadd_madd_smem, ACT.new; crush;
    try rewrite new_cache_not_shared in Heqval0; crush.
  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va,ma)); crush.
  rewrite get_remove_cache in Heqval0; crush.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval0; crush.

  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va,ma)); crush.
  rewrite get_remove_cache_old with (k':=(va,m)) (pg:=p) in Heqval1; crush.

  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va,ma)); crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush.
  unfold remove_new_va_mem; crush.
  rewrite <- get_add_mem; crush.
  unfold add_page_table_entry, madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p1); crush.
  rewrite get_remove_cache_distinct with (pg:=p) in Heqval0; crush.
  
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush.
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite new_mem_not_shared_invalid; crush.
Qed.

Lemma new_page_ne : forall s osi pt_addr va va' ma ma',
  valid_state s ->
  ma <> ma' ->
  pt_addr <> ma' ->
  (exists osi pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
  (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
  (exists osi' pt_padd' pt_addr', 
     osi_pa_to_ma s osi' pt_padd' = Some pt_addr' /\
     madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma') ->
  page_of_vadd_madd_smem (ACT.new s osi pt_addr va ma) va' ma' =
  page_of_vadd_madd_smem (st_memory s) va' ma'.
Proof.  
  intros.
  unfold new, page_of_vadd_madd_smem; crush.
  (* Shared page *)
  (* Cache, cache *)
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  destruct (madd_eq m ma'); crush.
  rewrite get_remove_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va',m)) in Heqval1; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va,m)) in Heqval1; crush.
  (* Mem, Cache *)
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  destruct (madd_eq m ma'); crush.
  erewrite get_remove_cache_old with (k':=(va',m)) in Heqval1; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va,m)) in Heqval1; crush.
  (* Cache, Mem *)
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  destruct (madd_eq m ma'); crush.
  destruct (vadd_eq va va'); crush.
  erewrite get_add_mem; crush.
  destruct no_cache_alias with s va va' ma' p0 p; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rename m0 into ma.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rename m0 into ma.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold remove_new_va_cache in *; crush.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold remove_new_va_cache in *; crush.
  
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rename m into ma.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold remove_new_va_cache in *; crush.
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold remove_new_va_cache in *; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in Heqo0; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold madd_type in *; crush.
  destruct (page_content x2); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  unfold madd_type in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p2; crush.
  unfold is_rw_page in *; crush.
  rewrite H11 in *; crush.
  unfold madd_type in *; crush.
  destruct (page_content x2); crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold  madd_type in *; crush.
  destruct (page_content p1); crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  unfold madd_type in *; crush.
  destruct (page_content p0); crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  destruct (madd_eq m ma); crush.
  rewrite <- get_add_mem in *; crush.
  rewrite get_add_mem_distinct in *; crush.

  (* Mem, Mem *)
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma'); crush.
  destruct (vadd_eq va' va); crush. 
  symmetry in Heqval3.
  destruct_valid_state.
  synonym_cached.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in Heqo0; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold madd_type in *; crush.
  destruct (page_content x2); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  unfold madd_type in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H9 in *; crush.
  unfold madd_type in *; crush.
  rewrite H11 in *; crush.
  unfold madd_type in *; crush.
  destruct (page_content x2); crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold  madd_type in *; crush.
  destruct (page_content p0); crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  unfold madd_type in *; crush.
  destruct (page_content p); crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.

  (* Not shared page *)
  (* Cache, Cache *)
  rewrite new_cache_not_shared in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  destruct (madd_eq m ma'); crush.
  rewrite get_remove_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va',m)) in Heqval1; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va,m)) in Heqval1; crush.
  (* Mem, Cache *)
  rewrite new_cache_not_shared in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  destruct (madd_eq m ma'); crush.
  erewrite get_remove_cache_old with (k':=(va',m)) in Heqval1; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_old with (k':=(va,m)) in Heqval1; crush.
  (* Cache, Mem *)
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  destruct (madd_eq m ma'); crush.
  destruct (vadd_eq va va'); crush.
  rewrite <- get_add_mem; crush.
  destruct no_cache_alias with s va va' ma' p0 p; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite new_cache_not_shared in *; crush.
  destruct (vadd_eq va va'); crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.
  unfold remove_new_va_cache in *; crush.
  erewrite get_remove_cache_distinct in Heqval0; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite new_cache_not_shared in *; crush.
  unfold remove_new_va_cache in *; crush.
  rewrite new_cache_not_shared in *; crush.
  unfold remove_new_va_cache in *; crush.
  
  unfold add_page_table_entry in *; crush.
  unfold madd_type in *; crush.
  destruct (page_content p0); crush.

  (* Mem, Mem *)
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma'); crush.
  destruct (vadd_eq va' va); crush. 
  symmetry in Heqval3.
  destruct_valid_state.
  synonym_cached.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold madd_type in *; crush.
  destruct (page_content p); crush.
Qed.  

Lemma new_get_mem_shared_pt_addr :
  forall s pt_addr pt_page pt va ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    get_mem (smem_mm (new_shared (st_memory s) pt_addr va ma)) pt_addr =
      Some (Page (PT_c (PT.set pt va ma)) (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.
  intros.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  rewrite new_mem_shared with (mm':=mm') (mm'':=mm'') (mm''':=mm'''); crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite <- get_add_mem; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m0 p3; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in Heqc; injection Heqc; intros; subst p1; clear Heqc.
  assert (page_cacheable pt_page = false); crush.
  destruct (page_cacheable pt_page) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page;
    unfold is_pt_page; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with va pt_addr p; crush.
  unfold is_rw_page in *; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m0 p2; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in Heqc; injection Heqc; crush; clear Heqc.
  assert (page_cacheable pt_page = false); crush.
  destruct (page_cacheable pt_page) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page;
    unfold is_pt_page; crush.
  rewrite <- get_add_mem; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite <- get_add_mem; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p2; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in Heqc; injection Heqc; crush; clear Heqc.
  assert (page_cacheable pt_page = false); crush.
  destruct (page_cacheable pt_page) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page;
    unfold is_pt_page; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H2 in *; crush.
  rewrite H2 in *; crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.

Lemma new_get_mem_not_shared_pt_addr :
  forall s pt_addr pt_page pt va ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    get_mem (smem_mm (new_not_shared (st_memory s) pt_addr va ma)) pt_addr =
      Some (Page (PT_c (PT.set pt va ma)) (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.
  intros.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|]; crush;
    [rewrite new_mem_not_shared with (mm':=mm')|]; crush;
    try (unfold add_page_table_entry in *; crush; rewrite H2 in *; crush).
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- get_add_mem; crush.
  destruct_valid_state.
  destruct H_valid_cache with va pt_addr p; crush.
  unfold is_rw_page in *; crush.
  rewrite <- get_add_mem; crush.
  rewrite <- get_add_mem; crush.
Qed.

Lemma new_page_pt_addr :
  forall s osi pt_addr pt_page pt va va' ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    page_of_vadd_madd_smem (st_memory s) va' pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    page_of_vadd_madd_smem (ACT.new s osi pt_addr va ma) va' pt_addr =
      Some (Page (PT_c (PT.set pt va ma)) (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.
  intros.
  unfold page_of_vadd_madd_smem in H1; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' pt_addr pt_page; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  (* shared page *)
  rewrite new_cache_shared in *; crush.
  rewrite get_remove_ma_cache_ne in Heqval1; crush.
  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va',pt_addr)); crush.
  erewrite get_remove_cache_old in Heqval; crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  apply new_get_mem_shared_pt_addr; crush.

  (* Not shared page *)
  rewrite new_cache_not_shared in *; crush.
  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va',pt_addr)); crush.
  erewrite get_remove_cache_old in Heqval; crush.
  apply new_get_mem_not_shared_pt_addr; crush.
Qed.

Lemma new_get_mem_pt_addr :
  forall s osi pt_addr pt_page pt va ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    get_mem (smem_mm (ACT.new s osi pt_addr va ma)) pt_addr =
      Some (Page (PT_c (PT.set pt va ma)) (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.
  intros.
  unfold new; crush;
    [apply new_get_mem_shared_pt_addr |
     apply new_get_mem_not_shared_pt_addr]; crush.
Qed.


Lemma new_get_mem_not_shared_distinct :
  forall s pt_addr va ma ma',
    valid_state s ->
    pt_addr <> ma' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = None ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (new_not_shared (st_memory s) pt_addr va ma)) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  intros.
  destruct (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma) eqn:?.
  rename m into mm'.
  rewrite new_mem_not_shared with (mm':=mm'); crush.
  unfold remove_new_va_mem; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite new_mem_not_shared_invalid; crush.
Qed.

Lemma new_get_mem_shared_distinct_pt :
  forall s pt_addr va ma ma',
    valid_state s ->
    pt_addr <> ma' ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (new_shared (st_memory s) pt_addr va ma)) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  unfold madd_type; crush.
  destruct (page_content p) eqn:?; crush.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr va ma).
  destruct o as [mm'''|].
  rewrite new_mem_shared with (mm':=mm') (mm'':=mm'') (mm''':=mm'''); crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p3); crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p2); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  unfold update_cache_entry in Heqopt_res2; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m0 p7; crush.
  unfold is_rw_page in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  assert (page_cacheable x = false); crush.
  destruct (page_cacheable x) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with ma' x;
    unfold is_pt_page; crush.
  rewrite <- H2; destruct x; auto.
  unfold update_cache_entry; crush.
  rewrite get_add_mem_distinct; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with va ma' p1; crush.
  unfold is_rw_page in *; crush.

  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p2) eqn:?; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res2; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m0 p6; crush.
  unfold is_rw_page in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  assert (page_cacheable x = false); crush.
  destruct (page_cacheable x) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with ma' x;
    unfold is_pt_page; crush.
  rewrite <- H2; destruct x; auto.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p1) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush.
  rewrite <- get_add_mem; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  unfold update_cache_entry in Heqopt_res1; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p4; crush.
  unfold is_rw_page in *; crush.
  assert (page_cacheable x = false); crush.
  destruct (page_cacheable x) eqn:?; crush.
  destruct_valid_state.
  destruct H_valid_pt_non_cacheable with ma' x;
    unfold is_pt_page; crush.
  rewrite <- H2; destruct x; auto.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  rewrite new_mem_shared_invalid; crush.
  unfold add_page_table_entry in Heqo0; crush.
  unfold mark_page_as_non_cacheable in Heqo; crush.
  destruct (madd_eq ma pt_addr); crush.
  rewrite <- get_add_mem in *; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  unfold add_page_table_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p3; crush.
  unfold is_rw_page in *; crush.
  rewrite H5 in *; crush.
  unfold add_page_table_entry; crush.
  destruct (page_content x); crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold add_page_table_entry; crush.
  destruct (get_mem (smem_mm (st_memory s)) pt_addr) eqn:?; crush.
  destruct (page_content p2); crush.
  unfold add_page_table_entry; crush.
  destruct (get_mem (smem_mm (st_memory s)) pt_addr) eqn:?; crush.
  destruct (page_content p1); crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.  
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.

Lemma new_get_mem_not_shared_distinct_pt :
  forall s pt_addr va ma ma',
    valid_state s ->
    pt_addr <> ma' ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (new_not_shared (st_memory s) pt_addr va ma)) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  unfold madd_type; crush.
  destruct (page_content p) eqn:?; crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr va ma).
  destruct o as [mm'|].
  rewrite new_mem_not_shared with (mm':=mm'); crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p3); crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p3); crush.
  rewrite get_add_mem_distinct; crush.
  destruct_valid_state.
  destruct H_valid_cache with va ma' p1; crush.
  unfold is_rw_page in *; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p1); crush.
  rewrite get_add_mem_distinct; crush.
  
  unfold add_page_table_entry in *; crush.
  destruct (page_content p1); crush.
  rewrite get_add_mem_distinct; crush.

  rewrite new_mem_not_shared_invalid; crush.
Qed.  
  
Lemma new_get_mem_distinct_pt :
  forall s osi pt_addr va ma ma',
    valid_state s ->
    pt_addr <> ma' ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  unfold new; crush;
    [apply new_get_mem_shared_distinct_pt |
     apply new_get_mem_not_shared_distinct_pt]; crush.
Qed.

(* jdcampo - Splitted new_get_mem_shared_distinct_ma in 3 lemmas *)
Lemma new_get_mem_shared_distinct_ma_some_entry :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma pg,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma' ->
    ma <> pt_addr' ->
    ma <> ma' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma ->
    get_cache (smem_c (st_memory s)) (va', ma) = Some pg ->
    
    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma = Some pg.
Proof.
  unfold new; crush.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma'))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma').
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma').
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite <- get_add_mem; crush.
Lemma get_cache_some_some_eq :
  forall s va ma pg pg',
    get_cache (smem_c (st_memory s)) (va, ma) = Some pg ->
    get_cache (smem_c (st_memory s)) (va, ma) = Some pg' ->
    pg = pg'.
Proof.
  crush.
Qed.
Lemma get_cache_some_none_abs :
  forall s va ma pg,
    get_cache (smem_c (st_memory s)) (va, ma) = Some pg ->
    get_cache (smem_c (st_memory s)) (va, ma) = None ->
    False.
Proof.
  crush.
Qed.
  rewrite get_cache_some_some_eq with s va' ma p pg; crush.
  elim get_cache_some_none_abs with s va' ma pg; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqo0; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  unfold madd_type in *; crush.
  destruct (page_content pt_page'); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  unfold madd_type in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr' p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H2 in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite H2 in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.

Lemma new_get_mem_shared_distinct_ma_not_mapped :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma' ->
    ma <> pt_addr' ->
    ma <> ma' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' <> Some ma ->

    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  unfold new; crush.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma'))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma').
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma').
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (page_content p0); crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (page_content p); crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (page_content p); crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqo0; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H2 in *; crush.
  rewrite H2 in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  destruct (madd_eq m ma'); crush.
  rewrite <- get_add_mem in *; crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
Qed.

Lemma new_get_mem_shared_distinct_ma_no_entry :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma,

    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma' ->
    ma <> pt_addr' ->
    ma <> ma' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma ->
    get_cache (smem_c (st_memory s)) (va', ma) = None ->

    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  unfold new; crush.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma'))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma').
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma').
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  elim get_cache_some_none_abs with s va' ma p; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (page_content p0); crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqopt_res; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma' pt_addr'); crush.
  rewrite <- get_add_mem in Heqo0; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H2 in *; crush.
  rewrite H2 in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
Qed.

Lemma new_get_mem_shared_new_ma :
  forall s osi' pt_addr' pt_page' pt' va' va_rep pg_rep ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma ->
    snd (remove_ma_cache (smem_c (st_memory s)) ma) = Some (va_rep,ma,pg_rep) ->
    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma)) ma =
    Some (Page (page_content pg_rep) (page_owned_by pg_rep) false).
Proof.
  crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  clear Heqval H5.
  destruct_valid_state.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma).
  destruct o as [mm'''|].
  rewrite new_mem_shared with (mm':=mm') (mm'':=mm'') (mm''':=mm'''); crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush.
  destruct (vadd_eq va_rep va'); crush.
  2:edestruct no_cache_alias with (va:=va_rep) (va':=va') (ma:=ma); crush.
  destruct H3; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 x4 = Some ma); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H5; crush.
  clear dependent x2.
  clear Heqval.
  assert (x4 <> va'); crush.
  synonym_cached.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  rewrite <- get_add_mem in *; crush.
  destruct H_valid_cache with va_rep ma pg_rep; crush.
  unfold is_rw_page in *; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  rewrite <- get_add_mem in *; crush.
  destruct H_valid_cache with va_rep ma pg_rep; crush.
  unfold is_rw_page in *; crush.
  
  unfold add_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  rewrite <- get_add_mem in *; crush.
  destruct H_valid_cache with va_rep ma pg_rep; crush.
  unfold is_rw_page in *; crush.

  unfold add_page_table_entry in Heqo0; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H2 in *; crush.
  destruct H_valid_cache with va_rep pt_addr' pg_rep; crush.
  unfold is_rw_page in *; crush.
  destruct H_valid_cache with va_rep pt_addr' pg_rep; crush.
  unfold is_rw_page in *; crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  rewrite <- get_add_mem in *; crush.
Qed.

Lemma new_get_mem_shared_new_ma_no_entry :
  forall s osi' pt_addr' pt_page' pt' va' ma pg,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ma <> pt_addr' ->           (* jdcampo - added this hypothesis *)
    shared_page s va' ma ->
    snd (remove_ma_cache (smem_c (st_memory s)) ma) = None ->
    get_mem (smem_mm (st_memory s)) ma = Some pg ->
    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma)) ma =
      Some (Page (page_content pg) (page_owned_by pg) false).
Proof.
  unfold new; crush.
  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma).
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma).
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush.
  destruct H4; crush.
  unfold osi_pa_to_pg in H4; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 x4 = Some ma).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H6; crush.
  assert (x4 <> va'); crush.
  symmetry in Heqval1.
  destruct_valid_state.
  clear dependent x2.
  synonym_cached.
  
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.
  
  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite <- get_add_mem; crush.
  unfold update_cache_entry in Heqopt_res0; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  rewrite H2 in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in *; crush.
Qed.

(* jdcampo - Splitted new_get_mem_not_shared_distinct_ma in 3 lemmas *)
Lemma new_get_mem_not_shared_distinct_ma_some_entry :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma pg,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ~ shared_page s va' ma' ->
    ma <> pt_addr' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma ->
    get_cache (smem_c (st_memory s)) (va', ma) = Some pg ->

    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    Some pg.
Proof.
  unfold new; crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'''|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite <- get_add_mem; crush.
  rewrite get_cache_some_some_eq with s va' ma p pg; crush.
  elim get_cache_some_none_abs with s va' ma pg; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
Qed.

Lemma new_get_mem_not_shared_distinct_ma_not_mapped :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ~ shared_page s va' ma' ->
    ma <> pt_addr' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' <> Some ma ->

    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  unfold new; crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
Qed.

Lemma new_get_mem_not_shared_distinct_ma_no_entry :
  forall s osi' pt_addr' pt_page' pt' va' ma' ma,

    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ~ shared_page s va' ma' ->
    ma <> pt_addr' ->
    ma <> ma' ->

    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma ->
    get_cache (smem_c (st_memory s)) (va', ma) = None ->

    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  unfold new; crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  elim get_cache_some_none_abs with s va' ma p; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
Qed.

Lemma new_get_mem_distinct_osi :
  forall s osi' pt_addr' pt_page' pt' va' ma' osi ma,
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg /\
      page_owned_by pg = Os_PO osi') ->                      (* jdcampo - added second condition *)
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    (exists pt_padd', osi_pa_to_ma s osi' pt_padd' = Some pt_addr') ->
    page_content pt_page' = PT_c pt' ->
    page_owned_by pt_page' = Os_PO osi' ->
    madd_owner (smem_mm (st_memory s)) ma = Os_PO osi ->
    osi <> osi' ->
    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  Opaque new.
  intros.
  assert (ma <> pt_addr'); crush.
  unfold madd_owner in *; crush.
  assert (ma <> ma'); crush.
  unfold madd_owner in *; crush.
  destruct (shared_page_dec s va' ma').
  apply new_get_mem_shared_distinct_ma_not_mapped with pt_page' pt'; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H3 in *; crush.
  destruct_valid_state.
  destruct H_valid_virtmap with osi' pt_addr' pt_page' pt' va' ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.
  
  apply new_get_mem_not_shared_distinct_ma_not_mapped with pt_page' pt'; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H3 in *; crush.
  destruct_valid_state.
  destruct H_valid_virtmap with osi' pt_addr' pt_page' pt' va' ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.
Qed.

Lemma new_shared_same_page_type :
  forall s ma pt_addr' pt_page' pt' osi va' ma' pg pg',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma' ->
    get_mem (smem_mm (st_memory s)) ma = Some pg ->
    get_mem (smem_mm (new s osi pt_addr' va' ma')) ma = Some pg' ->
    same_page_type pg pg' = true.
Proof.  
  unfold same_page_type; crush.
  destruct (madd_eq pt_addr' ma); crush.
  rewrite new_get_mem_pt_addr with (pt_page:=pg) (pt:=pt') in H5; crush.
  rewrite H2; crush.
  destruct (madd_eq ma' ma); crush.
  destruct (snd (remove_ma_cache (smem_c (st_memory s)) ma)) eqn:?; crush.
  destruct p.
  destruct c.
  
  erewrite new_get_mem_shared_new_ma with (va_rep:=v) (pg_rep:=p) in H5; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold is_rw_page in *; crush.
  rewrite H0, H7; crush.
  edestruct get_remove_ma_cache_some_entry; crush.

  erewrite new_get_mem_shared_new_ma_no_entry in H5; crush.
  destruct pg'; simpl in *.
  injection H0; intros.
  rewrite H7.
  destruct page_content0; auto.
  destruct o; auto.

  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?; crush.
  destruct (madd_eq m ma); crush.
  destruct (get_cache (smem_c (st_memory s)) (va', ma)) eqn:?; crush.
  erewrite new_get_mem_shared_distinct_ma_some_entry in H5; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page in *; crush.
  rewrite H5, H7; crush.
  erewrite new_get_mem_shared_distinct_ma_no_entry in H5; crush.
  destruct (page_content pg); simpl; auto.
  destruct o; auto.

  erewrite new_get_mem_shared_distinct_ma_not_mapped in H5; crush.
  destruct (page_content pg); simpl; auto.
  destruct o; auto.
  erewrite new_get_mem_shared_distinct_ma_not_mapped in H5; crush.
  destruct (page_content pg); simpl; auto.
  destruct o; auto.
Qed.  

Lemma new_not_shared_same_page_type :
  forall s ma osi pt_addr' pt_page' pt' va' ma' pg pg',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ~ shared_page s va' ma' ->

    get_mem (smem_mm (st_memory s)) ma = Some pg ->
    get_mem (smem_mm (new s osi pt_addr' va' ma')) ma = Some pg' ->
    same_page_type pg pg' = true.
Proof.  
  crush.
  (* unfold madd_type; crush. *)
  destruct (madd_eq pt_addr' ma); crush.
  rewrite new_get_mem_pt_addr with (pt_page:=pg) (pt:=pt') in H5; crush.
  unfold same_page_type; simpl.
  rewrite H2; crush.
  Transparent new.
    unfold new; crush.
  Opaque new.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'|].
  erewrite new_mem_not_shared in H5; crush. 
  unfold remove_new_va_mem in H5; crush.
  destruct (madd_eq m ma); crush.
  rewrite <- get_add_mem in H5; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma pg'; crush.
  unfold is_rw_page in *; crush.
  unfold same_page_type.
  rewrite H5, H7; crush.
  rewrite get_add_mem_distinct in H5; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in H5; crush.
  apply same_page_type_refl.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in H5; crush.
  apply same_page_type_refl.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct in H5; crush.
  apply same_page_type_refl.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
Qed.  

Lemma new_same_page_type :
  forall s osi pt_page' pt' ma pt_addr' va' ma' pg pg',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->

    get_mem (smem_mm (st_memory s)) ma = Some pg ->
    get_mem (smem_mm (ACT.new s osi pt_addr' va' ma')) ma = Some pg' ->
    same_page_type pg pg' = true.
Proof.  
  intros.
  destruct (shared_page_dec s va' ma'); crush.
  eapply new_shared_same_page_type with (pt_addr':=pt_addr') (ma':=ma'); crush.
  eapply new_not_shared_same_page_type with (pt_addr':=pt_addr') (ma':=ma'); crush.
Qed.  

Lemma new_shared_same_madd_type :
  forall s ma pt_addr' pt_page' pt' osi va' ma',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    shared_page s va' ma' ->
    madd_type (smem_mm (new s osi pt_addr' va' ma')) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.  
  unfold madd_type; crush.
  destruct (madd_eq pt_addr' ma); crush.
  rewrite new_get_mem_pt_addr with (pt_page:=pt_page') (pt:=pt'); crush.
  rewrite H2; crush.
  destruct (madd_eq ma' ma); crush.
  destruct (snd (remove_ma_cache (smem_c (st_memory s)) ma)) eqn:?; crush.
  destruct p.
  destruct c.
  
  erewrite new_get_mem_shared_new_ma with (va_rep:=v) (pg_rep:=p); crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold is_rw_page in *; crush.
  rewrite H4, H7; crush.
  edestruct get_remove_ma_cache_some_entry; crush.

  erewrite new_get_mem_shared_new_ma_no_entry; crush.
  
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?; crush.
  destruct (madd_eq m ma); crush.
  destruct (get_cache (smem_c (st_memory s)) (va', ma)) eqn:?; crush.
  erewrite new_get_mem_shared_distinct_ma_some_entry; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page in *; crush.
  rewrite H4, H7; crush.
  erewrite new_get_mem_shared_distinct_ma_no_entry; crush.
  erewrite new_get_mem_shared_distinct_ma_not_mapped; crush.
  erewrite new_get_mem_shared_distinct_ma_not_mapped; crush.
Qed.  

Lemma new_not_shared_same_madd_type :
  forall s ma osi pt_addr' pt_page' pt' va' ma',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    ~ shared_page s va' ma' ->

    madd_type (smem_mm (new s osi pt_addr' va' ma')) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.  
  unfold madd_type; crush.
  destruct (madd_eq pt_addr' ma); crush.
  rewrite new_get_mem_pt_addr with (pt_page:=pt_page') (pt:=pt'); crush.
  rewrite H2; crush.
  Transparent new.
    unfold new; crush.
  Opaque new.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush.
  rewrite <- get_add_mem; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page in *; crush.
  rewrite H4, H7; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H2 in *; crush.
Qed.  

Lemma new_same_madd_type :
  forall s osi pt_page' pt' ma pt_addr' va' ma',
    valid_state s ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->
    madd_type (smem_mm (ACT.new s osi pt_addr' va' ma')) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.  
  intros.
  destruct (shared_page_dec s va' ma'); crush.
  eapply new_shared_same_madd_type; crush.
  eapply new_not_shared_same_madd_type; crush.
Qed.  


Lemma new_same_madd_type_page :
  forall s osi va pt_addr' va' ma pg pg',
    valid_state s ->
    madd_type (smem_mm (st_memory s)) pt_addr' = Some PT ->
    page_of_vadd_madd_smem (ACT.new s osi pt_addr' va' ma) va ma = Some pg' ->
    get_mem (smem_mm (st_memory s)) ma = Some pg ->
    same_page_type pg pg' = true.
Proof.  
  intros.
  destruct (shared_page_dec s va' ma); crush.
  destruct (remove_ma_cache (smem_c (st_memory s)) ma) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  rewrite STAC.new_page_shared_entry_updated_page
    with (v:=v) (ma':=m) (pg':=p) in H1; crush.
  edestruct get_remove_ma_cache_some_entry
    with (c:=smem_c (st_memory s)) (ma:=ma); crush.
  rewrite Heqp; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma p; auto.
  destruct H5; repeat destruct H6.
  destruct H7.
  unfold is_rw_page in *; destruct H4; destruct H6.
  injection H2; intros; subst x; clear H2.
  unfold same_page_type; rewrite H4, H6; simpl; auto.
  edestruct get_remove_ma_cache_some_entry
    with (c:=smem_c (st_memory s)) (ma:=ma); crush.
  rewrite Heqp; crush.
  crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold is_rw_page, madd_type in *; crush.
  rewrite H6 in *; crush.
  rewrite Heqp; crush.
  destruct (madd_eq pt_addr' ma); crush.
  unfold madd_type, same_page_type in *.
  rewrite H2 in *; simpl in *.
  destruct (page_content pg) eqn:?; simpl in *;
    try (inversion H0; fail).
  rewrite STAC.new_page_pt_addr with (pt_page:=pg) (pt:=p) in H1; crush.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma p0; crush.
  unfold is_rw_page in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?; crush.
  rewrite STAC.new_page_shared_no_entry_updated_page
    with (pg:=p) in H1; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma p; auto.
  destruct H3 as [? [? [? [? ?]]]].
  unfold is_rw_page in *.
  destruct H1, H5.
  rewrite H2 in *; injection H4; intros; subst x.
  unfold same_page_type; rewrite H1, H5; auto.
  rewrite Heqp; crush.
  unfold page_of_vadd_madd_smem; crush.
  
  rewrite STAC.new_page_shared_no_entry_updated_page
    with (pg:=pg) in H1; crush.
  unfold same_page_type; destruct (page_content pg); simpl; auto;
    destruct o; simpl; auto.
  rewrite Heqp; crush.
  unfold page_of_vadd_madd_smem; crush.

  destruct (madd_eq ma pt_addr'); crush.
  unfold madd_type in *; crush.
  destruct (page_content pg) eqn:?; crush.
  rewrite STAC.new_page_pt_addr with (pt_page:=pg) (pt:=p) in H1; crush.
  unfold same_page_type; rewrite Heqc; auto.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va pt_addr' p0; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  Transparent new.
  crush.
  Opaque new.
  rewrite new_cache_not_shared in Heqval; crush.
  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va',m) (va,ma)); crush.
  rewrite get_remove_cache in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval; crush.
  destruct_valid_state.
  destruct H_valid_cache with va ma p0; auto.
  destruct H3; repeat destruct H4; destruct H5.
  unfold is_rw_page in *; destruct H1, H4.
  injection H2; intros; subst x.
  unfold same_page_type; rewrite H1, H4; auto.
  rewrite get_remove_cache_old with (k':=(va',m)) (pg:=pg') in Heqo; crush.
  destruct_valid_state.
  destruct H_valid_cache with va ma pg'; unfold is_rw_page in *; crush.
  unfold same_page_type; rewrite H1, H4; auto.
  destruct_valid_state.
  destruct H_valid_cache with va ma pg'; unfold is_rw_page in *; auto.
  destruct H1, H3; repeat destruct H4; destruct H5.
  destruct H4.
  injection H2; intros; subst x0.
  unfold same_page_type; rewrite H1, H4; auto.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma).
  destruct o as [mm'|].
  Transparent new.
  crush.
  Opaque new.
  erewrite new_mem_not_shared in H1; crush.
  unfold remove_new_va_mem in H1; crush.
  unfold add_page_table_entry in *; crush.
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq m ma).
  subst m.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma pg'; auto.
  destruct H1; repeat destruct H3; destruct H4.
  injection H2; intros; subst x.
  destruct H0, H3.
  unfold same_page_type; rewrite H0, H3; auto.
  rewrite get_add_mem_distinct in H1; auto.
  rewrite get_add_mem_distinct in H1; auto.
  rewrite H2 in H1; injection H1; intros;
    subst pg'; apply same_page_type_refl.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct in H1; crush.
  apply same_page_type_refl.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct in H1; crush.
  apply same_page_type_refl.
  unfold add_page_table_entry, madd_type in *; crush.
  destruct (page_content p); crush.
Qed.

Lemma new_no_va_mapped_same_mem :
  forall s osi' pt_addr' pt_page' pt' va' ma' osi pa ma,
    valid_state s ->

    pt_addr' <> ma ->           (* jdcampo - added these hyp *)
    (exists pg, get_mem (smem_mm (st_memory s)) ma' = Some pg) ->
    get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' ->
    page_content pt_page' = PT_c pt' ->

    osi_pa_to_ma s osi pa = Some ma ->
    ACT.no_va_mapped_to_ma osi ma s ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma.
Proof.
  intros.
  Transparent new.
  crush.
  Opaque new.

  remember (update_cache_entry (smem_mm (st_memory s))
    (snd (remove_ma_cache (smem_c (st_memory s)) ma'))) as mm'.
  remember (mark_page_as_non_cacheable mm' ma').
  destruct o as [mm'' | ].
  remember (add_page_table_entry mm'' pt_addr' va' ma').
  destruct o as [mm'''|].
  erewrite new_mem_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct (page_content p0); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush.
  clear Heqval.
  destruct s0; crush.
  destruct (os_ident_eq x1 osi); crush.
  (* x0 = osi *)
  edestruct H5; crush.
  (* x0 <> osi *)
  destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma' = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x1); crush.
  destruct H_valid_virtmap with x1 i1 x2 x3 x4 ma'; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in H10; crush.
  destruct (ctxt_vadd_accessible ctxt x4) eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  destruct_valid_state.
  destruct H_valid_cache_va_ma with va' ma p; crush.
  edestruct H5 with (pt_padd:=x1); crush.
  destruct (os_ident_eq x0 osi); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x0); crush.
  destruct H_valid_virtmap with x0 i1 x2 x3 va' ma; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in H9; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush.
  clear Heqval.
  destruct s0; crush.
  destruct (os_ident_eq x1 osi); crush.
  (* x1 = osi *)
  edestruct H5; crush.
  (* x1 <> osi *)
  destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma' = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x1); crush.
  destruct H_valid_virtmap with x1 i1 x2 x3 x4 ma'; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in H10; crush.
  destruct (ctxt_vadd_accessible ctxt x4) eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  destruct (page_content p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq ma ma'); crush.
  clear Heqval.
  destruct s0; crush.
  destruct (os_ident_eq x1 osi); crush.
  (* x1 = osi *)
  edestruct H5; crush.
  (* x1 <> osi *)
  destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma' = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x1); crush.
  destruct H_valid_virtmap with x1 i1 x2 x3 x4 ma'; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in H10; crush.
  destruct (ctxt_vadd_accessible ctxt x4) eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  unfold mark_page_as_non_cacheable in *; crush.
  destruct (madd_eq pt_addr' ma'); crush.
  rewrite <- get_add_mem in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with v m p1; crush.
  unfold is_rw_page in *; crush.
  rewrite H3 in *; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  unfold update_cache_entry in Heqo0; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in Heqo0; crush.
  rewrite H3 in *; crush.
  rewrite H3 in *; crush.

  unfold mark_page_as_non_cacheable in *; crush.
  unfold update_cache_entry in Heqopt_res; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.

  (* Not shared *)
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma').
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H3 in *; crush.
  rewrite get_add_mem_distinct; crush.

  destruct_valid_state.
  destruct H_valid_cache_va_ma with va' ma p; crush.
  edestruct H5 with (pt_padd:=x1); crush.
  destruct (os_ident_eq x0 osi); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x0); crush.
  destruct H_valid_virtmap with x0 i1 x2 x3 va' ma; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in H9; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H12 in *; crush.
  rewrite H13 in *; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H3 in *; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H3 in *; crush.
  rewrite get_add_mem_distinct; crush.

  unfold add_page_table_entry in *; crush.
  rewrite H3 in *; crush.
Qed.  

Lemma new_no_va_mapped_old :
  forall s s' osi' pt_addr' va' pa' osi ma,
    valid_state s ->
    ACT.pre_new osi' va' pa' s ->
    ACT.post_new osi' va' pa' s s' ->

    osi_current_page_table s osi' = Some pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' <> Some ma ->

    ACT.no_va_mapped_to_ma osi ma s' ->
    ACT.no_va_mapped_to_ma osi ma s.
Proof.
  unfold pre_new, post_new, post_all_new; crush.
  unfold no_va_mapped_to_ma, osi_pa_to_pg; crush.
  destruct (os_ident_eq osi osi'); crush.
  (* osi = osi' *)
  unfold osi_current_page_table in H6; crush.
  destruct (padd_eq pt_padd (os_current_page x1)); crush.
  (* pt_padd is current page *)

  destruct H4 with (os_current_page x1)
    (Page (PT_c (PT.set mp va' x2)) (page_owned_by pt_page) (page_cacheable pt_page))
    va (PT.set mp va' x2); crush.
  unfold osi_pa_to_pg; crush.
  apply new_get_mem_pt_addr; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x2 = Os_PO osi'); crush.
  unfold madd_owner in *; crush.
  rewrite PT.get_set_distinct; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H11 in *; crush.

  (* pt_padd is not the current page *)
  destruct H4 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite new_get_mem_distinct_pt; crush.
  destruct_valid_state; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x2 = Os_PO osi'); crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite H11; crush.

  (* osi <> osi' *)
  destruct H4 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite new_get_mem_distinct_pt; crush.
  destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  unfold osi_current_page_table in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi'); crush.
  destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x2 = Os_PO osi'); crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite H11; crush.
Qed.

Lemma madd_of_vadd_new_shared_distinct_va :
  forall s osi pt_addr pt_addr' va va' ma,
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    va <> va' ->
    (exists pg, get_mem (smem_mm (st_memory s)) ma = Some pg) ->
    madd_of_vadd (smem_mm (new_shared (st_memory s) pt_addr va ma)) pt_addr' va'
    = madd_of_vadd (smem_mm (st_memory s)) pt_addr' va'.
Proof.
  intros.
  destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq pt_addr' pt_addr); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite new_get_mem_shared_pt_addr with (pt_page:=pt_page) (pt:=pt); crush.
  rewrite PT.get_set_distinct; crush.
  rewrite H8; crush.
  apply madd_of_vadd_new_shared_distinct_pt with (osi:=osi); crush.
  unfold osi_current_page_table; crush.
Qed.

Lemma new_same_madd_of_vadd :
  forall s osi' pt_addr' va' ma' pt_addr va,
    valid_state s ->
    (va' <> va \/ pt_addr <> pt_addr') ->
    madd_type (smem_mm (st_memory s)) ma' <> None ->
    osi_current_page_table s osi' = Some pt_addr' ->
    madd_type (smem_mm (ACT.new s osi' pt_addr' va' ma')) pt_addr = Some PT ->
    madd_of_vadd (smem_mm (ACT.new s osi' pt_addr' va' ma')) pt_addr va =
    madd_of_vadd (smem_mm (st_memory s))pt_addr va.
Proof.    
  Local Opaque ACT.new.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold madd_type in *; crush.
  destruct (page_content p) eqn:?; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq pt_addr' pt_addr); crush.
  destruct H0; crush.
  (* va' <> va *)
  rewrite H8; crush.
  rewrite STAC.new_get_mem_pt_addr
    with (pt_page:=pt_page) (pt:=pt) in Heqopt_res; crush.
  injection Heqc; intros; subst p0; clear Heqc.
  rewrite PT.get_set_distinct; crush.
  rewrite STAC.new_get_mem_distinct_pt in *; crush.
  rewrite Heqc; crush.
  rewrite <- STAC.new_same_madd_type with (osi:=osi')
    (pt_page':=pt_page) (pt':=pt) (pt_addr':=pt_addr')
    (va':=va') (ma':=ma'); crush.
  unfold madd_type; crush.
  rewrite Heqc; crush.
Qed.

Lemma new_get_mem_complete_semantics : 
  forall s osi' pt_addr' va' ma' ma,
    valid_state s ->
    madd_type (smem_mm (st_memory s)) ma' <> None ->
    osi_current_page_table s osi' = Some pt_addr' ->

    (exists pg,  ma <> ma' /\ ma <> pt_addr' /\
     madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma /\
     get_cache (smem_c (st_memory s)) (va', ma) = Some pg /\
     get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma = Some pg) \/

    (ma = ma' /\
     (exists va pg, page_of_vadd_madd_smem (st_memory s) va ma = Some pg /\
               get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
               Some (Page (page_content pg) (page_owned_by pg) false))) \/

    (ma = ma' /\ ~ shared_page s va' ma' /\
     madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma /\    
     (exists pg, get_cache (smem_c (st_memory s)) (va', ma) = Some pg) /\
     (exists va pg, page_of_vadd_madd_smem (st_memory s) va ma = Some pg /\
               get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma = Some pg)) \/

    (exists pt_page' pt', 
       ma = pt_addr' /\
       get_mem (smem_mm (st_memory s)) pt_addr' = Some pt_page' /\
       page_content pt_page' = PT_c pt' /\
       get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
       Some (Page (PT_c (PT.set pt' va' ma')) (page_owned_by pt_page')
                  (page_cacheable pt_page'))) \/
    
    (ma <> pt_addr' /\ (~ shared_page s va' ma' \/ ma' <> ma) /\
     (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' <> Some ma \/
      get_cache (smem_c (st_memory s)) (va', ma) = None) /\
     get_mem (smem_mm (ACT.new s osi' pt_addr' va' ma')) ma =
     get_mem (smem_mm (st_memory s)) ma).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state; crush.
  unfold madd_type in H0; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq pt_addr' ma); crush.
  (* pt_addr' = ma *)
  right; right; right; left.
  exists pt_page; exists pt; crush.
  Local Transparent ACT.new.
  crush.
  apply STAC.new_get_mem_shared_pt_addr; crush.
  apply STAC.new_get_mem_not_shared_pt_addr; crush.
  (* pt_addr' <> ma *)
  Local Opaque ACT.new.
  destruct (madd_eq ma' ma); crush.
  (* ma' = ma *)
  destruct (shared_page_dec s va' ma); crush.
  (* shared *)
  destruct (snd (remove_ma_cache (smem_c (st_memory s)) ma)) eqn:?; crush.
  destruct p0.
  destruct c.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s)) v ma m p0; crush.
  right; left; crush.
  exists v; exists p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  apply STAC.new_get_mem_shared_new_ma with (pt_page':=pt_page) (pt':=pt) (va_rep:=v); 
    crush.
  right; left; crush.
  exists va'; exists p; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval; crush.
  apply STAC.new_get_mem_shared_new_ma_no_entry with (pt_page':=pt_page) (pt':=pt); crush.
  (* not shared *) 
  Local Transparent ACT.new.
  crush.
  remember (add_page_table_entry (smem_mm (st_memory s)) pt_addr' va' ma).
  destruct o as [mm'|].
  erewrite new_mem_not_shared; crush. 
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma); crush.
  right; right; left; crush.
  exists va'; exists p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite <- get_add_mem; crush.
  right; right; right; right.
  repeat split; try left; crush. 
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H7 in *; crush.
  rewrite get_add_mem_distinct; crush.
  right; right; right; right. 
  destruct (madd_eq m ma); crush;
    unfold add_page_table_entry in *; crush;
    rewrite H7 in *; crush;
    try rewrite get_add_mem_distinct; crush;
    try (left; crush; fail).
  right; right; right; right.
  repeat split; crush; try left; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H7 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H7 in *; crush.
  (* ma'<> ma *)
  Local Opaque ACT.new.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  destruct (madd_eq m ma); crush.
  destruct (get_cache (smem_c (st_memory s)) (va',ma)) eqn:?; crush.
  left; exists p0; crush.
  destruct (shared_page_dec s va' ma'); crush;
    [apply STAC.new_get_mem_shared_distinct_ma_some_entry 
       with (pt_page':=pt_page) (pt':=pt) |
     apply STAC.new_get_mem_not_shared_distinct_ma_some_entry 
     with (pt_page':=pt_page) (pt':=pt)]; crush.
  right; right; right; right.
  destruct (shared_page_dec s va' ma'); crush;
    [apply STAC.new_get_mem_shared_distinct_ma_no_entry 
       with (pt_page':=pt_page) (pt':=pt) | 
     apply STAC.new_get_mem_not_shared_distinct_ma_no_entry 
       with (pt_page':=pt_page) (pt':=pt)]; crush.
  right; right; right; right.
  repeat split; crush; try (right; crush; fail); try left; crush.
  destruct (shared_page_dec s va' ma'); crush;
    [apply STAC.new_get_mem_shared_distinct_ma_not_mapped 
       with (pt_page':=pt_page) (pt':=pt) |
     apply STAC.new_get_mem_not_shared_distinct_ma_not_mapped 
       with (pt_page':=pt_page) (pt':=pt)]; crush.
  right; right; right; right.
  repeat split; crush; try (right; crush; fail); try left; crush.
  destruct (shared_page_dec s va' ma'); crush;
    [apply STAC.new_get_mem_shared_distinct_ma_not_mapped 
       with (pt_page':=pt_page) (pt':=pt) |
     apply STAC.new_get_mem_not_shared_distinct_ma_not_mapped 
       with (pt_page':=pt_page) (pt':=pt)]; crush.
Qed.     


(* Del lemmas *)

Lemma del_osi_pa_to_ma :
  forall s s' osi pa osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_pa_to_ma s osi pa = osi_pa_to_ma s' osi pa. 
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del, osi_pa_to_ma; crush.
Qed.

Lemma del_madd_of_vadd_distinct_osi :
  forall s s' osi pt_padd pt_addr va ma osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi <> osi' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p1); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  destruct (vadd_eq va' va); crush.
  rewrite Heqc in Heqc0; injection Heqc0; intros;
    subst p3; clear Heqc0; crush.
  unfold osi_current_page_table in *; crush.
  assert (page_owned_by p0 = Os_PO osi').
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi' i (os_current_page x0) pt_addr; crush.
  unfold madd_owner; crush.
  assert (page_owned_by p0 = Os_PO osi); crush.
  unfold osi_pa_to_ma in H3; crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i0 pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  erewrite PT.get_remove_distinct; crush.
  rewrite Heqc in Heqc0; crush; injection Heqc0; crush.
  rewrite Heqc; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' pt_addr p1; crush; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc0; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  destruct (vadd_eq va' va); crush.
  rewrite Heqc in Heqc0; injection Heqc0; intros;
    subst p3; clear Heqc0; crush.
  unfold osi_current_page_table in *; crush.
  assert (page_owned_by p0 = Os_PO osi').
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi' i (os_current_page x0) pt_addr; crush.
  unfold madd_owner; crush.
  assert (page_owned_by p0 = Os_PO osi); crush.
  unfold osi_pa_to_ma in H3; crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i0 pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  erewrite PT.get_remove_distinct; crush.
  rewrite Heqc in Heqc0; crush; injection Heqc0; crush.
  rewrite Heqc; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc0; crush.
  unfold remove_page_table_entry in Heqo; crush.
  rewrite Heqc0 in *; crush.
Qed.

Lemma del_madd_of_vadd_distinct_pt :
  forall s s' osi pt_padd pt_addr va ma osi' pt_addr' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    pt_addr' <> pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del,
    madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p1); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite Heqc; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' pt_addr p1; crush.
  unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite Heqc0; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite Heqc; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite Heqc0; crush.
  unfold remove_page_table_entry in Heqo; crush.
  rewrite Heqc0 in *; crush.
Qed.

Lemma del_madd_of_vadd_distinct_va :
  forall s s' pt_addr va ma osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    va' <> va ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del,
    madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p1); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  erewrite PT.get_remove_distinct; crush.
  rewrite Heqc in Heqc0; crush; injection Heqc0; crush.
  rewrite Heqc; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' pt_addr p1; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc0; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  erewrite PT.get_remove_distinct; crush.
  rewrite Heqc in Heqc0; crush; injection Heqc0; crush.
  rewrite Heqc; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc0; crush.
  
  unfold remove_page_table_entry in Heqo; crush.
  rewrite Heqc0 in *; crush.
Qed.

Lemma del_madd_of_vadd_new_va_none :
  forall s s' pt_addr osi va ma,
    valid_state s ->
    ACT.pre_del osi va s ->
    ACT.post_del osi va s s' ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = None.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del,
    madd_of_vadd, madd_page_table; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va) eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va,x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem; crush.
  rewrite PT.get_remove; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va x1 p; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem; crush.
  rewrite PT.get_remove; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  
  unfold remove_page_table_entry in Heqo; crush.
  rewrite Heqc in *; crush.
Qed.

Lemma del_madd_of_vadd_old :
  forall s s' pt_addr va ma osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.
Proof.      
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del,
    madd_of_vadd, madd_page_table; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p) in Heqopt_res0; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  rewrite Heqc; crush.
  injection Heqc0; intros; subst p1; clear Heqc0.
  destruct (vadd_eq va' va); crush.
  rewrite PT.get_remove in *; crush. 
  apply PT.get_remove_old with va'; crush.
  rewrite Heqc0; crush.
  rewrite <- get_add_mem in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' pt_addr p; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2) in Heqopt_res0; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x1 pt_addr); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  rewrite Heqc; crush.
  injection Heqc0; intros; subst p1; clear Heqc0.
  destruct (vadd_eq va' va); crush.
  rewrite PT.get_remove in *; crush. 
  apply PT.get_remove_old with va'; crush.
  rewrite Heqc0; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
Qed.

Lemma del_page_of_vadd_madd_smem :
  forall s s' pt_addr va ma osi' pt_addr' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr' ->

    (exists osi pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->    
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->

    ma <> pt_addr' ->
    page_of_vadd_madd_smem (st_memory s) va ma =
    page_of_vadd_madd_smem (st_memory s') va ma.
Proof.      
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del,
    page_of_vadd_madd_smem; crush.
  rewrite  delete_cache with (ma:=x4) in Heqval0; crush.
  destruct (eq_cache_key (va,ma) (va',x4)); crush.
  rewrite get_remove_cache in Heqval0; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x4) in Heqval0; crush.
  destruct (eq_cache_key (va,ma) (va',x4)); crush.
  rewrite get_remove_cache in Heqval0; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x3 va') eqn:?; crush.
  rename m into mm'.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x4) (pg:=p); crush.
  rewrite <- get_add_mem; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H14 in *; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x4) in Heqval0; crush.
  destruct (eq_cache_key (va,ma) (va',x4)); crush.
  rewrite get_remove_cache in Heqval0; crush.
  rewrite get_remove_cache_old with (k':=(va',x4)) (pg:=p) in Heqval; crush.

  destruct (remove_page_table_entry (smem_mm (st_memory s)) x3 va') eqn:?; crush.
  rename m into mm'.
  assert (get_mem (smem_mm (st_memory s)) ma = get_mem mm' ma).
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite  delete_cache with (ma:=x4) in Heqval0; crush.
  destruct (eq_cache_key (va,ma) (va',x4)); crush.
  rewrite get_remove_cache in Heqval0; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x4); crush.
  destruct (madd_eq x4 ma); crush.
  assert (va <> va'); crush.
  destruct (get_cache (smem_c (st_memory s)) (va',ma)) eqn:?.
  destruct_valid_state.
  synonym_cached.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=ma); crush.
  destruct (get_cache (smem_c (st_memory s)) (va',x4)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x4) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x4); crush.

  unfold remove_page_table_entry in *; crush.
  destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H14 in *; crush.
Qed.

Lemma del_page_of_vadd_madd_smem_ne_osi :
  forall s s' osi va ma osi' pt_addr' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr' ->
    madd_owner (smem_mm (st_memory s)) ma = Os_PO osi ->
    osi <> osi' ->
    page_of_vadd_madd_smem (st_memory s) va ma = 
    page_of_vadd_madd_smem (st_memory s') va ma.
Proof.      
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H13 in *; crush.
  assert (ma <> x1 /\ ma <> x2); [|crush].
  destruct H_valid_virtmap with osi' x1 pt_page pt va' x2; crush;
    unfold madd_owner in *; crush.

  unfold page_of_vadd_madd_smem; crush.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite  delete_cache with (ma:=x2) in Heqval0;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H13; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x2) in Heqval0; crush;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H13; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x2) in Heqval0; crush;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H13; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in Heqval0; crush.
  rewrite  delete_cache with (ma:=x2) in Heqval0; crush;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H13; crush.
  rewrite get_remove_cache_old with (k':=(va',x2)) (pg:=p) in Heqval; crush.

  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  assert (get_mem (smem_mm (st_memory s)) ma = get_mem mm' ma).
  unfold remove_page_table_entry in *; crush.
  rewrite H13 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  
  unfold remove_page_table_entry in *; crush.
  rewrite H13 in *; crush.
Qed.  

Lemma del_page_of_vadd_madd_smem_ne_pt :
  forall s s' pt_addr va ma osi' pt_addr' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr' ->

    (exists osi pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->

    pt_addr <> pt_addr' ->
    ma <> pt_addr' ->
    page_of_vadd_madd_smem (st_memory s) va ma =
    page_of_vadd_madd_smem (st_memory s') va ma.
Proof.      
  intros.
  eapply del_page_of_vadd_madd_smem; crush.
Qed.

Lemma del_get_mem_distinct_pt :  
  forall s s' ma osi' pt_addr' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr' ->
    madd_type (smem_mm (st_memory s)) ma = Some PT ->
    ma <> pt_addr' ->
    get_mem (smem_mm (st_memory s')) ma = 
    get_mem (smem_mm (st_memory s)) ma.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  STAC.destruct_valid_state.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite MM.get_add_mem_distinct; crush.
  unfold madd_type in *; crush.
  destruct H_valid_cache with va' ma p; unfold is_rw_page in *; crush.
  destruct (page_content p0); crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct; crush.

  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
Qed.

Lemma del_get_mem_pt_page :
  forall s s' pt_addr pt_page pt osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    get_mem (smem_mm (st_memory s')) pt_addr =
      Some (Page (PT_c (PT.remove pt va'))
                 (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.      
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  STAC.destruct_valid_state.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H4 in *; crush.
  rewrite <- MM.get_add_mem; crush.
  destruct H_valid_cache with va' x1 p; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H4 in *; crush.
  rewrite <- MM.get_add_mem; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite H4 in *; crush.
Qed.

Lemma del_get_mem_same_page_but_value :
  forall s s' ma old_pg osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    get_mem (smem_mm (st_memory s)) ma = Some old_pg ->
    exists new_pg,
      get_mem (smem_mm (st_memory s')) ma = Some new_pg /\
      same_page_but_value old_pg new_pg.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H12 in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  assert (exists new_pg,
    get_mem mm' ma = Some new_pg /\ same_page_but_value old_pg new_pg); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H12 in *; crush.
  destruct (madd_eq ma x1); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  exists (Page (PT_c (PT.remove pt va')) (page_owned_by old_pg)
    (page_cacheable old_pg)); crush.
  unfold same_page_but_value, same_page_type; rewrite H12;
    repeat split; simpl; auto.
  unfold same_page_but_value, same_page_type; repeat split; simpl; auto.
  destruct (page_content old_pg); auto.
  destruct o; auto.

  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush. 
  destruct (madd_eq x2 ma); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  exists p; crush.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value, same_page_type; rewrite H14, H15, H16, H17, H18;
    repeat split; simpl; auto.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H12; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H12; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite H12 in *; crush.
Qed.

Lemma del_page_of_vadd_madd_smem_pt_page :
  forall s s' va pt_addr pt_page pt osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    osi_current_page_table s osi' = Some pt_addr ->
    page_of_vadd_madd_smem (st_memory s) va pt_addr = Some pt_page ->
    page_content pt_page = PT_c pt ->
    page_of_vadd_madd_smem (st_memory s') va pt_addr =
      Some (Page (PT_c (PT.remove pt va'))
                 (page_owned_by pt_page)
                 (page_cacheable pt_page)).
Proof.      
  unfold page_of_vadd_madd_smem; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va x1 pt_page; unfold is_rw_page in *; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va x1 pt_page; unfold is_rw_page in *; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  rewrite delete_cache with (ma:=x2) in *; crush.
  destruct (eq_cache_key (va,x1) (va',x2)); crush.
  rewrite get_remove_cache in *; crush.
  rewrite get_remove_cache_old with (k':=(va',x2)) (pg:=p) in Heqval; crush.
  eapply del_get_mem_pt_page; crush.
Qed.

Lemma del_same_madd_type:
  forall s s' ma osi' va',
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    madd_type (smem_mm (st_memory s')) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.  
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content pt_page) eqn:?; crush.
  injection H11; intros; subst p0; clear H11.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  assert (madd_type mm' ma = madd_type (smem_mm (st_memory s)) ma).
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  unfold madd_type; crush.
  destruct (madd_eq ma x1); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  rewrite Heqc; crush.

  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  destruct (madd_eq x2 ma); crush.
  unfold madd_type; crush.
  rewrite <- get_add_mem; crush.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page in *; crush.
  rewrite H10, H14; crush.
  unfold madd_type in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
Qed.

Lemma del_no_va_mapped_same_mem :
  forall s pt_addr' va' osi pa ma,
    valid_state s ->
    osi_pa_to_ma s osi pa = Some ma ->
    ACT.no_va_mapped_to_ma osi ma s ->
    ma <> pt_addr' ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (delete (st_memory s) pt_addr' va')) ma.
Proof.
  intros.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m into mm'.
  assert (get_mem (smem_mm (st_memory s)) ma = get_mem mm' ma).
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.

  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  destruct (madd_eq m ma); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.

  destruct (get_cache (smem_c (st_memory s)) (va',ma)) eqn:?.
  destruct_valid_state.
  destruct H_valid_cache_va_ma with va' ma p; crush.
  destruct H1 with (pt_padd:=x0) (pt_page:=x1) (va:=va') (mp:=x2); crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  destruct (os_ident_eq x osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  destruct H_valid_virtmap with x i1 x1 x2 va' ma; crush.
  assert (madd_owner (smem_mm (st_memory s)) i1 = Os_PO x); crush.
  unfold madd_owner in H8; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?.
  unfold madd_owner, os_accessible in *; crush.
  unfold madd_owner, os_accessible in *; crush.
  rewrite H10 in *; crush.

  rewrite delete_mem_no_entry with (mm':=mm') (ma:=ma); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.

  destruct (get_cache (smem_c (st_memory s)) (va',m)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=m) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=m); crush.
  rewrite delete_mem_not_mapped; crush.
  rewrite delete_mem_invalid; crush.
Qed.  
  

(* If ma was mapped in s, but not in s', then
   the page_of_vadd_madd_smem is in memory in s'. *)
Lemma del_no_va_mapped_smem_then_mem :
  forall s s' osi pt_addr' va' pt_padd pt_addr va ma pg,
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr' ->
    st_memory s' = ACT.delete_smem osi pt_addr' va' s ->
    st_hypervisor s' = st_hypervisor s ->
    ACT.no_va_mapped_to_ma osi ma s' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    ma <> pt_addr' ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg ->
    get_mem (smem_mm (st_memory s')) ma = Some pg.
Proof.
  crush.
  destruct_valid_state.
  unfold osi_pa_to_ma, madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite H1; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m into mm'.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  destruct (get_cache (smem_c (st_memory s)) (va',m)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=m) (pg:=p); crush.
  destruct (madd_eq m ma); crush.
  rewrite <- get_add_mem; crush.
  destruct (vadd_eq va va'); crush.
  unfold page_of_vadd_madd_smem in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  synonym_cached.

  destruct (madd_eq pt_addr pt_addr'); crush.
  destruct (vadd_eq va va'); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H3 with pt_padd
    (Page (PT_c (PT.remove p1 va'))
      (page_owned_by p0)
      (page_cacheable p0)) va (PT.remove p1 va'); crush.
  unfold osi_pa_to_pg; crush.
  rewrite H1, H2; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m0 into mm'.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=m) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- get_add_mem; crush.
  destruct H_valid_cache with va' pt_addr' p; unfold is_rw_page in *; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite PT.get_remove_distinct with (inf:=ma); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p2) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H3 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite H1, H2; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m0 into mm'.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=m) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H_valid_cache with va' pt_addr p; unfold is_rw_page in *; crush.
  rewrite delete_mem_invalid; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=m); crush.

  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite get_add_mem_distinct; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct (madd_eq pt_addr pt_addr'); crush.
  destruct (vadd_eq va va'); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite Heqc0 in Heqc; injection Heqc; intros; subst p2; clear Heqc.
  destruct H3 with pt_padd
    (Page (PT_c (PT.remove p1 va'))
      (page_owned_by p)
      (page_cacheable p)) va (PT.remove p1 va'); crush.
  unfold osi_pa_to_pg; crush.
  rewrite H1, H2; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m0 into mm'.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=m); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite <- get_add_mem; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite PT.get_remove_distinct with (inf:=ma); crush.
  destruct H3 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite H1, H2; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  rename m0 into mm'.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=m); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc0 in *; crush.

  rewrite delete_mem_not_mapped; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H3 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite H1, H2; crush.
  rewrite delete_mem_not_mapped; crush.

  unfold remove_page_table_entry, osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  get_osi_current_pt_page osi; crush.
  rewrite H13 in *; crush.
Qed.  


Lemma del_no_va_mapped_new :
  forall s s' osi' va' osi ma,
    valid_state s ->
    ACT.pre_del osi' va' s ->
    ACT.post_del osi' va' s s' ->
    ACT.no_va_mapped_to_ma osi ma s ->
    ACT.no_va_mapped_to_ma osi ma s'.
Proof.
  unfold pre_del, post_del, pre_all_del, no_va_mapped_to_ma;
    crush.
  unfold osi_pa_to_pg in *; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq x1 i0); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H2 with pt_padd p0 va p1; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) i0 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p) in *; crush.
  rewrite get_add_mem_distinct in H3; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem in *; crush.
  injection H4; intros; subst mp; crush.
  destruct (vadd_eq va va'); crush.
  rewrite PT.get_remove in *; crush.
  apply PT.get_remove_old with va'; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' i0 pt_page; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2) in *; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem in *; crush.
  injection H4; intros; subst mp; crush.
  destruct (vadd_eq va va'); crush.
  rewrite PT.get_remove in *; crush.
  apply PT.get_remove_old with va'; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.

  destruct H2 with pt_padd pt_page va mp; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p) in *; crush.
  rewrite get_add_mem_distinct in H3; crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite MM.get_add_mem_distinct in *; crush.
  rewrite <- get_add_mem in *; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' i0 pt_page; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2) in *; crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct in *; crush.

  rewrite delete_mem_invalid in *; crush.
Qed.

(* Del hyper lemmas *)
Lemma del_hyper_osi_pa_to_ma :
  forall s s' osi pa va',
    valid_state s ->
    ACT.pre_del_hyper va' s ->
    ACT.post_del_hyper va' s s' ->
    osi_pa_to_ma s osi pa = osi_pa_to_ma s' osi pa. 
Proof.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper, osi_pa_to_ma; crush.
Qed.
    
Lemma del_hyper_page_of_vadd_madd_smem_ne_osi :
  forall s s' osi va ma va',
    valid_state s ->
    ACT.pre_del_hyper va' s ->
    ACT.post_del_hyper va' s s' ->
    madd_owner (smem_mm (st_memory s)) ma = Os_PO osi ->
    madd_type (smem_mm (st_memory s)) ma = Some RW ->
    page_of_vadd_madd_smem (st_memory s) va ma = 
    page_of_vadd_madd_smem (st_memory s') va ma.
Proof.      
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H9 in *; crush.
  assert (ma <> x1); crush.
  destruct H_valid_virtmap with (st_active_os s) x0 pt_page pt va' x1; crush;
    unfold madd_owner in *; crush.
  assert (ma <> x0); crush.
  unfold madd_type in *; crush.
  rewrite H9 in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite  delete_cache with (ma:=x1) in Heqval0;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H9; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x1) in Heqval0; crush;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H9; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in *; crush.
  rewrite  delete_cache with (ma:=x1) in Heqval0; crush;
    unfold madd_of_vadd, madd_page_table; crush; try rewrite H9; crush.
  rewrite get_remove_cache_old with (k':=(va',x1)) (pg:=p) in Heqval; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x0 va') eqn:?; crush.
  rename m into mm'.
  assert (get_mem (smem_mm (st_memory s)) ma = get_mem mm' ma).
  unfold remove_page_table_entry in *; crush.
  rewrite H9 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  destruct (get_cache (smem_c (st_memory s)) (va',x1)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x1) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H9; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H9; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite H9 in *; crush.
Qed.  

Lemma del_hyper_same_madd_type:
  forall s ma pt_addr' va',
    valid_state s ->
    madd_type (smem_mm (delete (st_memory s) pt_addr' va')) ma = 
    madd_type (smem_mm (st_memory s)) ma.
Proof.  
  intros.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr' va') eqn:?; crush.
  rename m into mm'.
  assert (madd_type mm' ma = madd_type (smem_mm (st_memory s)) ma); crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  unfold madd_type; crush.
  destruct (madd_eq ma pt_addr'); crush.
  rewrite <- get_add_mem; crush.
  rewrite Heqc; crush.
  rewrite get_add_mem_distinct; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?.
  destruct (get_cache (smem_c (st_memory s)) (va',m)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=m) (pg:=p); crush.
  destruct (madd_eq m ma); crush.
  unfold madd_type; crush.
  rewrite <- get_add_mem; crush.
  destruct_valid_state.
  destruct H_valid_cache with va' ma p; unfold is_rw_page in *; crush.
  rewrite H1, H4; crush.
  unfold madd_type in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=m); crush.
  rewrite delete_mem_not_mapped; crush.
  rewrite delete_mem_invalid; crush.
Qed.

Lemma del_hyper_get_mem_same_page_but_value :
  forall s s' ma old_pg va',
    valid_state s ->
    ACT.pre_del_hyper  va' s ->
    ACT.post_del_hyper va' s s' ->
    get_mem (smem_mm (st_memory s)) ma = Some old_pg ->
    exists new_pg,
      get_mem (smem_mm (st_memory s')) ma = Some new_pg /\
      same_page_but_value old_pg new_pg.
Proof.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data; crush.
  STAC.get_aos_current_pt_page; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H8 in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x0 va') eqn:?; crush.
  rename m into mm'.
  assert (exists new_pg,
    get_mem mm' ma = Some new_pg /\ same_page_but_value old_pg new_pg); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H8 in *; crush.
  destruct (madd_eq ma x0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  exists (Page (PT_c (PT.remove pt va')) (page_owned_by old_pg)
    (page_cacheable old_pg)); crush.
  unfold same_page_but_value, same_page_type; rewrite H8;
    repeat split; simpl; auto.
  unfold same_page_but_value, same_page_type; repeat split; simpl; auto.
  destruct (page_content old_pg); auto.
  destruct o; auto.

  destruct (get_cache (smem_c (st_memory s)) (va',x1)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x1) (pg:=p); crush. 
  destruct (madd_eq x1 ma); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  exists p; crush.
  destruct H_valid_cache with va' ma p; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value, same_page_type; rewrite H10, H11, H12, H13, H14;
    repeat split; simpl; auto.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H8; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H8; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite H8 in *; crush.
Qed.
  
Lemma del_hyper_madd_of_vadd_old :
  forall s s' pt_addr va ma va',
    valid_state s ->
    ACT.pre_del_hyper va' s ->
    ACT.post_del_hyper va' s s' ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.
Proof.      
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper,
    madd_of_vadd, madd_page_table; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x0 va') eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va',x1)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x1) (pg:=p) in Heqopt_res0; crush.
  rewrite get_add_mem_distinct in Heqopt_res0; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x0 pt_addr); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  rewrite Heqc; crush.
  injection Heqc0; intros; subst p1; clear Heqc0.
  destruct (vadd_eq va' va); crush.
  rewrite PT.get_remove in *; crush. 
  apply PT.get_remove_old with va'; crush.
  rewrite Heqc0; crush.
  rewrite <- get_add_mem in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va' pt_addr p; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x1) in Heqopt_res0; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x0 pt_addr); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  rewrite Heqc; crush.
  injection Heqc0; intros; subst p1; clear Heqc0.
  destruct (vadd_eq va' va); crush.
  rewrite PT.get_remove in *; crush. 
  apply PT.get_remove_old with va'; crush.
  rewrite Heqc0; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.

  unfold remove_page_table_entry in *; crush.
  rewrite Heqc in *; crush.
Qed.



(* Pin lemmas *)
Lemma page_pin_osi_pa_to_ma :
  forall s s' osi pa ma osi' pa' type,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    osi_pa_to_ma s osi pa = Some ma ->
    osi_pa_to_ma s' osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, osi_pa_to_ma; crush.
  destruct (os_ident_eq osi' osi); crush;
  [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa' pa); crush;
  rewrite P2M.get_set_distinct in *; crush.
Qed.

Lemma page_pin_osi_pa_to_ma_distinct_osi :
  forall s s' osi pa ma osi' pa' type,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    osi <> osi' ->
    osi_pa_to_ma s' osi pa = Some ma ->
    osi_pa_to_ma s osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, osi_pa_to_ma; crush.
  rewrite HYP.get_set_distinct in *; crush.
Qed.

Lemma page_pin_osi_pa_to_ma_distinct_pa :
  forall s s' osi pa ma osi' pa' type,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    pa <> pa' ->
    osi_pa_to_ma s' osi pa = Some ma ->
    osi_pa_to_ma s osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, osi_pa_to_ma; crush.
  destruct (os_ident_eq osi' osi); crush;
  [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  rewrite P2M.get_set_distinct in *; crush.
Qed.

Lemma page_pin_madd_of_vadd :
  forall s s' pt_addr va ma osi' pa' type ,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq x pt_addr); crush;
      rewrite MM.get_add_mem_distinct; crush.
  rewrite Heqc; crush.
Qed.

Lemma page_pin_madd_of_vadd_old :
  forall s s' pt_addr va ma osi' pa' type ,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq x pt_addr); crush;
      [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  Transparent new_page.
  unfold new_page in Heqc; crush.
  Opaque new_page.
  injection Heqc; intros; subst p1.
  rewrite PT.get_empty in *; crush.
  rewrite Heqc; crush.
Qed.

Lemma page_pin_page_of_vadd_madd_smem :
  forall s s' osi pt_padd pt_addr va ma pg osi' pa' type ,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg ->
    page_of_vadd_madd_smem (st_memory s') va ma = Some pg.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin, page_of_vadd_madd_smem; crush.
  destruct (madd_eq x ma); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  generalize (H_valid_hyper osi i pt_padd pt_addr); intros.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H1 in *; crush.
  rewrite H17 in *; crush.
Qed.

Lemma page_pin_page_of_vadd_madd_smem_old :
  forall s s' osi pt_padd pt_addr va ma pg osi' pa' type ,
    valid_state s ->
    ACT.pre_pin osi' pa' type s ->
    ACT.post_pin osi' pa' type s s' ->
    osi_pa_to_ma s' osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    page_of_vadd_madd_smem (st_memory s') va ma = Some pg ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg.
Proof.
  Opaque new_page.
  unfold page_of_vadd_madd_smem; crush;
    try (unfold ACT.pre_pin, ACT.post_pin in *; crush; fail).
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma).
  eapply page_pin_madd_of_vadd_old; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr).
  destruct (os_ident_eq osi osi'); crush.
  eapply page_pin_osi_pa_to_ma_distinct_pa; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  unfold osi_pa_to_ma in H2; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite <- P2M.get_set in *; crush.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  destruct (page_content x2) eqn:?; crush.
  eapply page_pin_osi_pa_to_ma_distinct_osi; crush.

  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (madd_eq x ma); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (os_ident_eq osi osi'); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa' pt_padd); crush;
    rewrite P2M.get_set_distinct in *; crush.
  generalize (H_valid_hyper osi' x0 pt_padd pt_addr); intros.
  unfold madd_owner in *; crush.
  generalize (H_valid_hyper osi i0 pt_padd pt_addr); intros.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H4 in *; crush.
  rewrite H19 in *; crush.
Qed.

Lemma pin_same_get_mem :
  forall s s' osi' pa' type' osi pa ma,
    valid_state s ->
    ACT.pre_pin osi' pa' type' s ->
    ACT.post_pin osi' pa' type' s s' ->
    osi_pa_to_ma s osi pa = Some ma ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (st_memory s')) ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_pin, ACT.post_pin; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq x ma); crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  rewrite get_add_mem_distinct; crush.
Qed.

(* Unpin lemmas *)
Lemma page_unpin_osi_pa_to_ma_distinct_osi :
  forall s s' osi pa ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi <> osi' ->
    osi_pa_to_ma s osi pa = Some ma ->
    osi_pa_to_ma s' osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_ma; crush.
  rewrite HYP.get_set_distinct; crush.
Qed.

Lemma page_unpin_osi_pa_to_ma_distinct_pa :
  forall s s' osi pa ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    pa <> pa' ->
    osi_pa_to_ma s osi pa = Some ma ->
    osi_pa_to_ma s' osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_ma; crush.
  destruct (os_ident_eq osi' osi); crush;
    [rewrite <- HYP.get_set | rewrite HYP.get_set_distinct]; crush.
  erewrite P2M.get_remove_distinct; crush.
Qed.

Lemma page_unpin_osi_pa_to_ma_old :
  forall s s' osi pa ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi_pa_to_ma s' osi pa = Some ma ->
    osi_pa_to_ma s osi pa = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_ma; crush.
  destruct (os_ident_eq osi' osi); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa' pa); crush;
    [rewrite P2M.get_remove in * | eapply P2M.get_remove_old]; crush.
Qed.

Lemma page_unpin_madd_of_vadd_distinct_osi :
  forall s s' osi pt_padd pt_addr va ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi <> osi' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq x0 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  crush.
  unfold osi_pa_to_ma in *; crush.
  elim H2; crush.
  STAC.destruct_valid_state.
  assert (Os_PO osi = Os_PO osi'); crush.
  rewrite <- H_valid_hyper with osi' x pa' x3; crush.
  rewrite <- H_valid_hyper with osi i pt_padd x3; crush.
  rewrite Heqc; crush.
Qed.

Lemma page_unpin_madd_of_vadd_distinct_pt_padd :
  forall s s' osi pt_padd pt_addr va ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    pt_padd <> pa' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq x0 pt_addr); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (os_ident_eq osi' osi); crush.
  STAC.destruct_valid_state.
  elim H2; crush.
  assert (Os_PO osi = Os_PO osi'); crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi' x pa' x3; crush.
  rewrite <- H_valid_hyper with osi i pt_padd x3; crush.
  rewrite Heqc; crush.
Qed.

Lemma page_unpin_madd_of_vadd_old :
  forall s s' pt_addr va ma osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, madd_of_vadd, madd_page_table; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (madd_eq x0 pt_addr); crush;
      [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  Transparent new_page.
  unfold new_page in Heqc; crush.
  Opaque new_page.
  rewrite Heqc; crush.
Qed.

Lemma page_unpin_osi_pa_to_pg_distinct_osi :
  forall s s' osi pa pg osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi <> osi' ->
    osi_pa_to_pg s osi pa = Some pg ->
    osi_pa_to_pg s' osi pa = Some pg.
Proof.  
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_pg; crush.
  rewrite HYP.get_set_distinct in *; crush.
  rewrite get_add_mem_distinct; crush.
  assert (page_owned_by pg = Os_PO osi); crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i pa i0; crush.
  unfold madd_owner; crush.
  STAC.destruct_valid_state.
  assert (page_owned_by pg = Os_PO osi'); crush.
  rewrite <- H_valid_hyper with osi' x pa' i0; crush.
  unfold madd_owner; crush.
Qed.

Lemma page_unpin_osi_pa_to_pg_distinct_pa :
  forall s s' osi pa pg osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    pa <> pa' ->
    osi_pa_to_pg s osi pa = Some pg ->
    osi_pa_to_pg s' osi pa = Some pg.
Proof.  
  intros.
  destruct (os_ident_eq osi' osi); crush.
  2: eapply page_unpin_osi_pa_to_pg_distinct_osi; crush.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_pg in *; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=i0); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  elim H_injective_hyper with osi x pa pa' i0; crush.
Qed.

Lemma page_unpin_osi_pa_to_pg_old :
  forall s s' osi pa pg osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi_pa_to_pg s' osi pa = Some pg ->
    osi_pa_to_pg s osi pa = Some pg.
Proof.  
  intros.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_pg in *; crush.
  destruct (os_ident_eq osi' osi); crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq pa' pa); crush.
  rewrite P2M.get_remove in *; crush.
  assert (P2M.get x pa = Some i0); crush.
  rewrite P2M.get_remove_old with (k':=pa') (inf:=i0); crush.
  destruct (madd_eq x0 i0); crush.
  elim n.
  STAC.destruct_valid_state.
  rewrite (H_injective_hyper osi x pa' pa i0); crush.
  rewrite get_add_mem_distinct in H2; crush.

  rewrite HYP.get_set_distinct in *; crush.
  rewrite get_add_mem_distinct in H2; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i pa i0; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi'); crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi' x pa' i0; crush.
Qed.

Lemma page_unpin_page_of_vadd_madd_smem :
  forall s s' osi pt_padd pt_addr va ma pg osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg ->
    page_of_vadd_madd_smem (st_memory s') va ma = Some pg.
Proof.
  Opaque new_page.
  unfold ACT.pre_unpin, ACT.post_unpin, page_of_vadd_madd_smem; crush.
  destruct (madd_eq x0 ma); crush;
      [rewrite <- MM.get_add_mem | rewrite MM.get_add_mem_distinct]; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (os_ident_eq osi' osi); crush.
  unfold osi_pa_to_ma in *; crush.  
  elim H13 with pt_padd p0 va p1; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_pg; crush.
  elim n; crush; clear n.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in H2; crush.
  unfold valid_hypervisor in H_valid_hyper.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  assert (page_owned_by pg = Os_PO osi').
  rewrite <- H_valid_hyper with osi' x pa' ma; crush.
  unfold madd_owner; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H14 in *; crush.
  rewrite H15 in *; crush.
Qed.  

Lemma page_unpin_page_of_vadd_madd_smem_old :
  forall s s' osi pt_padd pt_addr va ma pg osi' pa',
    valid_state s ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi_pa_to_ma s' osi pt_padd = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s')) pt_addr va = Some ma ->
    page_of_vadd_madd_smem (st_memory s') va ma = Some pg ->
    page_of_vadd_madd_smem (st_memory s) va ma = Some pg.
Proof.
  Opaque new_page.
  unfold page_of_vadd_madd_smem; crush;
    try (unfold ACT.pre_unpin, ACT.post_unpin in *; crush; fail).
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma).
  eapply page_unpin_madd_of_vadd_old; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr).
  eapply page_unpin_osi_pa_to_ma_old; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in H6; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner in *; crush.
  destruct (os_ident_eq osi' osi); crush.
  unfold osi_pa_to_ma in *; crush.
  elim H15 with pt_padd p0 va p1; crush.
  unfold osi_pa_to_pg; crush.
  elim n; crush; clear n.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in H6; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  assert (page_owned_by x0 = Os_PO osi').
  rewrite <- H_valid_hyper with osi' x pa' ma; crush.
  unfold madd_owner; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H18 in *; crush.
  rewrite H19 in H20; crush.
Qed.

Lemma unpin_no_va_mapped_same_mem :
  forall s s' osi' pa' osi pa ma,
    valid_state s ->
    osi <> osi' ->
    ACT.pre_unpin osi' pa' s ->
    ACT.post_unpin osi' pa' s s' ->
    osi_pa_to_ma s osi pa = Some ma ->
    ACT.no_va_mapped_to_ma osi ma s ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (st_memory s')) ma.
Proof.
  unfold pre_unpin, post_unpin; crush.

  destruct (madd_eq ma x0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x3 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) x3 = Os_PO osi'); crush.
Qed.  

Lemma page_unpin_no_va_mapped :
  forall s s' osi pa ma,
    valid_state s ->
    ACT.pre_unpin osi pa s ->
    ACT.post_unpin osi pa s s' ->
    (ACT.no_va_mapped_to_ma osi ma s <->
    ACT.no_va_mapped_to_ma osi ma s').
Proof.
  Opaque new_page.
  split; intros; unfold no_va_mapped_to_ma; crush;
    unfold pre_unpin, post_unpin, osi_pa_to_pg in *; crush.
  rewrite <- HYP.get_set in *; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (padd_eq pa pt_padd); crush.
  rewrite P2M.get_remove in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  destruct H2 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite P2M.get_remove_old with (k':=pa) (inf:=i0); crush.  
  rewrite <- get_add_mem in H3; crush.
  Transparent new_page.
  crush.
  Opaque new_page.

  destruct H2 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite <- HYP.get_set; crush.
  destruct (padd_eq pa pt_padd); crush.
  rewrite P2M.get_remove; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite H12 with pt_page mp in H5; crush.
  rewrite PT.get_empty in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=i0); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  apply n.
  rewrite H_injective_hyper with osi x pt_padd pa i0; crush.
Qed.


End STAC.

