(***********************************************************************
 * File: Stealth_non_stealth_equivalence.v
 * $Revision: 12920 $
 * $Date: 2016-09-22 18:44:11 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * State equivalence for non stealth pages, and proof of invariance.
 *************************************************************************)
Require Import Stealth_actions Stealth_valid Stealth_isolation Exec_invariant
  Stealth_lemmas.

Module STEALTH_NON_STEALTH_EQ.
  Import PG C T MM SMEM ST ST_VALID OS_ISO
         STEALTH_ACTIONS VALID_SM_STATE
         OS_ISO STEQ STEALTH_ISOLATION STEALTH_LEMMAS.

  Inductive non_stealth_access : sm_action -> Prop :=
  | non_stealth_read: forall va, stealth_va <> va ->
    non_stealth_access (Action (Read va))
  | non_stealth_read_hyper: forall va, stealth_va <> va ->
    non_stealth_access (Action (Read_hyper va))
  | non_stealth_write: forall va val, stealth_va <> va ->
    non_stealth_access (Action (Write va val))
  | non_stealth_write_hyper: forall va val,
    non_stealth_access (Action (Write_hyper va val))
  | non_stealth_new: forall osi va pa,
    non_stealth_access (New_normal osi va pa)
  | non_stealth_del: forall osi va, va <> stealth_va ->
    non_stealth_access (Action (Del_untrusted osi va))
  | non_stealth_pin : forall osi pa type,
    non_stealth_access (Action (Page_pin_untrusted osi pa type))
  | non_stealth_unpin : forall osi pa,
    non_stealth_access (Action (Page_unpin_untrusted osi pa))
  | non_stealth_lswitch: forall osi pa,
    non_stealth_access (LSwitch_untrusted_sm osi pa).

Definition osi_non_stealth_equivalent_hyper_mapping (s1 s2:state) (osi:os_ident):
  Prop :=
  forall pa ma1 pg1,
    osi_pa_to_ma s1 osi pa = Some ma1 ->
    get_mem (smem_mm (st_memory s1)) ma1 = Some pg1 ->
    exists ma2, exists pg2,
      osi_pa_to_ma s2 osi pa = Some ma2 /\
      get_mem (smem_mm (st_memory s2)) ma2 = Some pg2 /\
      same_page_but_value pg1 pg2.

Definition osi_non_stealth_equivalent_hyper (s1 s2:state) (osi:os_ident) : Prop :=
  osi_non_stealth_equivalent_hyper_mapping s1 s2 osi /\
  osi_non_stealth_equivalent_hyper_mapping s2 s1 osi.

Definition osi_non_stealth_equivalent_pt_pages (s1 s2:state) (osi:os_ident):
  Prop :=
    forall pt_padd pt_page1 mp1 va ma1,
    (* pt_padd maps to pt_page1, which is a PT page with map mp1. *)
    osi_pa_to_pg s1 osi pt_padd = Some pt_page1 ->
    PT_c mp1 = (page_content pt_page1) ->
    (* In mp1, va maps to ma1 *)
    va <> stealth_va ->
    PT.get mp1 va = Some ma1 ->
    (* Then, for s1 and s2 to be equivalent, pt_padd must map
       to a pt_page2, of type PT *)
    (exists pt_page2, osi_pa_to_pg s2 osi pt_padd = Some pt_page2 /\
     exists mp2, PT_c mp2 = (page_content pt_page2) /\
     exists ma2, PT.get mp2 va = Some ma2 /\
     (forall pa,
        osi_pa_to_ma s1 osi pa = Some ma1 ->
        osi_pa_to_ma s2 osi pa = Some ma2 ) /\
     (forall pg1, page_of_vadd_madd_smem (st_memory s1) va ma1 = Some pg1 ->
      exists pg2, page_of_vadd_madd_smem (st_memory s2) va ma2 = Some pg2 /\
                  same_page_but_value pg1 pg2)).

Definition osi_non_stealth_equivalent_smem (s1 s2:state) (osi:os_ident) : Prop :=
  osi_non_stealth_equivalent_pt_pages s1 s2 osi /\
  osi_non_stealth_equivalent_pt_pages s2 s1 osi.

Definition same_current_pt_page_one (s1 s2:state) (osi:os_ident) :=
  (exists os1, OSS.get (st_oss s1) osi = Some os1) ->
   forall os1, OSS.get (st_oss s1) osi = Some os1 ->
     exists os2,
       OSS.get (st_oss s2) osi = Some os2 /\
       os_current_page os1 = os_current_page os2.

Definition same_current_pt_page (s1 s2:state) (osi:os_ident) :=
  same_current_pt_page_one s1 s2 osi /\
  same_current_pt_page_one s2 s1 osi.

(** Two states are osi non stealth equivalent, 
    if they are equal regarding the non stealth pages
    of all osis and the hypervisor.
*)
Definition osi_non_stealth_equivalent (s1 s2:state) :=
  forall osi,
     same_current_pt_page s1 s2 osi /\
     osi_non_stealth_equivalent_hyper s1 s2 osi /\
     osi_non_stealth_equivalent_smem s1 s2 osi.

Lemma same_page_but_value_refl : forall pg,
  same_page_but_value pg pg.
Proof.
  unfold same_page_but_value; crush.
  apply STAC.same_page_type_refl.
Qed.
Local Hint Resolve same_page_but_value_refl
  STAC.same_page_type_refl.

Lemma osi_non_stealth_equivalent_hyper_refl :
  forall (s:state) (osi:os_ident),
    osi_non_stealth_equivalent_hyper s s osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper,
  osi_non_stealth_equivalent_hyper_mapping;
  crush; repeat eexists; crush.
Qed.

Lemma osi_non_stealth_equivalent_smem_refl :
  forall (s:state) (osi:os_ident),
    osi_non_stealth_equivalent_smem s s osi.
Proof.
  unfold osi_non_stealth_equivalent_smem,
         osi_non_stealth_equivalent_pt_pages;
  crush; repeat eexists; crush.
Qed.
Local Hint Resolve osi_non_stealth_equivalent_hyper_refl
      osi_non_stealth_equivalent_smem_refl.

Lemma osi_non_stealth_equivalent_refl : forall (s:state),
  osi_non_stealth_equivalent s s.
Proof.
  intros; unfold osi_non_stealth_equivalent,
    same_current_pt_page,
    same_current_pt_page_one; split; intros;
  repeat eexists; crush.
Qed.
Local Hint Resolve osi_non_stealth_equivalent_refl.

Lemma osi_non_stealth_equivalent_hyper_symm :
  forall s1 s2 osi,
    osi_non_stealth_equivalent_hyper s1 s2 osi ->
    osi_non_stealth_equivalent_hyper s2 s1 osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper; crush.
Qed.

Lemma osi_non_stealth_equivalent_smem_symm :
  forall s1 s2 osi,
    osi_non_stealth_equivalent_smem s1 s2 osi ->
    osi_non_stealth_equivalent_smem s2 s1 osi.
Proof.
  unfold osi_non_stealth_equivalent_smem; crush.
Qed.
Local Hint Resolve osi_non_stealth_equivalent_hyper_symm
      osi_non_stealth_equivalent_smem_symm.

Lemma osi_non_stealth_equivalent_symm : forall (s1 s2:state),
  osi_non_stealth_equivalent s1 s2 -> osi_non_stealth_equivalent s2 s1.
Proof.
  unfold osi_non_stealth_equivalent, same_current_pt_page; crush;
    try destruct H with osi; crush.
Qed.
Local Hint Resolve osi_non_stealth_equivalent_symm.

Lemma same_page_but_value_trans : forall pg1 pg2 pg3,
  same_page_but_value pg1 pg2 ->
  same_page_but_value pg2 pg3 ->
  same_page_but_value pg1 pg3.
Proof.
  unfold same_page_but_value; intros.
  repeat (match goal with [H: _ /\ _ |-_ ] => destruct H end).
  repeat split.
  crush.
  crush.

  clear H H3 H0 H1.
  unfold same_page_type in *.
  destruct (page_content pg1);
  destruct (page_content pg2); try congruence;
  destruct (page_content pg3); try congruence;
    destruct o;
    destruct o0; try congruence;
    destruct o1; congruence.
Qed.

Lemma osi_non_stealth_equivalent_hyper_mapping_trans :
  forall (s1 s2 s3:state) (osi:os_ident),
    osi_non_stealth_equivalent_hyper_mapping s1 s2 osi ->
    osi_non_stealth_equivalent_hyper_mapping s2 s3 osi ->
    osi_non_stealth_equivalent_hyper_mapping s1 s3 osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  destruct H with pa ma1 pg1; crush.
  destruct H0 with pa x x0; crush.
  exists x1; exists x2; crush.
  apply same_page_but_value_trans with x0; crush.
Qed.

Lemma osi_non_stealth_equivalent_hyper_trans :
  forall (s1 s2 s3:state) (osi:os_ident),
    osi_non_stealth_equivalent_hyper s1 s2 osi ->
    osi_non_stealth_equivalent_hyper s2 s3 osi ->
    osi_non_stealth_equivalent_hyper s1 s3 osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper; crush;
    apply osi_non_stealth_equivalent_hyper_mapping_trans with s2;
    crush.
Qed.

Lemma osi_non_stealth_equivalent_pt_pages_trans :
  forall (s1 s2 s3:state) (osi:os_ident),
    osi_non_stealth_equivalent_pt_pages s1 s2 osi -> 
    osi_non_stealth_equivalent_pt_pages s2 s3 osi -> 
    osi_non_stealth_equivalent_pt_pages s1 s3 osi.
Proof.
  unfold osi_non_stealth_equivalent_pt_pages; crush.
  destruct H with pt_padd pt_page1 mp1 va ma1; crush.
  destruct H0 with pt_padd x x0 va x1; crush.
  exists x3; crush.
  destruct H9 with pg1; crush.
  destruct H14 with x5; crush.
  apply same_page_but_value_trans with x5; crush.
Qed.  

Lemma osi_non_stealth_equivalent_smem_trans :
  forall (s1 s2 s3:state) (osi:os_ident),
    osi_non_stealth_equivalent_smem s1 s2 osi ->
    osi_non_stealth_equivalent_smem s2 s3 osi ->
    osi_non_stealth_equivalent_smem s1 s3 osi.
Proof.
  unfold osi_non_stealth_equivalent_smem; crush;
    apply osi_non_stealth_equivalent_pt_pages_trans with s2;
    crush.
Qed.

Lemma same_current_pt_page_trans :
  forall (s1 s2 s3:state) (osi:os_ident),
    same_current_pt_page s1 s2 osi ->
    same_current_pt_page s2 s3 osi ->
    same_current_pt_page s1 s3 osi.
Proof.
  intros; unfold same_current_pt_page in *; crush.
  unfold same_current_pt_page_one; crush.
  rename x into os1.
  destruct H with os1; crush.
  rename x into os2.
  destruct H0 with os2; crush.
  unfold same_current_pt_page_one; crush.
  rename x into os3.
  destruct H1 with os3; crush.
  rename x into os2.
  destruct H2 with os2; crush.
Qed.

Lemma osi_non_stealth_equivalent_trans :
  forall (s1 s2 s3:state),
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s2 s3 ->
    osi_non_stealth_equivalent s1 s3.
Proof.
  intros; unfold osi_non_stealth_equivalent in *; crush;
    try specialize (H osi); try specialize (H0 osi); crush;
    try apply osi_non_stealth_equivalent_hyper_trans with s2; crush;
    try apply osi_non_stealth_equivalent_smem_trans with s2; crush;
    try apply same_current_pt_page_trans with s2; crush.
Qed.  

Lemma shared_page_non_stealth_eq:
  forall (s1 s2 : state) (osi : os_ident) (va : vadd) 
    (pa : padd) (ma1 ma2 : madd),
    valid_sm_state s1 ->
    ~ in_stealth_va_line va ->
    non_stealth_ma s1 osi ma1 ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_pa_to_ma s1 osi pa = Some ma1 ->
    shared_page s1 va ma1 ->
    osi_pa_to_ma s2 osi pa = Some ma2 ->
    shared_page s2 va ma2.
Proof.
  unfold shared_page; crush.
  rename x into pt_padd.
  rename x0 into osi'.
  rename x1 into pt_page1.
  rename x2 into mp1.
  rename x3 into va1.

  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi).
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  unfold valid_hypervisor in H_valid_hyper.
  unfold osi_pa_to_ma in H3; crush.

  assert (osi' = osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi'); crush.
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  unfold valid_virtual_mapping in H_valid_virtmap.
  unfold osi_pa_to_pg in H4; crush.
  rename i0 into pt_addr.
  destruct H_valid_virtmap with osi' pt_addr pt_page1 mp1 va1 ma1; crush.
  unfold valid_hypervisor in H_valid_hyper.
  rewrite <- H_valid_hyper with osi' i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold madd_owner; crush.
  destruct (ctxt_vadd_accessible ctxt va1) eqn:?; unfold os_accessible in *; auto.
  unfold madd_owner in H9.
  rewrite <- Heqval in H9.
  rewrite H18 in H9; congruence.

  exists pt_padd.
  exists osi.
  destruct H2 with osi; crush.
  destruct H12.
  destruct H12 with pt_padd pt_page1 mp1 va1 ma1; crush.
  unfold non_stealth_ma in *; crush.
  unfold osi_pa_to_pg in H4; crush.  
  destruct H1 with pt_padd i0; crush.
  unfold osi_pa_to_ma; crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state.
  generalize (H_valid_hyper osi i pt_padd i0); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H6; crush.
  unfold madd_owner in H9; crush.
  exists x. exists x0. exists va1; crush.
  rewrite H17 in H5; crush.
Qed.

(* Lemma no_va_mapped_non_stealth_eq : *)
(*   forall s1 s2 pa ma1 ma2 osi, *)
(*     osi_non_stealth_equivalent_pt_pages s2 s1 osi -> *)
(*     osi_pa_to_ma s1 osi pa = Some ma1 -> *)
(*     osi_pa_to_ma s2 osi pa = Some ma2 -> *)
(*     non_stealth_ma s2 osi ma2 -> *)
(*     ACT.no_va_mapped_to_ma osi ma1 s1 -> *)
(*     ACT.no_va_mapped_to_ma osi ma2 s2. *)
(* Proof. *)
(*   unfold ACT.no_va_mapped_to_ma; crush. *)
(*   unfold osi_pa_to_ma, osi_pa_to_pg in *; crush. *)
(*   destruct H with pt_padd pt_page mp va ma2; crush. *)
(*   unfold osi_pa_to_pg in *; crush. *)
(*   destruct H2 with pt_padd i0; crush. *)
(*   unfold osi_pa_to_ma; crush. *)
(*   unfold madd_of_vadd, madd_page_table; crush. *)
(*   rewrite H5; crush. *)
(*   destruct H3 with pt_padd x va x0; crush. *)
(*   unfold osi_pa_to_pg, osi_pa_to_ma in *; crush. *)
(*   specialize H10 with pa; crush. *)
(*   unfold osi_pa_to_pg, osi_pa_to_ma in *; crush. *)
(* Qed. *)


(********************************************)
(* Invariance Proof                         *)
(********************************************)
Lemma read_stealth_os_non_stealth_isolation :
  forall (s s':state) (va:vadd),
    valid_state s ->
    ACT.pre_all_read va s ->
    ACT.post_all_read va s s' ->
    osi_non_stealth_equivalent s s'.
Proof.
  unfold osi_non_stealth_equivalent,
    osi_non_stealth_equivalent_hyper,
    osi_non_stealth_equivalent_smem,
    same_current_pt_page, same_current_pt_page_one; crush;
  generalize H0 H1; unfold ACT.pre_all_read, ACT.post_all_read; crush;
    STAC.unify_read_post; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (st_memory s)) (va,m) x) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  destruct (madd_eq m0 ma1); crush.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma1 x1; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  exists ma1; exists x1; crush.
  apply get_cache_entry1 with (va':=v) (ma:=m) (pg:=x); crush.
  unfold same_page_but_value; rewrite H10, H12, H13;
    repeat split; auto.
  unfold same_page_type; rewrite H11, H7; auto.
  exists ma1; exists pg1; crush.
  rewrite get_cache_entry2 with (va':=v) (ma':=m0) (pg':=p) (ma:=m) (pg:=x);
    crush.
  rewrite Heqp; crush.
  exists ma1; exists pg1; crush.
  erewrite get_mem_unchanged with (ma:=m) (pg:=x); crush.
  rewrite Heqp; crush.
  
  unfold osi_non_stealth_equivalent_hyper_mapping;
    crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (st_memory s)) (va,m) x) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  destruct (madd_eq m0 ma1); crush.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma1 x1; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  exists ma1; exists x0; crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=m)
    (pg:=x) (pg':=x1) in H6; crush.
  unfold same_page_but_value; rewrite H13, H10, H14;
    repeat split; auto.
  unfold same_page_type; rewrite H12, H7; auto.
  exists ma1; exists pg1; crush.
  rewrite get_cache_entry2 with (va':=v) (ma':=m0) (pg':=p)
    (ma:=m) (pg:=x) in H6; crush.
  rewrite Heqp; crush.
  exists ma1; exists pg1; crush.
  erewrite get_mem_unchanged with (ma:=m) (pg:=x) in H6; crush.
  rewrite Heqp; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H2; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p; crush.
  destruct o; crush.
  destruct p. destruct c0.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p); crush.
  STAC.apply_absurd_cache_get_entry_pt.
  rewrite get_mem_unchanged with (ma:=m) (pg:=x); crush.
  exists mp1; crush.
  exists pg1; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_smem in *; crush.
  rewrite <- STAC.read_same_mem with (ma':=m) (pt_addr:=i0) (pg:=x) (val:=x0); crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush; 
    unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page i1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H6; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H2; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p; crush.
  destruct o; crush.
  destruct p. destruct c0.
  assert (m0 <> i0); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=m) (ma':=i0) (pg:=x) (pg':=p)
    in H3; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v i0 pt_page1; crush.
  symmetry; apply STAC.cache_deleted_entry_was_there with va m x; crush.
  unfold is_rw_page in *; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p)
    in H3; crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=x) in H3; crush.
  exists mp1; crush.
  exists pg1; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_smem in *; crush.
  rewrite STAC.read_same_mem with (va':=va) (pt_addr':=pt_addr0)
    (ma':=m) (pt_addr:=i0) (pg:=x) (val:=x0); crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush; 
    unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page i1); crush.

  rewrite STAC.read_same_madd_of_vadd with (va':=va) (pt_addr':=pt_addr0);
    crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H6; crush.
Qed.

Lemma write_stealth_os_non_stealth_isolation :
  forall (s s':state) (va:vadd) (val:value),
    valid_sm_state s ->
    ACT.pre_all_write va val s ->
    ACT.post_all_write va val s s' ->
    osi_non_stealth_equivalent s s'.
Proof.
  unfold osi_non_stealth_equivalent,
    osi_non_stealth_equivalent_hyper,
    osi_non_stealth_equivalent_smem,
    same_current_pt_page, same_current_pt_page_one,
    valid_sm_state; crush;
  generalize H0 H1; unfold ACT.pre_all_write, ACT.post_all_write; crush;
    STAC.unify_write_post; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (st_memory s)) (va,m)
    (Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)))
    eqn:?; crush.
  destruct o. destruct p. destruct c0.
  destruct (madd_eq m0 ma1); crush.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma1 x1; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  exists ma1; exists x1; crush.
  rewrite set_cache_entry1 with (va':=v) (ma:=m) (pg':=x1)
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)); crush.
  unfold same_page_but_value; rewrite H16, H18, H19;
    repeat split; auto.
  unfold same_page_type; rewrite H17, H13; auto.
  destruct (madd_eq ma1 m); crush.
  destruct set_new_mmem with (st_memory s)
    pt_addr0 va m val x; crush.
  unfold page_of_vadd_smem; crush.
  exists m; exists x0; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H15, H16, H19;
    repeat split; auto.
  unfold same_page_type in H14; rewrite H8 in H14; simpl in *.
  unfold same_page_type; rewrite H18; auto.
  unfold same_page_but_value; repeat split; auto.
  unfold is_rw_page in H10; destruct H10.
  unfold same_page_type in *; rewrite H10 in *; auto.
  exists ma1; exists pg1; crush.
  rewrite set_mem_unchanged1 with (ma:=m) (va':=v) (ma':=m0) (pg':=p)
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)); crush.
  rewrite Heqp; crush.
  destruct (madd_eq ma1 m); crush.
  destruct set_new_mmem with (st_memory s)
    pt_addr0 va m val x; crush.
  unfold page_of_vadd_smem; crush.
  exists m; exists x0; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H15, H16, H19;
    repeat split; auto.
  unfold same_page_type in H14; rewrite H8 in H14; simpl in *.
  unfold same_page_type; rewrite H18; auto.
  unfold same_page_but_value; rewrite H15, H16;
    repeat split; auto.
  unfold is_rw_page in H10; destruct H10.
  unfold same_page_type in *; rewrite H10 in *; auto.
  exists ma1; exists pg1; crush.
  rewrite set_mem_unchanged2 with (ma:=m)
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)); crush.
  rewrite Heqp; crush.
  
  unfold osi_non_stealth_equivalent_hyper_mapping;
    crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (st_memory s)) (va,m)
    (Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)))
    eqn:?; crush.
  destruct o. destruct p. destruct c0.
  destruct (madd_eq m0 ma1); crush.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma1 x1; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  exists ma1; exists x0; crush.
  rewrite set_cache_entry1 with (va':=v) (ma:=m) (pg':=x1)
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x))
    in H12; crush.
  unfold same_page_but_value; rewrite H16, H20;
    repeat split; auto.
  unfold same_page_type; rewrite H18, H13; auto.
  destruct (madd_eq ma1 m); crush.
  destruct set_new_mmem with (st_memory s)
    pt_addr0 va m val x; crush.
  unfold page_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  exists m; exists x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H14, H15, H16, H19;
    repeat split; auto.
  unfold same_page_type in H13; rewrite H8 in H13; simpl in *.
  unfold same_page_type; rewrite H18; auto.
  exists m; exists x; crush.
  unfold same_page_but_value; repeat split; auto.
  exists ma1; exists pg1; crush.
  rewrite set_mem_unchanged1 with (ma:=m) (va':=v) (ma':=m0) (pg':=p)
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)) in H12;
    crush.
  rewrite Heqp; crush.
  destruct (madd_eq ma1 m); crush.
  destruct set_new_mmem with (st_memory s)
    pt_addr0 va m val x; crush.
  unfold page_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  exists m; exists x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H14, H15, H16, H19;
    repeat split; auto.
  unfold same_page_type in H13; rewrite H8 in H13; simpl in *.
  unfold same_page_type; rewrite H18; auto.
  exists m; exists x; crush.
  unfold same_page_but_value; repeat split; auto.
  exists ma1; exists pg1; crush.
  erewrite set_mem_unchanged2 with
    (pg:=Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)) in H12;
    crush.
  rewrite Heqp; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H8; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p; crush.
  destruct o; crush.
  destruct p. destruct c0.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p); crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_madd_smem in H8; crush.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page in *; crush.
  unfold is_rw_page in *; crush.
  STAC.simpl_some_cache_entry; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m0 x1; crush.
  unfold is_rw_page in *; crush.
  rewrite set_mem_unchanged2 with (ma:=m) (pg:=x); crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_madd_smem in H8; crush.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page in *; crush.
  unfold is_rw_page in *; crush.
  exists mp1; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (madd_eq m ma1); crush.
  exists (Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)); crush.
  rewrite STAC.write_smem_mem_eq with (pt_addr':=i0) (old_pg:=pg1); crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct STAC.no_cache_alias with s va va0 ma1 x pg1; crush.
  destruct H_valid_cache with va ma1 x; crush.
  destruct H_valid_cache with va0 ma1 pg1; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H12; crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_madd_smem in H15; crush.
  destruct H_valid_cache with va0 ma1 pg1; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va ma1 x; crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct STAC.no_cache_alias with s va va0 ma1 x pg1; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold is_rw_page in H10; destruct H10.
  unfold same_page_type; rewrite H8; auto.
  destruct H_valid_cache with va ma1 x; crush.
  rewrite H19.
  unfold is_cacheable, is_rw_page in *; crush.
  unfold same_page_but_value; rewrite H18, H16; repeat split; auto.
  unfold same_page_type; rewrite H17; auto.
  destruct H_valid_cache with va0 ma1 pg1; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  unfold same_page_but_value; rewrite H18, H16; repeat split; auto.
  unfold same_page_type; rewrite H15; auto.
  unfold same_page_but_value; repeat split; auto.
  unfold is_rw_page in *; destruct H10.
  unfold same_page_type; rewrite H8; auto.
  rewrite STAC.write_smem_mem_neq with (ma':=m) (pg:=x) (pt_addr:=i0); crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H12; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_type (smem_mm (set (st_memory s) va pt_addr0 val)) i0 = Some PT).
  unfold madd_type; crush.
  rewrite <- H12; crush.
  assert (madd_type (smem_mm (st_memory s)) i0 = Some PT).
  erewrite <- STAC.write_same_madd_type; crush.
  unfold madd_type in H16; crush.
  destruct (page_content p) eqn:?; crush.
  unfold is_rw_page in *; crush.
  
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H8; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p1; crush.
  destruct o; crush.
  destruct p1. destruct c0.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p1)
    in H9; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page in *; crush.
  STAC.simpl_some_cache_entry; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m0 x2; crush.
  unfold is_rw_page in *; crush.
  rewrite set_mem_unchanged2 with (ma:=m) (pg:=x) in H9; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va m x; crush.
  unfold is_rw_page in *; crush.

  exists mp1; crush.
  STAC.destruct_valid_state.
  unfold page_of_vadd_smem in *; crush.
  destruct (madd_eq m ma1); crush.
  assert (exists old_pg, page_of_vadd_madd_smem
    (st_memory s) va0 ma1 = Some old_pg); crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  destruct H_valid_cache with va ma1 x; crush.
  rename x1 into old_pg.
  assert (is_rw_page old_pg); crush.
  unfold page_of_vadd_madd_smem in H17; crush.
  destruct H_valid_cache with va0 ma1 old_pg; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  destruct H_valid_cache with va ma1 x; crush.
  unfold is_rw_page; crush.
  unfold is_rw_page in *; crush.
  rewrite STAC.write_smem_mem_eq with (pt_addr':=i0) (old_pg:=old_pg) in H16; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; simpl; rewrite H18; auto.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  erewrite STAC.set_mem_old_pt_page with (va:=va)
    (pt_addr:=pt_addr0) (val:=val) in Heqopt_res1; crush.
  crush.
  rewrite <- H12 in Heqc; injection Heqc; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold is_rw_page; crush.
  unfold is_pt_page; crush.
  unfold is_rw_page; crush.

  erewrite STAC.write_smem_mem_neq with (pt_addr:=i0) in H16; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  erewrite STAC.set_mem_old_pt_page with (va:=va)
    (pt_addr:=pt_addr0) (val:=val) in Heqopt_res1; crush.
  crush.
  rewrite <- H12 in Heqc; injection Heqc; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold is_rw_page; crush.
  unfold is_pt_page; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold is_rw_page; crush.
Qed.  

Lemma new_stealth_os_non_stealth_isolation :
  forall (s s':state) (osi:os_ident) (pa:padd),
    valid_state s ->
    pre_new_sm osi pa s ->
    ACT.post_new osi stealth_va pa s s' ->
    osi_non_stealth_equivalent s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold osi_non_stealth_equivalent,
    same_current_pt_page, same_current_pt_page_one,
    osi_non_stealth_equivalent_smem, osi_non_stealth_equivalent_hyper; crush;
    unfold pre_new_sm, pre_all_new, ACT.post_new, ACT.post_all_new in *; crush.
  destruct (os_ident_eq osi osi0); crush.
  exists (OS None (os_current_page x)); rewrite <- OSS.get_set; crush.
  exists x; rewrite OSS.get_set_distinct; crush.
  destruct (os_ident_eq osi osi0); crush.
  rewrite <- OSS.get_set in *; crush.
  rewrite OSS.get_set_distinct in *; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping,
    osi_pa_to_ma in *; crush.
  destruct (madd_eq x2 ma1); crush.
  (* x2 = ma1 *)
  STAC.destruct_valid_state.
  STAC.get_osi_current_pt_page osi.
  unfold osi_current_page_table in *; crush.
  STAC.destruct_valid_state.
  exists ma1; exists (Page (PT_c (PT.set pt stealth_va x0))
    (page_owned_by pg1) (page_cacheable pg1)); crush.
  unfold osi_current_page_table in *; crush.
  apply STAC.new_get_mem_not_shared_pt_addr; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; simpl; auto.
  rewrite H17; auto.
  (* x2 <> ma1 *)
  exists ma1; exists pg1; crush.
  rewrite STAC.new_get_mem_not_shared_distinct; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping, osi_pa_to_ma in *;
    crush.
  destruct (madd_eq ma1 x2); crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  rewrite STAC.new_get_mem_not_shared_pt_addr
    with (pt_page:=pt_page) (pt:=pt) in H13; crush.
  exists x2; exists pt_page; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H18; auto.
  rewrite STAC.new_get_mem_not_shared_distinct in H13; crush.

  unfold osi_non_stealth_equivalent_pt_pages, osi_pa_to_pg; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq i0 x2); crush.
  (* i0 = x2 *)
  exists (Page (PT_c (PT.set mp1 stealth_va x0))
    (page_owned_by pt_page1) (page_cacheable pt_page1)); crush.
  apply STAC.new_get_mem_not_shared_pt_addr; crush.
  exists (PT.set mp1 stealth_va x0); crush.
  exists ma1; crush.
  rewrite PT.get_set_distinct; crush.
  destruct (madd_eq ma1 x2); crush.
  assert (pg1 = pt_page1); crush.
  unfold page_of_vadd_madd_smem in H17; crush.
  destruct H_valid_cache with va x2 pg1; crush.
  unfold is_rw_page in *; crush.
  rewrite STAC.new_page_pt_addr with (pt_page:=pt_page1) (pt:=pt); crush.
  eexists; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H20; auto.
  exists pg1; crush.
  rewrite STAC.new_page_ne; crush.
  apply n; unfold shared_page; crush.
  exists pt_padd; exists osi0;
    exists pt_page1; exists mp1; exists va; crush.
  unfold osi_pa_to_pg; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some x2); unfold osi_pa_to_ma in *; crush.
  unfold madd_type; crush.
  rewrite H20; crush.
  exists osi0; exists pt_padd; exists x2; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H13; crush.

  (* i0 <> x2 *)
  exists pt_page1; crush.
  rewrite STAC.new_get_mem_not_shared_distinct; crush.
  exists mp1; crush.
  destruct (madd_eq ma1 x2); crush.
  assert (pg1 = pt_page); crush.
  unfold page_of_vadd_madd_smem in H17; crush.
  destruct H_valid_cache with va x2 pg1; crush.
  unfold is_rw_page in *; crush.
  rewrite STAC.new_page_pt_addr with (pt_page:=pt_page) (pt:=pt); crush.
  eexists; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H20; auto.
  exists pg1; crush.
  rewrite STAC.new_page_ne; crush.
  apply n; unfold shared_page; crush.
  exists pt_padd; exists osi0; exists pt_page1; exists mp1; exists va; crush.
  unfold osi_pa_to_pg; crush.
  assert (osi_pa_to_ma s osi (os_current_page x1) = Some x2); 
    unfold osi_pa_to_ma in *; crush.
  unfold madd_type; crush.
  rewrite H20; crush.
  exists osi0; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H13; crush.
  
  unfold osi_non_stealth_equivalent_pt_pages, osi_pa_to_ma,
    osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  rewrite H20 in *; crush.
  assert (madd_of_vadd (smem_mm (new_not_shared (st_memory s) x2 stealth_va x0))
    i0 va = Some ma1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  
  rewrite <- H13; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) i0 va = Some ma1).
  rewrite STAC.madd_of_vadd_new_not_shared_distinct with (osi:=osi) in H17; crush.
  unfold osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H19; crush.
  destruct (page_content p0) eqn:?; crush.
  exists p0; crush.
  exists p1; crush.
  destruct (madd_eq ma1 x2); crush.
  exists pt_page; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va x2 p; crush.
  unfold is_rw_page in *; crush.
  rewrite STAC.new_page_pt_addr with (pt_page:=pt_page) (pt:=pt) in H22; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H20; auto.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va x2 p; crush.
  unfold is_rw_page in *; crush.
  exists pg1; crush.
  rewrite STAC.new_page_ne in H22; crush.
  apply n; unfold shared_page; crush.
  exists pt_padd; exists osi0; exists p0; exists p1; exists va; crush.
  unfold osi_pa_to_pg; crush.
  assert (osi_pa_to_ma s osi (os_current_page x1) = Some x2); 
    unfold osi_pa_to_ma in *; crush.
  unfold madd_type; crush.
  rewrite H20; crush.
  exists osi0; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
Qed.  

(* Del lemmas *)
Lemma del_non_stealth_isolation_hyper :
  forall (s s':state) (osi osi0:os_ident) (va:vadd),
    valid_sm_state s ->
    ACT.pre_del osi va s ->
    ACT.post_del osi va s s' ->
    osi_non_stealth_equivalent_hyper s s' osi0.
Proof.
  unfold valid_sm_state,
    osi_non_stealth_equivalent_hyper; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping in *; crush.
  generalize H0.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  destruct (madd_eq x ma1); crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in H14,H15; crush.
  destruct (page_content pg1) eqn:?; crush.
  exists ma1; exists
    (Page (PT_c (PT.remove p0 va))
          (page_owned_by pg1)
          (page_cacheable pg1)); crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x2 va) eqn:?;
    crush.
  rename m into mm'.
  assert (get_mem mm' ma1 = Some (Page (PT_c (PT.remove p0 va))
    (page_owned_by pg1) (page_cacheable pg1))).
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem; crush.

  assert (madd_type (smem_mm (delete (st_memory s) x2 va)) ma1 =
    madd_type (smem_mm (st_memory s)) ma1).
  rewrite STAC.del_hyper_same_madd_type; crush.
  unfold madd_type in *; crush.
  rewrite Heqc in *; crush.
  destruct (page_content p) eqn:?; crush.
  destruct (get_cache (smem_c (st_memory s)) (va, x0)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x0) (pg:=p2) in Heqopt_res;
    crush.
  rewrite get_add_mem_distinct in Heqopt_res; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 p2; unfold is_rw_page in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x0) in Heqopt_res; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.

  unfold same_page_but_value; repeat split; simpl; auto.
  unfold same_page_type; rewrite Heqc; auto.
  destruct STAC.del_get_mem_same_page_but_value with s s' ma1 pg1 osi
     va; crush.
  exists ma1; exists x1; crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.

  unfold osi_non_stealth_equivalent_hyper_mapping in *; crush.
  generalize H0.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in H14,H15; crush.
  destruct (page_content p0) eqn:?; crush.
  erewrite <- STAC.del_osi_pa_to_ma in H8; crush.
  unfold osi_pa_to_ma in H8; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi0).
  STAC.destruct_valid_state; crush.
  unfold madd_owner in *; crush.
  destruct STAC.del_get_mem_same_page_but_value
    with s s' ma1 p osi va; crush.
  unfold osi_current_page_table; crush.
  exists ma1; exists p; crush.
  unfold osi_pa_to_ma; crush.
  unfold same_page_but_value in *; destruct H18; destruct H18;
    repeat split; auto.
  unfold same_page_type in *; destruct (page_content p) eqn:?;
    simpl in *; destruct (page_content pg1) eqn:?; simpl in *; auto;
    try destruct o; try destruct o0; simpl in *; auto.
Qed.

Lemma del_stealth_os_non_stealth_isolation :
  forall (s s':state) (osi:os_ident),
    valid_sm_state s ->
    ACT.pre_del osi stealth_va s ->
    ACT.post_del osi stealth_va s s' ->
    osi_non_stealth_equivalent s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold osi_non_stealth_equivalent,
    same_current_pt_page, same_current_pt_page_one,
    osi_non_stealth_equivalent_smem; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  destruct (os_ident_eq osi osi0); crush.
  exists (OS None (os_current_page x)); rewrite <- OSS.get_set; crush.
  exists x; rewrite OSS.get_set_distinct; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  destruct (os_ident_eq osi osi0); crush.
  rewrite <- OSS.get_set in *; crush.
  rewrite OSS.get_set_distinct in *; crush.

  apply del_non_stealth_isolation_hyper with osi stealth_va;
    unfold valid_sm_state; crush.
  
  unfold osi_non_stealth_equivalent_pt_pages in *; crush.
  unfold osi_pa_to_pg in H8; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some i0); crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  generalize H0.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  destruct (madd_eq i0 x); crush.
  (* i0 = x *)
  exists (Page (PT_c (PT.remove mp1 stealth_va))
          (page_owned_by pt_page1)
          (page_cacheable pt_page1)); crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x stealth_va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' x = Some (Page (PT_c (PT.remove mp1 stealth_va))
    (page_owned_by pt_page1) (page_cacheable pt_page1))).
  unfold remove_page_table_entry in *; crush.
  rewrite <- H9 in *; crush.
  rewrite <- MM.get_add_mem; crush.
  destruct (get_cache (smem_c (st_memory s)) (stealth_va, x3)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x3) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with stealth_va x p; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x3); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite <- H9 in *; crush.
  exists (PT.remove mp1 stealth_va); crush.
  exists ma1; crush.
  rewrite PT.get_remove_distinct with (inf:=ma1); crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  destruct (madd_eq x ma1); crush.
  assert (pg1 = pt_page1); crush.
  unfold page_of_vadd_madd_smem in H20; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 pg1; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page with (s:=s) (pt:=mp1); crush.
  eexists; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite <- H9; auto.
  exists pg1; crush.
  erewrite <- STAC.del_page_of_vadd_madd_smem with
    (pt_addr:=x); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.
  (* i0 <> x *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 stealth_va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some pt_page1).
  unfold remove_page_table_entry in *; crush.
  unfold madd_of_vadd, madd_page_table in H19; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct; crush.
  unfold ACT.pre_del, ACT.pre_all_del in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (stealth_va, x3)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x3) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with stealth_va i0 p; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x3); crush.
  unfold remove_page_table_entry in *; crush.
  unfold madd_of_vadd, madd_page_table in H19; crush.
  destruct (page_content p0) eqn:?; crush.
  exists mp1; crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  destruct (madd_eq x ma1); crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (pg1 = pt_page); crush.
  unfold page_of_vadd_madd_smem in H20; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 pg1; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page
    with (s:=s) (pt_page:=pt_page) (pt:=pt); crush.
  eexists; crush.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H26; auto.
  unfold osi_current_page_table; crush.
  exists pg1; crush.
  erewrite <- STAC.del_page_of_vadd_madd_smem with (pt_addr:=i0); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.

  unfold osi_non_stealth_equivalent_pt_pages in *; crush.
  unfold osi_pa_to_pg in H8; crush.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some i0); crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  generalize H0.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  destruct (madd_eq i0 x); crush.
  (* i0 = x *)
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  exists p0; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in *; crush.
  exists p1; crush; exists ma1; crush.
  unfold ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 stealth_va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' x = Some pt_page1); crush.
  destruct (get_cache (smem_c (st_memory s)) (stealth_va, x0)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x0) (pg:=p) in H8; crush.
  rewrite get_add_mem_distinct in H8; crush. 
  STAC.destruct_valid_state.
  destruct H_valid_cache with stealth_va x p; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x0) in H8; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite <- MM.get_add_mem in *; crush.
  injection H9; intros; subst mp1; clear H9.
  apply PT.get_remove_old with (k':=stealth_va); crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite Heqc in *; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  destruct (madd_eq x ma1); crush.
  exists p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 p; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page in H20; crush.
  subst pg1.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite Heqc; auto.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 p; crush.
  unfold is_rw_page in *; crush.
  exists pg1; crush.
  erewrite <- STAC.del_page_of_vadd_madd_smem
    with (pt_addr:=x) in H20; crush.
  unfold osi_current_page_table; crush. 
  assert (madd_of_vadd (smem_mm (st_memory s')) x va = Some ma1).
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H9; crush.
  eapply STAC.del_madd_of_vadd_old; crush.
  (* i0 <> x *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 stealth_va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some pt_page1).
  unfold ACT.pre_del, ACT.pre_all_del in *; crush.
  assert (madd_type (smem_mm (delete (st_memory s) x stealth_va)) i0 =
    madd_type (smem_mm (st_memory s)) i0).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type in *; crush.
  rewrite <- H9 in *; crush.
  destruct (page_content p) eqn:?; crush.
  destruct (get_cache (smem_c (st_memory s)) (stealth_va, x3)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x3) (pg:=p1) in H8; crush.
  rewrite get_add_mem_distinct in H8; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with stealth_va i0 p1; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x3) in H8; crush.
  
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H19; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct in *; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H19; crush.
  destruct (page_content p0) eqn:?; crush.
  exists mp1; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq x ma1); crush.
  exists pt_page; crush.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 p; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page in H20; crush.
  subst pg1.
  unfold same_page_but_value; repeat split; auto.
  unfold same_page_type; rewrite H26; auto.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va ma1 p; crush.
  unfold is_rw_page in *; crush.
  exists pg1; crush.
  erewrite <- STAC.del_page_of_vadd_madd_smem in H20; crush.
  unfold osi_current_page_table; crush.
  assert (madd_of_vadd (smem_mm (st_memory s')) i0 va = Some ma1).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.
  eapply STAC.del_madd_of_vadd_old; crush.
Qed.  



Opaque new_page.
Lemma pin_os_non_stealth_isolation_smem :
  forall (s s':state) (osi osi0:os_ident) (pa:padd) (type:ptype),
    valid_sm_state s ->
    ACT.pre_pin osi pa type s ->
    ACT.post_pin osi pa type s s' ->
    osi_non_stealth_equivalent_smem s s' osi0.
Proof.
  unfold valid_sm_state, osi_non_stealth_equivalent_smem; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in H8; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s osi0 pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some pt_addr); crush.
  eapply STAC.page_pin_osi_pa_to_ma; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (madd_eq x pt_addr); crush;
    rewrite MM.get_add_mem_distinct; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (os_ident_eq osi0 osi); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa0 pa); crush;
    [rewrite <- P2M.get_set in * | rewrite P2M.get_set_distinct in *]; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  exists pg1; crush.
  eapply STAC.page_pin_page_of_vadd_madd_smem with (pt_addr:=pt_addr); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in H8; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some pt_addr); crush.
  destruct (os_ident_eq osi0 osi); crush.
  eapply STAC.page_pin_osi_pa_to_ma_distinct_pa; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  unfold osi_pa_to_ma in H10; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite <- P2M.get_set in *.
  injection Heqopt_res0; crush.
  rewrite <- MM.get_add_mem in *; crush.
  Transparent new_page.
  unfold new_page in H9; crush.
  Opaque new_page.
  injection H9; intros; subst mp1.
  rewrite PT.get_empty in *; crush.
  eapply STAC.page_pin_osi_pa_to_ma_distinct_osi; crush.
  assert (get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page1); crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (madd_eq x pt_addr); crush;
    [rewrite <- MM.get_add_mem in * | rewrite MM.get_add_mem_distinct in *]; crush.
  Transparent new_page.
  unfold new_page in H9; crush.
  Opaque new_page.
  injection H9; intros; subst mp1.
  rewrite PT.get_empty in *; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg; unfold osi_pa_to_ma in H13; crush.
  exists mp1; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (os_ident_eq osi0 osi); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa0 pa); crush;
    [rewrite <- P2M.get_set in * | rewrite P2M.get_set_distinct in *]; crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in H13; crush.
  destruct H_valid_virtmap with osi pt_addr pt_page1 mp1 va ma1; crush.
  unfold valid_hypervisor in H_valid_hyper.
  rewrite <- H_valid_hyper with osi x0 pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H15 in *; crush.
  rewrite H28 in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma1).
  eapply STAC.page_pin_madd_of_vadd_old; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some pt_addr).
  unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.
  exists pg1; crush.
  eapply STAC.page_pin_page_of_vadd_madd_smem_old with (pt_addr:=pt_addr); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.
Qed.

Lemma unpin_os_non_stealth_isolation_smem :
  forall (s s':state) (osi osi0:os_ident) (pa:padd),
    valid_sm_state s ->
    ACT.pre_unpin osi pa s ->
    ACT.post_unpin osi pa s s' ->
    osi_non_stealth_equivalent_smem s s' osi0.
Proof.
  unfold valid_sm_state, osi_non_stealth_equivalent_smem; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in H8; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s osi0 pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  generalize H0; unfold ACT.pre_unpin; crush.
  assert (pt_addr <> x0); crush.
  rewrite H20 with pt_page1 mp1 in H11; crush.
  rewrite PT.get_empty in *; crush.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some pt_addr); crush.
  destruct (os_ident_eq osi0 osi); crush.
  apply STAC.page_unpin_osi_pa_to_ma_distinct_pa
    with (s:=s) (osi':=osi) (pa':=pa); crush.
  apply STAC.page_unpin_osi_pa_to_ma_distinct_osi
    with (s:=s) (osi':=osi) (pa':=pa); crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.post_unpin in *; crush.
  rewrite get_add_mem_distinct; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.post_unpin in *; crush.
  destruct (os_ident_eq osi0 osi); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa0 pa); crush.
  destruct H21 with pt_padd pt_page1 va mp1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite P2M.get_remove_distinct with (inf:=ma1); crush.
  exists pg1; crush.
  eapply STAC.page_unpin_page_of_vadd_madd_smem with (pt_addr:=pt_addr); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.

  unfold osi_non_stealth_equivalent_pt_pages; crush.
  unfold osi_pa_to_pg in H8; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s' osi0 pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi0 pt_padd = Some pt_addr); crush.
  eapply STAC.page_unpin_osi_pa_to_ma_old; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  destruct (madd_eq x0 pt_addr); crush.
  rewrite <- get_add_mem in *; crush.
  Transparent new_page.
  crush.
  Opaque new_page.
  rewrite MM.get_add_mem_distinct in *; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  destruct (os_ident_eq osi0 osi); crush;
    [rewrite <- HYP.get_set in * | rewrite HYP.get_set_distinct in *]; crush.
  destruct (padd_eq pa0 pa); crush.
  rewrite P2M.get_remove in *; crush. 
  rewrite P2M.get_remove_old with (k':=pa) (inf:=ma1); crush.
  exists pg1; crush.
  eapply STAC.page_unpin_page_of_vadd_madd_smem_old with (pt_addr:=pt_addr);
    crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.
Qed.  

Lemma ctxt_switch_non_stealth_equivalent_hyper_mapping_1 :
  forall (s s':state) (osi osi':os_ident) (pa:padd) (pt_addr:madd),
    valid_sm_state s ->
    osi_pa_to_ma s osi' pa = Some pt_addr ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    osi_non_stealth_equivalent_hyper_mapping s s' osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper_mapping, valid_sm_state; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite H1,H2 in *; crush.
  clear dependent s'.
  unfold ctxt_switch_new_smem; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?.
  destruct (page_of_vadd_madd_smem (st_memory s) stealth_va m0) eqn:?.
  (* The new osi has a stealth page *)
  destruct (add_cache (smem_c (remove_osi_stealth_page (ctxt_switch (st_memory s)) m))
    (stealth_va,m0) p) eqn:?.
  (* The read stealth removes some entry *)
  destruct o. destruct p0. destruct c0.
  destruct (madd_eq m1 m0); crush.
  (* The removed entry is the new stealth page *)
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  unfold remove_osi_stealth_page in H2; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_cache in *; crush.
  destruct (madd_eq m0 i1); crush;
    [rewrite get_remove_ma_cache_eq in * |
     rewrite get_remove_ma_cache_ne in *]; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  destruct (madd_eq m0 ma1); crush.
  (* ma1 is the stealth page *)
  exists ma1; exists x0; crush.
  rewrite get_cache_entry1 with (ma:=ma1) (pg:=p)
    (va':=stealth_va) (pg':=x0); crush.
  destruct H_valid_cache with stealth_va ma1 x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H12, H16, H17;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H13; auto.
  destruct (madd_of_vadd (smem_mm (st_memory s)) m stealth_va) eqn:?.
  destruct (madd_eq m1 ma1); crush.
  destruct H5 with m ma1 pg1; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  exists ma1;
    exists (Page (RW_c (Some x)) (page_owned_by pg1) true); crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0); crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem; crush.
  destruct (vadd_eq v stealth_va); crush.
  destruct STAC.no_cache_alias with s v stealth_va m1 p4
    (Page (RW_c (Some x)) (page_owned_by pg1) true); crush.
  rewrite get_remove_ma_cache_no_entry in H13; crush.
  rewrite ctxt_switch_mem in *; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  unfold same_page_but_value; rewrite H1; repeat split; auto.
  unfold same_page_type; rewrite H12; auto.
  (* ma1 is not the old stealth page *)
  exists ma1; exists pg1; crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0); crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  (* No old stealth page *)
  exists ma1; exists pg1; crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0); crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H14 in *; crush.
  rewrite ctxt_switch_mem; crush.
  destruct H_valid_virtmap with (st_active_os s) m p0 pt stealth_va i1; crush.
  destruct H5 with m i1 x; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H14; crush.
  assert (stealth_va <> v); crush.
  STAC.synonym_cached.
  (* The old os had no stealth page *)
  rewrite ctxt_switch_cache in *; crush.
  unfold remove_osi_stealth_page in Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct (vadd_eq stealth_va v); crush.
  rewrite H6 with (pt_addr:=m) in H2; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H14; crush.
  STAC.synonym_cached.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  (* The removed entry is not the new stealth page *)
    STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold remove_osi_stealth_page in Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12 in *; crush.
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  rewrite flush_ma_cache in *; crush.
  destruct (madd_eq i1 m1); crush;
    [rewrite get_remove_ma_cache_eq in * |
     rewrite get_remove_ma_cache_ne in *]; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  destruct stealth_cached_is_aos with s m m1 x0; crush.
  unfold valid_sm_state; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H12 in *; crush.
  (* v <> stealth_va *)
  elim n1.
  unfold stealth_page_reserved_line in *.
  apply H7 with (ma:=m1) (pg:=x0); crush.
  symmetry.
  apply cache_inertia with
    (c:=fst (remove_ma_cache (smem_c (st_memory s)) i1))
    (va:=stealth_va) (va':=v) (ma:=m0) (ma':=m1) (pg:=p) (pg':=x0); crush.
  rewrite ctxt_switch_cache in *; crush.
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  destruct stealth_cached_is_aos with s m m1 x0; crush.
  unfold valid_sm_state; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H12 in *; crush.
  (* v <> stealth_va *)
  elim n0.
  unfold stealth_page_reserved_line in *.
  apply H7 with (ma:=m1) (pg:=x0); crush.
  symmetry.
  apply cache_inertia with (c:=smem_c (st_memory s))
    (va:=stealth_va) (va':=v) (ma:=m0) (ma':=m1) (pg:=p) (pg':=x0); crush.

  (* The read stealth does not remove an entry *)
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold remove_osi_stealth_page in Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.
  
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold remove_osi_stealth_page in Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12 in *; crush.
  destruct (madd_eq i1 ma1); crush.
  destruct H5 with m ma1 pg1; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H12; crush.
  exists ma1;
    exists (Page (RW_c (Some x)) (page_owned_by pg1) true); crush.
  rewrite get_mem_unchanged with (ma:=m0) (pg:=p); crush.
  unfold remove_osi_stealth_page; rewrite ctxt_switch_mem in *; crush.
  rewrite H12; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  rewrite <- get_add_mem; crush.
  destruct (vadd_eq stealth_va v); crush.
  destruct STAC.no_cache_alias with s stealth_va v m1
    (Page (RW_c (Some x)) (page_owned_by pg1) true) p2;
    crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite get_remove_ma_cache_no_entry in H11; crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.  
  rewrite H12; crush.
  rewrite Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.  
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H1; repeat split; auto.
  unfold same_page_type; rewrite H2; auto.
  exists ma1; exists pg1; crush.
  rewrite get_mem_unchanged with (ma:=m0) (pg:=p); crush.
  unfold remove_osi_stealth_page; rewrite ctxt_switch_mem in *; crush.
  rewrite H12; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.  
  rewrite H12; crush.
  rewrite Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.  
  rewrite ctxt_switch_cache in *; crush.  
  exists ma1; exists pg1; crush.
  rewrite get_mem_unchanged with (ma:=m0) (pg:=p); crush.
  unfold remove_osi_stealth_page; rewrite ctxt_switch_mem in *; crush.
  rewrite H12; crush.
  rewrite ctxt_switch_mem; crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12; crush.
  rewrite ctxt_switch_cache; crush.
  rewrite Heqp0; crush.
  rewrite ctxt_switch_cache; crush.
  rewrite Heqp0; crush.
  rewrite ctxt_switch_mem in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.

  destruct H_valid_virtmap with osi' pt_addr p0 p1 stealth_va m0; crush.
  rewrite <- H_valid_hyper with osi' i0 pa pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem in *; crush.

  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite H12; crush.
  destruct (madd_eq i1 ma1); crush.
  destruct H5 with m ma1 pg1; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H12; crush.
  exists ma1; exists
    (Page (RW_c (Some x)) (page_owned_by pg1) true); crush.
  rewrite get_invalid; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  rewrite <- get_add_mem; crush.
  destruct (vadd_eq stealth_va v); crush.
  destruct STAC.no_cache_alias with s v stealth_va m0 p1
    (Page (RW_c (Some x)) (page_owned_by pg1) true); crush.
  rewrite get_remove_ma_cache_no_entry in H11; crush.
  rewrite ctxt_switch_cache in *; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H1; repeat split; auto.
  unfold same_page_type; rewrite H2; auto.
  exists ma1; exists pg1; crush.
  rewrite get_invalid; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
  exists ma1; exists pg1; crush.
  rewrite get_invalid; crush.
  rewrite ctxt_switch_mem; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
Qed.

Lemma ctxt_switch_non_stealth_equivalent_hyper_mapping_2 :
  forall (s s':state) (osi osi':os_ident) (pa:padd) (pt_addr:madd),
    valid_sm_state s ->
    osi_pa_to_ma s osi' pa = Some pt_addr ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    osi_non_stealth_equivalent_hyper_mapping s' s osi.
Proof.
  unfold osi_non_stealth_equivalent_hyper_mapping, valid_sm_state; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite H1,H2 in *; crush.
  clear dependent s'.
  unfold ctxt_switch_new_smem in *; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?.
  destruct (page_of_vadd_madd_smem (st_memory s) stealth_va m0) eqn:?.
  (* The new osi has a stealth page *)
  destruct (add_cache (smem_c (remove_osi_stealth_page (ctxt_switch (st_memory s)) m))
    (stealth_va,m0) p) eqn:?.
  (* The read stealth removes some entry *)
  destruct o. destruct p0. destruct c0.
  destruct (madd_eq m1 m0); crush.
  (* The removed entry is the new stealth page *)
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  unfold remove_osi_stealth_page in H2; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_cache in *; crush.
  destruct (madd_eq m0 i1); crush;
    [rewrite get_remove_ma_cache_eq in * |
     rewrite get_remove_ma_cache_ne in *]; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  destruct (madd_eq m0 ma1); crush.
  (* ma1 is the stealth page *)
  destruct H_valid_cache with stealth_va ma1 x0; crush.
  exists ma1; exists x; crush.
  rewrite get_cache_entry1 with (ma:=ma1) (pg:=p)
    (va':=stealth_va) (pg':=x0) in H4; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H17, H18;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H4; auto.
  (* ma1 is not the stealth page *)
  destruct (madd_eq i1 ma1); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  destruct H5 with m ma1 p1; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H14; crush.
  exists ma1; exists p1; crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0) in H4; crush.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_mem in *; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- get_add_mem in *; crush.
  destruct (vadd_eq v stealth_va); crush.
  destruct H_valid_cache with stealth_va m1
    (Page (RW_c (Some x)) (page_owned_by p1) true); crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H19;
    repeat split; auto.
  unfold same_page_type; rewrite H18; auto.
  destruct STAC.no_cache_alias with s v stealth_va m1 pg1
    (Page (RW_c (Some x)) (page_owned_by p1) true); crush.
  rewrite ctxt_switch_mem in *; crush.
  (* ma1 is not the old stealth page *)
  exists ma1; exists pg1; crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0) in *; crush.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite flush_ma_mem in *; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite get_add_mem_distinct in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  assert (stealth_va <> v); crush.
  STAC.synonym_cached.
  (* No old stealth page *)
  rewrite ctxt_switch_cache in *; crush.
  destruct (madd_eq m0 ma1); crush.
  rewrite get_cache_entry1 with (ma:=ma1) (pg:=p)
    (va':=stealth_va) (pg':=x0) in H4; crush.
  destruct H_valid_cache with v ma1 pg1; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists ma1; exists x; crush.
  unfold same_page_but_value; rewrite H16, H17;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H13; auto.
  destruct (vadd_eq stealth_va v); crush.
  STAC.synonym_cached.
  (* ma1 is not the stealth page *)
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=stealth_va) (ma':=m0) (pg':=x0) in *; crush.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H14 in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  exists ma1; exists pg1; crush.
  rewrite ctxt_switch_mem in *; crush.
  destruct (vadd_eq stealth_va v); crush.
  STAC.synonym_cached.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  (* The removed entry is not the new stealth page *)
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct (madd_eq m1 ma1); crush.
  rewrite get_cache_entry1 with (ma:=m0) (pg:=p)
    (va':=v) (pg':=p0) in H4; crush.
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12 in *; crush.
  rewrite flush_ma_cache in *; crush.
  destruct (madd_eq i1 ma1); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache with v ma1 x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists ma1; exists x; crush.
  unfold same_page_but_value; rewrite H17, H18;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H16; auto.
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache with v ma1 x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists ma1; exists x; crush.
  unfold same_page_but_value; rewrite H17, H18;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H16; auto.
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache with v ma1 x0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists ma1; exists x; crush.
  unfold same_page_but_value; rewrite H17, H18;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H16; auto.
  rewrite Heqp0; crush.
  rewrite get_cache_entry2 with (ma:=m0) (pg:=p)
    (va':=v) (ma':=m1) (pg':=p0) in *; crush.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12 in *; crush.
  rewrite flush_ma_mem in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  destruct (madd_eq m2 ma1); crush.
  rewrite <- get_add_mem in *; crush.
  destruct H_valid_cache with v0 ma1 pg1; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists ma1; exists x; crush.
  unfold same_page_but_value; rewrite H15, H16;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H14; auto.
  rewrite get_add_mem_distinct in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite Heqp0; crush.
  
  (* The read stealth does not remove an entry *)
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite get_mem_unchanged with (ma:=m0) (pg:=p) in H4; crush.
  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12 in *; crush.
  rewrite flush_ma_mem in *; crush.
  unfold update_cache_entry in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  destruct (madd_eq m1 ma1); crush.
  rewrite <- get_add_mem in *; crush.
  destruct H_valid_cache with v ma1 pg1; crush.
  exists ma1; exists x; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H15, H16;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H14; auto.
  rewrite get_add_mem_distinct in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite Heqp0; crush.

  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush. 
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with osi' pt_addr p0 p1 stealth_va m0; crush.
  rewrite <- H_valid_hyper with osi' i0 pa pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem in *; crush.

  unfold remove_osi_stealth_page in H4; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite H12 in *; crush.
  rewrite get_invalid in *; crush.
  rewrite flush_ma_mem in *; crush.
  unfold update_cache_entry in *; crush.
  destruct (madd_eq m0 ma1); crush.
  rewrite <- get_add_mem in *; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache with v ma1 pg1; crush.
  exists ma1; exists x; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value; rewrite H15, H16;
    repeat split; auto.
  unfold same_page_type; rewrite H1, H14; auto.
  rewrite get_add_mem_distinct in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
  rewrite get_invalid in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite get_invalid in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
Qed.

Lemma ctxt_switch_non_stealth_equivalent_pt_pages_1 :
  forall (s s':state) (osi osi':os_ident) (pa:padd) (pt_addr:madd),
    valid_sm_state s ->
    osi_pa_to_ma s osi' pa = Some pt_addr ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    osi_non_stealth_equivalent_pt_pages s s' osi.
Proof.
  intros.
  generalize H.
  unfold osi_non_stealth_equivalent_pt_pages, valid_sm_state; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H2 in *; crush.
  rewrite H1 in *; crush.
  rewrite osi_pa_to_pg_ctxt_switch_new_smem
    with (osi:=osi') (pt_padd:=pa) (osi':=osi) (pa':=pt_padd); crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite <- H5; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H2; crush.
  exists pg1; crush.
  rewrite H1.
  unfold osi_pa_to_pg in *; crush.
  rewrite page_of_vadd_madd_smem_ctxt_switch_new_smem
    with (osi:=osi') (osi':=osi) (pt_padd:=pa) (pt_padd':=pt_padd) 
                     (pt_addr':=i0); crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H5; crush.
Qed.

Lemma ctxt_switch_non_stealth_equivalent_pt_pages_2 :
  forall (s s':state) (osi osi':os_ident) (pa:padd) (pt_addr:madd),
    valid_sm_state s ->
    osi_pa_to_ma s osi' pa = Some pt_addr ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    osi_non_stealth_equivalent_pt_pages s' s osi.
Proof.
  intros.
  generalize H.
  unfold osi_non_stealth_equivalent_pt_pages, valid_sm_state; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H2 in *; crush.
  rewrite H1 in *; crush.
  assert (madd_type (smem_mm (ctxt_switch_new_smem s pt_addr)) i0 = Some PT).
  unfold madd_type; crush.
  rewrite <- H5; crush.
  rewrite madd_type_ctxt_switch_new_smem
    with (osi:=osi') (pt_padd:=pa) (osi':=osi) (pa':=pt_padd) in H14; unfold osi_pa_to_ma in *; crush.
  erewrite osi_pa_to_pg_ctxt_switch_new_smem
    with (osi:=osi') (pt_padd:=pa) (pt_addr:=pt_addr) (osi':=osi) (pa':=pt_padd) in H4; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  exists mp1; unfold osi_pa_to_ma in *; crush.
  exists pg1; crush.
  rewrite H1 in *.
  unfold osi_pa_to_pg in *; crush.
  rewrite page_of_vadd_madd_smem_ctxt_switch_new_smem
    with (osi:=osi') (osi':=osi) (pt_padd:=pa) (pt_padd':=pt_padd)
                     (pt_addr':=i1) in H14; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H2 in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H1 in *.
  rewrite osi_pa_to_pg_ctxt_switch_new_smem
    with (osi:=osi') (pt_padd:=pa) (osi':=osi) (pa':=pt_padd) in H4; crush.
  rewrite <- H5; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H2 in *; crush.
  unfold osi_pa_to_ma; crush.
  rewrite <- madd_type_ctxt_switch_new_smem
    with (osi:=osi') (pt_padd:=pa) (pt_addr:=pt_addr)
         (osi':=osi) (pa':=pt_padd); crush.
  unfold madd_type; crush.
  rewrite <- H5; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H2 in *; crush.
Qed.

Lemma switch_stealth_os_non_stealth_isolation :
  forall (s s':state) (osi:os_ident),
    valid_sm_state s ->
    ACT.pre_switch osi s ->
    post_sm_switch osi s s' ->
    osi_non_stealth_equivalent s s'.
Proof.
  unfold ACT.pre_switch, post_sm_switch; crush.
  generalize H; unfold valid_sm_state; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite H1 with pt_addr in *; crush; clear H1.
  2: unfold osi_current_page_table; crush.

  assert (osi_pa_to_ma s osi (os_current_page x) = Some pt_addr).
  unfold osi_pa_to_ma; crush.
  unfold osi_non_stealth_equivalent, osi_non_stealth_equivalent_smem,
    osi_non_stealth_equivalent_hyper; crush.
  unfold same_current_pt_page, same_current_pt_page_one; crush.
  eapply ctxt_switch_non_stealth_equivalent_hyper_mapping_1; crush.
  eapply ctxt_switch_non_stealth_equivalent_hyper_mapping_2; crush.
  eapply ctxt_switch_non_stealth_equivalent_pt_pages_1; crush.
  eapply ctxt_switch_non_stealth_equivalent_pt_pages_2; crush.
Qed.

Theorem os_non_stealth_isolation_not_osi :
  forall (s s':state) (a:sm_action),
    valid_sm_exec s a s' ->
    valid_sm_state s ->
    ~ non_stealth_access a ->
    osi_non_stealth_equivalent s s'.
Proof.
  intros.
  Local Hint Resolve STAC.same_page_type_refl osi_non_stealth_equivalent_refl.
  Local Hint Constructors non_stealth_access osi_action.
  assert (valid_state s); [unfold valid_sm_state in *; crush|].

  sm_action_induction a; [clear Case; STAC.action_induction a | | | | ];
    unfold valid_sm_exec in *; crush;
    unfold ACT.post_silent, ACT.pre_read, ACT.pre_read_hyper,
      ACT.post_read, ACT.post_read_hyper, ACT.pre_write, ACT.pre_write_hyper,
      ACT.post_write, ACT.post_write_hyper in *;
    crush; try (elim H1; crush; fail).

  (* Read *)
  eapply read_stealth_os_non_stealth_isolation; crush.

  (* Read hyper *)
  eapply read_stealth_os_non_stealth_isolation; crush.

  (* Write *)
  eapply write_stealth_os_non_stealth_isolation; crush.

  (* Del *)
  destruct (vadd_eq v stealth_va); crush;
    try (elim H1; crush; fail).
  eapply del_stealth_os_non_stealth_isolation; crush.

  (* Hcall *)
  unfold ACT.pre_hcall, ACT.post_hcall in *;
  unfold osi_non_stealth_equivalent, same_current_pt_page,
    same_current_pt_page_one, osi_non_stealth_equivalent_hyper_mapping,
    osi_pa_to_ma, osi_non_stealth_equivalent_pt_pages;
    crush; try (repeat eexists; crush; fail).
  destruct (os_ident_eq osi (st_active_os s)); crush.
  exists (OS (Some h) (os_current_page x0));
    rewrite <- OSS.get_set; crush.
  exists x; rewrite OSS.get_set_distinct; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush.
  rewrite <- OSS.get_set in *; crush.
  exists x; rewrite OSS.get_set_distinct in *; crush.

  (* Ret ctrl *)
  unfold ACT.pre_ret_ctrl, ACT.post_ret_ctrl in *;
  unfold osi_non_stealth_equivalent, same_current_pt_page,
    same_current_pt_page_one, osi_non_stealth_equivalent_hyper_mapping,
    osi_pa_to_ma, osi_non_stealth_equivalent_pt_pages;
    crush; try (repeat eexists; crush; fail).

  (* Chmod *)
  unfold ACT.pre_chmod, ACT.post_chmod in *;
  unfold osi_non_stealth_equivalent, same_current_pt_page,
    same_current_pt_page_one, osi_non_stealth_equivalent_hyper_mapping,
    osi_pa_to_ma, osi_non_stealth_equivalent_pt_pages;
    crush; try (repeat eexists; crush; fail).

  (* New sm *)
  destruct (os_ident_eq o (st_active_os s)); crush.
  destruct new_sm_eq_old_new_active with s s' p;
    unfold ACT.valid_exec,valid_sm_exec in *; crush;
    unfold ACT.pre_read_hyper, ACT.post_read_hyper in *; crush.
  apply osi_non_stealth_equivalent_trans with x; crush.
  eapply new_stealth_os_non_stealth_isolation; crush.
  eapply read_stealth_os_non_stealth_isolation; crush.
  apply EXEC_INV.valid_state_invariant with s
    (New_untrusted (st_active_os s) stealth_va p);
    unfold ACT.valid_exec; crush.
  destruct new_sm_eq_old_new with s s' o p;
    unfold valid_sm_exec; crush.
  eapply new_stealth_os_non_stealth_isolation; crush.

  (* Switch *)
  eapply switch_stealth_os_non_stealth_isolation; crush.
Qed.


  Lemma same_page_but_value_symm : forall pg1 pg2,
    same_page_but_value pg1 pg2 ->
    same_page_but_value pg2 pg1.
  Proof.
    intros.
    unfold same_page_but_value in *; crush.
    unfold same_page_type in *; crush.
    destruct (page_content pg1); destruct (page_content pg2); crush.
  Qed.

Opaque ACT.new.
Lemma new_non_stealth_hyper_eq :
  forall (s1 s2 s1' s2':state) (osi osi0:os_ident) (va:vadd) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    pre_new_normal osi va pa s1 ->
    ACT.post_new osi va pa s1 s1' ->
    pre_new_normal osi va pa s2 ->
    ACT.post_new osi va pa s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent_hyper_mapping s1' s2' osi0.
Proof.
  unfold osi_non_stealth_equivalent; crush.
  unfold valid_sm_state in H; crush.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  unfold pre_new_normal, pre_all_new, ACT.post_new,
    ACT.post_all_new in *; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct H5 with osi0; crush.
  
  destruct (os_ident_eq osi0 osi); crush.
  (* osi0 = osi *)
  destruct (padd_eq pa0 (os_current_page x4)); crush.
  (* pa0 is the current page *)
  erewrite STAC.new_get_mem_pt_addr in H13; crush.
  2: assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi);
    crush; unfold madd_owner in *; crush.
  subst pg1.
  destruct H33.
  destruct H12 with (os_current_page x4) x3 pt_page; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  destruct H31.
  destruct H31 with x4; crush.
  rewrite H51 in *; crush.
  exists x0; exists
    (Page (PT_c (PT.set pt0 va x2)) (page_owned_by x7) (page_cacheable x7));
    crush.
  erewrite STAC.new_get_mem_pt_addr; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold same_page_but_value in *;
    destruct H38; destruct H38; simpl in *;
    rewrite H33, H38, H50 in *; repeat split; auto.
  (* pa0 is not the current page *)
  destruct (shared_page_dec s1 va x5).
  (* new shared page *)
  destruct (padd_eq pa0 pa); crush.
  destruct (remove_ma_cache (smem_c (st_memory s1)) x5) eqn:?; crush.
  (* Removed cache entry from s1 *)
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  destruct get_remove_ma_cache_some_entry
    with (smem_c (st_memory s1)) v x5 m p; crush.
  rewrite STAC.new_get_mem_shared_new_ma 
    with (pt_page':=pt_page) (pt':=pt) (va_rep:=v) (pg_rep:=p)
    in H13; crush.
  destruct H_valid_cache with v m p; crush.
  destruct H33.
  destruct H33 with pa m x5; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold valid_sm_state in *; crush.  
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (remove_ma_cache (smem_c (st_memory s2)) x2) eqn:?; crush.
  destruct o. destruct p0. destruct c1.
  symmetry in Heqp.
  destruct get_remove_ma_cache_some_entry
    with (smem_c (st_memory s2)) v0 x2 m0 p0; crush.
  rewrite Heqp0; crush.
  destruct H_valid_cache0 with v0 m0 p0; crush.
  exists m0; exists (Page (page_content p0) (page_owned_by p0) false); crush.
  rewrite STAC.new_get_mem_shared_new_ma 
    with (pt_page':=pt_page0) (pt':=pt0) (va_rep:=v0) (pg_rep:=p0); crush.
  apply shared_page_non_stealth_eq with (s1:=s1) (osi:=osi)
    (pa:=pa) (ma1:=m); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp0; crush.
  unfold is_rw_page in *; crush.
  unfold same_page_but_value in *; destruct H45; destruct H61;
  simpl in *; rewrite H41, H60; repeat split; auto.
  unfold same_page_type; rewrite H12, H52; auto.
  exists x2; exists (Page (page_content x7) (page_owned_by x7) false); crush.
  rewrite STAC.new_get_mem_shared_new_ma_no_entry
    with (pt_page':=pt_page0) (pt':=pt0) (pg:=x7); crush.
  destruct H45; crush.
  unfold is_rw_page in H39; crush.
  unfold same_page_type in H53; rewrite H39, H55 in H53; simpl in H53; crush.
  apply shared_page_non_stealth_eq with (s1:=s1) (osi:=osi)
    (pa:=pa) (ma1:=m); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp0; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value in *; destruct H45; destruct H52;
  simpl in *; rewrite H41; repeat split; auto.
  unfold same_page_type in H54; rewrite H39 in H54; simpl in *.
  unfold same_page_type; destruct (page_content x7) eqn:?;
   rewrite H12; simpl in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  (* No removed cache entry from s1 *)
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  rewrite STAC.new_get_mem_shared_new_ma_no_entry
    with (pt_page':=pt_page) (pt':=pt) (pg:=p) in H13; crush.
  destruct H33.
  destruct H13 with pa x5 p; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold valid_sm_state in *; crush.  
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (remove_ma_cache (smem_c (st_memory s2)) x2) eqn:?; crush.
  destruct o. destruct p0. destruct c1.
  symmetry in Heqp.
  destruct get_remove_ma_cache_some_entry
    with (smem_c (st_memory s2)) v x2 m p0; crush.
  rewrite Heqp0; crush.
  destruct H_valid_cache0 with v m p0; crush.
  exists m; exists (Page (page_content p0) (page_owned_by p0) false); crush.
  rewrite STAC.new_get_mem_shared_new_ma 
    with (pt_page':=pt_page0) (pt':=pt0) (va_rep:=v) (pg_rep:=p0); crush.
  apply shared_page_non_stealth_eq with (s1:=s1) (osi:=osi)
    (pa:=pa) (ma1:=x5); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp0; crush.
  unfold is_rw_page in *; crush.
  unfold same_page_but_value in *; destruct H39; destruct H55;
  simpl in *; rewrite H12, H54, <- H39; repeat split; auto.
  unfold same_page_type in *; rewrite H46 in *; simpl in *.
  destruct (page_content p);
    rewrite H52 in *; try (destruct o; simpl in *; auto); auto.
  exists x2; exists (Page (page_content x7) (page_owned_by x7) false); crush.
  rewrite STAC.new_get_mem_shared_new_ma_no_entry
    with (pt_page':=pt_page0) (pt':=pt0) (pg:=x7); crush.
  destruct H5 with osi; crush.
  destruct H46; crush.
  destruct H46 with x4; crush.
  rewrite <- H53 in H16; crush.
  apply shared_page_non_stealth_eq with (s1:=s1) (osi:=osi)
    (pa:=pa) (ma1:=x5); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp0; crush.
  unfold same_page_but_value in *; destruct H39; destruct H46;
    simpl in *; repeat split; auto.
  rewrite Heqp; crush.

  (* pa0 <> pa *)
  assert (exists pg1_old, get_mem (smem_mm (st_memory s1)) ma1 = Some pg1_old /\
            same_page_but_value pg1_old pg1); crush.
  destruct (madd_of_vadd (smem_mm (st_memory s1)) x3 va) eqn:?; crush.
  destruct (madd_eq m ma1); crush.
  destruct (get_cache (smem_c (st_memory s1)) (va, ma1)) eqn:?; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_some_entry
    with (pt_page':=pt_page) (pt':=pt) (pg:=p) in H13; crush.
  destruct H_valid_cache with va ma1 pg1; crush.
  destruct H33.
  destruct H33 with pa0 ma1 x6; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold same_page_but_value.
  unfold is_cacheable in *.
  rewrite H37, H40, H41; repeat split; auto.
  unfold is_rw_page in *; destruct H13, H39.
  unfold same_page_type.
  rewrite H13, H39; simpl; auto.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_no_entry
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rename x6 into pg1_old.

  destruct H33.
  destruct H33 with pa0 ma1 pg1_old; crush.
  unfold osi_pa_to_ma; crush.
  rename x6 into ma2.
  rename x7 into pg2_old.
  destruct H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s2)) x0 va) eqn:?; crush.
  destruct (madd_eq m ma2); crush.
  destruct (get_cache (smem_c (st_memory s2)) (va, ma2)) eqn:?; crush.
  exists ma2; exists p; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_some_entry
    with (pt_page':=pt_page0) (pt':=pt0) (pg:=p); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_non_stealth_eq with s1 osi pa x5; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  destruct H_valid_cache0 with va x0 p; crush.
  unfold is_rw_page in *; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_cache0 with va ma2 p; crush.
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  apply same_page_but_value_trans with pg2_old; crush.
  apply same_page_but_value_symm; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  unfold same_page_but_value; rewrite H51, H55, H56; repeat split; auto.
  unfold same_page_type; rewrite H49, H54; simpl; auto.
  exists ma2; exists pg2_old; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_no_entry
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_non_stealth_eq with s1 osi pa x5; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H5 with osi; crush.
  destruct H49.
  destruct H49 with x4; crush.
  rewrite H56 in *; crush.
  unfold osi_pa_to_ma in *; crush.
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  exists ma2; exists pg2_old; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_non_stealth_eq with s1 osi pa x5; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H5 with osi; crush.
  destruct H49.
  destruct H49 with x4; crush.
  rewrite H56 in *; crush.
  unfold osi_pa_to_ma in *; crush.
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  exists ma2; exists pg2_old; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_non_stealth_eq with s1 osi pa x5; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H5 with osi; crush.
  destruct H49.
  destruct H49 with x4; crush.
  rewrite H56 in *; crush.
  unfold osi_pa_to_ma in *; crush.
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  
  (* Not shared page *) 
  destruct (padd_eq pa pa0); crush.
  (* pa = pa0 *)
  assert (exists pg1_old, get_mem (smem_mm (st_memory s1)) ma1 = Some pg1_old /\
            same_page_but_value pg1_old pg1); crush.
  Transparent ACT.new.
  unfold ACT.new in H13; crush.
  Opaque ACT.new.
  destruct (add_page_table_entry (smem_mm (st_memory s1)) x3 va ma1) eqn:?; crush.
  rename m into mm'.
  rewrite SMEM.new_mem_not_shared with (mm':=mm') in H13; crush.
  unfold remove_new_va_mem in H13; crush.
  destruct (madd_eq m ma1); crush.
  destruct H_valid_cache with va ma1 p; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H34 in *; crush.  
  rewrite <- get_add_mem in H13; crush.
  unfold is_rw_page, is_cacheable in *; destruct H39, H15.
  unfold same_page_but_value, same_page_type; rewrite H13, H15, H40, H37, H41;
    repeat split; simpl; auto.
  unfold add_page_table_entry in *; crush.
  rewrite H34 in *; crush.
  rewrite get_add_mem_distinct in H13; crush.
  rewrite get_add_mem_distinct in H13; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H34 in *; crush.
  rewrite get_add_mem_distinct in H13; crush.
  unfold add_page_table_entry in *; crush.
  rewrite H34 in *; crush.
  rewrite get_add_mem_distinct in H13; crush.
  rewrite new_mem_not_shared_invalid in H13; crush.
  rename x5 into pg1_old.

  destruct H33; crush.
  destruct H33 with pa0 ma1 pg1_old; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  rename x2 into ma2, x6 into pg2_old.
  assert (same_page_but_value pg1 pg2_old).
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  exists ma2.
  Transparent ACT.new.
  unfold ACT.new.
  Opaque ACT.new.
  crush.
  elim n0; crush.
  apply shared_page_non_stealth_eq with s2 osi pa0 ma2; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  destruct (add_page_table_entry (smem_mm (st_memory s2)) x0 va ma2) eqn:?; crush.
  rename m into mm'.
  rewrite SMEM.new_mem_not_shared with (mm':=mm'); crush.
  unfold remove_new_va_mem; crush.
  destruct (madd_eq m ma2); crush.
  exists p; crush.
  rewrite <- get_add_mem; crush.
  destruct H0; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_cache0 with va ma2 p; crush.
  unfold is_rw_page, is_cacheable in *; destruct H48, H50.
  apply same_page_but_value_trans with pg2_old; crush.
  unfold same_page_but_value, same_page_type; rewrite H48, H49, H50, H51, H52;
    repeat split; simpl; auto.
  rewrite get_add_mem_distinct; crush.
  unfold add_page_table_entry in *; crush.
  destruct H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush. 
  rewrite H51 in *; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H5 with osi; crush.
  destruct H48.
  destruct H48 with x4; crush.
  rewrite H55 in *; crush.
  unfold add_page_table_entry in *; crush.
  destruct H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush. 
  rewrite H51 in *; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H5 with osi; crush.
  destruct H48.
  destruct H48 with x4; crush.
  rewrite H55 in *; crush.
  unfold add_page_table_entry in *; crush.
  destruct H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush. 
  rewrite H51 in *; crush.
  rewrite get_add_mem_distinct; crush.
  destruct H5 with osi; crush.
  destruct H48.
  destruct H48 with x4; crush.
  rewrite H55 in *; crush.
  rewrite new_mem_not_shared_invalid; crush.

  (* pa <> pa0 *)
  assert (exists pg1_old, get_mem (smem_mm (st_memory s1)) ma1 = Some pg1_old /\
            same_page_but_value pg1_old pg1); crush.
  destruct (madd_of_vadd (smem_mm (st_memory s1)) x3 va) eqn:?; crush.
  destruct (madd_eq m ma1); crush.
  destruct (get_cache (smem_c (st_memory s1)) (va, ma1)) eqn:?; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_some_entry
    with (pt_page':=pt_page) (pt':=pt) (pg:=p) in H13; crush.
  destruct H_valid_cache with va ma1 pg1; crush.
  destruct H33.
  destruct H33 with pa0 ma1 x6; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold same_page_but_value.
  unfold is_cacheable in *.
  rewrite H37, H40, H41; repeat split; auto.
  unfold is_rw_page in *; destruct H13, H39.
  unfold same_page_type.
  rewrite H13, H39; simpl; auto.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_no_entry
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt) in H13; crush.
  unfold madd_type in H2; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.

  rename x6 into pg1_old.
  destruct H33.
  destruct H33 with pa0 ma1 pg1_old; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  rename x6 into ma2, x7 into pg2_old.
  assert (same_page_but_value pg1 pg2_old).
  apply same_page_but_value_trans with pg1_old; crush.
  apply same_page_but_value_symm; crush.
  assert (~ shared_page s2 va x2); crush.
  elim n0.
  apply shared_page_non_stealth_eq with s2 osi pa x2; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  destruct H5 with osi; crush.
  destruct H45.
  destruct H45 with x4; crush.
  rewrite H50 in *; crush.
  clear dependent s1.
  destruct H0; crush.
  STAC.destruct_valid_state.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s2)) x0 va) eqn:?; crush.
  destruct (madd_eq m ma2); crush.
  destruct (get_cache (smem_c (st_memory s2)) (va, ma2)) eqn:?; crush.
  exists ma2; exists p; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_some_entry
    with (pt_page':=pt_page0) (pt':=pt0) (pg:=p); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  destruct H_valid_cache with va ma2 p; crush.
  apply same_page_but_value_trans with pg2_old; crush.
  unfold is_cacheable, is_rw_page in *; crush.
  unfold same_page_but_value; rewrite H11, H28, H29; repeat split; auto.
  unfold same_page_type; rewrite H9, H25; simpl; auto.
  exists ma2; exists pg2_old; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_no_entry
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  exists ma2; exists pg2_old; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  exists ma2; exists pg2_old; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in H4; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.

  (* osi0 <> osi *)
  rewrite STAC.new_get_mem_distinct_osi
    with (pt_page':=pt_page) (pt':=pt) (osi:=osi0)
    in H13; crush.
  destruct H33.
  destruct H33 with pa0 ma1 pg1; crush.
  unfold osi_pa_to_ma; crush.
  exists x6, x7; crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite STAC.new_get_mem_distinct_osi
    with (pt_page':=pt_page0) (pt':=pt0) (osi:=osi0); crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists (os_current_page x4); unfold osi_pa_to_ma; crush.
Qed.

Lemma non_stealth_eq_same_page_but_value :
  forall s1 s2 osi va pa ma1 ma2 pg1,
    valid_state s1 ->
    valid_state s2 ->
    osi_non_stealth_equivalent_hyper_mapping s1 s2 osi ->
    osi_pa_to_ma s1 osi pa = Some ma1 ->
    osi_pa_to_ma s2 osi pa = Some ma2 ->
    page_of_vadd_madd_smem (st_memory s1) va ma1 = Some pg1 ->
    (exists pg2, 
      page_of_vadd_madd_smem (st_memory s2) va ma2 = Some pg2 /\
      same_page_but_value pg1 pg2).
Proof.
  intros.
  STAC.destruct_valid_state_H H.
  STAC.destruct_valid_state_H H0.
  unfold page_of_vadd_madd_smem in H4; crush.
  destruct H_valid_cache with va ma1 pg1; crush.
  destruct H1 with pa ma1 x; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache0 with va ma2 p; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists p; crush.
  unfold same_page_but_value in *; destruct H11; destruct H17;
    rewrite H15, H16, H13, H9 in *; simpl in *; repeat split; auto.
  unfold same_page_type in *; rewrite H4, H7, H12, H14 in *;
    simpl in *; auto.
  unfold is_rw_page, is_cacheable in *; crush.
  unfold same_page_but_value in *; destruct H11; destruct H12;
    rewrite H9, H8, H5 in *; simpl in *; repeat split; auto.
  unfold same_page_type in *; rewrite H4, H7 in *;
    simpl in *; auto.
  destruct H1 with pa ma1 pg1; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache0 with va ma2 p; crush.
  unfold is_rw_page, is_cacheable in *; crush.
  exists p; crush.
  unfold same_page_but_value in *; destruct H6; destruct H12;
    rewrite H10, H11, H8 in *; simpl in *; repeat split; auto.
  unfold same_page_type in *; rewrite H9, H7 in *;
    simpl in *; auto.
Qed.

Lemma new_non_stealth_pt_pages_eq :
  forall (s1 s2 s1' s2':state) (osi osi0:os_ident) (va:vadd) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    pre_new_normal osi va pa s1 ->
    ACT.post_new osi va pa s1 s1' ->
    pre_new_normal osi va pa s2 ->
    ACT.post_new osi va pa s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent_pt_pages s1' s2' osi0.
Proof.
  unfold osi_non_stealth_equivalent; crush.
  unfold valid_sm_state in H; crush.
  unfold osi_non_stealth_equivalent_pt_pages; crush.
  generalize H1 H2 H3 H4.
  unfold pre_new_normal, pre_all_new, ACT.post_new,
    ACT.post_all_new; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct H5 with osi0; crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush. 

  destruct (os_ident_eq osi0 osi); crush.
  (* osi0 = osi *)
  destruct H37.
  destruct H37 with x4; crush.
  match goal with
      | [H1: ACT.post_new ?osi ?va ?pa s1 ?s1',
         H2: ACT.post_new ?osi ?va ?pa s2 ?s2' |-_]=>
        assert (valid_state s1'); [|assert (valid_state s2');
        [|assert (osi_non_stealth_equivalent_hyper_mapping s1' s2' osi)]]
  end.
  apply EXEC_INV.valid_state_invariant with s1
    (New_untrusted osi va pa);
    unfold ACT.valid_exec; crush.
  unfold pre_new_normal, pre_all_new in H1; unfold ACT.pre_new;
    crush.
  exists x5; crush.
  unfold osi_pa_to_ma; crush.
  apply EXEC_INV.valid_state_invariant with s2
    (New_untrusted osi va pa);
    unfold ACT.valid_exec; crush.
  unfold pre_new_normal, pre_all_new in H3; unfold ACT.pre_new;
    crush.
  exists x2; crush.
  unfold osi_pa_to_ma; crush.
  apply new_non_stealth_hyper_eq with s1 s2 osi va pa; crush.
  unfold valid_sm_state; crush.
  unfold valid_sm_state; crush.  
  
  destruct (padd_eq pt_padd (os_current_page x4)); crush.
  (* pt_padd is the current page table *)
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page) (pt:=pt) in H12;
    crush.
  injection H13; intros; subst mp1; clear H13.
  rewrite H55 in *; crush.
  exists (Page (PT_c (PT.set pt0 va x2)) (page_owned_by pt_page0)
                     (page_cacheable pt_page0)); crush.
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page0) (pt:=pt0); crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists (PT.set pt0 va x2); crush.
  destruct (vadd_eq va0 va); crush.
  (* va0 = va *)
  rewrite <- PT.get_set in *; crush.
  exists x2; crush.
  destruct (padd_eq pa0 pa); crush.
  elim n.
  rewrite <- H_injective_hyper with osi x pa pa0 ma1; crush.
  replace (ACT.new s2 osi x0 va x2) with (st_memory (State
    (st_active_os s2)
         (OSS.set (st_oss s2) osi
         {| os_hcall := None; os_current_page := os_current_page x1 |})
         (st_exec_mode s2)
          (st_aos_activity s2)
          (st_hypervisor s2)
          (ACT.new s2 osi x0 va x2))); [|crush].
  eapply non_stealth_eq_same_page_but_value with
    (s1:=((State
         (st_active_os s1)
         (OSS.set (st_oss s1) osi
         {| os_hcall := None; os_current_page := os_current_page x4 |})
         (st_exec_mode s1)
         (st_aos_activity s1)
         (st_hypervisor s1)
         (ACT.new s1 osi x3 va ma1)))) (osi:=osi); crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  (* va0 <> va *)
  rewrite PT.get_set_distinct in *; crush.
  destruct H42.
  destruct H12 with (os_current_page x1) pt_page pt va0 ma1; crush.
  unfold osi_pa_to_pg; crush.
  exists x9; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (osi_pa_to_ma s2 osi pa0 = Some x9).
  apply H60; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_pg in *; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?.
  (* va0 accessible *)
  destruct H_va_has_pa with osi x x3 pt_page pt va0 ma1; crush.
  rename x10 into pa0.
  replace (ACT.new s2 osi x0 va x2) with (st_memory (State
    (st_active_os s2)
         (OSS.set (st_oss s2) osi
         {| os_hcall := None; os_current_page := os_current_page x1 |})
         (st_exec_mode s2)
          (st_aos_activity s2)
          (st_hypervisor s2)
          (ACT.new s2 osi x0 va x2))); [|crush].
  eapply non_stealth_eq_same_page_but_value with
    (s1:=((State
         (st_active_os s1)
         (OSS.set (st_oss s1) osi
         {| os_hcall := None; os_current_page := os_current_page x4 |})
         (st_exec_mode s1)
         (st_aos_activity s1)
         (st_hypervisor s1)
         (ACT.new s1 osi x3 va x5)))) (osi:=osi); crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s2 osi pa0 = Some x9).
  apply H60; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  (* va0 not accessible *)
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Hyp_PO).
  destruct H_valid_virtmap with osi x3 pt_page pt va0 ma1; crush.
  unfold os_accessible in *.
  rewrite <- H64; unfold madd_owner; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x9 = Hyp_PO).
  destruct H_valid_virtmap0 with osi x0 x7 x8 va0 x9; crush.
  unfold os_accessible in *.
  rewrite <- H65; unfold madd_owner; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  rewrite STAC.new_page_ne in H62; crush.  
  rewrite STAC.new_page_ne; crush.  
  unfold madd_owner in H63; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite <- H58; crush.
  unfold madd_owner in H65; crush.
  exists osi; exists (os_current_page x1); exists x0; 
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H58; crush.
  unfold madd_owner in H51; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite H40; crush.
  unfold madd_owner in H64; crush.
  exists osi; exists (os_current_page x1); exists x3; 
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite H40; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.

  (* pt_padd is not the current page table *)
  rewrite STAC.new_get_mem_distinct_pt in H12; crush.
  destruct H42.
  destruct H42 with pt_padd pt_page1 mp1 va0 ma1; crush.
  unfold osi_pa_to_pg; crush.
  unfold osi_pa_to_pg in *; crush.
  exists x7; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  apply n.
  rewrite H_injective_hyper0 with osi x6 pt_padd (os_current_page x1) i; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite <- H60; crush.
  exists x8; crush.
  assert (osi_pa_to_ma s2 osi pa0 = Some x9).
  apply H62; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?.
  (* va0 accessible *)
  destruct H_va_has_pa with osi x i0 pt_page1 mp1 va0 ma1; crush.
  (* rewrite <- H_valid_hyper with osi x pt_padd i0; crush. *)
  (* unfold madd_owner; crush. *)
  rename x10 into pa0.
  replace (ACT.new s2 osi x0 va x2) with (st_memory (State
    (st_active_os s2)
         (OSS.set (st_oss s2) osi
         {| os_hcall := None; os_current_page := os_current_page x1 |})
         (st_exec_mode s2)
          (st_aos_activity s2)
          (st_hypervisor s2)
          (ACT.new s2 osi x0 va x2))); [|crush].
  eapply non_stealth_eq_same_page_but_value with
    (s1:=((State
         (st_active_os s1)
         (OSS.set (st_oss s1) osi
         {| os_hcall := None; os_current_page := os_current_page x4 |})
         (st_exec_mode s1)
         (st_aos_activity s1)
         (st_hypervisor s1)
         (ACT.new s1 osi x3 va x5)))) (osi:=osi); crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s2 osi pa0 = Some x9).
  apply H62; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  (* va0 not accessible *)
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Hyp_PO).
  destruct H_valid_virtmap with osi i0 pt_page1 mp1 va0 ma1; crush.
  rewrite <- H_valid_hyper with osi x pt_padd i0; crush.
  unfold madd_owner; crush.
  unfold os_accessible in *.
  rewrite <- H67; unfold madd_owner; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x9 = Hyp_PO).
  destruct H_valid_virtmap0 with osi i x7 x8 va0 x9; crush.
  rewrite <- H_valid_hyper0 with osi x6 pt_padd i; crush.
  unfold madd_owner; crush.
  unfold os_accessible in *.
  rewrite <- H68; unfold madd_owner; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  rewrite STAC.new_page_ne in H64; crush.  
  rewrite STAC.new_page_ne; crush.  
  unfold madd_owner in H66; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite H53; crush.
  unfold madd_owner in H68; crush.
  exists osi; exists pt_padd; exists i;
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush;
      rewrite <- H60; crush.
  unfold madd_owner in H65; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite H40; crush.
  unfold madd_owner in H67; crush.
  exists osi; exists pt_padd; exists i0;
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush;
    rewrite <- H13; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=osi) (pt_addr':=x3) (va':=va) (ma':=x5); crush.
  unfold madd_type; crush.
  rewrite <- H13; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.

  (* osi0 <> osi *)
  rewrite STAC.new_get_mem_distinct_pt in H12; crush.
  destruct H42.
  destruct H42 with pt_padd pt_page1 mp1 va0 ma1; crush.
  unfold osi_pa_to_pg; crush.
  unfold osi_pa_to_pg in *; crush.
  exists x7; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i2 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s2)) i2 = Os_PO osi0); crush.
  unfold madd_type in H19; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2) eqn:?; crush.
  unfold madd_type; crush.
  rewrite <- H55; crush.
  exists x8; crush.
  assert (osi_pa_to_ma s2 osi0 pa0 = Some x9).
  apply H57; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_page_ne in H59; crush.
  destruct H58 with pg1; crush.
  exists x10; crush.
  rewrite STAC.new_page_ne; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x9 = Os_PO osi); crush.
  destruct H_valid_virtmap0 with osi0 i2 x7 x8 va0 x9; crush.
  rewrite <- H_valid_hyper0 with osi0 i1 pt_padd i2; crush.
  unfold madd_owner; crush.
  unfold madd_owner in H62; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; unfold os_accessible in *; auto.
  rewrite H64 in H62; crush.
  rewrite H65 in H62; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x9 = Os_PO osi); crush.
  destruct H_valid_virtmap0 with osi0 i2 x7 x8 va0 x9; crush.
  rewrite <- H_valid_hyper0 with osi0 i1 pt_padd i2; crush.
  unfold madd_owner; crush.
  unfold madd_owner in H62; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; unfold os_accessible in *; auto.
  rewrite H63 in H62; crush.
  rewrite H64 in H62; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite H53; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists osi0; exists pt_padd; exists i2;
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush;
    rewrite <- H55; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi); crush.
  destruct H_valid_virtmap with osi0 i0 pt_page1 mp1 va0 ma1; crush.
  rewrite <- H_valid_hyper with osi0 i pt_padd i0; crush.
  unfold madd_owner; crush.
  unfold madd_owner in H60; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; unfold os_accessible in *; auto.
  rewrite H62 in H60; crush.
  rewrite H63 in H60; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi); crush.
  destruct H_valid_virtmap with osi0 i0 pt_page1 mp1 va0 ma1; crush.
  rewrite <- H_valid_hyper with osi0 i pt_padd i0; crush.
  unfold madd_owner; crush.
  unfold madd_owner in H60; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; unfold os_accessible in *; auto.
  rewrite H61 in H60; crush.
  rewrite H62 in H60; crush.
  exists osi; exists (os_current_page x4); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush; rewrite H40; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists osi0; exists pt_padd; exists i0;
    unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush;
    rewrite <- H13; crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type
    with (osi:=osi) (pt_addr':=x3) (va':=va) (ma':=x5); crush.
  unfold madd_type; crush.
  rewrite <- H13; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
Qed.

Lemma new_osi_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (va:vadd) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    pre_new_normal osi va pa s1 ->
    ACT.post_new osi va pa s1 s1' ->
    pre_new_normal osi va pa s2 ->
    ACT.post_new osi va pa s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  intros; unfold osi_non_stealth_equivalent; crush.

  unfold osi_non_stealth_equivalent in *; crush.
  destruct H5 with osi0; crush.
  unfold same_current_pt_page, same_current_pt_page_one,
    pre_new_normal, ACT.post_new, ACT.post_all_new in *; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  exists (OS None (os_current_page x2)); crush.
  destruct H6 with x5; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  destruct H9 with x2; crush.

  unfold osi_non_stealth_equivalent_hyper; crush.
  eapply new_non_stealth_hyper_eq with (s1:=s1) (s2:=s2); crush.
  eapply new_non_stealth_hyper_eq with (s1:=s2) (s2:=s1); crush.
  unfold osi_non_stealth_equivalent_smem; crush.
  eapply new_non_stealth_pt_pages_eq with (s1:=s1) (s2:=s2); crush.
  eapply new_non_stealth_pt_pages_eq with (s1:=s2) (s2:=s1); crush.
Qed.


(* Two step executions *)
Lemma del_non_stealth_pt_pages_eq :
  forall (s1 s2 s1' s2':state) (osi osi0:os_ident) (va:vadd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_del osi va s1 ->
    ACT.post_del osi va s1 s1' ->
    ACT.pre_del osi va s2 ->
    ACT.post_del osi va s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent_pt_pages s1' s2' osi0.
Proof.
  unfold osi_non_stealth_equivalent; crush.
  unfold valid_sm_state in H; crush.
  destruct H5 with osi0; crush; clear H5.
  unfold osi_non_stealth_equivalent_pt_pages; crush.
  destruct (os_ident_eq osi osi0); crush.
  (* osi = osi0 *)
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi0; crush.
  STAC.get_osi_current_pt_page osi0; crush.
  assert (page_of_vadd_madd_smem (st_memory s1') va0 pt_addr =
          Some (Page (PT_c (PT.remove pt va))
                     (page_owned_by pt_page)
                     (page_cacheable pt_page))); crush.
  clear dependent s2.
  apply STAC.del_page_of_vadd_madd_smem_pt_page with (s:=s1) (osi':=osi0); crush.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 pt_addr p; crush.
  unfold is_rw_page in *; crush.

  destruct (padd_eq pt_padd (os_current_page x)); crush.
  (* pt_padd = os_current_page *)
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr va0 = Some ma1).
  eapply STAC.del_madd_of_vadd_old; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  unfold osi_pa_to_pg in H5; crush.
  unfold ACT.post_del in H2; crush.
  crush.
  rewrite <- H15; crush.
  destruct H14.
  destruct H14 with (os_current_page x) pt_page pt va0 ma1; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_of_vadd, madd_page_table in H25; crush.
  rewrite H23 in *; crush.
  assert (st_hypervisor s2 = st_hypervisor s2').
  unfold ACT.post_del in H4; crush.
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi0; crush.
  STAC.get_osi_current_pt_page osi0; crush.
  assert (os_current_page x = os_current_page x4); crush.
  unfold same_current_pt_page in *; crush.
  destruct H12 with x; crush.
  exists (Page (PT_c (PT.remove x2 va))
               (page_owned_by x1)
               (page_cacheable x1)); crush.
  unfold osi_pa_to_pg; crush.
  rewrite <- H32; crush.
  rewrite H43; crush.
  erewrite STAC.del_get_mem_pt_page; crush.
  unfold osi_current_page_table; crush.
  unfold osi_pa_to_pg in H27; crush.
  exists (PT.remove x2 va); crush.
  exists x3; crush.
  rewrite PT.get_remove_distinct with (inf:=x3); crush.
  clear dependent s2.
  unfold osi_pa_to_pg in H5; crush.
  rewrite STAC.del_get_mem_pt_page with
    (s:=s1) (pt_page:=pt_page) (pt:=pt) (osi':=osi0) (va':=va) in H5; crush.
  injection H15; intros; subst mp1.
  rewrite PT.get_remove in H17; crush.
  unfold osi_current_page_table; crush.
  unfold ACT.post_del in *; crush.
  unfold ACT.post_del in *; crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  apply H30; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?.
  (* va0 accessible *)
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H23 in *; crush.
  destruct H_va_has_pa with osi0 x0
    pt_addr pt_page pt va0 ma1; crush.
  unfold os_accessible; crush.
  rename x6 into pa.
  destruct (padd_eq (os_current_page x) pa); crush.
  (* ma1 = pt_addr *)
  assert (osi_pa_to_ma s2 osi0 (os_current_page x) = Some x3).
  apply H30; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in H20; crush.
  unfold osi_pa_to_ma in H20; crush.
  rewrite <- H43 in *; crush.
  rewrite STAC.del_page_of_vadd_madd_smem_pt_page
    with (s:=s1) (pt_page:=pt_page) (pt:=pt) (osi':=osi0) (va':=va)
    in H22; crush.
  exists (Page (PT_c (PT.remove pt0 va)) (page_owned_by pt_page0) (page_cacheable pt_page0));
    crush.
  rewrite STAC.del_page_of_vadd_madd_smem_pt_page
    with (s:=s2) (pt_page:=pt_page0) (pt:=pt0) (osi':=osi0) (va':=va); crush.
  unfold osi_current_page_table; crush.
  rewrite H43 in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache0 with va0 x3 p; crush.
  unfold is_rw_page in *; crush.
  destruct H31 with pt_page; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in H22; crush.
  destruct H_valid_cache0 with va0 x3 x6; crush.
  unfold is_rw_page in *; crush.
  rewrite H43 in *; crush.
  unfold same_page_but_value in *; destruct H41; destruct H42;
    repeat split; simpl; auto.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  (* ma <> pt_addr *)
  destruct (vadd_eq va0 va); crush.
  (* va0 = va *)
  unfold osi_pa_to_pg in H5; crush.
  rewrite STAC.del_get_mem_pt_page
    with (s:=s1) (pt_page:=pt_page) (pt:=pt) (osi':=osi0) (va':=va)
    in H5; crush.
  injection H15; intros; subst mp1.
  rewrite PT.get_remove in *; crush.
  unfold ACT.post_del in H2; crush.
  unfold osi_current_page_table; crush.
  unfold ACT.post_del in H2; crush.
  (* va0 <> va *)
  rewrite <- STAC.del_page_of_vadd_madd_smem 
    with (s:=s1) (osi':=osi0) (pt_addr:=pt_addr) (pt_addr':=pt_addr) (va':=va)
    in H46; crush.
  destruct H31 with pg1; crush.
  exists x6; crush.
  unfold osi_pa_to_pg in H27; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem
    with (s:=s2) (osi':=osi0) (pt_addr:=i)
      (pt_addr':=pt_addr0) (va':=va); crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists (os_current_page x); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H28; crush.
  assert (osi_pa_to_ma s2 osi0 pa = Some pt_addr0).
  apply H30; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H12; crush.
  destruct H12 with x; crush.
  rewrite H53 in *; crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists (os_current_page x); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H23; crush.
  (* va0 not accessible *)
  assert (va <> va0); crush.
  unfold ACT.pre_del in H1; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem
    with (s:=s1) (osi':=osi0) (pt_addr:=pt_addr) (pt_addr':=pt_addr) (va':=va)
     in H46; crush.
  destruct H31 with pg1; crush.
  exists x6; crush.
  destruct H12; crush.
  destruct H12 with x; crush.
  rewrite H52 in *; crush.
  unfold osi_pa_to_pg in H27; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem
    with (s:=s2) (osi':=osi0) (pt_addr:=pt_addr0)
      (pt_addr':=pt_addr0) (va':=va); crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists (os_current_page x4); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H44; crush.
  rewrite <- H28 in H44; injection H44; crush.
  unfold madd_of_vadd, madd_page_table in H25; crush.
  rewrite H23 in *; crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr0 = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr0 = Hyp_PO); crush.
  destruct H_valid_virtmap0 with osi0 pt_addr0 x1 x2 va0 pt_addr0; crush.
  unfold madd_owner, os_accessible in *; crush.
  rewrite H53; crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists (os_current_page x); unfold osi_pa_to_ma; crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Hyp_PO); crush.
  unfold madd_of_vadd, madd_page_table in H25; crush.
  rewrite H23 in *; crush.
  destruct H_valid_virtmap with osi0 pt_addr pt_page pt va0 pt_addr; crush.
  unfold madd_owner, os_accessible in *; crush.
  rewrite H49; crush.

  (* pt_padd <> os_current_page *)
  unfold osi_pa_to_pg in H5; crush.
  rename i0 into pt_addr1.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr1 va0 = Some ma1).
  eapply STAC.del_madd_of_vadd_old; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H15; crush.
  unfold madd_of_vadd, madd_page_table in H25; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H14.
  destruct H14 with pt_padd p0 p1 va0 ma1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in H2; crush.
  assert (st_hypervisor s2 = st_hypervisor s2').
  unfold ACT.post_del in H4; crush.
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi0; crush.
  STAC.get_osi_current_pt_page osi0; crush.
  assert (os_current_page x = os_current_page x4); crush.
  unfold same_current_pt_page in *; crush.
  destruct H12 with x; crush.
  exists x1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite <- H32; crush.
  unfold osi_pa_to_pg in H27; crush.
  generalize H3.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold osi_current_page_table in H50; crush.
  unfold madd_of_vadd, madd_page_table in H51; crush.
  rewrite H44 in *; crush.
  rewrite STAC.del_get_mem_distinct_pt with s2 s2' i0 osi0 x6 va; auto.
  unfold osi_current_page_table; crush.
  unfold madd_type; crush.
  rewrite <- H28; crush.
  crush.
  apply n; crush.
  rewrite H_injective_hyper0 with osi0 x5 pt_padd (os_current_page x4) x6;
    crush.
  exists x2; crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  apply H30; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.

  destruct (ctxt_vadd_accessible ctxt va0) eqn:?.
  (* va0 accessible *)
  destruct H_va_has_pa with osi0 x0
    pt_addr1 p0 p1 va0 ma1; crush.
  unfold ACT.post_del in H2; crush.
  rename x6 into pa.
  destruct (padd_eq (os_current_page x) pa); crush.
  (* ma1 = pt_addr *)
  assert (osi_pa_to_ma s2 osi0 (os_current_page x) = Some x3).
  apply H30; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in H20; crush.
  rewrite <- H43 in *; crush.
  rewrite STAC.del_page_of_vadd_madd_smem_pt_page
    with (s:=s1) (pt_page:=pt_page) (pt:=pt) (osi':=osi0) (va':=va)
    in H22; crush.
  exists (Page (PT_c (PT.remove pt0 va)) (page_owned_by pt_page0) (page_cacheable pt_page0));
    crush.
  rewrite STAC.del_page_of_vadd_madd_smem_pt_page
    with (s:=s2) (pt_page:=pt_page0) (pt:=pt0) (osi':=osi0) (va':=va); crush.
  unfold osi_current_page_table; crush.
  rewrite H43 in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache0 with va0 x3 p; crush.
  unfold is_rw_page in *; crush.
  destruct H31 with pt_page; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in H22; crush.
  destruct H_valid_cache0 with va0 x3 x6; crush.
  unfold is_rw_page in *; crush.
  rewrite H43 in *; crush.
  unfold same_page_but_value in *; destruct H41; destruct H42;
    repeat split; simpl; auto.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  (* ma <> pt_addr *)
  rewrite <- STAC.del_page_of_vadd_madd_smem_ne_pt 
    with (s:=s1) (pt_addr:=pt_addr1) (osi':=osi0) (pt_addr':=pt_addr) (va':=va)
    in H46; auto.
  destruct H31 with pg1; crush.
  exists x6; crush.
  unfold osi_pa_to_pg in H27; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem_ne_pt 
    with (s:=s2) (pt_addr:=i0) (osi':=osi0) (pt_addr':=pt_addr0) (va':=va); auto.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H28; crush.
  intro; subst i0.
  apply n.
  rewrite H_injective_hyper0 with osi0 x5 pt_padd (os_current_page x4) pt_addr0; crush.
  intro; subst x3.
  apply n0.
  rewrite H_injective_hyper0 with osi0 x5 pa (os_current_page x) pt_addr0; crush.
  assert (osi_pa_to_ma s2 osi0 pa = Some pt_addr0).
  apply H30.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in H50; crush.
  rewrite H43 in *; crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold ACT.post_del in H2; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  intro; subst pt_addr1.
  apply n.
  rewrite H_injective_hyper with osi0 x0 pt_padd (os_current_page x4) pt_addr; crush.
  unfold ACT.post_del in H2; crush.
  unfold ACT.post_del in H2; crush.
  rewrite H43 in *; crush.
  intro; subst ma1.
  apply n0.
  rewrite H_injective_hyper with osi0 x0 pa (os_current_page x) pt_addr; crush.
  (* va0 not accessible *)
  assert (va <> va0); crush.
  unfold ACT.pre_del in H1; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem
    with (s:=s1) (osi':=osi0) (pt_addr:=pt_addr1) (pt_addr':=pt_addr) (va':=va)
    in H46; crush.
  destruct H31 with pg1; crush.
  exists x6; crush.
  unfold osi_pa_to_pg in H27; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem
    with (s:=s2) (osi':=osi0) (pt_addr:=i0) (pt_addr':=pt_addr0) (va':=va); crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H28; crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr0 = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr0 = Hyp_PO); crush.
  destruct H_valid_virtmap0 with osi0 i0 x1 x2 va0 pt_addr0; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i0 = Os_PO osi0); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner, os_accessible in *; crush.
  rewrite H52; crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold ACT.post_del in H2; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Hyp_PO); crush.
  destruct H_valid_virtmap with osi0 pt_addr1 p0 p1 va0 pt_addr; crush.
  unfold ACT.post_del in H2; crush.
  unfold ACT.post_del in H2; crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr1 = Os_PO osi0); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner, os_accessible in *; crush.
  rewrite H49; crush.

  (* osi <> osi0 *)
  destruct H14.
  destruct H14 with pt_padd pt_page1 mp1 va0 ma1; crush.
  unfold osi_pa_to_pg in *; crush.
  assert (st_hypervisor s1 = st_hypervisor s1').
  unfold ACT.post_del in H2; crush.
  rewrite H19; crush.
  generalize H1.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  rewrite <- STAC.del_get_mem_distinct_pt with s1 s1' i0 osi x va; auto.
  rewrite <- H19 in Heqopt_res; crush.
  erewrite <- STAC.del_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite <- H15; crush.
  intro; subst x.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H24; crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO osi0); crush.
  rewrite H_valid_hyper with osi0 i pt_padd i0; auto.
  rewrite H19; crush.
  unfold madd_owner in *; crush.
  exists x; crush.
  unfold osi_pa_to_pg in H19; unfold osi_pa_to_pg; crush.
  assert (st_hypervisor s2 = st_hypervisor s2').
  unfold ACT.post_del in H4; crush.
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite <- H24; crush.
  rewrite STAC.del_get_mem_distinct_pt with s2 s2' i0 osi pt_addr va; auto.
  unfold osi_current_page_table; crush.
  unfold madd_type; crush.
  rewrite <- H20; crush.
  intro; subst i0.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr = Os_PO osi0).
  rewrite H_valid_hyper with osi0 i pt_padd pt_addr; auto.
  unfold madd_owner in H35; crush.
  exists x0; crush.
  rewrite <- STAC.del_osi_pa_to_ma with (s:=s2) (osi':=osi) (va':=va); crush.
  apply H22; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  unfold valid_sm_state in *; crush.

  destruct H23 with pg1; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; crush.
  rewrite STAC.del_page_of_vadd_madd_smem_ne_osi
    with (s':=s1') (osi:=osi0) (osi':=osi) (pt_addr':=pt_addr) (va':=va);
    crush.
  unfold osi_current_page_table; crush.
  unfold osi_pa_to_pg in H19; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_pg in H5; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1')) i2 va0 = Some ma1).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H15; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) i2 va0 = Some ma1).
  erewrite STAC.del_madd_of_vadd_old; crush.
  unfold madd_of_vadd, madd_page_table in H32; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H_valid_virtmap with osi0 i2 p0 p1 va0 ma1; crush.
  unfold ACT.post_del in H2; crush.
  rewrite <- H_valid_hyper0 with osi0 i1 pt_padd i2; crush.
  unfold madd_owner; crush.
  unfold ACT.post_del in H2; crush.
  unfold madd_owner; crush.
  (* va0 is not accessible *)
  unfold osi_pa_to_pg in H5; crush.
  rewrite STAC.del_page_of_vadd_madd_smem_ne_pt
    with (s':=s1') (pt_addr:=i0) (osi':=osi) (pt_addr':=pt_addr) (va':=va);
    crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold ACT.post_del in H2; crush.  
  erewrite STAC.del_madd_of_vadd_old; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H15; crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Os_PO osi0); crush.
  unfold ACT.post_del in H2; crush.
  assert (madd_owner (smem_mm (st_memory s1)) pt_addr = Os_PO osi); crush.
  assert (madd_of_vadd (smem_mm (st_memory s1')) i0 va0 = Some pt_addr).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H15; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) i0 va0 = Some pt_addr).
  erewrite STAC.del_madd_of_vadd_old; crush.
  unfold madd_of_vadd, madd_page_table in H33; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H_valid_virtmap with osi0 i0 p0 p1 va0 pt_addr; crush.
  unfold ACT.post_del in H2; crush.
  rewrite <- H_valid_hyper with osi0 i pt_padd i0; crush.
  unfold madd_owner; crush.
  unfold ACT.post_del in H2; crush.
  unfold madd_owner in H29; crush.
  rewrite H35 in H29; crush.

  exists x2; crush.
  unfold osi_pa_to_pg in H19; crush.
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem_ne_pt
    with (s:=s2) (pt_addr:=i0) (osi':=osi) (pt_addr':=pt_addr) (va':=va);
    crush.
  unfold osi_current_page_table; crush.
  exists osi0; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H20; crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr = Os_PO osi0); crush.
  unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s2)) pt_addr = Os_PO osi); crush.
  destruct H_valid_virtmap with osi0 i0 x x0 va0 pt_addr; crush.
  rewrite <- H_valid_hyper with osi0 i pt_padd i0; crush.
  unfold madd_owner; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va0) eqn:?; unfold os_accessible in *; auto.
  rewrite H40 in *; crush.
  rewrite H41 in *; crush.
Qed.

Lemma del_osi_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (va:vadd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_del osi va s1 ->
    ACT.post_del osi va s1 s1' ->
    ACT.pre_del osi va s2 ->
    ACT.post_del osi va s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  intros; unfold osi_non_stealth_equivalent; crush.

  unfold osi_non_stealth_equivalent in *; crush.
  destruct H5 with osi0; crush.
  unfold same_current_pt_page, same_current_pt_page_one,
    ACT.pre_del, ACT.post_del in *; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  exists (OS None (os_current_page x3)); crush.
  destruct H6 with x3; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  destruct H6 with x3; crush.

  apply osi_non_stealth_equivalent_hyper_trans with s1.
  apply osi_non_stealth_equivalent_hyper_symm.
  apply del_non_stealth_isolation_hyper with (osi:=osi) (va:=va); crush.
  apply osi_non_stealth_equivalent_hyper_trans with s2.
  unfold osi_non_stealth_equivalent in *; crush.
  destruct H5 with osi0; crush.
  apply del_non_stealth_isolation_hyper with (osi:=osi) (va:=va); crush.

  unfold osi_non_stealth_equivalent_smem; crush.
  eapply del_non_stealth_pt_pages_eq with (s1:=s1) (s2:=s2); crush.
  eapply del_non_stealth_pt_pages_eq with (s1:=s2) (s2:=s1); crush.
Qed.

Lemma pin_non_stealth_hyper_eq :
  forall (s1 s2 s1' s2':state) (osi osi0:os_ident) (pa:padd) (t:ptype),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_pin osi pa t s1 ->
    ACT.post_pin osi pa t s1 s1' ->
    ACT.pre_pin osi pa t s2 ->
    ACT.post_pin osi pa t s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent_hyper_mapping s1' s2' osi0.
Proof.
  Opaque new_page.
  unfold osi_non_stealth_equivalent; crush.
  unfold valid_sm_state in *; crush.
  destruct H5 with osi0; crush; clear H5.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  destruct (os_ident_eq osi0 osi); crush.
  destruct (padd_eq pa0 pa); crush.
  unfold ACT.post_pin in H4; crush.
  exists x; exists (new_page (Some t) (Os_PO osi));
    crush.
  unfold osi_pa_to_ma; crush.
  rewrite <- HYP.get_set; crush.
  rewrite <- P2M.get_set; crush.
  rewrite <- get_add_mem; crush.
  clear dependent s2.
  unfold ACT.post_pin in *; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite <- P2M.get_set in *; crush.
  rewrite <- get_add_mem in *; crush.

  (* pa0 <> pa *)
  destruct H19.
  destruct H19 with pa0 ma1 pg1; crush.
  erewrite STAC.page_pin_osi_pa_to_ma_distinct_pa; crush.
  erewrite STAC.pin_same_get_mem with (osi':=osi) (osi:=osi) (pa:=pa0); crush.
  erewrite STAC.page_pin_osi_pa_to_ma_distinct_pa; crush.
  exists x; exists x0; crush.
  erewrite STAC.page_pin_osi_pa_to_ma; crush.
  erewrite <- STAC.pin_same_get_mem with (osi':=osi) (osi:=osi) (pa:=pa0); crush.

  (* osi0 <> osi *)
  destruct H19.
  destruct H19 with pa0 ma1 pg1; crush.
  erewrite STAC.page_pin_osi_pa_to_ma_distinct_osi; crush.
  erewrite STAC.pin_same_get_mem with (osi':=osi) (osi:=osi0) (pa:=pa0); crush.
  erewrite STAC.page_pin_osi_pa_to_ma_distinct_osi; crush.
  exists x; exists x0; crush.
  erewrite STAC.page_pin_osi_pa_to_ma; crush.
  erewrite <- STAC.pin_same_get_mem with (osi':=osi) (osi:=osi0) (pa:=pa0); crush.
Qed.  

Lemma pin_osi_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (pa:padd) (t:ptype),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_pin osi pa t s1 ->
    ACT.post_pin osi pa t s1 s1' ->
    ACT.pre_pin osi pa t s2 ->
    ACT.post_pin osi pa t s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  Opaque new_page.
  intros; unfold osi_non_stealth_equivalent; crush.

  unfold osi_non_stealth_equivalent in *; crush.
  destruct H5 with osi0; crush.
  unfold same_current_pt_page, same_current_pt_page_one,
    ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  exists (OS None (os_current_page x2)); crush.
  destruct H6 with x8; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  destruct H6 with x8; crush.

  unfold osi_non_stealth_equivalent_hyper; crush.
  eapply pin_non_stealth_hyper_eq with (s1:=s1) (s2:=s2); crush.
  eapply pin_non_stealth_hyper_eq with (s1:=s2) (s2:=s1); crush.

  apply osi_non_stealth_equivalent_smem_trans with s1.
  apply osi_non_stealth_equivalent_smem_symm.
  apply pin_os_non_stealth_isolation_smem with (osi:=osi) (pa:=pa) (type:=t); crush.
  apply osi_non_stealth_equivalent_smem_trans with s2.
  destruct H5 with osi0; crush.
  apply pin_os_non_stealth_isolation_smem with (osi:=osi) (pa:=pa) (type:=t); crush.
Qed.

Lemma unpin_non_stealth_hyper_eq :
  forall (s1 s2 s1' s2':state) (osi osi0:os_ident) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_unpin osi pa s1 ->
    ACT.post_unpin osi pa s1 s1' ->
    ACT.pre_unpin osi pa s2 ->
    ACT.post_unpin osi pa s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent_hyper_mapping s1' s2' osi0.
Proof.
  Opaque new_page.
  unfold osi_non_stealth_equivalent; crush.
  unfold valid_sm_state in *; crush.
  destruct H5 with osi0; crush; clear H5.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush.
  destruct (os_ident_eq osi0 osi); crush.
  destruct (padd_eq pa0 pa); crush.
  cut False; crush.
  clear dependent s2.
  unfold ACT.post_unpin in *; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite P2M.get_remove in *; crush.
  (* pa0 <> pa *)
  destruct H19.
  destruct H19 with pa0 ma1 pg1; crush.
  erewrite STAC.page_unpin_osi_pa_to_ma_old; crush.
  assert (osi_pa_to_pg s1 osi pa0 = Some pg1).
  eapply STAC.page_unpin_osi_pa_to_pg_old; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  unfold ACT.post_unpin in *; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=i0) in H5; crush.
  exists x; exists x0; crush.
  erewrite STAC.page_unpin_osi_pa_to_ma_distinct_pa; crush.
  assert (osi_pa_to_pg s2' osi pa0 = Some x0); crush.
  eapply STAC.page_unpin_osi_pa_to_pg_distinct_pa; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.post_unpin in H4; crush.
  rewrite <- HYP.get_set in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=x) in Heqopt_res0; crush.

  (* osi0 <> osi *)
  destruct H19.
  destruct H19 with pa0 ma1 pg1; crush.
  erewrite STAC.page_unpin_osi_pa_to_ma_old; crush.
  assert (osi_pa_to_pg s1 osi0 pa0 = Some pg1).
  eapply STAC.page_unpin_osi_pa_to_pg_old; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  unfold ACT.post_unpin in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  exists x; exists x0; crush.
  erewrite STAC.page_unpin_osi_pa_to_ma_distinct_osi; crush.
  assert (osi_pa_to_pg s2' osi0 pa0 = Some x0); crush.
  eapply STAC.page_unpin_osi_pa_to_pg_distinct_osi; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.post_unpin in H4; crush.
  rewrite HYP.get_set_distinct in *; crush.
Qed.

Lemma unpin_osi_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_unpin osi pa s1 ->
    ACT.post_unpin osi pa s1 s1' ->
    ACT.pre_unpin osi pa s2 ->
    ACT.post_unpin osi pa s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  Opaque new_page.
  intros; unfold osi_non_stealth_equivalent; crush.

  unfold osi_non_stealth_equivalent in *; crush.
  destruct H5 with osi0; crush.
  unfold same_current_pt_page, same_current_pt_page_one,
    ACT.pre_unpin, ACT.post_unpin in *; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  exists (OS None (os_current_page x3)); crush.
  destruct H6 with x7; crush.
  destruct (os_ident_eq osi osi0); crush;
    [rewrite <- OSS.get_set in * | rewrite OSS.get_set_distinct in *]; crush.
  destruct H6 with x7; crush.

  unfold osi_non_stealth_equivalent_hyper; crush.
  eapply unpin_non_stealth_hyper_eq with (s1:=s1) (s2:=s2); crush.
  eapply unpin_non_stealth_hyper_eq with (s1:=s2) (s2:=s1); crush.

  apply osi_non_stealth_equivalent_smem_trans with s1.
  apply osi_non_stealth_equivalent_smem_symm.
  apply unpin_os_non_stealth_isolation_smem with (osi:=osi) (pa:=pa); crush.
  apply osi_non_stealth_equivalent_smem_trans with s2.
  destruct H5 with osi0; crush.
  apply unpin_os_non_stealth_isolation_smem with (osi:=osi) (pa:=pa); crush.
Qed.

Lemma lswitch_stealth_os_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (pa:padd),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    ACT.pre_lswitch pa osi s1 ->
    post_sm_lswitch pa osi s1 s1' ->
    ACT.pre_lswitch pa osi s2 ->
    post_sm_lswitch pa osi s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  Local Hint Resolve
        ctxt_switch_non_stealth_equivalent_hyper_mapping_1
        ctxt_switch_non_stealth_equivalent_hyper_mapping_2
        ctxt_switch_non_stealth_equivalent_pt_pages_1
        ctxt_switch_non_stealth_equivalent_pt_pages_2.
  intros.
  generalize H, H0; unfold valid_sm_state; crush.
  unfold ACT.pre_lswitch, post_sm_lswitch in *; crush.
  rewrite H2 with x0 in *; crush; clear H2.
  rewrite H4 with x in *; crush; clear H4.

  unfold osi_non_stealth_equivalent in *; crush;
    try destruct H5 with osi; crush.
  unfold same_current_pt_page, same_current_pt_page_one; crush;
  destruct (os_ident_eq osi (st_active_os s2)); crush;
    try (rewrite <- OSS.get_set in *; crush; fail);
    rewrite OSS.get_set_distinct in *; crush;
    destruct H2;
    try (destruct H2 with x1; crush; fail);
    try (destruct H30 with x1; crush; fail).

  unfold osi_non_stealth_equivalent_hyper in *; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s1; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s2; crush.
  unfold osi_non_stealth_equivalent_hyper in *; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s2; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s1; crush.
  unfold osi_non_stealth_equivalent_smem in *; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s1; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s2; crush.
  unfold osi_non_stealth_equivalent_smem in *; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s2; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s1; crush.

  rewrite H2 with x0 in *; crush; clear H2.
  rewrite H4 with x in *; crush; clear H4.
  unfold osi_non_stealth_equivalent in *; crush;
    try destruct H5 with osi; crush.
  unfold same_current_pt_page, same_current_pt_page_one; crush;
  destruct (os_ident_eq osi (st_active_os s2)); crush;
    try (rewrite <- OSS.get_set in *; crush; fail);
    rewrite OSS.get_set_distinct in *; crush;
    destruct H2;
    try (destruct H2 with x1; crush; fail);
    try (destruct H30 with x1; crush; fail).
  
  unfold osi_non_stealth_equivalent_hyper in *; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s1; crush.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush;
    repeat eexists; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s2; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s2; crush.

  unfold osi_non_stealth_equivalent_smem in *; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s2; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s2; crush.

  rewrite H2 with x0 in *; crush; clear H2.
  rewrite H4 with x in *; crush; clear H4.
  unfold osi_non_stealth_equivalent in *; crush;
    try destruct H5 with osi; crush.
  unfold same_current_pt_page, same_current_pt_page_one; crush;
  destruct (os_ident_eq osi (st_active_os s1)); crush;
    try (rewrite <- OSS.get_set in *; crush; fail);
    rewrite OSS.get_set_distinct in *; crush;
    destruct H2;
    try (destruct H2 with x1; crush; fail);
    try (destruct H30 with x1; crush; fail).

  unfold osi_non_stealth_equivalent_hyper in *; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s1; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s2; crush.
  unfold osi_non_stealth_equivalent_hyper_mapping; crush;
    repeat eexists; crush.
  apply osi_non_stealth_equivalent_hyper_mapping_trans with s1; crush.

  unfold osi_non_stealth_equivalent_smem in *; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s1; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s2; crush.
  unfold osi_non_stealth_equivalent_pt_pages; crush;
    repeat eexists; crush.
  apply osi_non_stealth_equivalent_pt_pages_trans with s1; crush.

  rewrite H2 with x0 in *; crush; clear H2.
  rewrite H4 with x in *; crush; clear H4.
  unfold osi_non_stealth_equivalent in *; crush;
    try destruct H5 with osi0; crush.
  unfold same_current_pt_page, same_current_pt_page_one; crush;
  destruct (os_ident_eq osi osi0); crush;
    try (rewrite <- OSS.get_set in *; crush; fail);
    rewrite OSS.get_set_distinct in *; crush;
    destruct H2;
    try (destruct H2 with x1; crush; fail);
    try (destruct H30 with x1; crush; fail).
Qed.


Ltac eq_one_step_non_stealth :=
  match goal with
    | [H1: osi_non_stealth_equivalent ?s1 ?s2
       |-  osi_non_stealth_equivalent ?s1' ?s2'] =>
      apply osi_non_stealth_equivalent_trans with s1;
        [apply osi_non_stealth_equivalent_symm |
         apply osi_non_stealth_equivalent_trans with s2]; crush;
      try (eapply os_non_stealth_isolation_not_osi; crush;
           match goal with
             | [H4: non_stealth_access ?a |- _ ] => inversion H4; crush; fail
           end; fail)
  end.

Theorem os_non_stealth_isolation :
  forall (s1 s2 s1' s2':state) (a:sm_action),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    valid_sm_exec s1 a s1' ->
    valid_sm_exec s2 a s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    osi_non_stealth_equivalent s1' s2'.
Proof.
  Local Hint Resolve osi_non_stealth_equivalent_refl
    read_stealth_os_non_stealth_isolation
    write_stealth_os_non_stealth_isolation
    new_stealth_os_non_stealth_isolation
    new_osi_non_stealth_isolation
    del_stealth_os_non_stealth_isolation
    del_osi_non_stealth_isolation
    pin_osi_non_stealth_isolation
    unpin_osi_non_stealth_isolation
    switch_stealth_os_non_stealth_isolation
    lswitch_stealth_os_non_stealth_isolation.
  Local Hint Constructors non_stealth_access osi_action.

  intros.
  assert (valid_state s1).
  unfold valid_sm_state in *; crush.
  assert (valid_state s2).
  unfold valid_sm_state in *; crush.

  sm_action_induction a; [clear Case; STAC.action_induction a | | | | ];
    generalize H1 H2; unfold valid_sm_exec; crush;
    unfold ACT.post_silent, ACT.pre_read, ACT.pre_read_hyper,
      ACT.post_read, ACT.post_read_hyper, ACT.pre_write, ACT.pre_write_hyper,
      ACT.post_write, ACT.post_write_hyper in *;
    crush; try (eq_one_step_non_stealth; fail).
Qed.

End STEALTH_NON_STEALTH_EQ.

